"""Forge subcommands as well as the main entry point for the forge tools"""
import argparse
import logging
import os
from os import path
import shutil
import sys
import traceback
import json
import multiprocessing
import Queue
import threading
import urlparse

import forge
from forge import defaults, build_config, ForgeError
from forge import build as forge_build
from forge.generate import Generate
from forge.remote import Remote
from forge.templates import Manager
from forge.lib import try_a_few_times, AccidentHandler, FilterHandler, CurrentThreadHandler, classify_platform
from forge import async
from forge import cli
from forge import updater
import forge.reload
import forge.modules

LOG = logging.getLogger(__name__)
ENTRY_POINT_NAME = 'forge'
TARGETS_WE_CAN_RUN_FOR = ('firefox', 'ios', 'android', 'web', 'wp', 'chrome', 'osx', 'ios-native', 'android-native', 'crosswalk')
TARGETS_WE_CAN_PACKAGE_FOR = ('ios', 'android', 'web', 'wp', 'chrome', 'osx', 'ios-native', 'android-native', 'crosswalk')
ERROR_LOG_FILE = os.path.join(os.getcwd(), 'forge-error.log')

class RunningInForgeRoot(Exception):
    pass

class ArgumentError(Exception):
    pass

def _assert_outside_of_forge_root(app_path):
    if app_path == defaults.FORGE_ROOT:
        raise RunningInForgeRoot

def _assert_not_in_subdirectory_of_forge_root(app_path):
    cwd = str(app_path)
    if cwd.startswith(defaults.FORGE_ROOT + os.sep):
        raise RunningInForgeRoot

def _assert_have_target_folder(directory, target):
    if not os.path.isdir(path.join(directory, target)):
        raise ForgeError(
                "Can't run build for '{target}', because you haven't built it!\n"
                "If you're interested in targetting {target}, please contact support@trigger.io to sign up for one of our paid plans.".format(target=target)
        )

def _assert_have_development_folder():
    if not os.path.exists('development'):
        message = (
                "No folder called 'development' found. You're trying to run your app but you haven't built it yet!\n"
                "Try {prog} build first."
        ).format(
                prog=ENTRY_POINT_NAME
        )
        raise ForgeError(message)

def _check_working_directory_is_safe(app_path=None):
    if app_path is None:
        app_path = os.getcwd()

    _assert_outside_of_forge_root(app_path)
    try:
        _assert_not_in_subdirectory_of_forge_root(app_path)
    except RunningInForgeRoot:
        LOG.warning(
                "Working directory is a subdirectory of the forge build tools.\n"
                "This is probably a bad idea! Please do your app development in a folder outside\n"
                "of the build tools installation directory.\n"
        )

def _setup_error_logging_to_file():
    """Creates a handler which logs to a file in the case of an error, and
    attaches it to the root logger.
    """
    file_handler = logging.FileHandler(ERROR_LOG_FILE, delay=True)
    file_handler.setFormatter(logging.Formatter('%(asctime)s [%(levelname)7s] %(message)s'))
    file_handler.setLevel(logging.DEBUG)

    accident_handler = AccidentHandler(target=file_handler, capacity=9999, flush_level='ERROR')
    accident_handler.setLevel(logging.DEBUG)

    thread_handler = CurrentThreadHandler(target_handler=accident_handler)
    thread_handler.setLevel(logging.DEBUG)

    logging.root.addHandler(thread_handler)


def _setup_logging_to_stdout(stdout_log_level):
    """Creates a stream handler at the given log level and attaches it to
    the root logger.
    """
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(stdout_log_level)
    stream_handler.setFormatter(logging.Formatter('[%(levelname)7s] %(message)s'))

    handler = CurrentThreadHandler(target_handler=stream_handler)
    handler.setLevel(stdout_log_level)
    logging.root.addHandler(handler)

def _filter_requests_logging():
    """Stops requests from logging to info!"""
    logging.getLogger('requests').setLevel(logging.ERROR)

def setup_logging(settings):
    'Adjust logging parameters according to command line switches'
    verbose = settings.get('verbose')
    quiet = settings.get('quiet')
    if verbose and quiet:
        raise ArgumentError('Cannot run in quiet and verbose mode at the same time')
    if verbose:
        stdout_log_level = logging.DEBUG
    elif quiet:
        stdout_log_level = logging.WARNING
    else:
        stdout_log_level = logging.INFO

    logging.root.setLevel(logging.DEBUG)

    _setup_logging_to_stdout(stdout_log_level)
    _setup_error_logging_to_file()
    _filter_requests_logging()

    LOG.info('Forge tools version: %s' % forge.get_version())

def add_primary_options(parser):
    '''Top-level command-line arguments for settings which affect the running of
    any command for any platform
    '''
    parser.add_argument('command', choices=COMMANDS.keys())
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-q', '--quiet', action='store_true')
    parser.add_argument('--username', help='your email address used to login to the forge website')
    parser.add_argument('--password', help='password used to login to the forge website')

def handle_primary_options(args):
    'Parameterise our option based on common command-line arguments'
    parser = argparse.ArgumentParser(prog='forge', add_help=False)
    add_primary_options(parser)

    handled_args, other_args = parser.parse_known_args(args)

    # TODO setup given user/password somewhere accessible by remote.py
    forge.settings['command'] = handled_args.command
    if handled_args.username:
        forge.settings['username'] = handled_args.username
    if handled_args.password:
        forge.settings['password'] = handled_args.password
    forge.settings['verbose'] = bool(handled_args.verbose)
    forge.settings['quiet'] = bool(handled_args.quiet)

    try:
        setup_logging(forge.settings)
    except ArgumentError as e:
        parser.error(e)

    return other_args


def _add_create_options(parser):
    parser.description = 'create a new application'
    parser.add_argument('--name', help='name of the application to create')
    parser.add_argument('--project', help='ID of the project to create the app in')

def _handle_create_options(handled):
    if handled.name:
        forge.settings['name'] = handled.name
    if handled.project:
        forge.settings['project'] = handled.project


def _add_build_options(parser):
    parser.description = 'Creates new local, unzipped development add-ons with your source and configuration'
    parser.add_argument('-f', '--full', action='store_true', help='force a complete rebuild of your app')

def _handle_build_options(handled):
    forge.settings['full'] = bool(handled.full)


def _add_run_options(parser):
    parser.description = 'Run a built app on a particular platform'
    parser.add_argument('platform', choices=TARGETS_WE_CAN_RUN_FOR)

def _handle_run_options(handled):
    forge.settings['platform'] = handled.platform


def _add_package_options(parser):
    parser.description='Package up a build for distribution'
    parser.add_argument('platform', choices=TARGETS_WE_CAN_PACKAGE_FOR)

def _handle_package_options(handled):
    forge.settings['platform'] = handled.platform


def _add_check_options(parser):
    parser.description='Do some testing on the current local configuration settings'

def _handle_check_options(handled):
    pass


def _add_reload_options(parser):
    parser.description='Run reload commands'

def _handle_reload_options(handled):
    pass


def _add_module_options(parser):
    parser.description="Run module commands"
    parser.add_argument("--name", help="name of the module")
    parser.add_argument("--project", help="id of the module's project")

def _handle_module_options(handled):
    if handled.name:
        forge.settings["name"] = handled.name
    if handled.project:
        forge.settings["project"] = handled.project


def _add_version_options(parser):
    parser.description='Output forge versions'

def _handle_version_options(handled):
    pass


def _add_serve_options(parser):
    parser.description='Run serve commands'
    parser.add_argument('platform', choices=TARGETS_WE_CAN_RUN_FOR)

def _handle_serve_options(handled):
    forge.settings['platform'] = handled.platform


def handle_secondary_options(command, args):
    parser = argparse.ArgumentParser(
            prog="{entry} {command}".format(entry=ENTRY_POINT_NAME, command=command),
            epilog="For more detailed information, see http://current-docs.trigger.io/tools/commands.html",
    )
    options_handlers = {
            "create": (_add_create_options, _handle_create_options),
            "build": (_add_build_options, _handle_build_options),
            "run": (_add_run_options, _handle_run_options),
            "package": (_add_package_options, _handle_package_options),
            "check": (_add_check_options, _handle_check_options),
            "reload": (_add_reload_options, _handle_reload_options),
            "serve": (_add_serve_options, _handle_serve_options),
            "module": (_add_module_options, _handle_module_options),
            "version": (_add_version_options, _handle_version_options),
    }

    # add command-specific arguments
    options_handlers[command][0](parser)

    handled, other = parser.parse_known_args(args)

    # parse command-specific arguments
    options_handlers[command][1](handled)

    return other

def create(unhandled_args):
    'Create a new app'
    _check_working_directory_is_safe()
    config = build_config.load()
    remote = Remote(config)
    updater.distribution_aware_updater().check_version(remote)
    remote._authenticate()

    if os.path.exists(defaults.SRC_DIR):
        raise ForgeError('Source folder "%s" already exists, if you really want to create a new app you will need to remove it!' % defaults.SRC_DIR)

    # schema for async.emit
    questions_to_ask = {
            'description': 'Enter details for app',
            'properties': {
                    'name': {
                            'type': 'string',
                            'title': 'App Name',
                            'description': 'This name is what your application will be called on devices. You can change it later through config.json.'
                    },
                    'project': {
                            'type': 'string',
                            'title': 'Project ID',
                            'description': 'Choose the project ID for your app:\n\n' +
                                    "\n".join([
                                            "  " + str(project['id']) + "\t" + project['name']
                                            for project in remote._api_get('project')['data']['projects']
                                    ]) + "\n"
                    }
            }
    }
    answers_to_questions = {}

    if "name" in forge.settings and forge.settings["name"]:
        del questions_to_ask["properties"]["name"]
        answers_to_questions["name"] = forge.settings["name"]
    if "project" in forge.settings and forge.settings["project"]:
        del questions_to_ask["properties"]["project"]
        answers_to_questions["project"] = forge.settings["project"]

    if questions_to_ask["properties"]:
        event_id = async.current_call().emit('question', schema=questions_to_ask)

        answers = async.current_call().wait_for_response(event_id)
        for key in questions_to_ask["properties"]:
            answers_to_questions[key] = answers['data'][key]
    uuid = remote.create(name=answers_to_questions["name"], project=answers_to_questions["project"])
    remote.fetch_initial(uuid)
    LOG.info("Building app for the first time...")
    development_build([], has_target=False)
    LOG.info('App structure created. To proceed:')
    LOG.info('1) Put your code in the "%s" folder' % defaults.SRC_DIR)
    LOG.info('2) Run %s build to make a build' % ENTRY_POINT_NAME)
    LOG.info('3) Run %s run to test out your build' % ENTRY_POINT_NAME)

FORGE_BUILD_NEEDS_TARGET = """Target required for 'forge build', e.g. 'forge build android'"""

def development_build(unhandled_args, has_target=True):
    '''Pull down new version of platform code in a customised build, and create unpacked development add-on.

    :param has_target: If this is False, just fetch the generation instructions, don't build any targets.
    '''
    _check_working_directory_is_safe()

    if not os.path.isdir(defaults.SRC_DIR):
        raise ForgeError(
                'Source folder "{src}" does not exist - have you run {prog} create yet?'.format(
                        src=defaults.SRC_DIR,
                        prog=ENTRY_POINT_NAME,
                )
        )

    config = build_config.load()
    remote = Remote(config)
    updater.distribution_aware_updater().check_version(remote)
    remote._authenticate()
    manager = Manager(config)

    instructions_dir = defaults.INSTRUCTIONS_DIR
    if forge.settings.get('full', False):
        # do this first, so that bugs in generate_dynamic can always be nuked with a -f
        LOG.debug("Full rebuild requested: removing previous templates")
        shutil.rmtree(instructions_dir, ignore_errors=True)

    app_config = build_config.load_app()

    should_rebuild = remote.server_says_should_rebuild()
    server_changed = should_rebuild['should_rebuild']
    reason = should_rebuild['reason']
    stable_platform = should_rebuild['stable_platform']

    if server_changed:
        # Need new generate dynamic - download it
        LOG.debug("Server requires rebuild: {reason}".format(reason=reason))
        LOG.info("Your Forge platform has been updated, downloading updated build instructions.")
        manager.fetch_instructions()

        # repeat the whole procedure, as we may have migrated the app in some way
        forge.settings['full'] = False
        return development_build(unhandled_args, has_target)

    # Try to do a build using generate_dynamic.toolkit if it's available
    try:
        generate_dynamic = forge_build.import_generate_dynamic()
    except ForgeError as e:
        # don't have generate_dynamic available yet
        LOG.info("Error: could not locate generate_dynamic: %s" % e)
        return

    if hasattr(generate_dynamic, "buildtools"):
        LOG.info("Using generate_dynamic.buildtools for build")
        return generate_dynamic.buildtools.development_build_dynamic(unhandled_args, has_target, manager, remote, app_config, stable_platform, development_build)

    # otherwise use deprecated static instructions instead
    config_changed = manager.need_new_templates_for_config()
    if config_changed:
        # Need new builds due to local config change
        LOG.info("Your local config has been changed, downloading updated build instructions.")
        manager.fetch_instructions()

        # repeat the whole procedure, as we may have migrated the app in some way
        forge.settings['full'] = False
        return development_build(unhandled_args, has_target)

    reload_result = remote.create_buildevent(app_config)['data']
    if not has_target:
        # No need to go further if we aren't building a target
        return

    try:
        target = unhandled_args.pop(0)
        if target.startswith("-"):
            raise ForgeError(FORGE_BUILD_NEEDS_TARGET)
    except IndexError:
        raise ForgeError(FORGE_BUILD_NEEDS_TARGET)

    # Not all targets output into a folder by the same name.
    target_dirs = {
            'safari': 'forge.safariextension',
    }
    target_dir = target
    if target in target_dirs:
        target_dir = target_dirs[target]

    reload_config = json.loads(reload_result['config'])
    reload_config_hash = reload_result['config_hash']


    if target != "reload": # Don't do a server side build for reload
        if not path.exists(path.join('.template', target_dir)):
            LOG.info("Your app configuration has changed since your last build of this platform, performing a remote build of your app. Once this is downloaded future builds will be faster.")

            build = remote.build(config=reload_config, target=target)
            remote.fetch_unpackaged(build, to_dir=defaults.TEMPLATE_DIR, target=target)
    else:
        LOG.info('Config matches previously downloaded build, performing local build.')

    current_platform = app_config['platform_version']

    # Advise user about state of their current platform
    platform_category = classify_platform(stable_platform, current_platform)
    if platform_category == 'nonstandard':
        LOG.warning("Platform version: %s is a non-standard platform version, it may not be receiving updates and it is recommended you update to the stable platform version: %s" % (current_platform, stable_platform))

    elif platform_category == 'minor':
        # do nothing: not an issue to be on a minor platform since v2.0.0
        pass

    elif platform_category == 'old':
        LOG.warning("Platform version: %s is no longer the current platform version, it is recommended you switch to a newer version." % current_platform)

    def move_files_across():
        shutil.rmtree(path.join('development', target_dir), ignore_errors=True)
        if target != "reload":
            # Delete reload as other targets may build it
            shutil.rmtree(path.join('development', 'reload'), ignore_errors=True)
            # No reload server template
            shutil.copytree(path.join(defaults.TEMPLATE_DIR, target_dir), path.join('development', target_dir), symlinks=True)

    # Windows often gives a permission error without a small wait
    try_a_few_times(move_files_across)

    # Put config hash in config object for local generation
    # copy first as mutating dict makes assertions about previous uses tricky
    reload_config_for_local = reload_config.copy()
    reload_config_for_local['config_hash'] = reload_config_hash

    # have templates and instructions - inject code
    generator = Generate()
    generator.all('development', defaults.SRC_DIR, extra_args=unhandled_args, config=reload_config_for_local, target=target)

    LOG.info("Development build created. Use {prog} run to run your app.".format(
            prog=ENTRY_POINT_NAME
    ))

def run(unhandled_args):
    _check_working_directory_is_safe()
    build_type_dir = 'development'
    _assert_have_development_folder()
    _assert_have_target_folder(build_type_dir, forge.settings['platform'])

    generate_dynamic = forge_build.import_generate_dynamic()

    build_to_run = forge_build.create_build(
            build_type_dir,
            targets=[forge.settings['platform']],
            extra_args=unhandled_args,
    )

    generate_dynamic.customer_goals.run_app(
            generate_module=generate_dynamic,
            build_to_run=build_to_run,
    )

def package(unhandled_args):
    #TODO: ensure dev build has been done first (probably lower down?)
    generate_dynamic = forge_build.import_generate_dynamic()
    build_type_dir = 'development'
    build_to_run = forge_build.create_build(
            build_type_dir,
            targets=[forge.settings['platform']],
            extra_args=unhandled_args,
    )

    generate_dynamic.customer_goals.package_app(
            generate_module=generate_dynamic,
            build_to_run=build_to_run,
    )

def check(unhandled_args):
    '''
    Run basic linting on project JS to save the user some trouble.
    '''
    if not os.path.isdir(defaults.SRC_DIR):
        raise ForgeError(
                'Source folder "{src}" does not exist - have you run {prog} create yet?'.format(
                        src=defaults.SRC_DIR,
                        prog=ENTRY_POINT_NAME,
                )
        )

    try:
        generate_dynamic = forge_build.import_generate_dynamic()
    except ForgeError:
        # don't have generate_dynamic available yet
        LOG.info("Unable to check local settings until a build has completed")
        return
    build_to_run = forge_build.create_build(
            "development",
            targets=[],
            extra_args=unhandled_args,
    )
    generate_dynamic.customer_goals.check_settings(
            generate_dynamic,
            build_to_run,
    )

def reload(unhandled_args):
    '''
    Run reload module command
    '''
    if not os.path.isdir(defaults.SRC_DIR):
        raise ForgeError(
                'Source folder "{src}" does not exist - have you run {prog} create yet?'.format(
                        src=defaults.SRC_DIR,
                        prog=ENTRY_POINT_NAME,
                )
        )

    try:
        generate_dynamic = forge_build.import_generate_dynamic()
    except ForgeError:
        # don't have generate_dynamic available yet
        raise ForgeError("Unable to use reload until a build has completed")

    build_to_run = forge_build.create_build(
            "development",
            targets=["reload"],
    )
    config_dict = build_config.load()
    remote = Remote(config_dict)
    forge.reload.run_command(build_to_run, remote, unhandled_args)

def serve(unhandled_args):
    '''
    Update app code changes instantly to device
    '''
    if not os.path.isdir(defaults.SRC_DIR):
        raise ForgeError(
                'Source folder "{src}" does not exist - have you run {prog} create yet?'.format(
                        src=defaults.SRC_DIR,
                        prog=ENTRY_POINT_NAME,
                )
        )

    try:
        generate_dynamic = forge_build.import_generate_dynamic()
    except ForgeError:
        # don't have generate_dynamic available yet
        raise ForgeError("Unable to serve changes until a build has completed")

    build_to_run = forge_build.create_build(
            "development",
            targets=[forge.settings['platform']],
            extra_args=unhandled_args,
    )

    generate_dynamic.customer_goals.serve_app(
            generate_module=generate_dynamic,
            build_to_run=build_to_run,
    )


def native_module(unhandled_args):
    '''
    Manage native modules
    '''
    module_path = os.getcwd()
    _check_working_directory_is_safe(module_path)

    config = build_config.load()
    forge.modules.run_command(config, module_path, unhandled_args)


def version(unhandled_args):
    '''
    Output forge versions
    '''
    config = build_config.load()
    remote = Remote(config)
    response = remote._api_get("platform")
    platforms = response["data"]
    LOG.info("Forge platform stable version: %s" % platforms["latest_stable"])


def _dispatch_command(command, other_args):
    """Runs our subcommand in a separate thread, and handles events emitted by it"""
    call = None
    task_thread = None
    try:
        other_other_args = handle_secondary_options(command, other_args)

        subcommand = COMMANDS[command]

        # setup enough stuff so the target function can communicate back using events
        call = async.Call(
                call_id=0,
                target=subcommand,
                args=(other_other_args, ),
                input=Queue.Queue(),
                output=Queue.Queue(),
                io_lock=multiprocessing.RLock()
        )
        async.set_current_call(call, thread_local=False)

        # capture logging on any thread but this one and turn it into events
        handler = async.CallHandler(call)
        handler.setLevel(logging.DEBUG)

        current_thread = threading.current_thread().name
        filtered_handler = FilterHandler(handler, lambda r: r.threadName != current_thread)
        filtered_handler.setLevel(logging.DEBUG)

        logging.root.addHandler(filtered_handler)
        logging.root.setLevel(logging.DEBUG)

        task_thread = threading.Thread(target=call.run)
        task_thread.daemon = True
        task_thread.start()

        while True:
            try:
                # KeyboardInterrupts aren't seen until the .get() completes :S
                # So we set a timeout here to make sure we receive it
                next_event = call._output.get(block=True, timeout=1)
            except Queue.Empty:
                continue
            event_type = next_event['type']

            if event_type == 'question':
                response = cli.ask_question(next_event)
                response.update({
                        'eventId': next_event['eventId'],
                })
                call.input(response)

            if event_type == 'confirm':
                answer = cli.ask_confirm(next_event)
                call.input({
                        'eventId': next_event['eventId'],
                        'data': answer
                })

            if event_type == 'progressStart':
                cli.start_progress(next_event)

            if event_type == 'progressEnd':
                cli.end_progress(next_event)

            if event_type == 'progress':
                cli.progress_bar(next_event)

            # TODO: handle situation of logging while progress bar is running
            # e.g. extra newline before using LOG.log
            if event_type == 'log':
                # all logging in our task thread comes out as events, which we then
                # plug back into the logging system, which then directs it to file/console output
                logging_level = getattr(logging, next_event.get('level', 'DEBUG'))
                LOG.log(logging_level, next_event.get('message', ''))

            elif event_type == 'success':
                return 0

            elif event_type == 'error':
                # re-raise exception originally from other thread/process
                try:
                    raise call.exception

                except RunningInForgeRoot:
                    LOG.error(
                            "You're trying to run commands in the build tools directory.\n"
                            "You need to move to another directory outside of this one first.\n"
                    )

                except updater.UpdateRequired:
                    LOG.info("An update to these command line tools is required, downloading...")

                    config = build_config.load()
                    remote = Remote(config)
                    try:
                        updater.distribution_aware_updater().get_update_zip(remote)
                        _restart_if_supported_by_os("Downloaded update zip")
                    except Exception as e:
                        LOG.error("Upgrade process failed: %s" % e)
                        LOG.debug("%s" % traceback.format_exc(e))
                        LOG.error("You can get the tools from https://trigger.io/api/latest_tools and extract them yourself")
                        LOG.error("Contact support@trigger.io if you have any further issues")

                except ForgeError as e:
                    # thrown by us, expected
                    LOG.error(next_event.get('message'))
                    LOG.debug(str(next_event.get('traceback')))

                except Exception:
                    LOG.error("Something went wrong that we didn't expect:")
                    LOG.error(next_event.get('message'))
                    LOG.debug(str(next_event.get('traceback')))

                    LOG.error("See %s for more details" % ERROR_LOG_FILE)
                    LOG.error("Please contact support@trigger.io")

                return 1
    except KeyboardInterrupt:
        sys.stdout.write('\n')
        LOG.info('Exiting...')
        if call:
            call.interrupt()
            task_thread.join(timeout=5)
        return 1

def _restart_if_supported_by_os(restart_reason):
    """If we "restart" the commandline tools using os.execv on windows, the original process
    returns, so instead we tell the user to re-run the same command.
    """
    if sys.platform.startswith("win"):
        LOG.info("%s: please re-execute your command." % restart_reason)
        sys.exit(1)
    LOG.info("%s: restarting..." % restart_reason)
    updater.restart()

def main():
    """The main entry point for the commandline program.

    Parses enough to figure out what subparser to hand off to, sets up logging and error handling
    for the chosen sub-command.

    Checks for an update zip and applies it if it's there.
    """
    # this is used in the forge package to figure out if we're running from the commandline
    forge.settings['commandline'] = True

    # need to handle primary options such as -v to setup logging
    other_args = handle_primary_options(sys.argv[1:])

    LOG.debug('Checking for update zip...')
    if updater.distribution_aware_updater().extract_update_zip_if_present():
        _restart_if_supported_by_os("Applied update")
    sys.exit(_dispatch_command(forge.settings['command'], other_args))

COMMANDS = {
        'create'  : create,
        'build'   : development_build,
        'run'     : run,
        'package' : package,
        'check'   : check,
        'reload'  : reload,
        'serve'   : serve,
        'module'  : native_module,
        'version'  : version,
}
