import json
import logging
import os
import re
import sys
import traceback
import validictory
import zipfile

from cookielib import LWPCookieJar
from distutils.version import StrictVersion
from send2trash import send2trash

from forge import async
from forge import build as forge_build
from forge import defaults, remote
from forge import lib
from forge import settings
from forge import updater

from forge.remote import Remote


# - diagnostics ---------------------------------------------------------------

from forge import ForgeError

LOG = logging.getLogger(__name__)

class ModulesError(ForgeError):
    pass


# - entry point ---------------------------------------------------------------

def run_command(build_config, module_path, args):
    remote = Remote(build_config)
    updater.distribution_aware_updater().check_version(remote)
    remote._authenticate()

    # TODO how do we force update modules in .trigger/cache when they change in inspector_config.json?

    # download or refresh .trigger/module_dynamic if needed
    manifest = _manifest(module_path)
    if "platform_version" in manifest:
        platform_version = manifest["platform_version"]
    else:
        platforms = remote._api_get("platform")["data"]
        platform_version = platforms["latest_stable"]
    update_module_dynamic(remote, module_path, platform_version)

    # parse arguments and dispatch commands
    LOG.info("")
    if len(args) == 0:
        raise ModulesError("Expecting command after 'forge module'")
    if args[0] == "create":
        if len(args) != 1:
            raise ModulesError("Invalid number of arguments passed to 'forge module create'")
        create_module(remote, module_path)
    if args[0] == "init":
        if len(args) != 1:
            raise ModulesError("Invalid number of arguments passed to 'forge module init'")
        create_module(remote, module_path, only_create_files=True)
    elif args[0] == "inspector":
        if len(args) != 2:
            raise ModulesError("Invalid number of arguments passed to 'forge module inspector <target>'")
        update_inspector(remote, build_config, module_path, args[1])
    elif args[0] == "check":
        if len(args) != 1:
            raise ModulesError("Invalid number of arguments passed to 'forge module check'")
        check_module(remote, module_path, manifest)
    elif args[0] == "upload":
        if len(args) != 1:
            raise ModulesError("Invalid number of arguments passed to 'forge module upload'")
        upload_module(remote, module_path, manifest)
    elif args[0] == "versions":
        if len(args) != 1:
            raise ModulesError("Invalid number of arguments passed to 'forge module versions'")
        module_versions(remote, module_path)
    elif args[0] == "publish":
        if len(args) != 2:
            raise ModulesError("Invalid number of arguments passed to 'forge module publish'")
        publish_version(remote, module_path, args[1])
    elif args[0] == "unpublish":
        if len(args) != 2:
            raise ModulesError("Invalid number of arguments passed to 'forge module unpublish'")
        unpublish_version(remote, module_path, args[1])
    elif args[0] == "set_current":
        if len(args) != 2:
            raise ModulesError("Invalid number of arguments passed to 'forge module set_current'")
        set_current_version(remote, module_path, args[1])
    else:
        raise ModulesError("Unknown command '%s'. (choose from 'create', 'init', 'inspector', 'check', 'upload', 'versions', 'publish', 'unpublish', 'set_current')" % args[0])


# - create module -------------------------------------------------------------

def create_module(remote, module_path, only_create_files=False):
    if _manifest_exists(module_path):
        raise ForgeError(
            "Cannot create a new module here, the module manifest '{manifest}' already exists.".format(
                manifest=manifest_path
            )
        )

    # get project and module name from user
    projects = remote._api_get("project")["data"]["projects"]
    project_list = ["  " + str(project["id"]) + "\t" + project["name"] for project in projects]
    questions = {
        "description": "Enter details for module",
        "properties": {
            "module_name": {
                "type": "string",
                "title": "Module Name",
                "description": "The name of your module, used as the identifier for module specific files and when communicating between native and javascript code:\n"
            },
            "project_id": {
                "type": "string",
                "title": "Project ID",
                "description": "Choose the project ID for your module:\n\n" + "\n".join(project_list) + "\n"
            }
        }
    }
    answers = {}
    if "name" in settings and settings["name"]:
        del questions["properties"]["module_name"]
        answers["module_name"] = settings["name"]
    if "project" in settings and settings["project"]:
        del questions["properties"]["project_id"]
        answers["project_id"] = settings["project"]
    event_id = async.current_call().emit("question", schema=questions)
    response = async.current_call().wait_for_response(event_id)
    for key in questions["properties"]:
        answers[key] = response["data"][key]

    # validate project id
    project_id = answers["project_id"]
    if not re.match("^[0-9]+$", project_id):
        raise ModulesError("Invalid project id '%s'. A project ID must be a number." % project_id)
    project = next(iter(filter(lambda project: project["id"] == int(project_id), projects)), None)
    if not project:
        raise ModulesError("Invalid project id '%s'. Please make sure you selected the correct id." % project_id)

    # validate module name
    module_name = answers["module_name"]
    if not re.match("^[a-z][a-z0-9_]+$", module_name):
        raise ModulesError("Invalid module name '%s'. Name must start with a lowercase letter and only contain lowercase letters, numbers and _." % module_name)

    # create module in backend
    if only_create_files == False:
        LOG.info("Create module '%s' in project '%s'..." % (module_name, project["name"]))

        response = remote._api_post("module/", data = {
            "name": module_name,
            "project": project_id
        })
        LOG.info("Module created with response: %s" % response["result"])

    # create module files
    LOG.info("Create module files")
    module_dynamic = forge_build.import_module_dynamic()
    module_dynamic.module.create_template(name=module_name, path=module_path)

    # create inspector_config file
    inspector_config = _inspector_config(module_path)
    inspector_config_path = os.path.join(module_path, defaults.MODULE_DIR, "inspector_config.json")
    with open(inspector_config_path, "w") as f:
        json.dump(inspector_config, f, indent=4)

    LOG.info("Module structure created. To proceed:")
    LOG.info("1) Run 'forge inspector <target>' to create the inspector projects for this module. Supported targets are: android ios")
    LOG.info("2) Open the 'inspector/an-inspector' project in Android Studio or 'inspector/ios-inspector' in Xcode.")
    LOG.info("3) Remember to shut down your IDE and update the inspector projects by running 'forge inspector <target> every time you make changes to the files in the 'module/' directory.")
    LOG.info("You can find documentation for native module development here: https://trigger.io/docs/current/api/native_modules/index.html")


# - update inspector ----------------------------------------------------------

def update_inspector(remote, build_config, module_path, target):
    if not _manifest_exists(module_path):
        raise ForgeError(
            "Module manifest '{manifest_path}' does not exist - have you run 'forge module create' yet?".format(
                manifest_path=manifest_path
            )
        )

    # initialize forge_tool
    try:
        import trigger
        from trigger import forge_tool
        singleton_params = {
            "config": build_config
        }
        forge_tool.init(singleton_params)
    except ImportError as e:
        LOG.info("trigger.forge_tool is not available: %s" % e)
        return

    # get cookies - TODO toolkit and build-tools should use a unified backend driver
    import cookielib
    path_to_cookies = remote._path_to_cookies()
    cookies = LWPCookieJar(path_to_cookies)
    cookies.load(path_to_cookies)
    cookies = _cookies_for_domain(remote.hostname, cookies)

    # get module dependencies
    inspector_config = _inspector_config(module_path)
    modules = inspector_config["modules"]
    dependencies = map(lambda name: {
        "name": name, "version": modules[name]["version"]
    }, modules)

    # get system config
    # https://trigger.io/api/v1/modules/archive_links/?project=13659&modules=%7B%22file%22%3A%20%222.23%22%7D
    # where modules is a uri encoded object of format: {"file": "2.23", "foo": "1.0", "bar": "0.4.2"} ?
    system_config = {
        "module_urls": "https://s3.amazonaws.com/trigger-module-build/%s/%s.zip"
    }

    # update inspector
    module_dynamic = forge_build.import_module_dynamic()
    try:
        update_inspector = getattr(module_dynamic.inspector, "update_" + target)
    except:
        LOG.error("Invalid target '%s'. Valid targets are 'ios' and 'android'." % target)
        return
    LOG.info("Update inspector for target '%s'." % target)
    update_inspector(cookies, dependencies, system_config)


# - check module --------------------------------------------------------------

def check_module(remote, module_path, manifest):
    if not _manifest_exists(module_path):
        raise ForgeError(
            "Module manifest '{manifest_path}' does not exist - have you run 'forge module create' yet?".format(
                manifest_path=manifest_path
            )
        )

    identity = _identity(module_path)
    LOG.info("Validate module %s v%s", identity["name"], manifest["version"])

    # get inspector_config
    try:
        inspector_config = _inspector_config(module_path)
    except ValueError as error:
        LOG.error(error)
        return False

    # validate config for 'this_module'
    config = inspector_config["this_module"]["config"]
    config_schema = _config_schema(module_path)
    try:
        validictory.validate(config, config_schema)
    except ValueError as error:
        LOG.error("Error validating module/inspector_config.json: %s" % str(error))
        return False

    # check any dependent modules included in inspector
    modules = inspector_config["modules"]
    dependencies = map(lambda name: {
        "name": name, "version": modules[name]["version"]
    }, modules)
    errors = filter(lambda dependency: _check_module_dependency(module_path, manifest, inspector_config, dependency), dependencies)
    for dependency in errors:
        LOG.error("Error validating dependency '%s': %s" % (dependency["name"], dependency["error"]))
    if len(errors):
        return False

    # check for native module binaries
    binary_path = os.path.join(module_path, defaults.MODULE_DIR, "android", "ForgeModule-vanilla-release.aar")
    if not os.path.isfile(binary_path):
        LOG.warning("No Android binary found for module. Have you built it and copied it to the 'module/android' directory?")
    binary_path = os.path.join(module_path, defaults.MODULE_DIR, "ios", "module.a")
    if not os.path.isfile(binary_path):
        LOG.warning("No iOS binary found for module. Have you built it and copied it to the 'module/ios' directory?")

    # validate module directory structure
    messages = _validate_module_directory_structure(module_path)
    if len(messages):
        for message in messages:
            LOG.error(message)
        return False

    # check if we've already uploaded this version of the module
    if _existing_version(remote, module_path, identity["name"], manifest):
        LOG.warning("You have already uploaded a build of this module with version '%s'." % manifest["version"])
        LOG.warning("Further uploads will only modify the documentation, description, changes and minimum / maximum supported platform versions.");
        LOG.warning("To upload new code you will first need to increment the module version number in 'module/manifest.json'.")

    LOG.info("Finished validating module")

    return True


def _check_module_dependency(module_path, manifest, inspector_config, dependency):
    LOG.info("Validate module dependency: {name} v{version}".format(**dependency))

    name = dependency["name"]
    version = dependency["version"]
    directory = os.path.join(".trigger", "cache", "%s-%s" % (name, version))

    # validate config for dependency
    config = inspector_config["modules"][name]["config"]
    config_schema = _config_schema(module_path, directory)
    try:
        validictory.validate(config, config_schema)
    except ValueError as error:
        dependency["error"] = str(error)
        return True

    # check min platform version for dependency
    module_platform = manifest["platform_version"]
    dependency_manifest = _manifest(module_path, directory)
    dependency_min_platform = dependency_manifest["min_platform_version"]
    comparison = _compare_versions(module_platform, dependency_min_platform)
    if comparison < 0:
        dependency["error"] = "Module dependency '{name} v{version}' requires platform version %s or higher but this module is on platform version %s".format(**dependency) % (dependency_min_platform, module_platform)
        return True

    return False


# - upload module -------------------------------------------------------------

def upload_module(remote, module_path, manifest):
    if not check_module(remote, module_path, manifest):
        LOG.error("Cannot upload module. Please fix all reported errors first.")
        return

    # zip module files
    LOG.info("Zip module files")
    module_dynamic = forge_build.import_module_dynamic()
    module_zip = module_dynamic.module.create_upload_zip(path=module_path)

    # check if this is an update or an upload
    identity = _identity(module_path)
    is_update = _existing_version(remote,  module_path, identity["name"], manifest)
    if is_update:
        LOG.info("Updating module: %s v%s", identity["name"], manifest["version"])
    else:
        LOG.info("Uploading module: %s v%s", identity["name"], manifest["version"])

    # upload .trigger/upload_tmp.zip
    with open(module_zip, "rb") as f:
        response = remote._api_post("module/{name}/version".format(**identity), data = {
            "is_update": is_update
        }, files = {
            "module_zip": f
        })
        LOG.info("Action completed with response: %s" % response["result"])


# - module versions -----------------------------------------------------------

def module_versions(remote, module_path):
    if not _manifest_exists(module_path):
        raise ForgeError(
            "Module manifest '{manifest_path}' does not exist - have you run 'forge module create' yet?".format(
                manifest_path=manifest_path
            )
        )

    # get module versions
    identity = _identity(module_path)
    response = remote._api_get("module/{name}".format(**identity))
    module = response["data"]
    versions = module["versions"]

    # print version information
    for version in versions:
        LOG.info("# {name} v{version} {published} {current}".format(
            name=identity["name"],
            version=version["version"],
            published=version["approval"] == "pending" and "[pending approval]" or
                      version["published"] and "[published]" or "[unpublished]",
            current=version["current"] and "[current]" or ""
        ))

        changes = version["changes"].split("\n")
        changes = filter(lambda c: len(c) > 0, changes)
        if not len(changes):
            LOG.info("Changes: No changes were documented for this version.")
        else:
            LOG.info("Changes:")
            map(lambda c: LOG.info("  %s", c.strip()), changes)

        LOG.info("")


# - publish version -----------------------------------------------------------

def publish_version(remote, module_path, version):
    LOG.info("MODULE LICENCING")
    LOG.info("")
    LOG.info("Publishing this version will make it publically available, pending review by the Trigger.io team.")
    LOG.info("")
    LOG.info("If this is the first version you are publishing it will also cause your module to be publically listed on Trigger.io.")
    LOG.info("")
    LOG.info("When you publish a module, you are agreeing to make the module source code and compiled binary available under the 2-clause BSD license (as recorded at: https://opensource.org/licenses/BSD-2-Clause), and that you have made the source and binary available publicly on the Github account associated with this project.")

    # confirm action
    questions = {
        "description": "Confirmation",
        "properties": {
            "confirm": {
                "type": "string",
                "title": "(yes/no)",
                "description": "Do you accept this agreement?"
            }
        }
    }
    event_id = async.current_call().emit("question", schema=questions)
    response = async.current_call().wait_for_response(event_id)
    if response["data"]["confirm"] != "yes":
        LOG.info("Action cancelled.")
        return

    # perform action
    identity = _identity(module_path)
    LOG.info("Publish module %s v%s", identity["name"], version)
    response = _update_version(remote, module_path, identity["name"], version, "publish")
    LOG.info("Action completed with response: %s %s", response["result"], response["approval"] or "")
    if response["approval"] == "pending":
        LOG.info("Your module version will be approved as soon as a Trigger.io team member has reviewed it. (Usually within 24 hours during the week.)")



# - unpublish version ---------------------------------------------------------

def unpublish_version(remote, module_path, version):
    LOG.warning("Unpublishing a module version will mean any users currently using it will no longer be able to build their app. Please only unpublish versions if there is a good reason such as a major security issue.")

    # confirm action
    questions = {
        "description": "Confirmation",
        "properties": {
            "confirm": {
                "type": "string",
                "title": "(yes/no)",
                "description": "Please confirm whether you wish to proceed."
            }
        }
    }
    event_id = async.current_call().emit("question", schema=questions)
    response = async.current_call().wait_for_response(event_id)
    if response["data"]["confirm"] != "yes":
        LOG.info("Action cancelled.")
        return

    # perform action
    identity = _identity(module_path)
    LOG.info("Unpublish module %s v%s", identity["name"], version)
    response = _update_version(remote, module_path, identity["name"], version, "publish")
    LOG.info("Action completed with response: %s", response["result"])



# - set current version -------------------------------------------------------

def set_current_version(remote, module_path, version):
    identity = _identity(module_path)
    LOG.info("Set module %s v%s to be the current version", identity["name"], version)
    response = _update_version(remote, module_path, identity["name"], version, "set_current")
    LOG.info("Action completed with response: %s" % response["result"])


# - module_dynamic ------------------------------------------------------------

def update_module_dynamic(remote, module_path, manifest_platform_version):
    current_platform_version = ""
    current_platform_version_path = os.path.join(module_path, ".trigger", "platform_version.txt")
    if os.path.exists(current_platform_version_path):
        with async.current_call().io_lock:
            with open(current_platform_version_path, "r") as f:
                current_platform_version = f.read()

    if current_platform_version != manifest_platform_version:
        _update_module_dynamic(remote, module_path, manifest_platform_version)


def _update_module_dynamic(remote, module_path, platform_version):
    if os.path.exists(os.path.join(module_path, ".trigger")):
        for dirpath, dirnames, filenames in os.walk(os.path.join(module_path, ".trigger"), topdown=False):
            for filename in filenames:
                send2trash(os.path.join(dirpath, filename))
            for dirname in dirnames:
                send2trash(os.path.join(dirpath, dirname))
        send2trash(os.path.join(module_path, ".trigger"))

    os.makedirs(os.path.join(module_path, ".trigger"))

    module_dynamic_path = os.path.join(module_path, ".trigger", "module_instructions.zip")
    remote._get_file(remote.make_api_url("platform/%s/module_instructions" % platform_version), module_dynamic_path)

    try:
        with zipfile.ZipFile(module_dynamic_path) as z:
            z.extractall(os.path.join(module_path, ".trigger"))
    finally:
        os.unlink(module_dynamic_path)


# - helpers -------------------------------------------------------------------

def _identity(module_path):
    identity_path = os.path.join(module_path, defaults.MODULE_DIR, "identity.json")
    try:
        with open(identity_path) as f:
            return json.load(f)
    except ValueError:
        raise ValueError("Couldn't parse module/identity.json file: %s" % identity_path)


def _manifest_exists(module_path):
    manifest_path = os.path.join(module_path, defaults.MODULE_DIR, "manifest.json")
    return os.path.isfile(manifest_path)


def _manifest(module_path, directory=None):
    if not directory:
        manifest_path = os.path.join(module_path, defaults.MODULE_DIR, "manifest.json")
    else:
        manifest_path = os.path.join(module_path, directory, "manifest.json")

    manifest = {}
    try:
        with open(manifest_path) as f:
            try:
                manifest = json.load(f)
            except ValueError:
                raise ValueError("Couldn't parse module/manifest.json file: %s" % manifest_path)
    except IOError:
        return manifest

    return manifest


def _inspector_config(module_path, directory=None):
    if not directory:
        inspector_config_path = os.path.join(module_path, defaults.MODULE_DIR, "inspector_config.json")
    else:
        inspector_config_path = os.path.join(module_path, directory, "inspector_config.json")

    inspector_config = {
        "this_module": {
            "config": {}
        },
        "modules": {}
    }
    try:
        with open(inspector_config_path) as f:
            try:
                inspector_config = json.load(f)
            except ValueError:
                raise ValueError("Couldn't parse module/inspector_config.json file: %s" % inspector_config_path)
    except IOError:
        return inspector_config

    if not "this_module" in inspector_config:
        raise ValueError("Missing field in module/inspector_config.json: 'this_module'")
    elif not "config" in inspector_config["this_module"]:
        raise ValueError("Missing field in module/inspector_config.json: 'this_module.config'")
    elif not "modules" in inspector_config:
        raise ValueError("Missing field in module/inspector_config.json: 'modules'")

    return inspector_config


def _config_schema(module_path, directory=None):
    if not directory:
        config_schema_path = os.path.join(module_path, defaults.MODULE_DIR, "config_schema.json")
    else:
        config_schema_path = os.path.join(module_path, directory, "config_schema.json")

    config_schema = {}
    try:
        with open(config_schema_path) as f:
            try:
                config_schema = json.load(f)
            except ValueError:
                raise ValueError("Couldn't parse module/config_schema.json file: %s" % config_schema_path)
    except IOError:
        return config_schema

    return config_schema


def _validate_module_directory_structure(module_path):
    schema_path = os.path.join(".trigger", "schema", "module_structure.json")
    try:
        with open(schema_path) as f:
            schema = json.load(f)
    except ValueError:
        raise ValueError("Couldn't parse .trigger/schema/module_structure.json file: %s" % schema_path)

    # build up module directory structure
    files_path = os.path.join(module_path, defaults.MODULE_DIR)
    structure = {}
    files_path = files_path.rstrip(os.sep)
    start = files_path.rfind(os.sep) + 1
    for dirname, dirnames, filenames in os.walk(files_path):
        if '.placeholder' in filenames:
            filenames.remove('.placeholder')
        folders = dirname[start:].split(os.sep)
        subdir = dict.fromkeys(filenames, "file")
        parent = reduce(dict.get, folders[:-1], structure)
        parent[folders[-1]] = subdir

    # validate module directory structure
    def validate_structure(structure, schema, path, messages):
        if "validators" in schema:
            # TODO validate json_schema for identity.json, inspector_config.json, manifest.json
            return messages

        schema_type = schema["type"]
        if schema_type == "file":
            if structure == "file":
                return messages
            else:
                messages.append("The file '%s' was expected to be a directory." % path)
                return messages
        elif schema_type == "directory":
            if structure == "file":
                messages.append("The file '%s' was expected to be a directory." % path)
                return messages
            if "children" in schema:
                schema_children = schema["children"]
                for schema_child in schema_children:
                    # TODO hacky way of improving missing .a / .jar message
                    # this *should* be handled better with improved logic for what
                    # is and is not a valid structure for a module.
                    # TODO as of 10/2019 this can probably be deprecated as we check for it separately
                    #      and we may still want modules with an android or ios directory because
                    #      they might have build_steps even if they have no binaries.
                    if "required" in schema_children[schema_child] and schema_children[schema_child]["required"]\
                                                                   and not schema_child in structure:
                        message = "The path '{child_path}' is required and was not found."
                        if schema_child.endswith(".aar"):
                            message = "Module binary '{child_path}' is required. If you don't want to build for Android remove the '{path}' directory."
                        elif schema_child.endswith(".a"):
                            message = "Module binary '{child_path}' is required. If you don't want to build for iOS remove the '{path}' directory."
                        messages.append(message.format(child_path=os.path.join(defaults.MODULE_DIR, path, schema_child), path=os.path.join(defaults.MODULE_DIR, path)))

            if path != "":
                path += "/"

            for structure_child in structure:
                schema_child = schema["children"][structure_child] if "children" in schema and structure_child in schema["children"] else None
                if schema_child:
                    messages = validate_structure(structure[structure_child], schema["children"][structure_child], os.path.join(path, structure_child), messages)
                elif "additionalChildren" in schema:
                    additionalChildren = schema["additionalChildren"]
                    if isinstance(additionalChildren, bool) and additionalChildren == False and structure_child[0] != '.':
                        messages.append("The path '{child_path}' was found but not expected. You will need to remove it.".format(
                            child_path=os.path.join(defaults.MODULE_DIR, path, structure_child)
                        ))
                    elif isinstance(additionalChildren, bool) and additionalChildren == True:
                        pass # keep going
                    elif structure_child[0] != '.':
                        messages = validate_structure(structure_child, schema["additionalChildren"], os.path.join(path, structure_child), messages)
        else:
            LOG.warning("Encountered unknown schema type while validating module directory structure: %s", schema_type)

        return messages

    return validate_structure(structure["module"], schema, "", [])


def _cookies_for_domain(domain, cookies):
    return dict((c.name, c.value) for c in cookies if c.domain == domain)


def _existing_version(remote, module_path, name, manifest):
    response = remote._api_get("module/{name}".format(name=name))
    module = response["data"]
    existing_version = next(iter(filter(lambda v: v["version"] == manifest["version"], module["versions"])), None)
    if existing_version:
        return True
    return False


def _update_version(remote, module_path, name, version, action):
    '''supported actions are: publish, unpublish, set_current'''
    response = remote._api_get("module/{name}/version/{version}/{action}".format(
        name=name,
        version=version,
        action=action,
    ))
    return response


def _compare_versions(a, b):
    if a == b:
        return 0

    a = a[a.startswith('v') and 1:]
    b = b[b.startswith('v') and 1:]

    if StrictVersion(a) < StrictVersion(b):
        return -1
    elif StrictVersion(a) > StrictVersion(b):
        return 1
    else:
        return 0
