"""Logic for performing reload actions"""
from StringIO import StringIO
import hashlib
import json

import logging
from zipfile import ZipFile
from forge import ForgeError, lib
from forge import build_config
import os
import re

LOG = logging.getLogger(__name__)

class ReloadError(ForgeError):
    pass

def run_command(build, remote, args):
    remote._authenticate()

    if len(args) == 0:
        raise ReloadError("Expecting command after 'forge reload'")
    if args[0] == 'list':
        if len(args) != 1:
            raise ReloadError("Invalid number of arguments passed to 'forge reload list'")
        list_streams(build, remote)
    elif args[0] == 'create':
        if len(args) != 2:
            raise ReloadError("Invalid number of arguments passed to 'forge reload create'")
        create_stream(build, remote, args[1])
    elif args[0] == 'push':
        if len(args) != 2:
            raise ReloadError("Invalid number of arguments passed to 'forge reload push'")
        push_stream(build, remote, args[1])
    elif args[0] == 'pushcdn':
        if len(args) != 3:
            raise ReloadError("Invalid number of arguments passed to 'forge reload pushcdn'")
        pushcdn_stream(build, remote, args[1], args[2])
    else:
        raise ReloadError("Unknown command 'forge reload %s'" % args[0])

def list_streams(build, remote):
    LOG.info("List of streams from website:")
    content_dict = remote._api_get('reload/%s/stream' % build.config['uuid'])

    for stream in content_dict['data']['streams']:
        LOG.info('  %s' % stream['name'])

def create_stream(build, remote, stream_id):
    # Validate id
    if not re.match("^[a-z0-9_-]+$", stream_id):
        raise ReloadError("Invalid stream id, must only contain lowercase a-z, 0-9, - and _.")

    LOG.info("Creating stream with website")
    created = remote._api_post('reload/%s/stream' % build.config['uuid'], data={'name': stream_id})
    LOG.info("Stream created, use 'forge reload push %s' to push a snapshot" % stream_id)

def push_stream(build, remote, stream_id):
    manifest = dict()
    file_for_hash = dict()

    src = os.path.join('development', 'reload', 'src')
    for root, dirs, files in os.walk(src):
        for filename in files:
            filename = os.path.join(root, filename)
            with open(filename, 'rb') as file_to_be_hashed:
                hash_of_file = hashlib.sha1(file_to_be_hashed.read()).hexdigest()
                manifest[filename[len(src) + 1:].replace('\\', '/')] = hash_of_file
                file_for_hash[hash_of_file] = filename

    content_dict = remote._api_post('reload/snapshot/filter', files={'manifest': StringIO(json.dumps(manifest))})
    hashes_to_upload = set(content_dict['data']['manifest'].values())

    with lib.temp_file() as zip_file_path:
        with ZipFile(zip_file_path, 'w') as zip_file:
            for hash_of_file in hashes_to_upload:
                zip_file.write(file_for_hash[hash_of_file], hash_of_file)

        with open(zip_file_path, 'rb') as zip_file:
            remote._api_post('reload/snapshot/%s/%s' % (build.config['uuid'], stream_id), files={
                    'config': StringIO(json.dumps(build_config.load_app())),
                    'manifest': StringIO(json.dumps(manifest)),
                    'forge-deploy': zip_file
            })
    LOG.info("Pushed snapshot to stream '%s'" % stream_id)

def pushcdn_stream(build, remote, stream_id, manifest_url):
    external = dict()
    external["manifest_url"] = manifest_url
    remote._api_post('reload/snapshot/%s/%s' % (build.config['uuid'], stream_id), files={
            'config': StringIO(json.dumps(build_config.load_app())),
            'external': StringIO(json.dumps(external))
    })

    LOG.info("Pushed snapshot to stream '%s' with manifest url '%s'" % (stream_id, manifest_url))
