'Operations which require involvement of the remote Forge build servers'
from cookielib import LWPCookieJar
import json
import logging
import os
from os import path
import shutil
import time
import urlparse
from urlparse import urljoin, urlsplit
from threading import Lock, Thread
from Queue import Queue
import thread
import webbrowser
import hashlib
from copy import copy
import weakref
from collections import deque

import requests

import forge
from forge import build_config, defaults
from forge import ForgeError, lib

LOG = logging.getLogger(__name__)

cookie_lock = Lock()
cache_lock = Lock()
read_lock = Lock()
read_count_lock = Lock()
read_count = 0

# Set to force specific cookies to be used rather than read in from file
# Used to override older code which can't be changed
override_cookies = None

class RequestError(ForgeError):
    def __init__(self, response, message, errors=None):
        ForgeError.__init__(self, message)
        self.response = response
        self.errors = errors

    def extra(self):
        return dict(
                content=self.response.content,
                errors=self.errors,
        )

class FormError(ForgeError):
    def __init__(self, errors, *args, **kw):
        ForgeError.__init__(self, *args, **kw)
        self.errors = errors

    def extra(self):
        return dict(
                errors=self.errors
        )

def _format_form_errors(form_errors):
    lines = []
    for field_name, field_errors in form_errors.items():
        lines.append("%s: %s" % (field_name, ", ".join(field_errors)))
    return "There was a problem submitting a form:\n" + "\n".join(lines)


def _raise_if_json_error(content_dict, error_template, resp, url):
    if content_dict and content_dict['result'] == 'error':
        text = content_dict.get('text')
        errors = content_dict.get('errors')

        reason = text
        if not reason:
            if errors:
                reason = _format_form_errors(errors)

        if not reason:
            reason = "Unknown error"

        error_message = error_template.format(url=url, reason=reason)
        raise RequestError(resp, error_message, errors=errors)

def _check_api_response_for_error(url, method, resp, text=None):
    '''Check an API response from the website to see if there was an error. Checks for one of the following:

    No status code, as in, no valid response from the server.

    HTTP status code is not 200 but get a JSON response
    HTTP status code is not 200 and the response is not a valid JSON response

    Code 200 with a valid JSON response and the 'result' property is set to 'error'
    Code 200 with no JSON response

    :param url: The API url used in the request
    :param method: The HTTP method used in the request (e.g. GET or POST)
    :param resp: The response from the API call
    :raises: RequestException if the response is an error or malformed
    '''
    if text is None:
        text = method + ' to %(url)s failed: status code %(status_code)s\n%(content)s'

    LOG.debug("Checking API response for success or error")
    error_template = "Forge API call to {url} went wrong: {reason}"

    content_dict = None
    try:
        content_dict = json.loads(resp.content)
    except ValueError:
        pass

    # result was critical: show a message and abort
    if content_dict and 'result' in content_dict and content_dict['result'] == 'critical':
        LOG.debug("Got critical result %s" % json.dumps(content_dict, indent=4))

        if 'commandline' not in forge.settings:
            webbrowser.open(content_dict['url'])
            thread.interrupt_main()

        raise RequestError(resp, "There is a problem with your tools, follow the instructions at %s" % content_dict['url'])

    _raise_if_json_error(content_dict, error_template, resp, url)

    # server didn't return 2xx status code, but may have returned JSON
    if not resp.ok:
        if resp.status_code is None:
            raise RequestError(resp, "Request to {url} got no response".format(url=url))

        try:
            msg = text % resp.__dict__
        except KeyError:
            # in case error_message is looking for something resp doesn't have
            msg = method + ' to %s failed: %s' % (
                    url,
                    str(resp)
            )
        raise RequestError(resp, msg)

    if not content_dict:
        reason = 'Server meant to respond with JSON, but response content was: %s' % resp.content
        text = error_template.format(reason=reason, url=url)
        raise RequestError(resp, text)


def _check_response_for_error(url, method, resp):
    if not resp.ok:
        raise RequestError(resp, "Request to {url} went wrong, error code: {code}".format(url=url, code=resp.status_code))

def _cookies_for_domain(domain, cookies):
    return dict((c.name, c.value) for c in cookies if c.domain == domain)

class Remote(object):
    'Wrap remote operations'
    POLL_DELAY = 10

    def __init__(self, config, use_cookie_file=True, enable_caching=False):
        'Start new remote Forge builds'
        self.session = requests.session()
        self.session.prefetch = True
        self.session.store_cookies = False
        self.config = config
        self.async_queue = None
        if override_cookies:
            use_cookie_file = False
        self.use_cookie_file = use_cookie_file
        if use_cookie_file:
            cookie_path = self._path_to_cookies()
            self.cookies = LWPCookieJar(cookie_path)
            with cookie_lock:
                if not os.path.exists(cookie_path):
                    self.cookies.save()
                else:
                    self.cookies.load()
        self.enable_caching = enable_caching
        if enable_caching:
            self.cache = {}
            self.invalidate_last = 0
            self.invalidate_on = {}
            self.invalidate_thread_queue = None
            self._add_invalidate_on([], "first")
            self.invalidator_monitors = []
            self.recently_invalidated = deque()
        self.validated_email = None
        self.thirdparty_module_warning = None

    def _path_to_cookies(self):
        return path.join(defaults.FORGE_ROOT, 'cookies.txt')

    def _invalidate_cache(self, invalidate):
        with cache_lock:
            for pair in invalidate:
                name, id = pair.split(":")

                # Make sure we haven't already handled this invalidation message
                if id in self.recently_invalidated and id != 0:
                    continue
                self.recently_invalidated.append(id)
                while len(self.recently_invalidated) > 20:
                    self.recently_invalidated.popleft()

                to_delete = []

                for monitorref in self.invalidator_monitors:
                    monitor = monitorref()
                    if monitor is None:
                        to_delete.append(monitorref)
                    else:
                        monitor.emit('invalidate', name=name)

                for monitorref in to_delete:
                    self.invalidator_monitors.remove(monitorref)

                if name == "*":
                    # Invalidate everything
                    self.cache = {}
                    self.invalidate_on = {}
                    LOG.debug("Invalidated whole cache")
                elif name in self.invalidate_on:
                    hashes = self.invalidate_on.pop(name)
                    for hash in hashes:
                        if hash in self.cache:
                            LOG.debug("Invalidated cache for: "+name)
                            self.cache.pop(hash)

    def _add_invalidate_on(self, invalidate_on, request_hash):
        """Add a list of names to invalid a particular request hash cache on"""

        with cache_lock:
            new_names = False

            if request_hash == "first": # Send an initial invalidator request with no names.
                new_names = True

            for name in invalidate_on:
                if name in self.invalidate_on:
                    self.invalidate_on[name].append(request_hash)
                else:
                    self.invalidate_on[name] = [request_hash]
                    new_names = True

            if new_names:
                # If we have new invalidation names to watch kick off a new invalidation thread immediately.

                if self.invalidate_thread_queue is not None:
                    # Stop any existing long polling thread
                    self.invalidate_thread_queue.put(True)


                def invalidate_cache(remote, stop_queue):
                    while True:
                        kw = {}
                        if remote.config.get('main', {}).get('authentication'):
                            kw['auth'] = (
                                    remote.config['main']['authentication'].get('username'),
                                    remote.config['main']['authentication'].get('password')
                            )

                        if remote.config.get('main', {}).get('proxies'):
                            kw['proxies'] = remote.config['main']['proxies']

                        kw['prefetch'] = True

                        try:
                            with cache_lock: # Make sure we're not currently modifying the cache
                                stop_queue.get(block=False)
                            return
                        except Exception:
                            pass

                        try:
                            resp = remote.session.get(urljoin(remote.server, 'invalidator/?last=%d&names=%s' % (remote.invalidate_last, ",".join(remote.invalidate_on.keys()))), **kw)
                            last = json.loads(resp.content)['last']
                        except Exception:
                            # Something went wrong, clear the cache, invalidator will restart on next request.
                            LOG.debug("Something went wrong in invalidator request, about to clear cache")
                            with read_lock: # Wait till any current requests are done
                                LOG.debug("Got into lock")
                                self.cache = {}
                                self.invalidate_last = 0
                                self.invalidate_on = {}
                                self.invalidate_thread_queue = None
                            LOG.debug("Got out of lock")
                            return

                        with read_lock: # Only invalidate once all ongoing requests have finished
                            try:
                                stop_queue.get(block=False)
                                return
                            except Exception:
                                pass

                            remote.invalidate_last = last

                            if 'x-trigger-invalidate' in resp.headers:
                                invalidate = resp.headers['x-trigger-invalidate'].split(',')
                                remote._invalidate_cache(invalidate)

                        # Brief sleep in case this goes wrong
                        time.sleep(1)

                # If we need to watch any keys start a new thread to longpoll for them
                queue = Queue()
                self.invalidate_thread_queue = queue
                thread = Thread(target=invalidate_cache, args=(self, queue))
                thread.start()

    def add_invalidator_monitor(self, call):
        self.invalidator_monitors.append(weakref.ref(call))

    @property
    def server(self):
        'The URL of the build server to use (default https://trigger.io/api/)'
        return self.config.get('main', {}).get('server', 'https://trigger.io/api/')

    @property
    def hostname(self):
        'The hostname of the build server'
        return urlparse.urlparse(self.server).hostname

    def __queue_request(self, method, url, *args, **kw):
        '''A way to queue requests we don't need a response for'''
        if self.async_queue is None:
            self.async_queue = Queue()

            def handle_async_requests(queue, session):
                while True:
                    method, url, args, kw = queue.get()
                    getattr(session, method.lower())(url, *args, **kw)

            async_request_thread = Thread(target=handle_async_requests, args=(self.async_queue, self.session))
            async_request_thread.start()

        self.async_queue.put((method, url, args, kw))

    def __get_or_post(self, url, async=False, *args, **kw):
        global read_count
        '''Expects ``__method`` entry in :param:`**kw`
        '''

        with read_count_lock:
            read_count += 1
            if read_count == 1:
                read_lock.acquire()

        try:
            method = kw['__method']
            del kw['__method']

            kw['verify'] = self.config.get('verify_certificates', True)

            if self.use_cookie_file:
                kw['cookies'] = _cookies_for_domain(self.hostname, self.cookies)
            elif 'cookies' in kw and kw['cookies'] is not None:
                kw['cookies'] = dict((k.encode('utf8'),v.encode('utf8')) for k,v in kw['cookies'].items() if v is not None)
            elif override_cookies:
                kw['cookies'] = dict((k.encode('utf8'),v.encode('utf8')) for k,v in override_cookies.items() if v is not None)

            if method == "POST":
                # must have CSRF token
                data = kw.get("data", {})
                data['build_tools_version'] = forge.get_version()

                if 'cookies' in kw and 'csrftoken' in kw['cookies']:
                    data["csrfmiddlewaretoken"] = kw['cookies']['csrftoken']
                else:
                    raise Exception("We don't have a CSRF token")

                kw["data"] = data

            kw['headers'] = {
                    'REFERER': url,
                    'TRIGGER_TOOLS_VERSION': forge.get_version()
            }

            if self.config.get('main', {}).get('authentication'):
                if urlparse.urlparse(url).hostname == self.hostname:
                    kw['auth'] = (
                            self.config['main']['authentication'].get('username'),
                            self.config['main']['authentication'].get('password')
                    )

            if self.config.get('main', {}).get('proxies'):
                kw['proxies'] = self.config['main']['proxies']

            LOG.debug('{method} {url}'.format(method=method.upper(), url=url))

            try:
                if "files" in kw:
                    kw_copy = copy(kw)
                    del kw_copy['files']
                    request_hash = hashlib.sha1("%s%s%s" % (json.dumps(kw_copy), url, method))
                    for file in kw['files']:
                        request_hash.update(file)
                        request_hash.update(kw['files'][file].read())
                        kw['files'][file].seek(0)
                    request_hash = request_hash.hexdigest()
                else:
                    request_hash = hashlib.sha1("%s%s%s" % (json.dumps(kw), url, method)).hexdigest()
            except Exception:
                request_hash = None

            if self.enable_caching and request_hash is not None and request_hash in self.cache:
                LOG.debug('Cache hit for: {method} {url}'.format(method=method.upper(), url=url))
                resp = self.cache[request_hash]
            else:
                if async:
                    self.__queue_request(method, url, *args, **kw)
                    resp = None
                else:
                    # Make sure cookies are cleared before making a session request
                    self.session.cookies.clear()
                    resp = getattr(self.session, method.lower())(url, *args, **kw)

                    if self.use_cookie_file and resp:
                        lib.load_cookies_from_response(resp, self.cookies)
                        with cookie_lock:
                            self.cookies.save()

                    if self.enable_caching:
                        if 'x-trigger-cache' in resp.headers:
                            invalidate_on = []
                            if 'x-trigger-invalidate-on' in resp.headers:
                                invalidate_on = resp.headers['x-trigger-invalidate-on'].split(',')

                            if request_hash is not None:
                                self._add_invalidate_on(invalidate_on, request_hash)
                                with cache_lock:
                                    self.cache[request_hash] = resp

                        if 'x-trigger-invalidate' in resp.headers:
                            invalidate = resp.headers['x-trigger-invalidate'].split(',')
                            self._invalidate_cache(invalidate)

            # XXX: response is definitely json at this point?
            # guaranteed if we're only making calls to the api
            return resp

        finally:
            with read_count_lock:
                read_count -= 1
                if read_count == 0:
                    read_lock.release()

    def make_api_url(self, url):
        if url.startswith('/'):
            prefix = 'v1'
        else:
            prefix = 'v1/'
        return urljoin(self.server, prefix + url)

    def _api_response_to_dict(self, resp):
        result = json.loads(resp.content)
        if 'x-trigger-cache' in resp.headers and self.enable_caching and 'x-trigger-invalidate-on' in resp.headers:
            result['_invalidate_on'] = resp.headers['x-trigger-invalidate-on'].split(',')

        return result

    def _api_post(self, url, *args, **kw):
        '''Make a POST request.

        :param url: see :module:`requests`
        :param *args: see :module:`requests`
        '''
        kw['__method'] = 'POST'

        # Force tracking requests to be async
        if url == 'track/':
            kw['async'] = True

        absolute_url = self.make_api_url(url)
        resp = self.__get_or_post(absolute_url, *args, **kw)

        if 'async' in kw and kw['async']:
            return
        else:
            _check_api_response_for_error(url, 'POST', resp)

            return self._api_response_to_dict(resp)

    def _api_get(self, url, *args, **kw):
        '''Make a GET request.

        :param url: see :module:`requests`
        :param *args: see :module:`requests`
        '''
        kw['__method'] = 'GET'

        absolute_url = self.make_api_url(url)
        resp = self.__get_or_post(absolute_url, *args, **kw)

        if 'async' in kw and kw['async']:
            return
        else:
            _check_api_response_for_error(url, 'GET', resp)

            return self._api_response_to_dict(resp)

    def _get(self, url, *args, **kw):
        '''Make a GET request.

        :param url: see :module:`requests`
        :param *args: see :module:`requests`
        '''
        kw['__method'] = 'GET'
        resp = self.__get_or_post(url, *args, **kw)
        _check_response_for_error(url, 'GET', resp)
        return resp

    def _post(self, url, *args, **kw):
        '''Make a GET request.

        :param url: see :module:`requests`
        :param *args: see :module:`requests`
        '''
        kw['__method'] = 'POST'
        resp = self.__get_or_post(url, *args, **kw)
        _check_response_for_error(url, 'POST', resp)
        return resp

    # Required for old platform version
    def _authenticate(self):
        '''Authentication handshake with server
        '''
        # Don't check authentication if we're not using the cookie file
        if self.use_cookie_file:
            resp = self._api_get('auth/loggedin')
            if resp.get('loggedin'):
                LOG.debug('already authenticated via cookie - continuing')
                self.validated_email = resp.get('validated_email')
                self.thirdparty_module_warning = resp.get('thirdparty_module_warning')
                return

            email = forge.request_username()
            password = forge.request_password()

            self.login(email, password)

    def upload_plugin_version(self, unused_uuid, version, description, manifest, targets, zip_file, cookies):
        manifest_dict = json.loads(manifest)
        self._api_post('plugin/%s/version/' % manifest_dict['uuid'], data={
                'version': version,
                'description': description,
                'manifest': manifest,
                'targets': ",".join(targets)
        }, files={
                'built_file': zip_file
        }, cookies=cookies)

    def create(self, name, project, cookies=None):
        data = {
                'name': name,
                'project': project
        }
        LOG.info('Registering new app "{name}" with {hostname}...'.format(
                name=name,
                hostname=self.hostname,
                project=project
        ))
        return self._api_post('app', data=data, cookies=cookies)['data']['uuid']

    # TODO: currently this method seems to corrupt zip files downloaded,
    # not sure why.
    def _get_file_with_progress_bar(self, response, write_to_path, progress_bar_title):
        if progress_bar_title is None:
            progress_bar_title = 'Download'
        content_length = response.headers.get('content-length')
        message = 'Fetching ({content_length}) into {out_file}'.format(
                content_length=content_length,
                out_file=write_to_path
        )
        LOG.debug(message)

        with lib.ProgressBar(progress_bar_title) as bar:

            bytes_written = 0
            with open(write_to_path, 'wb') as write_to_file:
                # TODO: upgrade requests, use Response.iter_content
                for chunk in response.iter_content(chunk_size=102400):
                    if content_length:
                        content_length = int(content_length)
                        write_to_file.write(chunk)
                        bytes_written += len(chunk)
                        fraction_complete = float(bytes_written) / content_length
                        bar.progress(fraction_complete)

                    write_to_file.write(chunk)

    def _get_file(self, url, write_to_path, progress_bar_title=None, cookies=None):
        if cookies:
            response = self._get(url, cookies=cookies)
        else:
            response = self._get(url)
        try:
            content_length = float(response.headers.get('Content-length'))
        except Exception:
            content_length = None

        # TODO: fix usage of iter_content for fetching files with progress bar.
        if False and content_length:
            self._get_file_with_progress_bar(response, write_to_path, progress_bar_title)
        else:
            with open(write_to_path, 'wb') as write_to_file:
                write_to_file.write(response.content)

    def fetch_initial(self, uuid, app_path=".", cookies=None):
        '''Retrieves the initial project template

        :param uuid: project uuid
        '''
        LOG.info('Fetching initial project template')

        initial_zip_filename = path.join(app_path, 'initial.zip')

        self._get_file(
                self.make_api_url('app/{uuid}/initial_files'.format(uuid=uuid)),
                write_to_path=initial_zip_filename,
                progress_bar_title='Fetching initial files',
                cookies=cookies
        )
        lib.unzip_with_permissions(initial_zip_filename, app_path)
        LOG.debug('Extracted initial project template')

        os.remove(initial_zip_filename)
        LOG.debug('Removed downloaded file "%s"' % initial_zip_filename)

    def _handle_unpackaged(self, platform, filename):
        '''De-compress a built output tree.

        :param platform: e.g. "chrome", "ios" - we expect the contents of the ZIP file to
                contain a directory named after the relevant platform
        :param filename: the ZIP file to extract
        '''
        shutil.rmtree(platform, ignore_errors=True)
        LOG.debug('removed "%s" directory' % platform)

        lib.unzip_with_permissions(filename)
        LOG.debug('Extracted unpackaged build for %s' % platform)

        os.remove(filename)
        LOG.debug('removed downloaded file "%s"' % filename)

    def fetch_unpackaged(self, build, to_dir, target):
        '''Retrieves the unpackaged artefacts for a particular build.

        :param build: the build to fetch
        :param to_dir: directory that will hold all the unpackged build trees
        '''
        LOG.info('Fetching Forge templates for %s into "%s"' % (build["id"], to_dir))

        filenames = []
        if not path.isdir(to_dir):
            LOG.debug('Creating output directory "%s"' % to_dir)
            os.mkdir(to_dir)

        with lib.cd(to_dir):
            location = build['file_output']

            filename = urlsplit(location).path.split('/')[-1]

            LOG.debug('writing %s to %s' % (location, path.abspath(filename)))
            self._get_file(
                    location,
                    write_to_path=filename,
                    progress_bar_title=target,
            )

            self._handle_unpackaged(target, filename)

            filenames.append(path.abspath(target))

        LOG.debug('Fetched build into "%s"' % '", "'.join(filenames))
        return filenames

    def fetch_generate_instructions(self, to_dir):
        '''Retreive the generation instructions for our current environment.

        Rather than hard-coding these instructions - how to inject customer
        code into the apps - they are loaded dynamically from the server to
        allow for different platforms versions to work with a larger number
        of build-tools versions.

        :param to_dir: where the instructions will be put
        '''

        platform_version = build_config.load_app()['platform_version']
        temp_instructions_file = 'instructions.zip'

        LOG.info("Fetching generation instructions for {platform_version} "
                        "into \"{to_dir}\"".format(**locals()))

        try:
            # ensure generate_dynamic dir is there before extracting instructions into it
            if not path.isdir(to_dir):
                os.makedirs(to_dir)

            with lib.cd(to_dir):
                self._get_file(
                        self.make_api_url('platform/{platform_version}/core_instructions'.format(platform_version=platform_version)),
                        temp_instructions_file
                )
                lib.unzip_with_permissions(temp_instructions_file)

        finally:
            if path.isfile(path.join(to_dir, temp_instructions_file)):
                os.remove(path.join(to_dir, temp_instructions_file))

        return to_dir

    def _request_development_build(self, config, target, id=None):
        data = {}
        app_config = config

        data['config'] = json.dumps(app_config)
        data['target'] = target
        if not id is None:
            data['id'] = id

        url = 'app/{uuid}/build'.format(uuid=app_config['uuid'])
        return self._api_post(url, data=data)

    def build(self, config, target):
        '''Start a build on the remote server.

        **NB:** this method blocks until the remote build has completed!

        :return: the primary key of the build
        :raises Exception: if any errors occur during the build
        '''
        LOG.info('Starting new build')

        src_dir = defaults.SRC_DIR

        if not path.isdir(src_dir):
            raise ForgeError("No {0} directory found: are you currently in the right directory?".format(src_dir))

        LOG.info('This could take a while, but will only happen again if you modify config.json')
        build = {
                "state": "pending"
        }
        while build['state'] in ('pending', 'working'):
            build = self._request_development_build(config, target, id=build.get('id', None))

            messages = build.get('messages', None)
            if messages:
                LOG.warning(messages)

            if build["state"] == 'complete':
                return build

            if not build['state'] in ('pending', 'working'):
                raise ForgeError('build failed: %s' % build['log_output'])

            LOG.debug('build {id} is {state}...'.format(id=build['id'], state=build['state']))
            time.sleep(self.POLL_DELAY)

    def server_says_should_rebuild(self, path_to_app=".", cookies=None):
        app_config = build_config.load_app(path_to_app)
        url = 'app/{uuid}/should_fetch_core_instructions'.format(uuid=app_config['uuid'])
        resp = self._api_get(url,
                        params = dict(
                                platform_version=app_config['platform_version'],
                                platform_changeset=lib.platform_changeset(path_to_app)
                        ),
                        cookies = cookies
        )
        return resp

    def login(self, email, password):
        LOG.info('authenticating as "%s"' % email)
        credentials = {
                'email': email,
                'password': password
        }

        self._api_get('auth/hello')

        self._api_post('auth/verify', data=credentials)
        LOG.info('authentication successful')

    def logout(self, cookies=None):
        if cookies:
            result = self._api_post('auth/logout', cookies=cookies)
        else:
            result = self._api_post('auth/logout')
        if self.use_cookie_file:
            self.cookies.clear_session_cookies()
            try:
                os.remove(self._path_to_cookies())
            except OSError:
                LOG.debug('Error deleting cookie file', exc_info=True)
            except IOError:
                LOG.debug('Error deleting cookie file', exc_info=True)

        return result

    def create_buildevent(self, app_config):
        return self._api_post('reload/buildevent/create', data={
                'config': json.dumps(app_config)
        })
