"""Generic logic for updating a tools distribution from the trigger website."""
import logging
from urlparse import urljoin
import zipfile
import tarfile
import errno
import sys
import os
import re

import forge
from forge import defaults

LOG = logging.getLogger(__name__)

def unzip_with_permissions(filename, out_path=".", path_filter=None):
    LOG.debug('Unzipping {zip} to {dest}'.format(zip=filename, dest=out_path))

    if filename.endswith(".tar.gz"):
        tar_to_extract = tarfile.open(filename, 'r:gz')
        tar_to_extract.extractall(out_path)
    elif filename.endswith(".zip"):
        zip_to_extract = zipfile.ZipFile(filename)

        for info in sorted(zip_to_extract.infolist()):
            if path_filter:
                info.filename = path_filter(info.filename)

            abs_out_path = os.path.join(out_path, info.filename)
            if info.filename.endswith('/'):
                try:
                    os.makedirs(abs_out_path)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise

                    if os.path.isfile(os.path.normpath(abs_out_path)):
                        raise Exception(
                                "Couldn't extract folder '%s' from %s, as a file of the same name already exists" % (info.filename, filename)
                        )
            else:
                zip_to_extract.extract(info, path=out_path)
                if not sys.platform.startswith("win"):
                    # http://unix.stackexchange.com/questions/14705/the-zip-formats-external-file-attribute
                    unix_permission = (info.external_attr >> 16) & 0777
                    os.chmod(abs_out_path, unix_permission)

        zip_to_extract.close()
    else:
        raise Exception("Unknown archive file type for file '%s'" % filename)

class UpdateRequired(Exception):
    pass

class Updater(object):
    def __init__(self, zip_url, extract_to, download_to=None, path_filter=None):
        """Class for pulling down and extracting an update zip.

        :type zip_url: str
        :param zip_url: URL to request the zip from, relative to the base URL for the trigger API.

        :type extract_to: str
        :param extract_to: Path to extract the update zip over.

        :type download_to: str
        :param download_to: Folder to download the update zip into.

        :type path_filter: str -> str
        :param path_filter: Transformation to apply to the path each file is extracted to.
        """
        self._zip_url = zip_url
        self._extract_to = extract_to
        if download_to is None:
            self._download_to = extract_to
        else:
            self._download_to = download_to
        self._update_zip_path = os.path.join(self._download_to, zip_url + '.zip')
        self._path_filter = path_filter

    def get_update_zip(self, remote):
        """
        :type remote: forge.remote.Remote
        :param remote: Interface for making the remote requests to the trigger API.
        """
        update_zip_path = self._update_zip_path
        remote._get_file(
                remote.make_api_url(self._zip_url),
                update_zip_path
        )

    def extract_update_zip_if_present(self):
        """Apply update if necessary and return whether an update was applied.

        :return bool: True if an update was applied
        """
        if os.path.exists(self._update_zip_path):
            self.extract_update_zip()
            return True
        return False

    def extract_update_zip(self):
        if os.path.exists(os.path.join(self._download_to, 'no_update')):
            raise Exception("Tools tried to update themselves during development")
        try:
            unzip_with_permissions(self._update_zip_path, self._extract_to, path_filter=self._path_filter)
        finally:
            try:
                if os.path.isfile(self._update_zip_path):
                    os.remove(self._update_zip_path)
            except Exception as e:
                LOG.debug('Error occured when cleaning up after trying to update: %s' % e)

    def force_check_version(self, remote):
        result = remote._api_get(
                'version_check/{version}/'.format(version=forge.get_version().replace('.', '/'))
        )

        if result['result'] == 'ok':
            if 'upgrade' in result:
                LOG.info('Update result: %s' % result['message'])
            else:
                LOG.debug('Update result: %s' % result['message'])

            if result.get('upgrade') == 'required':
                raise UpdateRequired
        else:
            LOG.info('Upgrade check failed.')

    def check_version(self, remote):
        # Don't worry if we're not using a cookie file, the toolkit will have already checked
        if remote.use_cookie_file:
            return self.force_check_version(remote)

# specific updater implementations
def toolkit_updater():
    """Construct an Updater which can update the toolkit distribution."""
    import workspace
    return Updater(
            zip_url='latest_toolkit',
            extract_to=workspace.path()
    )

def buildtools_updater():
    """Construct an Updater which can update the build tools distribution."""
    above_forge_root = os.path.normpath(os.path.join(defaults.FORGE_ROOT, '..'))
    build_tools_folder_name = os.path.basename(os.path.normpath(defaults.FORGE_ROOT))

    return Updater(
            zip_url='latest_tools',
            extract_to=above_forge_root,
            download_to=defaults.FORGE_ROOT,
            path_filter=lambda p: re.sub(r'^forge-tools', build_tools_folder_name, p)
    )

def distribution_aware_updater():
    """Check if we're running in the toolkit distribution or build tools distribution
    and return the appropriate updater."""
    try:
        import trigger
    except ImportError:
        return buildtools_updater()
    else:
        return toolkit_updater()

def restart():
    """Re-execute the currently running executable with the same arguments.

    On UNIX this replaces the current executing process with a new one.

    On Windows a new process is spawned and the original one immediately dies. This means
    that if the currently running executable was spawned from the terminal, the terminal
    resumes (but with mixed input/output with the new process!).
    """
    args = sys.argv[:]
    # execv requires args to have the executable to run as the first parameter
    # when running in a binary, args will currently look like: ['TriggerToolkit.exe']
    # when running using the system python interpreter args will look like ['run_trigger_toolkit.py']
    if not hasattr(sys, 'frozen'):
        # insert python interpreter as the first arg to make os.execv happy
        args.insert(0, sys.executable)

    if sys.platform == 'win32':
        # spaces in args passed to execv on windows need escaping since it backs onto
        # the CreateProcess API
        # http://msdn.microsoft.com/en-us/library/431x4c1w(v=vs.80).aspx
        args = ['"%s"' % arg for arg in args]
    LOG.debug("Restarting executable: %s, %s" % (sys.executable, args))
    os.execv(sys.executable, args)

def update_and_restart_if_necessary():
    if distribution_aware_updater().extract_update_zip_if_present():
        LOG.info("Update applied, restarting.")
        restart()
    LOG.info("No update zip to extract, carrying on.")
