#!/usr/bin/env python
"""BE WARY OF UPDATING THIS FILE:

This file is built into TriggerToolkit.exe and the OSX TriggerToolkit binary.
Currently we aren't able to update those two files, which means that any logic put
in here will not make it's way to users through an update.

This file calls trigger.web.main, you can put logic in or after there instead.

The same applies to anything imported in this file!
"""
import logging
import os
from os import path
import shutil
import sys
import multiprocessing
import exceptions
import codecs

import include_everything
import workspace

logging.basicConfig(level='DEBUG', format='[%(levelname)7s] %(message)s')
LOG = logging.getLogger(__name__)


def _read_file(p):
    with codecs.open(p, encoding='utf8') as f:
        return f.read()


def _install_to_library(user_library_dir):
    _copy_ontop_of(workspace.path(), user_library_dir)


def _windows_safe_copystat(src, dst):
    WindowsError = getattr(exceptions, 'WindowsError', None)

    if WindowsError is not None:
        try:
            shutil.copystat(src, dst)
        except WindowsError:
            # can't copy file access times on windows
            pass
    else:
        shutil.copystat(src, dst)


def _copy_ontop_of(src, dst):
    # modified shutil.copytree to only create folders if they aren't already there
    if not path.isdir(dst):
        os.makedirs(dst)

    errors = []
    for name in os.listdir(src):
        srcname = path.join(src, name)
        dstname = path.join(dst, name)

        try:
            if os.path.isdir(srcname):
                _copy_ontop_of(srcname, dstname)
            else:
                shutil.copy2(srcname, dstname)
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, str(why)))
        except shutil.Error as err:
            errors.extend(err.args[0])

    try:
        _windows_safe_copystat(src, dst)
    except OSError as why:
        errors.append((src, dst, src(why)))

    if errors:
        raise shutil.Error(errors)


def _version_tuple(version_string):
    return tuple(map(int, version_string.strip().split('.')))


def main():
    try:
        if 'frozen' in dir(sys) and sys.platform.startswith("darwin"):
            user_library_dir = path.expanduser('~/Library/Trigger Toolkit')
            app_version_path = workspace.path('trigger/version.txt')
            installed_version_path = path.join(user_library_dir, 'trigger/version.txt')

            if not path.exists(user_library_dir):
                LOG.debug('First time run, copying toolkit code into %s' % user_library_dir)
                _install_to_library(user_library_dir)

            if not path.exists(installed_version_path):
                LOG.debug('No version file, copying toolkit code into %s' % user_library_dir)
                _install_to_library(user_library_dir)

            else:
                installed_version = _version_tuple(_read_file(installed_version_path))
                app_version = _version_tuple(_read_file(app_version_path))

                if app_version > installed_version:
                    LOG.debug('App version newer than installed version, copying toolkit code into %s' % user_library_dir)
                    _install_to_library(user_library_dir)

                elif app_version == installed_version:
                    LOG.debug("Installed version is same as App version, using code from %s" % user_library_dir)

                else:
                    LOG.debug("Installed version is newer than App version, using code from %s" % user_library_dir)

            LOG.debug('Changing workspace to %s' % user_library_dir)
            workspace.WORKSPACE = user_library_dir

        # Import system openssl on linux before adding our library paths.
        if sys.platform.startswith("linux"):
            try:
                import OpenSSL
            except ImportError:
                raise Exception("PyOpenSSL is required to use Trigger Toolkit.")

        root = workspace.path()
        sys.path.insert(0, root)
        sys.path.insert(0, os.path.join(root, 'trigger-dependencies'))
        sys.path.insert(0, os.path.join(root, 'build-tools'))
        sys.path.insert(0, os.path.join(root, 'build-tools', 'forge-dependencies'))

        multiprocessing.freeze_support()
        # import using a string here so pyinstaller doesn't try to include
        # trigger code inside the final binary
        trigger = __import__('trigger.web')
        launch_browser = '--no-launch' not in sys.argv
        trigger.web.main(launch_browser)
    except Exception as e:
        import traceback
        # TODO: show user something useful when not running in a console setting
        print "Unhandled exception in Trigger Toolkit: %s" % e
        traceback.print_exc()
        print "Please contact support@trigger.io"
        raw_input("Press enter to continue")


if __name__ == "__main__":
    main()
