/*global Toolkit*/
/*global Backbone*/
/*global _, $*/
/*global Mustache*/
/*global _superApply*/
/*global moment*/
/*global NO_TIMEZONE_FORMAT*/
/*global RpcCall*/
(function () {
	// represents an app on the file system
	Toolkit.Models.LocalApp = Toolkit.Models.WatchingModel.extend({
		idAttribute: 'path',
		cacheName: 'LocalApp',
		watchDirPath: function () {
			return this.get('path');
		},

		initialize: function () {
			var pathToApp = this.get('path');
			this.set({
				files: {
					localConfig: pathToApp + '/local_config.json',
					config: pathToApp + '/src/config.json',
					identity: pathToApp + '/src/identity.json',
					configSchema: pathToApp + "/.template/lib/schema.json"
				},
				schemas: {
					localConfig: pathToApp + "/.template/lib/local_config_schema.json",
					config: pathToApp + "/.template/lib/schema.json",
					plugins: pathToApp + "/.template/lib/plugins_config_schema.json"
				}
			});

			var me = this;
			this.on('change:identity', function () {
				if (me.get("identity") && me.get("identity").uuid) {
					new Toolkit.Models.App({uuid: this.get("identity").uuid}).set("local", me);
				}
			});

			this.listenTo(this, 'change:config', this.updatePlatform);
			Toolkit.Models.WatchingModel.prototype.initialize.apply(this, arguments);
		},

		updatePlatform: function () {
			var config = this.get('config');
			var platformVersion = config && config['platform_version'];
			var platforms = Toolkit.platforms.get('platforms');
			var platform = platforms && platforms.get(platformVersion);

			if (platform) {
				platform.getLocalConfigSchema();
			}

			this.set("platform", platform);
		},

		getPlatformVersion: function () {
			return this.get("config").platform_version;
		},
		getConfig: function () {
			return this.get("config");
		},
		getConfigPath: function () {
			return this.get("files").config;
		},
		saveConfig: function (config) {
			if (config === undefined) {
				config = this.getConfig();
			}
			var rpc = Toolkit.rpcSession.remoteCall('file.write', {
				path: this.getConfigPath(),
				data: JSON.stringify(config, undefined, "\t")
			});

			rpc.on('success', function () {
				$(".saved").fadeIn();
				setTimeout(function () {
					$(".saved").fadeOut();
				}, 3000);
			});
		},
		getProject: function () {
			return new Toolkit.Models.App({uuid: this.get("identity").uuid}).get("project");
		},

		clean: function () {
			var me = this;
			var rpc = Toolkit.rpcSession.remoteCall('app.clean', {
				path: me.get('path')
			});

			rpc.on('success', function () {
				Toolkit.showNotification('Successfully removed templates for ' + me.get('config').name);
				me.buildIfNecessary();
			});
		},

		_buildCbs: [],

		_bubbleChange: function () {
			this.trigger('change');
		},

		_build: function () {
			// remove old bubbling
			var oldBuildTask = this.get('buildtask');
			if (oldBuildTask) {
				oldBuildTask.off(null, this._bubbleChange);
			}

			// create rpc
			var buildRpc = Toolkit.rpcSession.remoteCall('app.build', {
				cookies: Toolkit.cookies,
				path: this.get('path'),
				target: null
			});

			var me = this;

			buildRpc.on("success", function () {
				_.each(me._buildCbs, function (cb) {
					cb && cb();
				});
			});

			// wrap rpc in Model
			this.setTemp('buildtask', new Toolkit.Models.ConsoleTask({
				rpc: buildRpc
			}, {
				preventNavigation: false
			}));

			// bubble change events from ConsoleTask to LocalApp
			this.get('buildtask').on('change', this._bubbleChange, this);
		},

		// ask local python/trigger.io if we should fetch
		// down new generate_dynamic, and start doing so if so.
		buildIfNecessary: function (cb) {
			var me = this;

			// do nothing if build currently in progress
			var buildTask = me.get('buildtask');
			if (buildTask && buildTask.get('status') === 'running') {
				if (cb) {
					this._buildCbs.push(cb);
				}
				return;
			}

			// do nothing if should rebuild check in progress
			var rpc = me.get('staterpc');
			if (rpc) {
				if (cb) {
					this._buildCbs.push(cb);
				}
				return;
			}

			this._buildCbs = [];

			if (cb) {
				this._buildCbs.push(cb);
			}

			rpc = Toolkit.rpcSession.remoteCall('app.state', {
				cookies: Toolkit.cookies,
				path: this.get('path')
			});

			rpc.on('success', function (e) {
				if (e.data['should_rebuild']) {
					me._build();
				} else {
					_.each(me._buildCbs, function (cb) {
						cb && cb();
					});
				}
				delete e.data['should_rebuild'];
				me.set('staterpc', undefined);
				me.set(e.data);
			});

			// allow observers to see that we're checking the state
			me.set('staterpc', rpc);
		},

		pushReload: function (stream, externalUrl) {
			var streamName = stream.get('name');
			var oldPushTask = this.get('pushtask');
			if (oldPushTask) {
				oldPushTask.off(null, this._bubbleChange);
			}

			// create rpc
			var pushRpc;
			if (externalUrl) {
				pushRpc = Toolkit.rpcSession.remoteCall('reload.pushcdn', {
					cookies: Toolkit.cookies,
					path: this.get('path'),
					stream_name: streamName,
					manifest_url: externalUrl
				});
			} else {
				pushRpc = Toolkit.rpcSession.remoteCall('reload.push', {
					cookies: Toolkit.cookies,
					path: this.get('path'),
					stream_name: streamName
				});
			}

			// wrap rpc in Model
			this.set('pushtask', new Toolkit.Models.ConsoleTask({
				rpc: pushRpc,
				messageWhenPreventingNavigation: 'Currently pushing to ' + this.get('config').name + ', navigating away will interrupt this.'
			}));

			pushRpc.on('success', function () {
				stream.set('last_pushed', moment().format(NO_TIMEZONE_FORMAT));
			});

			// bubble change events from ConsoleTask to LocalApp
			this.get('pushtask').on('change', this._bubbleChange, this);
		}
	});

	_.extend(Toolkit.Models.LocalApp.prototype, Toolkit.Mixins.Dependencies);

	// collection of local apps on the file system
	Toolkit.Collections.LocalAppCollection = Backbone.Collection.extend({
		model: Toolkit.Models.LocalApp
	});

	// model which watches local_apps.json and maintains a collection of
	// all LocalApps on the file system
	Toolkit.Models.LocalAppsFile = Toolkit.Models.WatchingModel.extend({

		// return LocalApp for App
		getLocal: function (app) {
			var uuid = app.get('uuid');
			var available = this.availableUuids();
			return available[uuid];
		},

		removeLocal: function (app) {
			this.get('collection').remove(this.getLocal(app));
		},

		availableUuids: function () {
			var available = {};
			this.get('collection').map(function (localApp) {
				var identity = localApp.get('identity');
				// might be in the middle of fetching the identity file,
				// can't say what it's uuid is yet in that case
				if (identity) {
					available[identity.uuid] = localApp;
				}
			});
			return available;
		},

		initialize: function () {
			Toolkit.Models.WatchingModel.prototype.initialize.apply(this, arguments);
			this.set({
				collection: new Toolkit.Collections.LocalAppCollection(),
				files: {
					localApps: 'build-tools/local_apps.json'
				}
			});
			this.listenTo(this, 'change:localApps', function () {
				// read local_apps.json into a collection of LocalApps
				// TODO: change format of local_apps.json to not include UUIDs
				this.get('collection').update(
					_(this.get('localApps')).values()
				);
			});
		}
	});

	Toolkit.events.once('rpc', function () {
		Toolkit.registerPreReadyFunction(function (done) {
			Toolkit.rpcSession.remoteCall('app.remove_local_apps_not_on_filesystem').on('success', function () {
				Toolkit.localAppsFile = new Toolkit.Models.LocalAppsFile();
				Toolkit.localAppsFile.once('change:localApps', done);
			});
		});
	});

	Toolkit.Collections.ModuleVersionCollection = Backbone.Collection.extend({
		model: Toolkit.Models.ModuleVersion,
		comparator: function (model) {
			return model.id;
		}
	});

	// represents an app from the server
	Toolkit.Models.App = Toolkit.Models.RpcModel.extend({
		idAttribute: 'uuid',

		cacheName: 'App',

		method: 'api.get',

		args: function () {
			return {
				cookies: Toolkit.cookies,
				method: 'app/'+this.get('uuid')
			};
		},

		initialize: function () {
			Toolkit.Models.App.__super__.initialize.call(this);
			this.set("status", {});
			this.set("config_modules", new Toolkit.Collections.ModuleVersionCollection());
			var me = this;

			me.doWhenAvailable(me, 'local', function () {
				var local = me.get("local");
				var checkPlatformVersion = function () {
					if (local.get("problem")) {
						var problem = local.get("problem");
						switch (problem.type) {
						case 'platform_version_missing':
							me.setStatus('forge', 'local_file', 'error', "There was no platform_version property in <code>src/config.json</code>.");
							break;

						case 'platform_status_timeout':
							me.setStatus('forge', 'local_file', 'error', "Couldn't check status for Trigger.io platform version, timed out after 10 seconds.");
							break;

						case 'config_missing':
							// No need to report this as the javascript side will pick it up
							me.clearStatus('forge', 'local_file');
							break;

						case 'remote_error':
							me.setStatus('forge', 'local_file', 'error', "Problem requesting information about this App from "+problem.data.server+": "+problem.data.message);
							break;
						}

					} else {
						me.clearStatus('forge', 'local_file');
					}

					if (!local.get('config')) {
						me.setStatus('forge', 'platform_version', 'error', 'No config.json could be read for this app, make sure config.json exists and is valid json.');
					} else {
						var platform_version = local.get('config').platform_version;
						if (!platform_version) {
							me.setStatus('forge', 'platform_version', 'error', 'No Trigger.io platform version could be found in config.json, this is required to use the Toolkit.');
						} else {
							if (Toolkit.platforms.get("latest_stable") === platform_version) {
								me.setStatus('forge', 'platform_version', 'good', 'You are currently using the latest stable Trigger.io version: '+platform_version+'.');
							} else {
								if (!Toolkit.platforms.get("platforms") || !Toolkit.platforms.get("platforms").get(platform_version)) {
									me.setStatus('forge', 'platform_version', 'error', 'You are currently on an unknown Trigger.io platform version ('+platform_version+'): you will need to switch to a known version to use the Toolkit.');
								} else if (platform_version === "v1.4") {
									me.setStatus('forge', 'platform_version', 'error', '<strong>Trigger.io version v1.4 will not receive any more major updates!</strong><p>In the <a href="#apps/'+me.id+'/config">App Config</a> view, there\'s an easy, automatic migration to v2.0.0.</p><p>If you need to use v1.4 to <a href="'+Toolkit.websiteUrl()+'reload/">push updates</a> to existing apps, see the <a href="#apps/'+me.id+'/reload">Reload</a> tab.</p>');
								} else if (platform_version.substr(0,4) === "v1.3") {
									me.setStatus('forge', 'platform_version', 'error', '<strong>Trigger.io version v1.3 is no longer supported!</strong><p>Please update to a newer Trigger.io platform version.</p>');
								} else {
									me.setStatus('forge', 'platform_version', 'warning', 'Your current Trigger.io platform version is not the latest stable. Unless you have a particular reason to use an old version, it is recommended you go to the <a href="#apps/' + me.id + '/config">App config</a> page and update.');
								}
							}
						}
					}
				};
				me.listenTo(local, 'change', checkPlatformVersion);
				me.listenTo(Toolkit.platforms, 'change', checkPlatformVersion);
				checkPlatformVersion();

				me.listenTo(local, 'change:config', function () {
					me.updateConfigModules();
					me.checkModuleStatus();
					me.checkConfigStatus();
				});
				me.listenTo(local, 'change:localConfig change:platform', function () {
					me.checkToolConfigStatus();
				});
				me.listenTo(local, 'change:configSchema', function () {
					me.checkConfigStatus();
				});
				me.doWhenAvailable(me, 'project_id', function () {
					me.updateConfigModules();
					me.checkModuleStatus();
				});
				me.updateConfigModules();
				me.checkModuleStatus();
				me.checkConfigStatus();
				me.checkToolConfigStatus();

				me.listenToWhenAvailable(me, 'project', 'change:limits', me.checkLimitStatus);
				me.checkLimitStatus();

			});
		},

		parent: {
			attribute: 'project',
			idAttribute: 'project_id',
			model: 'Project',
			collectionAttribute: 'apps'
		},

		children: {
			buildevents: Toolkit.Models.BuildEvent
		},

		inject: {
			buildevents: function (buildevents) {
				return new Toolkit.Collections.BuildEventCollection(buildevents);
			}
		},

		// method to update the buildevents attribute of this app
		loadBuildEvents: function (force) {
			if (this.get('_loadingBuildevents')) {
				return;
			}

			if (this.get('_loadedBuildevents') && !force) {
				return;
			}

			this.set('_loadedBuildevents', false);
			this.set('_loadingBuildevents', true);
			var local = this.get("local");
			var config = local.get('config');
			var configWithIdentity = _.extend({}, config, local.get('identity'));
			var rpc = Toolkit.rpcSession.remoteCall('api.post', {
				cookies: Toolkit.cookies,
				method: 'reload/buildevent/create_and_list',
				data: {
					config: JSON.stringify(configWithIdentity)
				}
			});

			var me = this;
			rpc.on('success', function (e) {
				me.loadSet('buildevents', e.data['data']['configs']);
				var localBuildevent = me.setActiveConfig(config, e.data['data']['config_hash']);
				localBuildevent.load();
				me.set('_loadedBuildevents', true);
			});

			rpc.on('finish', function () {
				me.set('_loadingBuildevents', false);
			});
		},

		// update models to reflect the "active local config" for this app:
		setActiveConfig: function (config, configHash) {
			var local = this.get("local");
			local.set('config', config);
			local.set('config_hash', configHash);

			// update collection to mark the correct buildevent as active
			var activeBuildEvent = this.get('buildevents').where({active: true})[0];
			if (activeBuildEvent) {
				activeBuildEvent.set('active', undefined);
			}
			var localBuildevent = this.get('buildevents').get(local.get('config_hash'));
			return localBuildevent.set('active', true);
		},

		checkModuleStatus: function () {
			var app = this;
			var local = app.get("local");
			if (!local || !local.get("config") || local.get("config")["platform_version"].substr(0,2) == "v1") {
				// Ignore for pre v2
				return;
			}
			var moduleVersions = app.get("config_modules");
			this.doWhenAll(moduleVersions.models, '_loaded', function () {
				app.clearAllStatusesWithType('module');
				moduleVersions.each(function (moduleVersion) {
					if (local.get("config").modules[moduleVersion.get("module__name")].disabled) {
						return;
					}

					if (moduleVersion.get("module__name").substr(0,6) === 'debug_') {
						app.setStatus({
							'type': 'module',
							'id': moduleVersion.get("module__name")+"-debug",
							'severity': "warning",
							'message': "<strong>" + moduleVersion.get("module__name") + "</strong> module Is a debug module, make sure you disable it before creating a production build.</a>",
							'tags': ['module-'+moduleVersion.get("module__name")]
						});
					}

					if (Toolkit.premium.get("modules") && _.findWhere(Toolkit.premium.get("modules").toJSON(), {"name": moduleVersion.get("module__name")}) !== undefined) {
						var features = app.get("project").get("features");
						if (!features) {
							return;
						}
						
						var available = features['modules'];

						if (_.indexOf(available, moduleVersion.get("module__name")) == -1) {
							app.setStatus({
								'type': 'module',
								'id': moduleVersion.get("module__name")+"-version",
								'severity': "error",
								'message': "<strong>" + moduleVersion.get("module__name") + "</strong> is a premium module. Please check out <a href=\"https://trigger.io/pricing\" target=\"_blank\">our paid plans</a> for more info!",
								'tags': ['module-'+moduleVersion.get("module__name")]
							});
							return;
						}
					}

					if (moduleVersion.get("type") == "public" && moduleVersion.get("version") != moduleVersion.get("current_version")) {
						app.setStatus({
							'type': 'module',
							'id': moduleVersion.get("module__name")+"-version",
							'severity': "warning",
							'message': "<strong>" + moduleVersion.get("module__name") + "</strong> module is not set to the current version, current version is <strong>"+moduleVersion.get("current_version")+"</strong>, your version is <strong>"+moduleVersion.get("version")+"</strong>. <a href='#apps/"+app.id+"/browseModuleDetails/"+moduleVersion.get("module__name")+"/v"+moduleVersion.get("current_version")+"'>View current version details.</a>",
							'tags': ['module-'+moduleVersion.get("module__name")]
						});
					}

					var schema = moduleVersion.get("schema");

					if (_.isEmpty(schema)) {
						// No config for module.
						return;
					}

					var config = local.get("config").modules[moduleVersion.get("module__name")].config;
					var validate = Toolkit.validateConfig(config, schema);

					if (validate.changed) {
						local.getConfig().modules[moduleVersion.get("module__name")].config = validate.config;
						local.saveConfig();
					}

					if (!validate.valid) {
						var message = "<strong>" + moduleVersion.get("module__name") + "</strong> module config issues:<ul><li>";
						message += _.map(validate.problems, function (problem) {
							return "<strong>"+problem.field.join(".")+"</strong> "+problem.message;
						}).join("</li><li>")+"</li></ul>";

						app.setStatus({
							'type': 'module',
							'id': moduleVersion.get("module__name")+"-config",
							'severity': "error",
							'message': message,
							'tags': ['module-'+moduleVersion.get("module__name")]
						});
					}
				});
				if (moduleVersions.length === 0) {
					var modulePath = "#apps/"+app.get('uuid')+"/publicModules";
					app.setStatus(
						'module',
						'',
						'warning',
						'You have not enabled any modules. See the <a href="' +
							modulePath +
							'">Modules</a> tab to get started!'
					);
				}
				app.checkDependencies();
			});
		},

		checkConfigStatus: function () {
			var local = this.get("local");
			var config = local.get("config");
			var schema = local.get("configSchema");
			
			// no-op if we don't have all the files yet, expect to be called later
			if (!schema || !config) {
				return;
			}

			var validate = Toolkit.validateConfig(config, schema);
			if (validate.changed) {
				local.saveConfig();
			}
			if (!validate.valid) {
				// Set status
				var message = "App config validation failed due to the following errors:<ul><li>";
				message += _.map(validate.problems, function (problem) {
					return "<strong>"+problem.field.join(".")+"</strong> "+problem.message;
				}).join("</li><li>")+"</li></ul>";
				this.setStatus(
					'forge',
					'config',
					'error',
					message
				);
			} else {
				this.clearStatus('forge', 'config');
			}
		},

		checkToolConfigStatus: function () {
			var local = this.get("local");
			var config = local.get("localConfig");
			var platform = local.get("platform");
			var schema = platform && platform.get("localConfigSchema");
			
			if (!config || !schema) {
				return;
			}
			var validate = Toolkit.validateConfig(config, schema);

			if (validate.changed) {
				//local.saveConfig();
			}
			if (!validate.valid) {
				// Set status
				var message = "Tool config validation failed due to the following errors:<ul><li>";
				message += _.map(validate.problems, function (problem) {
					return "<strong>"+problem.field.join(".")+"</strong> "+problem.message;
				}).join("</li><li>")+"</li></ul>";
				this.setStatus(
					'forge',
					'tool-config',
					'error',
					message
				);
			} else {
				this.clearStatus('forge', 'tool-config');
			}
		},
		checkDependencies: function () {
			var app = this;
			var local = app.get("local");
			var config = local ? local.get("config") : {};
			var project = app.get("project");
			
			if (!project || !config.modules || config.config_version < 4) {
				return;
			}

			project.loadDependencies(function () {
				local.checkDependenciesAndSetStatus(
					app,
					'module',
					project.get("dependencies"),
					[],
					local.get("config")["platform_version"]
				);
			});
		},

		checkLimitStatus: function () {
			this.clearAllStatusesWithType('limits');
			var project = this.get("project");
			if (project && project.get("limits")) {
				var limits = project.get("limits");
				this.setStatus('limits', 'reload', (limits.reload.current_threshold === 0 ? 'info' : 'warning'),
					'You are currently using '+limits.reload.percentage_used+'% of your reload limit for this project ('+limits.reload.used+'/'+limits.reload.allowed+' reloads)');
			}
		},

		updateConfigModules: function () {
			var local = Toolkit.localAppsFile.getLocal(this);
			var projectId = this.get("project_id");
			var config_modules = this.get("config_modules");
			var config = local.get("config");
			if (!config || !config.modules || !projectId || config.config_version < 4) {
				return;
			}
			var to_remove = config_modules.pluck('id');
			config_modules.add(_.map(config.modules, function (data, module) {
				var id = projectId+"-"+module+"-"+data.version;
				to_remove = _.without(to_remove, id);
				return new Toolkit.Models.ModuleVersion({
					id: id,
					module__name: module,
					version: data.version,
					project_id: projectId
				}).load();
			}));
			config_modules.remove(_.map(to_remove, function (id) {
				return config_modules.get(id);
			}));
		}
	});

	_.extend(Toolkit.Models.App.prototype, Toolkit.Mixins.Status, {
		getStatusForTemplating: function () {
			// provide sensible default descriptions for when there are no
			// statuses for these
			var status = Toolkit.Mixins.Status.getStatusForTemplating.call(this);
			if (!status.forge || !status.forge.length) {
				status.forge = [{
					icon: 'ok-sign',
					class: 'success',
					message: 'No problems found with your app.'
				}];
			}
			if (!status.module || !status.module.length) {
				status.module = [{
					icon: 'ok-sign',
					class: 'success',
					message: 'No problems found with modules.'
				}];
			}

			if (!status.limits || !status.limits.length) {
				status.limits = [{
					icon: 'ok-sign',
					class: 'success',
					message: 'No problems found with limits.'
				}];
			}
			
			return status;
		}
	});

	// collection of apps fetched down from the server side
	Toolkit.Collections.AppCollection = Backbone.Collection.extend({
		model: Toolkit.Models.App,
		withLocalApps: function () {
			return this.filter(function (app) {
				var availableUuids = Toolkit.localAppsFile.availableUuids();
				return app.get('uuid') in availableUuids;
			});
		},
		comparator: function (app) {
			// Maybe sorting by name would be better? But name is harder to get access to here.
			return app.get("uuid");
		}
	});


	Toolkit.Views.AppImportView = Toolkit.Views.FormView.extend({
		className: 'import',
		template: 'projects/import_app',

		templateAttributes: function () {
			var superAttrs = Toolkit.Views.AppImportView.__super__.templateAttributes.call(this);
			return _.extend({
				project: this.model.attributes
			}, superAttrs);
		},

		submit: function (form) {
			var me = this;

			form.cookies = Toolkit.cookies;
			form.project = this.model.id;
			var rpc = Toolkit.rpcSession.remoteCall('app.import_app', form);

			rpc.on('success', function () {
				me.trigger("cancel");
			});

			rpc.on('error', function (e) {
				me._submitFinish();
				me.$(".alert-error").text('Failed to import app: ' + e.message).show();
			});
		}
	});

	var AppCreateView = Toolkit.Views.AppCreateView = Toolkit.Views.FormView.extend({
		className: 'create',
		template: 'projects/create_app',

		// mixin events from FormView
		events: _.extend({}, Toolkit.Views.FormView.prototype.events, {
			"change input#name": "_changeName",
			"keypress input#name": "_changeName",
			"keyup input#name": "_changeName"
		}),

		submit: function (form) {
			// TODO: need to not be able to submit if apps collection isn't there?
			var me = this;
			var rpc = Toolkit.rpcSession.remoteCall('app.create', {
				cookies: Toolkit.cookies,
				name: form.name,
				path: form.path,
				project: this.model.id
			});

			rpc.on('success', function (e) {
				var apps = me.model.get('apps');
				var uuid = e.data['data']['uuid'];
				apps.add({
					uuid: uuid,
					name: form.name
				});
				me.trigger("cancel");
			});

			rpc.on('error', function (e) {
				me._submitFinish();
				this.defaults.error(e);
			});
		},

		_changeName: function () {
			if (this.tempModel.get('userdir')) {
				var name = this.$('#name').val();
				this.tempModel.set('name', name);
				this.tempModel.set('path', this.tempModel.get('userdir')+name.toLowerCase().replace(/ /g, '-').replace(/[^a-z0-9-]/g, ''));
			}
		},

		_renderPath: function () {
			this.$('#path').val(this.tempModel.get('path'));
		},

		initialize: function () {
			_superApply(AppCreateView, this, 'initialize', arguments);

			// create our own model to store stuff not passed to us
			this.tempModel = new Backbone.Model();
			this.tempModel.set('userdir', false);
			this.listenTo(this.tempModel, 'change:path', this._renderPath);
			this.listenTo(this.tempModel, 'change:userdir', this._renderPath);

			// get userdir
			var rpc = Toolkit.rpcSession.remoteCall('directory.get_home_directory');
			var me = this;
			rpc.on('success', function (e) {
				var sep = e.data.separator;
				me.tempModel.set('userdir', e.data.userdir + sep + 'forge-workspace' + sep);
			});
		},

		templateAttributes: function () {
			return {
				tempModel: this.tempModel.attributes,
				model: this.model.attributes
			};
		}
	});

	/**
	 * Lays out sidebar/statusbar and content, using a new this.ContentView for the content.
	 * Observes an App to update sidebar/statusbar/breadcrumbs.
	 * 
	 * @param model {Backbone.Model} should either be a Toolkit.Models.App, or have an 'app' attribute which is.
	 */
	var GenericAppView = Toolkit.Views.GenericAppView = Toolkit.Views.ViewWithTemplate.extend({
		breadcrumbs: function () {
			return this.appBreadcrumbs(this.getApp());
		},

		getApp: function () {
			if (this.model.has("app")) {
				return this.model.get("app");
			} else {
				return this.model;
			}
		},

		initialize: function () {
			var app = this.getApp();

			// ContentView should be overidden by subclasses
			this.content = new RequiresBuiltAppView({
				content: new this.ContentView(this.options),
				model: app
			});
			this.autoClose(this.content);

			// create component views
			this.statusbar = new Toolkit.Views.StatusBarView({model: app, url: 'apps/'+app.id});
			this.autoClose(this.statusbar);

			var sidebarModel = new Backbone.Model(this.sidebarContents());
			this.sidebar = new Toolkit.Views.SideBarView({
				model: sidebarModel
			});
			this.sidebarlayout = new Toolkit.Views.SideBarLayoutView({
				sidebar: this.sidebar,
				content: this.content
			});
			this.autoClose(this.sidebarlayout);

			// keep breadcrumbs up to date
			this.listenTo(app, 'change:name', this.updateBreadcrumbs);
			this.listenToWhenAvailable(app, 'project', 'change:name', this.updateBreadcrumbs);
			this.listenToWhenAvailable(app, 'project', 'change:modules', function () {
				sidebarModel.set(me.sidebarContents());
				me.sidebar.render();
			});

			var me = this;

			sidebarModel.set(me.sidebarContents());

			this.listenToWhenAvailable(app, 'local', 'change:config', function () {
				sidebarModel.set(me.sidebarContents());
				me.sidebar.render();
				me.render();
			});
		},

		sidebarContents: function () {
			var app = this.getApp();
			var local = app.get("local");

			var contents = [
				{
					"item": true,
					"text": "Status"
				},
				{
					"section": true,
					"text": "Actions"
				},
				{
					"item": true,
					"text": "Forge",
					"link": "/forge"
				},
				{
					"item": true,
					"text": "Reload",
					"link": "/reload"
				},
				{
					"section": true,
					"text": "Config"
				},
				{
					"item": true,
					"text": "App",
					"link": "/config"
				},
				{
					"item": true,
					"text": "Tools",
					"link": "/localConfig"
				}
			];

			if (local && local.has('config') && local.get("config").platform_version) {
				var features = Toolkit.platforms.get("platforms").get(local.get("config").platform_version).get("features");
				if (_.contains(features, "plugins")) {
					contents.push({
						"item": true,
						"text": "Plugins",
						"link": "/plugins"
					});
				}
				if (_.contains(features, "modules")) {
					if (app.get('project') &&
							app.get('project').get('modules') &&
							app.get('project').get('modules').length > 0) {
						contents.push({
							"section": true,
							"text": "Modules"
						});
						contents.push({
							"item": true,
							"text": "Public",
							"link": "/publicModules"
						});
						contents.push({
							"item": true,
							"text": "Private",
							"link": "/projectModules"
						});
					} else {
						contents.push({
							"item": true,
							"text": "Modules",
							"link": "/publicModules"
						});
					}
				}
			}

			if (!_.isArray(this.active)) {
				this.active = [this.active];
			}
			_.each(this.active, function (active) {
				var matches = _.where(contents, {text: active});
				if (matches.length > 0) {
					matches[0].active = true;
				}
			});

			return {
				prefix: "#apps/"+app.get('uuid'),
				contents: contents
			};
		},

		render: function () {
			// Don't show the app view until we know what the sidebar will look like
			if (!this.getApp() || !this.getApp().get("local") || !this.getApp().get("local").get("config") || !this.getApp().get("local").get("config").platform_version) {
				this.$el.html('<p class="loading"><i class="icon-cog icon-spin"></i> Loading...</p>');
				return this;
			} else if (this.rendered) {
				return this;
			} else {
				this.$el.html('');
			}
			// complete render, attach subviews to this.el
			this.$el.append(this.statusbar.render().el);
			this.$el.append(this.sidebarlayout.render().el);
			this.updateBreadcrumbs();
			this.rendered = true;
			return this;
		}
	});

	/**
	 * Wrap a view in this to automatically show a "build in progress/build failed" message.
	 *
	 * @param model: the App to observe for build tasks
	 * @param options.content: view to attach to the DOM when no build is occuring
	 */
	var RequiresBuiltAppView = Toolkit.Views.ViewWithTemplate.extend({
		initialize: function () {
			this.alreadyRendered = false;
			RequiresBuiltAppView.__super__.initialize.call(this);
			if (this.options.content) {
				this.content = this.options.content;
			}
			this.autoClose(this.content);
			this.content.render();

			this.listenToWhenAvailable(this.model, 'local', 'change', this.render);
		},

		render: function () {
			var alreadyRendered = this.alreadyRendered;
			this.alreadyRendered = false;
			var local = this.model.get("local");

			if (!local) {
				this.renderTemplate('apps/loading');
				return this;
			}

			var buildTask = local.get('buildtask');
			var status = buildTask ? buildTask.get('status') : undefined;

			if (status === 'running') {
				this.renderTemplate('apps/building');
				return this;
			}

			if (status === 'failed') {
				this.renderTemplate('apps/build_failed');
				return this;
			}

			// Don't re-render if we don't need to
			if (!alreadyRendered) {
				this.$el.html('');
				this.$el.append(this.content.el);
				this.content.delegateSubViewEvents();
				this.content.delegateEvents();
			}
			this.alreadyRendered = true;
			return this;
		}
	});


	Toolkit.Views.BaseConfigView = Toolkit.Views.ViewWithTemplate.extend({
		className: 'configuration',
		initialize: function () {
			Toolkit.Views.ViewWithTemplate.prototype.initialize.call(this);
			this.listenTo(this.model, 'change:config', this.render);
			this.listenTo(this.model, 'change:schema', this.render);
			this.listenToWhenAvailable(this.model, 'project', 'change:name', this.render);
		},
		render: function () {
			this.renderTemplate();
			if (!this.model.get('config') || !this.model.get('schema')) {
				return undefined;
			}
			return this;
		},
		save: function () {
			var local = this.model.get('app').get("local");
			local.saveConfig(this.model.get("config"));
		}
	});


	Toolkit.Views.AppConfigView = GenericAppView.extend({
		active: 'App',
		ContentView: Toolkit.Views.BaseConfigView.extend({
			template: 'apps/config/index',
			events: {
				"click .build": "_clickedBuild", // lose ?
				"click .change-version": "_clickedChangeVersion"
			},
			initialize: function () {
				Toolkit.Views.BaseConfigView.prototype.initialize.call(this);
				this.listenTo(this.model, 'change:premium_modules', this.render);
			},
			_clickedBuild: function (e) {
				e.preventDefault();
				this.model.set('should_rebuild', true);
				Toolkit.router.navigate('apps/'+this.model.id, true);
			},
			_clickedChangeVersion: function () {
				var me = this;
				var view = new (Toolkit.Views.ViewWithTemplate.extend({
					template: 'apps/config/change_platform_version',
					events: {
						"click .update-stable": "updateStable",
						"click .update-selection": "updateSelection"
					},
					updateToVersion: function (version) {
						var self = this;
						var changeVersion = function () {
							me.model.get('config').platform_version = version;
							var rpc = Toolkit.rpcSession.remoteCall('file.write', {
								path: me.model.get('path'),
								data: JSON.stringify(me.model.get('config'), undefined, "\t")
							});
							
							rpc.on('success', function () {
								me.model.get("app").get("local").buildIfNecessary();
								me.render();
								self.trigger("cancel");
							});
						};

						var dependencies = me.model.get("app").get("project").get("dependencies");

						var config = me.model.get('config');
						var modules = [];
						_.each(config.modules, function (data, module) {
							modules.push({
								"module": module,
								"version": data.version
							});
						});
						var before = Toolkit.dependencies.check_dependencies_and_clashes(dependencies, _.clone(modules), config.platform_version);
						var after = Toolkit.dependencies.check_dependencies_and_clashes(dependencies, _.clone(modules), version);
						
						if (before.status == "ok" && after.status != "ok") {
							if (after.status == "missing") {
								after.problems = [];
								_.each(after.missing, function (missing) {
									after.problems.push("Missing dependency: "+missing.module+" v"+missing.version);
								});
							}

							var modal = new Toolkit.Views.ModalDialog({
								title: 'Are you sure?',
								content: "<p>After updating your platform version builds will not work due to the following problems:<br>"+after.problems.join("<br>")+"</p>",
								actions: [{
									type: "cancel",
									text: "Cancel"
								}, {
									type: "danger",
									text: "Change platform version",
									onClicked: function () {
										changeVersion();
										modal.hide();
									}
								}]
							});
							modal.show();
						} else {
							changeVersion();
						}
					},
					updateStable: function () {
						this.updateToVersion(Toolkit.platforms.get('latest_stable'));
					},
					updateSelection: function () {
						this.updateToVersion(this.$('select').val());
					},
					templateAttributes: function () {
						var current_version = me.model.get('config').platform_version;
						var stable_version = Toolkit.platforms.get('latest_stable');
						return {
							'current_version': current_version,
							'stable_version': stable_version,
							'current_is_stable': current_version == stable_version,
							'available_platforms': Toolkit.platforms.platformsLocalAppCanSwitchTo(current_version)
						};
					}
				}))();
				var modalView = new Toolkit.Views.ModalDialog({
					title: 'Change Trigger.io platform version',
					contentView: view
				});
				modalView.show();
			},
			templateAttributes: function () {
				return {
					'platform_version': this.model.get('config') && this.model.get('config').platform_version
				};
			},
			delegateSubViewEvents: function () {
				this.configView && this.configView.delegateEvents();
			},
			render: function () {
				var local = this.model.get("app").get("local");
				if (local && local.get("problem")) {
					if (local.get("problem").type == 'remote_error') {
						// Allow platform version to be updated if it is only a remote error
						Toolkit.Views.BaseConfigView.prototype.render.call(this);
						this.$el.append(Mustache.to_html(Toolkit.Templates['apps/serious_problem'], {}));
					} else {
						this.$el.html(Mustache.to_html(Toolkit.Templates['apps/serious_problem'], {}));
					}
				} else {
					if (!Toolkit.Views.BaseConfigView.prototype.render.call(this) ||
						!this.model.get('premium_modules')) {
						this.$el.html(Mustache.to_html(Toolkit.Templates['apps/config/loading'], {}));
						return this;
					}

					var self = this;

					// tag objects we want to hide
					var tagged_schema = _.clone(this.model.get('schema'));
					_.each([ "config_version", "plugins", "platform_version" ], function (property) {
						if (_.has(tagged_schema.properties, property)) {
							tagged_schema.properties[property]._hide = true;
						}
					});

					// tag v1.4 modules
					_.each(tagged_schema.properties.modules.properties, function (schema, name) {
						if (_.contains(self.model.get("premium_modules"), name)) {
							schema._v14_premium_module = true;
						} else {
							//schema._v14_module = true;
						}
					});

					var html = Mustache.to_html(Toolkit.Templates['apps/config_saved'], {});
					self.$el.append(html);

					var configView = this.configView = new Toolkit.Views.ConfigurationView({
						schema: tagged_schema,
						model: this.model.get('config'),
						name: null,
						pickerStart: local.get('path')+'/src',
						parent: self
					});

					this.autoClose(configView);

					self.$el.append(configView.render().el);
				}
				return this;
			}
		})
	});

	Toolkit.Views.AppLocalConfigView = GenericAppView.extend({
		active: 'Tools',
		ContentView: Toolkit.Views.BaseConfigView.extend({
			template: 'apps/local_config',

			initialize: function () {
				Toolkit.Views.BaseConfigView.prototype.initialize.call(this);
				if (this.model.get('platform') && this.model.get('platform').has('localConfigSchema')) {
					this.model.set('schema', JSON.parse(this.model.get('platform').get('localConfigSchema')));
				}
				var me = this;
				this.model.get('platform') && this.listenTo(this.model.get('platform'), "change:localConfigSchema", function () {
					me.model.set('schema', JSON.parse(me.model.get('platform').get('localConfigSchema')));
					me.render();
				});
				this.listenToWhenAvailable(this.model, 'local', 'change:problem', this.render);
			},

			render: function () {
				var local = this.model.get("app").get("local");
				if (local && local.get("problem")) {
					this.$el.html(Mustache.to_html(Toolkit.Templates['apps/serious_problem'], {}));
				} else {
					if (!Toolkit.Views.BaseConfigView.prototype.render.call(this)) {
						return this;
					}

					var self = this;

					var config_view = this.config_view = new Toolkit.Views.ConfigurationView({
						schema: this.model.get('schema'),
						model: this.model.get('config'),
						name: false,
						parent: self
					}).render();

					this.autoClose(config_view);

					var html = Mustache.to_html(Toolkit.Templates['apps/config_saved'], {});
					self.$el.append(html);
					self.$el.append(config_view.el);
				}

				return this;
			},

			delegateSubViewEvents: function () {
				this.config_view && this.config_view.delegateEvents();
			},

			save: function () {
				var rpc = Toolkit.rpcSession.remoteCall('file.write', {
					path: this.model.get('path'),
					data: JSON.stringify(this.model.get('config'), undefined, "\t")
				});
				var self = this;
				rpc.on('success', function () {
					self.$(".saved").fadeIn();
					setTimeout(function () {
						self.$(".saved").fadeOut();
					}, 3000);
				});
			}
		})
	});

	Toolkit.Views.AppForgeView = GenericAppView.extend({
		active: 'Forge',
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: 'apps/forge',

			events: {
				"click a.forget": "_forgetClicked",
				"click a.clean": "_cleanClicked",
				"click a.disabled": "_disabledClicked"
			},

			_disabledClicked: function (e) {
				e.preventDefault();
				alert("That target is currently disabled, contact support@trigger.io to add more targets to your app.");
			},

			_cleanClicked: function (e) {
				e.preventDefault();
				this.model.get("local").clean();
			},

			_forgetClicked: function (e) {
				e.preventDefault();
				var me = this;
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Forget app?',
					content: '<p>This will cause the Toolkit to no longer show the app, you can reimport it into the Toolkit at a later point using the "Import an app" button.</p>',
					actions: [{
						type: "cancel",
						text: "Cancel"
					}, {
						type: "danger",
						text: "Forget",
						onClicked: function () {
							var rpc = Toolkit.rpcSession.remoteCall('app.forget', {
								uuid: me.model.id
							});

							rpc.on('success', function () {
								// TODO: Could do better than a forced refresh?
								window.location.hash = "#";
								window.location.reload();
								modal.hide();
							});
						}
					}]
				});
				modal.show();
				this.autoClose(modal);
			},

			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
				this.listenTo(this.model, "change", this.render);
				this.listenToWhenAvailable(this.model, 'project', this.render);

				this.listenToWhenAvailable(this.model, 'local', 'change', this.render);
			},

			templateAttributes: function () {
				var local = this.model.get("local");
				var targets = this.model.get('project') && this.model.get('project').get('features') && this.model.get('project').get('features').targets;
				var platform_targets = this.model.get('platform_targets');
				var schema = local.get("configSchema");
				var serve_targets = [];
				if (schema && hasNestedProperty(schema, "properties.core.properties.general.properties.live")) {
					serve_targets = _.map(_.intersection(['android', 'ios', 'crosswalk'], targets, platform_targets).sort(), Toolkit.preprocess.target);
				}
				return {
					loading: !(local && targets && schema),
					breadcrumbs: this.appBreadcrumbs(this.model),
					project: this.model.getAttributes('project'),
					app: this.model.attributes,
					local: local ? local.attributes : undefined,
					run_targets: _.map(_.intersection(['android', 'ios', 'chrome', 'safari', 'firefox', 'ie', 'web', 'osx', 'crosswalk'], targets, platform_targets).sort(), Toolkit.preprocess.target),
					package_targets: _.map(_.intersection(['android', 'ios', 'web', 'crosswalk'], targets, platform_targets).sort(), Toolkit.preprocess.target),
					serve_targets: serve_targets,
					validated_email: Toolkit.authStatus.get('validated_email')
				};
			},

			render: function () {
				var local = this.model.get("local");
				if (local && local.get("problem")) {
					this.$el.html(Mustache.to_html(Toolkit.Templates['apps/serious_problem'], {}));
				} else {
					this.renderTemplate();
					this.renderFeatureStyle(this.model.get('project'));
					this._updateTargets();
				}
				return this;
			},

			_updateTargets: function () {
				var me = this;
				me.$('.platformbttn').addClass('disabled');

				// leave all disabled while this app is orphaned
				var parentProject = this.model.get('project');
				if (!parentProject) {
					return;
				}

				var features = parentProject.get('features');
				features && _(features.targets).each(function (target) {
					me.$('.platformbttn.'+target).addClass('enabled').removeClass('disabled');
				});
			},

			delegateSubViewEvents: function () {
				// no subviews to manage
			}
		})
	});

	Toolkit.Views.AppStatusView = GenericAppView.extend({
		active: 'Status',
		ContentView: Toolkit.Views.StatusPageView.extend({
			template: 'apps/status',

			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
				this.listenTo(this.model, "change", this.render);
				this.listenToWhenAvailable(this.model, 'project', this.render);
				this.listenToWhenAvailable(this.model, 'local', 'change', this.render);
			}
		})
	});

	Toolkit.Views.AppPluginsView = GenericAppView.extend({
		active: 'Plugins',
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: "apps/plugins",
			events: {
				"click .active": "_clickedActive",
				"click .removebroken": "_clickedRemoveBroken",
				"change select": "_changedVersion"
			},

			_configUpdated: function () {
				var local = this.model.get("local");
				local.get('config').plugins = this.options.configPlugins;

				var rpc = Toolkit.rpcSession.remoteCall('file.write', {
					path: local.get('files').config,
					data: JSON.stringify(local.get('config'), undefined, "\t")
				});
				var me = this;
				rpc.on('success', function () {
					me.$(".saved").fadeIn();
					setTimeout(function () {
						me.$(".saved").fadeOut();
					}, 3000);
				});
				this._updateFromConfig();
				this.render();
			},

			_clickedRemoveBroken: function (e) {
				delete this.options.configPlugins[this.$(e.target).data('pluginname')];
				this._configUpdated();
				return false;
			},

			_updatePluginWithLabel: function (label) {
				var plugin = null, hash = null;
				var type = label.data('plugintype');

				var name = label.data('pluginname');
				if (label.find('input[type="checkbox"]').is(':checked')) {
					// Enable
					plugin = _.find(this.options.projectPluginsClean, function (plugin) {
						return plugin.name === name;
					});
					hash = label.find('select').val();
					this.options.configPlugins[name] = {
						hash: hash,
						type: type
					};
				} else {
					// Disable
					delete this.options.configPlugins[name];
				}

				this._configUpdated();
			},

			_changedVersion: function (e) {
				var active = this.$(e.target).parent().find('input[type="checkbox"]').is(':checked');
				if (active) {
					this._updatePluginWithLabel(this.$(e.target).parent());
				}
			},

			_clickedActive: function (e) {
				this._updatePluginWithLabel(this.$(e.target).parent().parent());
			},

			_updateFromConfig: function () {
				this.options.projectPlugins = JSON.parse(JSON.stringify(this.options.projectPluginsClean));
				var configPlugins = this.options.configPlugins;

				// for each project plugin, see if there is an entry in config.json that refers to
				// a version for the plugin
				var matched = [];
				_.each(this.options.projectPlugins, function (projectPlugin) {
					// check to see if there's a plugin in config.json with the right name
					if (configPlugins[projectPlugin.name]) {
						matched.push(projectPlugin.name);
						projectPlugin.active = true;

						// mark the version that is currently in the config
						var foundVersion = false;
						_.each(projectPlugin.versions, function (version) {
							if (version.filename === configPlugins[projectPlugin.name].hash) {
								foundVersion = true;
								version.active = true;
							} else {
								version.active = false;
							}
						});

						// the hash in config.json doesn't belong to a version for this project plugin
						if (!foundVersion) {
							projectPlugin.versions.push({
								active: true,
								filename: configPlugins[projectPlugin.name].hash,
								version: "Unknown version in config.json, please select another."
							});
						}
					}
				});

				// List any unmatched plugins as broken
				this.options.brokenPlugins = _.map(_.filter(_.keys(configPlugins), function (name) {
					return matched.indexOf(name) === -1;
				}), function (name) {
					return {
						name: name
					};
				});
			},

			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
				this.options.configPlugins = this.options.config.plugins || {};

				this.doWhenAvailable(this.model, 'project', function (project) {
					this.doWhenAvailable(project, 'plugins', function (plugins) {
						var fired = false;
						var updateWhenAllPluginsLoaded = function () {
							if (fired) {
								return;
							}
							var allLoaded = plugins.every(function (p) {
								return p.get('_loaded');
							});
							if (allLoaded) {
								fired = true;
								this.options.projectPlugins = plugins;
								this.options.projectPluginsClean = JSON.parse(JSON.stringify(this.options.projectPlugins));
								this._updateFromConfig();
								this.render();
								return true;
							}
						};
						// force all plugins to load
						plugins.each(function (p) {p.load();});
						if (!updateWhenAllPluginsLoaded.call(this)) {
							this.listenTo(plugins, 'change', updateWhenAllPluginsLoaded);
						}
					});
				});
			},
			templateAttributes: function () {
				var project_plugins = this.options.projectPlugins;
				return _.extend(this.modelAttributes(), {
					plugins_loaded: project_plugins && _(project_plugins).every(function (p) {
						return p._loaded;
					}),
					project_plugins: project_plugins,
					has_broken_plugins: _.size(this.options.brokenPlugins) !== 0,
					broken_plugins: this.options.brokenPlugins
				});
			},
			render: function () {
				var local = this.model.get("local");
				if (local && local.get("problem")) {
					this.$el.html(Mustache.to_html(Toolkit.Templates['apps/serious_problem'], {}));
				} else {
					Toolkit.Views.ViewWithTemplate.prototype.render.apply(this, arguments);
					var html = Mustache.to_html(Toolkit.Templates['apps/config_saved'], {});
					this.$el.append(html);
				}
				return this;
			},

			delegateSubViewEvents: function () {
				// no subviews to manage
			}
		})
	});

	Toolkit.Views.AppConsoleView = GenericAppView.extend({
		active: 'Forge',
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: 'component/console/page',

			templateAttributes: function () {
				var command = this.model.get('command');
				return {
					target: Toolkit.preprocess.target(this.model.get('target')),
					command: command[0].toUpperCase() + command.slice(1)
				};
			},

			delegateSubViewEvents: function () {
				this.consoleView && this.consoleView.delegateEvents();
			},

			render: function () {
				this.renderTemplate();

				var consoleView = this.consoleView = new Toolkit.Views.ConsoleView({
					model: this.model.get('consoleModel')
				}).render();

				this.$('#console').html(consoleView.el);
				return this;
			}
		})
	});

	// add a handler for app migration events
	RpcCall.prototype.defaults.migration = function () {
		// migration of app has occured:
		// various bits of app may have been changed
		// easiest to reload to get models up to date
		window.location.reload();
	};
}());
