/*global Toolkit*/
/*global _*/
/*global Backbone*/
/*global Mustache*/
(function () {
	/**
	 * Base view for configuring/toggling/removing an individual module.
	 *
	 * @param model.attributes.app {Toolkit.Models.App} App model being configured
	 * @param model.attributes.moduleVersion {Toolkit.Models.ModuleVersion} Details about the configured module version
	 */
	var GenericModuleConfigView = Toolkit.Views.GenericModuleConfigView = Toolkit.Views.ViewWithTemplate.extend({
		initialize: function () {
			GenericModuleConfigView.__super__.initialize.apply(this, arguments);
			this.listenTo(this.model, 'change', this.renderIfNecessary);
			this.listenTo(this.model.get('moduleVersion'), 'change', this.renderIfNecessary);
			this.doWhenAvailable(this.model.get("object"), 'local', function () {
				this.model.set("local", this.model.get("object").get("local"));
			});
			this.doWhenAvailable(this.model, 'local', this.renderIfNecessary);
			this.listenToWhenAvailable(this.model, 'local', 'change:inspector_config', this.renderIfNecessary);
			if (this.model.get("module")) {
				this.listenToWhenAvailable(this.model.get("module"), 'project', 'change:dependencies', this.renderIfNecessary);
			}
		},

		events: {
			"click .module-toggle": "clickModuleToggle",
			"click .module-remove": "clickModuleRemove"
		},

		clickModuleToggle: function () {
			this.model.get("local").toggleModule(this.model.get("moduleVersion").get("module__name"));
		},

		clickModuleRemove: function () {
			this.model.get("local").removeModule(this.model.get("moduleVersion").get("module__name"));
		},

		modelAttributes: function () {
			return {
				moduleVersion: this.model.attributes.moduleVersion.toJSON()
			};
		},

		templateAttributes: function () {
			var module_name = this.model.get("moduleVersion").get("module__name");
			var config = this.model.get("local") && this.model.get("local").getConfig();
			if (!config) {
				return {};
			}
			var module_config = config.modules[module_name] && {version: config.modules[module_name].version, disabled: config.modules[module_name].disabled};
			var superAttrs = GenericModuleConfigView.__super__.templateAttributes.apply(this, arguments);
			var warning = false;
			if (this.model.get("module")) {
				// figure out if the ModuleVersion for this View is a dependency for the module
				var module = this.model.get("module");
				var local = this.model.get("local");
				var dependencies = _.clone(module.get("project").get("dependencies"));
				if (dependencies) {
					var indexOfExisting;
					for (var i = 0; i < dependencies.length; ++i) {
						if (dependencies[i].module === module.id) {
							indexOfExisting = i;
							break;
						}
					}

					var dependencyInfoFromManifest = {
						module: module.id,
						version: local.get("manifest").version,
						dependencies: local.get("manifest").dependencies,
						clashes: []
					};

					if (indexOfExisting !== undefined) {
						dependencies[indexOfExisting] = dependencyInfoFromManifest;
					} else {
						dependencies.push(dependencyInfoFromManifest);
					}

					var modules = [{
						"module": module.id,
						"version": local.get("manifest").version
					}];
					var without = [{
						"module": module.id,
						"version": local.get("manifest").version
					}];
					_.each(config.modules, function (data, module) {
						if (data.disabled) {
							return;
						}
						modules.push({
							"module": module,
							"version": data.version
						});
						if (module != module_name) {
							without.push({
								"module": module,
								"version": data.version
							});
						}
					});
					var result = Toolkit.dependencies.check_dependencies_and_clashes(dependencies, modules, local.get("manifest").platform_version);
					var resultWithout = Toolkit.dependencies.check_dependencies_and_clashes(dependencies, without, local.get("manifest").platform_version);
					if (result.status === "ok" && resultWithout.status === "ok") {
						warning = "This module is not a dependency, while it will be included in your inspector project it will not necessarily be available in a users application.";
					}
				}
			}
			return _.extend(superAttrs, {
				config: module_config,
				app: this.model.get("app") ? this.model.get("app").toJSON() : false,
				warning: warning
			});
		},

		render: function () {
			this.renderTemplate();

			if (!this.model.get("local") || !this.model.get("local").getConfig().modules[this.model.get("moduleVersion").get("module__name")]) {
				return this;
			}

			var schema = this.model.get('moduleVersion').get('schema');
			if (schema && !_.isEmpty(schema)) {
				var configView = new Toolkit.Views.ConfigurationView({
					model: this.model.get("local").getConfig().modules[this.model.get("moduleVersion").get("module__name")].config,
					pickerStart: this.model.get("local").get("path") + "/src",
					schema: schema,
					parent: this
				});
				configView.render();

				// dispose old config view
				if (this._configView) {
					this._configView.close();
				}
				this._configView = configView;

				this.autoClose(configView);
				this.$('.module-config').append(configView.el);
				this.$('.config-panel').show();
			} else {
				this.$('.config-panel').hide();
			}

			if (this.model.get("local").getConfig().modules[this.model.get("moduleVersion").get("module__name")].disabled) {
				this.$('input,select').prop('disabled', true);
			}

			return this;
		},

		save: function (moduleConfig) {
			var config = this.model.get("local").getConfig();
			config['modules'][this.model.get('moduleVersion').get('module__name')]['config'] = moduleConfig;
			this.model.get("local").saveConfig();
		},

		delegateSubViewEvents: function () {
			this._configView && this._configView.delegateEvents();
		}
	});


	/**
	 * Base view  for views which show multiple project/public modules with configuration. Views which extend this should use
	 * addModuleView for each individual GenericModuleConfigView they want to display.
	 */
	var GenericModulesConfigView = Toolkit.Views.GenericModulesConfigView = Toolkit.Views.ViewWithTemplate.extend({
		initialize: function () {
			GenericModulesConfigView.__super__.initialize.apply(this, arguments);
			this.model.set('_loaded', false);
			this.listenTo(this.model, 'change:_loaded', this.renderIfNecessary);
			this.listenTo(this.model.get("config_modules"), 'remove', this.removeModuleView);
			this._moduleViews = {};

			var object = this.model.get("app") || this.model.get("module");

			this.doWhenAvailable(object, 'local', this.renderIfNecessary);

			var me = this;
			var moduleVersions = this.model.get("config_modules");
			var local = object.get("local");
			var addModuleView = function (moduleVersion) {
				me.addModuleView(moduleVersion.id, new me.ModuleView({
					parent: me,
					model: new Backbone.Model({
						moduleVersion: moduleVersion,
						local: local,
						app: me.model.get("app"),
						module: me.model.get("module"),
						object: object
					})
				}).render());
			};

			// listen for moduleVersions being added to collection
			this.listenTo(moduleVersions, 'add', function (moduleVersion) {
				if (moduleVersion.get('_loaded')) {
					if (typeof(moduleVersion.get('type')) !== "undefined" &&
							moduleVersion.get('type') !== me.moduleType) {
						return;
					}
					addModuleView(moduleVersion);
					me.render();
				} else {
					this.doWhenAvailable(moduleVersion, '_loaded', function () {
						if (typeof(moduleVersion.get('type')) !== "undefined" &&
								moduleVersion.get('type') !== me.moduleType) {
							return;
						}
						addModuleView(moduleVersion);
						me.render();
					});
				}
			});

			this.doWhenAll(moduleVersions.models, '_loaded', function () {
				// create subview for module entries from config.json
				moduleVersions.each(function (moduleVersion) {
					if (typeof(moduleVersion.get('type')) !== "undefined" &&
							moduleVersion.get('type') !== me.moduleType) {
						return;
					}
					addModuleView(moduleVersion);
				});

				this.model.set('_loaded', true);
			});
		},

		delegateSubViewEvents: function () {
			_.each(this._moduleViews, function (v) {
				v.delegateSubViewEvents();
				v.delegateEvents();
			});
		},

		addModuleView: function (name, view) {
			this._moduleViews[name] = view;
		},

		removeModuleView: function (moduleVersion) {
			var moduleId = moduleVersion.id;
			if (this._moduleViews[moduleId]) {
				var view = this._moduleViews[moduleId];
				view.close();
				view.remove();
				delete this._moduleViews[moduleId];
			}
			this.render();
		},

		modelAttributes: function () {
			return _.omit(this.model.attributes, 'local');
		},

		render: function () {
			var html,
				me = this;

			var object = this.model.get("app") || this.model.get("module");
			if (object.get("local")) {
				GenericModulesConfigView.__super__.render.apply(this, arguments);
				if (this.model.get('_loaded')) {
					if (_.isEmpty(this._moduleViews)) {
						html = Mustache.to_html(Toolkit.Templates['modules/get_started'], {
							"public": this.moduleType === "public",
							"project": this.moduleType === "project"
						});
						if (!this.templateAttributes().disabled) {
							this.$el.append(html);
						}
						return this;
					} else {
						_.each(_.keys(this._moduleViews).sort(), function (key) {
							me.$el.append(me._moduleViews[key].el);
						});
						this.delegateSubViewEvents();
					}
				}
				html = Mustache.to_html(Toolkit.Templates['apps/config_saved'], {});
				if (!this.templateAttributes().disabled) {
					this.$el.append(html);
				}
			} else {
				this.$el.append("No local version, see status tab.");
			}
			return this;
		}
	});
}());
