/*global Toolkit*/
/*global _*/
/*global $*/
(function () {

	/**
	 * Displays a list of project modules that can be used in an app build.
	 *
	 * @param model.attributes.app {Toolkit.Models.App} app which is having project modules configured.
	 * @param model.attributes.moduleVersions {Backbone.Collection<Toolkit.Models.ModuleVersion>} module versions configured for this app.
	 */
	var AddProjectModuleView = Toolkit.Views.ViewWithTemplate.extend({
		template: 'apps/projectModules/add',

		initialize: function () {
			AddProjectModuleView.__super__.initialize.apply(this, arguments);
			this.doWhenAvailable(this.model, 'local', function (local) {
				this.listenTo(local, 'change:config change:inspector_config', this.renderIfNecessary);
			});
			this.doWhenAvailable(this.model, 'project', function (project) {
				this.doWhenAvailable(project, 'modules', this.renderIfNecessary);
				this.listenToWhenAvailable(project, 'modules', this.renderIfNecessary);
			});
		},

		events: {
			'click .module-add': 'clickModuleAdd'
		},

		clickModuleAdd: function (e) {
			e.preventDefault();
			var moduleName = this.$(e.target).text();

			// need to make sure Module model is loaded so we can query for versions
			var projectModule = this.model.get('project').get('modules').get(moduleName).load();

			this.doWhenAvailable(projectModule, 'versions', function (versions) {
				// get latest version for module
				var versionStrings = _.pluck(versions, 'version');
				versionStrings.sort(function (a, b) {
					var aChunks = a.split('.'), bChunks = b.split('.');
					for (var i = 0; i < Math.max(aChunks.length, bChunks.length); i++) {
						if (aChunks[i] === undefined) {
							return -1;
						}
						if (bChunks[i] === undefined) {
							return 1;
						}
						if (aChunks[i] !== bChunks[i]) {
							return aChunks[i] !== bChunks[i];
						}
					}
				});
				var latestVersion = versionStrings[versionStrings.length - 1];

				if (!latestVersion) {
					this.autoClose(Toolkit.modal.alert({
						title: "Module has no versions",
						content:"You cannot activate this module as it has no uploaded versions."
					}));
					return;
				}

				this.model.get("local").addModule(moduleName, latestVersion);
			});
		},

		templateAttributes: function () {
			var project = this.model.get('project');
			var local = this.model.get("local");
			if (!local || !project) {
				return {};
			}
			var config = local.getConfig();
			var modules = project &&
					project.get("modules") &&
					config &&
					_.filter(project.get("modules").toJSON(), function (module) {
				return !config['modules'][module.name];
			});
			return {
				modules: modules
			};
		}
	});

	var ProjectModuleView = Toolkit.Views.ProjectModuleView = Toolkit.Views.GenericModuleConfigView.extend({
		template: 'apps/projectModules/module',

		events: {
			"click .update-version": "updateVersion"
		},

		initialize: function (options) {
			ProjectModuleView.__super__.initialize.apply(this, arguments);
			this.parent = options.parent;
			this.listenToWhenAvailable(this.model, 'object', 'change:status', this.render);
		},

		templateAttributes: function () {
			return _.extend(Toolkit.Views.GenericModuleConfigView.prototype.templateAttributes.call(this), {
				'status': this.model.get("object").getStatusesWithTag('module-'+this.model.get("moduleVersion").get("module__name"))
			});
		},

		updateVersion: function (e) {
			e.preventDefault();
			this.model.get("local").updateModule(this.model.get("moduleVersion").get("module__name"), $(e.target).attr('data-version'));
		}
	});

	_.extend(ProjectModuleView.prototype.events, ProjectModuleView.__super__.events);

	var ProjectModulesContentView = Toolkit.Views.GenericModulesConfigView.extend({
		template: 'apps/projectModules/config',
		ModuleView: ProjectModuleView,
		moduleType: 'project',

		initialize: function () {
			// subview with controls for selecting/adding a project module
			this.addView = new AddProjectModuleView({parent: this, model: this.model}).render();
			Toolkit.Views.GenericModulesConfigView.prototype.initialize.apply(this, arguments);
		},

		delegateSubViewEvents: function () {
			Toolkit.Views.GenericModulesConfigView.prototype.delegateSubViewEvents.apply(this, arguments);
			this.addView && this.addView.delegateEvents();
		},

		render: function () {
			Toolkit.Views.GenericModulesConfigView.prototype.render.apply(this, arguments);
			this.$('.module-adder').append(this.addView.el);
			this.delegateSubViewEvents();
			return this;
		}
	});

	Toolkit.Views.ProjectModulesView = Toolkit.Views.GenericAppView.extend({
		active: 'Private',
		ContentView: ProjectModulesContentView
	});

	Toolkit.Views.InspectorProjectModulesView = Toolkit.Views.GenericModuleView.extend({
		active: 'Private',
		ContentView: ProjectModulesContentView.extend({
			templateAttributes: function () {
				var attr = ProjectModulesContentView.prototype.templateAttributes.apply(this, arguments);
				var local = this.model.get("local");
				var disabled = !!local && !local.canUseFeature("modules-inspector-config");

				return _.extend(attr, {
					disabled: disabled
				});
			}
		})
	});

	Toolkit.Views.InspectorThisModuleContentView = Toolkit.Views.BaseConfigView.extend({
		template: 'modules/inspector_this_module_config',

		initialize: function () {
			Toolkit.Views.BaseConfigView.prototype.initialize.apply(this, arguments);
			this.doWhenAvailable(this.model, 'module', function (module) {
				this.listenTo(module, 'change:status', this.render);
			});
			this.doWhenAvailable(this.model, 'local', function (local) {
				this.listenTo(local, 'change:config_schema', this.render);
			});
		},

		save: function () {
			var local = this.model.get('local');
			local.saveConfig();
		},

		delegateSubViewEvents: function () {
			this.configView && this.configView.delegateEvents();
		},

		templateAttributes: function () {
			var local = this.model.get('local');
			var configSchemaPath = local && local.relativeFilePath('config_schema');
			var moduleName = local && local.getName();
			var manifest = local && local.get('manifest');
			var description = manifest && manifest['description'];
			var shortDescription = description ? description.split('.')[0] : '';
			var module = this.model.get('module');
			var status = module && module.getStatusesWithTag('module-'+moduleName);

			return {
				configSchemaPath: configSchemaPath,
				status: status,
				moduleVersion: {
					module__name: moduleName,
					short_description: shortDescription
				},
				hideVersionSwitcher: true,
				hideControls: true
			};
		},

		render: function () {
			Toolkit.Views.GenericModulesConfigView.prototype.render.apply(this, arguments);

			var local = this.model.get('local');
			var moduleConfig = local && local.getConfig();
			var moduleName = local && local.getName();
			var moduleConfigSchema = local && local.get('config_schema');

			// give up if enough hasn't loaded
			if (!local || !moduleName || !moduleConfigSchema || !moduleConfig) {
				return this;
			}

			// create, render and attach config subview, and clean up previous
			if (this.configView) {
				this.configView.close();
			}

			this.configView = new Toolkit.Views.ConfigurationView({
				schema: moduleConfigSchema,
				model: moduleConfig['this_module']['config'],
				name: false,
				parent: this
			});

			this.configView.render();
			this.autoClose(this.configView);
			this.$('.module-config').html(this.configView.el);

			return this;
		}
	});

	Toolkit.Views.InspectorThisModuleView = Toolkit.Views.GenericModuleView.extend({
		active: 'This module',
		ContentView: Toolkit.Views.InspectorThisModuleContentView
	});
}());
