/*global $*/
/*global _*/
/*global Backbone*/
/*global Toolkit*/
/*global moment*/
/*global NO_TIMEZONE_FORMAT*/
/*global Mustache*/
(function () {
	/**
	 * Fill out a sparse array of weekly update counts.
	 *
	 * @param updates_weekly {Array.<{start: string, end: string, count: number}>}
	 * @param updates_gt {string}
	 * @param countAttribute {string} Name of attribute that contains count, e.g. 'count' or 'total'
	 * @param nowString {string=}
	 * @returns {Array.<{start: string, end: string, count: number}>}
	 */
	function filledUpdatesWeekly(updates_weekly, updates_gt, countAttribute, nowString) {
		var nowMoment;
		if (nowString) {
			nowMoment = moment(nowString, NO_TIMEZONE_FORMAT);
		} else {
			nowMoment = moment();
		}

		var now = nowMoment.toDate().getTime();

		var gt = new Date(updates_gt).getTime();

		var start;
		var existing;
		var i = 0;
		if (_.isEmpty(updates_weekly)) {
			start = moment(updates_gt);
		} else {
			start = moment(updates_weekly[0].start);
			existing = updates_weekly[i];
		}
		var start2 = start.clone();

		var filledUpdates = [];

		for (; start.toDate().getTime() < now;) {
			var startFormat = start.format(NO_TIMEZONE_FORMAT);
			// .add mutates inplace
			start.add('days', 7);

			if (existing && existing.start === startFormat) {
				filledUpdates.push(existing);
				existing = updates_weekly[++i];
				continue;
			}

			var update = {
				'start': startFormat,
				'end': start.format(NO_TIMEZONE_FORMAT)
			};
			update[countAttribute] = 0;
			filledUpdates.push(update);
		}

		for (; start2.toDate().getTime() > gt;) {
			var update2 = {
				'end': start2.format(NO_TIMEZONE_FORMAT),
				// .subtract mutates inplace
				'start': start2.subtract('days', 7).format(NO_TIMEZONE_FORMAT)
			};
			update2[countAttribute] = 0;

			filledUpdates.unshift(update2);
		}

		return filledUpdates;
	}

	// exposing this for testing
	Toolkit.filledUpdatesWeekly = filledUpdatesWeekly;

	/**
	 * "active users" for a stream/buildevent doesn't come down with server responses
	 * but we can derive it from what does using this
	 */
	function computeActiveUserTotal() {
		var updatesWeekly = this.get('weekly_totals');
		if (!updatesWeekly || _.isEmpty(updatesWeekly)) {
			this.set('active_user_total', 0);
			return;
		}

		this.set('active_user_total', _.last(updatesWeekly));
	}

	function computeActiveUserCount() {
		var updatesWeekly = this.get('weekly_counts');
		if (!updatesWeekly || _.isEmpty(updatesWeekly)) {
			this.set('active_user_count', 0);
			return;
		}

		this.set('active_user_count', _.last(updatesWeekly));
	}

	/**
	 * Use the sparsely populated 'updates_weekly' property to compute the 'filled_updates_weekly' property.
	 */
	function computeWeeklyTotals() {
		var updatesWeekly = this.get('updates_weekly');
		if (updatesWeekly) {
			var filled = filledUpdatesWeekly(updatesWeekly, this.get('updates_gt'), 'total');
			this.set('weekly_totals', _(filled).pluck('total'));
		}
	}

	function computeWeeklyCounts() {
		var updatesWeekly = this.get('updates_weekly');
		if (updatesWeekly) {
			var filled = filledUpdatesWeekly(updatesWeekly, this.get('updates_gt'), 'count');
			this.set('weekly_counts', _(filled).pluck('count'));
		}
	}

	/**
	 * Represents a stream for a particular BuildEvent.
	 * @constructor
	 */
	Toolkit.Models.Stream = Backbone.Model.extend({
		idAttribute: 'name',

		initialize: function () {
			Toolkit.Models.Stream.__super__.initialize.apply(this, arguments);
			this.computedAttribute('updates_weekly', computeWeeklyCounts);
			this.computedAttribute('weekly_counts', computeActiveUserCount);
		}
	});

	/**
	 * Represents a built config for a particular app.
	 * @constructor
	 */
	Toolkit.Models.BuildEvent = Toolkit.Models.RpcModel.extend({
		idAttribute: 'config_hash', // does this include uuid?
		cacheName: 'BuildEvent',
		method: 'api.get',
		args: function () {
			return {
				cookies: Toolkit.cookies,
				method: 'reload/buildevent/'+this.get('config_hash')
			};
		},
		parent: {
			attribute: 'app',
			idAttribute: 'app__uuid',
			model: 'App',
			collectionAttribute: 'buildevents'
		},
		inject: {
			streams: function (streams) {
				var collection = new Backbone.Collection(streams, {
					model: Toolkit.Models.Stream,
					comparator: function (b) {
						return -(new Date(b.get('last_pushed')).getTime());
					}
				});

				// bubble changes to streams
				this.listenTo(collection, 'all', function () {
					return this.trigger('change:streams');
				});
				return collection;
			}
		},
		toJSON: function () {
			var superAttrs = Toolkit.Models.BuildEvent.__super__.toJSON.call(this);
			// need to serialize child collection to an array
			if (superAttrs['streams']) { superAttrs['streams'] = superAttrs['streams'].toJSON(); }
			return superAttrs;
		},

		initialize: function () {
			Toolkit.Models.Stream.__super__.initialize.apply(this, arguments);
			this.computedAttribute('updates_weekly', computeWeeklyTotals);
			this.computedAttribute('weekly_totals', computeActiveUserTotal);
		}
	});

	/**
	 * Represent a collection of BuildEvents for an individual App.
	 * @constructor
	 */
	Toolkit.Collections.BuildEventCollection = Backbone.Collection.extend({
		model: Toolkit.Models.BuildEvent,
		/**
		 * Sort descending by "first built".
		 * @param buildevent {BuildEvent}
		 * @returns {number}
		 */
		comparator: function (buildevent) {
			return -(new Date(buildevent.get('first_built')).getTime());
		}
	});

	Toolkit.Views.ReloadView = Toolkit.Views.GenericAppView.extend({
		active: 'Reload',
		breadcrumbs: function () {
			return this.appBreadcrumbs(this.model);
		},
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: 'apps/reload/index',
			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);

				// use this to store state specific to the page
				// i.e. whether extra configs should be shown
				this.tempModel = new Backbone.Model();
				this.tempModel.set('hideconfigs', true);
				this.listenTo(this.tempModel, 'change', this.render);

				var app = this.model;
				this.listenTo(app, "change", this.render);
				this.listenTo(app, "change:_loadingBuildevents", this.render);
				this.listenToWhenAvailable(app, 'project', this.render);
				this.listenToWhenAvailable(app, 'buildevents', 'change', this.render);
				this.listenToWhenAvailable(app, 'local', 'change', this.render);

				this.waitingToRenderSparklines = false;
			},
			activeBuildevent: function () {
				var app = this.model;
				var local = app.get("local");
				var localConfigHash = local.get('config_hash');
				var buildevents = app.get('buildevents');
				return buildevents && buildevents.get(localConfigHash);
			},
			templateAttributes: function () {
				var activeBuildevent = this.activeBuildevent();
				var limits = (this.model.get("project") && this.model.get("project").get('limits')) ? this.model.get("project").get('limits') : {};
				return _({
					hideconfigs: this.tempModel.get('hideconfigs'),
					activeBuildevent: activeBuildevent && activeBuildevent.toJSON(),
					overLimit: this.model.get("project") && this.model.get("project").get("limits") && this.model.get("project").get("limits").reload.grant === false,
					limits: limits
				}).extend(this.modelAttributes());
			},
			applyTimeago: function () {
				this.$('.timeago').timeago();
			},

			render: function () {
				var local = this.model.get("local");
				if (local && local.get("problem")) {
					this.$el.html(Mustache.to_html(Toolkit.Templates['apps/serious_problem'], {}));
				} else {
					this.renderTemplate();
					this.applyTimeago();
					this.applySparklines();
				}
				return this;
			},
			applySparklines: function () {
				// prevent excessive setTimeouts which seems to cause the sparklines to disappear
				if (!this.waitingToRenderSparklines) {
					this.waitingToRenderSparklines = true;
					var me = this;
					// setTimeout to wait until sparkline data is in the DOM
					setTimeout(function () {
						me.$('.sparkline').sparkline('html', {width:125, height:40});
						me.waitingToRenderSparklines = false;
					}, 0);
				}
			},
			events: {
				"click .toggle-configs": "toggleConfigs",
				"click .restore-config": "restoreConfig",
				"click .create-stream": "createStream",
				"click .push-to-stream": "pushToStream"
			},
			toggleConfigs: function (e) {
				e.preventDefault();
				e.stopPropagation();
				this.tempModel.set('hideconfigs', !this.tempModel.get('hideconfigs'));
			},
			restoreConfig: function (e) {
				e.preventDefault();
				e.stopPropagation();

				var configHash = $(e.currentTarget).attr('data-confighash');
				var app = this.model;
				var buildevent = app.get('buildevents').get(configHash);
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Restore previously used App Config',
					contentView: new Toolkit.Views.ReloadRestoreConfigView({
						model: app,
						buildevent: buildevent.load()
					})
				});
				modal.show();
				this.autoClose(modal);
			},
			createStream: function (e) {
				e.preventDefault();
				e.stopPropagation();
				var app = this.model;
				var buildevent = this.activeBuildevent();
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Create stream',
					contentView: new Toolkit.Views.ReloadCreateStreamView({
						model: app,
						buildevent: buildevent
					})
				});
				modal.show();
				this.autoClose(modal);
			},
			pushToStream: function (e) {
				e.preventDefault();
				e.stopPropagation();
				var stream = $(e.currentTarget).attr('data-stream');
				var app = this.model;
				var buildevent = this.activeBuildevent();
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Confirm push to "'+stream+'" stream',
					contentView: new Toolkit.Views.ReloadPushView({
						model: app,
						buildevent: buildevent,
						stream: stream
					})
				});
				modal.show();
				this.autoClose(modal);
			},

			delegateSubViewEvents: function () {
				// no subviews to manage
			}
		})
	});

	Toolkit.Views.ReloadRestoreConfigView = Toolkit.Views.ViewWithTemplate.extend({
		template: 'apps/reload/restore',
		initialize: function () {
			var app = this.model;
			var buildevent = this.options.buildevent;
			this.listenTo(buildevent, 'change', this.render);
			this.listenTo(app, 'change', this.render);
			this.listenToWhenAvailable(app, 'project', this.render);
		},
		templateAttributes: function () {
			return this.options.buildevent.toJSON();
		},
		events: {
			"click .submit": "_restore"
		},
		_restore: function (e) {
			e.preventDefault();
			e.stopPropagation();

			var app = this.model;
			var buildevent = this.options.buildevent;
			var local = app.get("local");

			var rpc = Toolkit.rpcSession.remoteCall('reload.restore_config', {
				path: local.get('path'),
				config: buildevent.get('config')
			});

			var me = this;
			rpc.on('success', function (e) {
				// This will be overridden by file watching APIs, but we need to make
				// sure the active config hash is set before going back to the reload view
				app.setActiveConfig(e.data, buildevent.get('config_hash'));
				Toolkit.showNotification('Restored src/config.json for '+local.get('config').name);
				Toolkit.router.navigate('apps/'+app.get('uuid')+'/reload', true);
				me.trigger('cancel');
			});

			rpc.on('error', function (e) {
				this.defaults.error(e);
			});
		}
	});

	Toolkit.Views.ReloadPushView = Toolkit.Views.ViewWithTemplate.extend({
		className: 'app-reload-push',
		template: 'apps/reload/push',

		events: {
			"click .submit": "submit",
			"click .generate": "_clickedGenerate"
		},

		initialize: function () {
			Toolkit.Views.ReloadPushView.__super__.initialize.apply(this, arguments);
			var app = this.model;
			var buildevent = this.options.buildevent;

			this.listenTo(buildevent, 'change', this.render);
			this.listenToWhenAvailable(app, 'project', this.render);
			this.listenTo(app, 'change', this.render);
			this.listenToWhenAvailable(app, "local", "change", this.render);
		},

		render: function () {
			if (this.templateAttributes().status.success) {
				Toolkit.showNotification("Push completed successfully.");
				this.trigger("cancel");
			}
			Toolkit.Views.ReloadPushView.__super__.render.apply(this, arguments);
		},

		close: function () {
			var local = this.model.get("local");

			// clear out any build task
			var pushtask = local.get('pushtask');
			if (pushtask && pushtask.get('status') !== 'running') {
				local.set({pushtask: undefined});
			}

			Toolkit.Views.ReloadPushView.__super__.close.call(this);
		},


		_clickedGenerate: function (e) {
			e.preventDefault();
			this.$('a.generate').button('loading');
			this.$(".cdn_url.error").html("");
			this.$(".manifest_url.error").html("");
			var cdn_url = this.$('#cdn_url').val();
			if (cdn_url === "") {
				this.$(".cdn_url.error").html("You need to specify a valid URL");
				return;
			}
			var rpc = Toolkit.rpcSession.remoteCall("reload.generate_manifest", {
				cookies: Toolkit.cookies,
				url: cdn_url,
				path: this.options.buildevent.get('app').get("local").get('path')
			});
			var self = this;
			rpc.on("success", function (e) {
				self.$("#manifest_url_container").show(100);
				self.$("#generate_button_container").hide();
				self.$("#manifest_url").text(e.data.manifest_url);
				Toolkit.rpcSession.remoteCall('directory.open_in_filebrowser', {
					path: e.data.output_path
				});
			});
		},

		submit: function (e) {
			e.preventDefault();
			this.$(".cdn_url.error").html("");
			this.$(".manifest_url.error").html("");

			var externalUrl;
			var localConfig = this._getLocalConfig();
			var useExternalUrl = localConfig && localConfig.general && localConfig.general.reload && localConfig.general.reload.external;
			if (useExternalUrl) {
				externalUrl = this.$('#manifest_url').text();
				if (externalUrl === "") {
					this.$(".cdn_url.error").html("You need to generate a manifest first");
					return;
				}
			}

			var buildevent = this.options.buildevent;
			var local = this.options.buildevent.get('app').get("local");
			var streamModel = buildevent.get('streams').get(this.options.stream);
			local.pushReload(streamModel, externalUrl);
		},

		_getLocalConfig: function () {
			var buildevent = this.options.buildevent;
			var app = buildevent.get('app');
			var local = app && app.get("local");
			return local && local.get('localConfig');
		},

		templateAttributes: function () {
			var app = this.model;
			var local = app.get("local");
			var pushtask = local.get('pushtask');

			var buildevent = this.options.buildevent;
			var streams = buildevent.get('streams');
			var streamToPushTo = streams && streams.where({name: this.options.stream})[0];
			var status = {};
			if (pushtask) {
				status[pushtask.get('status')] = true;
			}
			var limits = (this.model.get("project") && this.model.get("project").get('limits')) ? this.model.get("project").get('limits') : {};

			return {
				status: status,
				pushtask: pushtask && pushtask.toJSON(),
				buildevent: buildevent.toJSON(),
				stream: streamToPushTo && streamToPushTo.toJSON(),
				localConfig: this._getLocalConfig(),
				limits: limits
			};
		}
	});

	Toolkit.Views.ReloadCreateStreamView = Toolkit.Views.FormView.extend({
		template: 'apps/reload/create_stream',

		submit: function (form) {
			var rpc = Toolkit.rpcSession.remoteCall('api.post', {
				cookies: Toolkit.cookies,
				method: 'reload/'+this.model.get('uuid')+'/stream',
				data: {
					name: form.name
				}
			});

			var me = this;
			rpc.on('success', function (e) {
				// streams are global to all configs
				me.model.get('buildevents').each(function (buildevent) {
					var streams = buildevent.get('streams');
					var stream = e.data['data'];
					stream['updates_weekly'] = [];
					stream['updates_gt'] = moment().subtract(35, 'days');
					if (streams) {
						streams.add(stream);
					} else {
						buildevent.loadSet('streams', [stream]);
					}
				});
				Toolkit.showNotification('Added stream to app');
				me.options.buildevent.get('streams').add(e.data['data']);
				me.trigger("cancel");
			});

			rpc.on('error', this._handleFormErrors());
		}
	});

}());