/*global Toolkit*/
/*global _*/
/*global Mustache*/
/*global $*/
/*global CloseableView*/
/*global _superApply*/
/*global fixBootstrapButtonEvent*/
/*global Backbone*/
(function () {

	Toolkit.validateConfig = function (config, schema) {
		var original = JSON.stringify(config);
		// Preprocess schema to make validation less effort
		var preprocessSchema = function (schema) {
			if (!schema.type) {
				// Not sure about this, since it doesn't quite follow validictory, but we don't support "any"
				schema.type = "object";
			}
			schema.required = schema.required !== false;
			if (schema.pattern) {
				schema.pattern = schema.pattern.replace(/\{,5\}/g, "{0,5}");
			}
			schema.blank = schema.blank === true;
			if (schema.properties) {
				_.each(schema.properties, function (value) {
					preprocessSchema(value);
				});
			}
			if (schema.additionalProperties) {
				preprocessSchema(schema.additionalProperties);
			}
			// TODO: patternproperties?
			if (schema.items) {
				// TODO: support array of item schemas?
				preprocessSchema(schema.items);
			}
			// TODO: additionalItems?
		};
		preprocessSchema(schema);

		config = sanitizeValueForSchema(schema, config);

		var problems = [];

		// Recursively validate
		// Some validation errors never actually occur because of the previous sanitization of config.
		var validateConfig = function (config, schema, path) {
			if (schema._hide) {
				return true;
			}
			if (config === undefined) {
				if (schema.required) {
					problems.push({
						message: "Missing required field",
						field: path
					});
					return false;
				}
				return true;
			} else {
				if (schema.type === "object") {
					var result = true;
					// Clone config object so we can mutate it with causing problems
					config = _.clone(config);
					// Loop through properties in schema
					if (schema.properties) {
						_.each(schema.properties, function (schema, property) {
							result = validateConfig(config[property], schema, _.flatten([path, property])) && result;
							delete config[property];
						});
					}
					if (schema.additionalProperties) {
						_.each(config, function (config, property) {
							result = result && validateConfig(config, schema.additionalProperties, _.flatten([path, property]));
						});
					} else if (schema.additionalProperties === false) {
						if (_.keys(config).length !== 0) {
							problems.push({
								message: "Additional properties not allowed",
								field: path
							});
							return false;
						}
					}
					return result;
					// Check any additional properties against additionalProperties
				} else if (schema.type === "array") {
					if (schema.items) {
						return _.reduce(_.map(config, function (config, index) {
							return validateConfig(config, schema.items, _.flatten([path, '['+index+']']));
						}), function (memo, result) {
							return memo && result;
						}, true);
					} else {
						return true;
					}
				} else if (schema.type === "string") {
					if (!schema.blank && config === "") {
						problems.push({
							message: "Field should not be blank",
							field: path
						});
						return false;
					}
					if (schema.pattern && !(new RegExp(schema.pattern).test(config))) {
						problems.push({
							message: "Field fails pattern validation",
							field: path
						});
						return false;
					}
					return true;
				} else if (schema.type === "boolean") {
					// Should have already been made valid by the config preprocessor
					return true;
				} else {
					// TODO: Unknown type, return true for now
					return true;
				}
			}
		};
		var valid = validateConfig(config, schema, []);
		return {
			problems: problems,
			valid: valid,
			changed: original != JSON.stringify(config),
			config: config
		};
	};

	/** the look */
	// Config templates live under: static/templates/config

	/** the feel */
	var ConfigurationView = CloseableView.extend({
		className: "configuration",

		initialize: function () {
			_superApply(ConfigurationView, this, 'initialize', arguments);
			// Make sure the top level of the schema makes sense
			if (!_.has(this.options.schema, "type")) {
				this.options.schema.type = "object";
			}
			if (!_.has(this.options.schema, "required")) {
				this.options.schema.required = true;
			}
			// Make sure we have a valid config to start with
			var newModel = sanitizeValueForSchema(this.options.schema, this.model);
			if (! _.isEqual(newModel, this.model)) {
				this.model = newModel;
				this.options.parent.save(this.model);
			}

			var sensitive = this.options.sensitive = {};
			var schemaProperties = this.options.schema.properties || {};
			_.each(schemaProperties, function (schema, name) {
				if (schema._sensitive) {
					sensitive[name] = true;
				}
			});
		},

		render: function () {
			this.$el.html("");
			this.$el.append(Toolkit.Templates['component/config/container']);
			this.walk(this.$(".form"),
					  this.options.schema,
					  this.model,
					  this.options.schema.name,
					  []);
			return this;
		},

		walk: function (context, schema, config, name, path) {
			if (schema._hide) { return; }

			path = path.concat(name);
			var self = this;

			if (_.isArray(schema.type)) { schema.type = "object"; } // TODO support schema.type polymorphism

			context = self.v14.dirty_walk(context, schema, config, name, path, self);
			if (!context) {
				return;
			}

			// {,5} appears to not be supported in JS, replace here so old config patterns work.
			if (schema.pattern) {
				schema.pattern = schema.pattern.replace(/\{,5\}/g, "{0,5}");
			}

			if (schema.type === "object" && schema.properties !== undefined) {
				context = self.renderObject(context, schema, config, name, path);
				if (_.isObject(config)) {
					_.each(ordered(schema), function (name) {
						self.walk(context, schema.properties[name], config[name] ? config[name] : "", name, path);
					});
				}
				if (schema.additionalProperties) {
					_.each(config, function (entry, n) {
						if (schema.properties && _.has(schema.properties, n)) {
							return; // skip fixed entries
						}
						// Make sure additionalProperties aren't made optional.
						schema.additionalProperties.required = true;
						var _context = self.renderAdditionalProperty(context, schema.additionalProperties, entry, n, path.concat(n));
						self.walk(_context, schema.additionalProperties, entry, n, path);
					});
				}
			} else if (schema.type === "array") {
				context = self.renderArray(context, schema, config, name, path);
				_.each(config, function (entry, n) {
					var _context = self.renderArrayItem(context, schema.items, entry, n, path.concat(n));
					// Make sure items aren't displayed as optional
					schema.items.required = true;
					self.walk(_context, schema.items, entry, n, path);
				});
			} else {
				self.renderProperty(context, schema, config ? config : "", name, path);
			}
		},

		titleForProperty: function (name, schema) {
			if (schema.title !== undefined) {
				return schema.title.toString();
			} else if (name !== undefined) {
				return name.toString().replace(/[_-]/g, ' ');
			} else {
				return undefined;
			}
		},

		renderObject: function (context, schema, config, name, path) {
			path = path.join("__");

			var node = $(Mustache.to_html(Toolkit.Templates['component/config/object'], {
				schema: schema,
				id: path,
				name: this.titleForProperty(name, schema),
				enabled: _.isObject(config)
			}));

			context.append(node);
			return $("div.theproperties", node);
		},

		renderAdditionalProperty: function (context) {
			var node = $(Mustache.to_html(Toolkit.Templates['component/config/additional_property'], {}));

			context.append(node);

			return $(".theitem", node);
		},

		renderArray: function (context, schema, config, name, path) {
			path = path.join("__");

			var node = $(Mustache.to_html(Toolkit.Templates['component/config/array'], {
				schema: schema,
				id: path,
				name: this.titleForProperty(name, schema)
			}));

			context.append(node);

			return $(".items", node);
		},

		renderArrayItem: function (context, schema, config, name, path) {
			path = path.join("__");

			var node = $(Mustache.to_html(Toolkit.Templates['component/config/item'], {
				schema: schema,
				id: path
			}));

			context.append(node);

			return $(".theitem", node);
		},

		renderProperty: function (context, schema, config, name, path) {
			path = path.join("__");
			var input = "";
			if (schema['enum']) {
				input = Mustache.to_html(Toolkit.Templates['component/config/input_enum'], {
					schema: schema,
					id: path,
					value: config,
					options: _.map(schema['enum'], function (option) {
						return { option: option, selected: option === config };
					})
				});
			} else if (schema.type == "object" && schema.properties === undefined) {
				input = Mustache.to_html(Toolkit.Templates['component/config/input_textarea'], {
					id: path,
					value: JSON.stringify(config, undefined, "\t"),
					required: schema.required !== false && schema.blank !== true
				});
			} else if (schema.type) {
				input = Mustache.to_html(Toolkit.Templates["component/config/input_" + schema.type], {
					schema: schema,
					id: path,
					value: config,
					required: schema.required !== false && schema.blank !== true,
					filepicker: (schema._filepicker || schema._dirpicker) ? {
						inputId: path,
						root: this.options.pickerStart,
						type: schema._filepicker ? "file" : "dir"
					} : undefined
				});
			}

			var propertyTemplate = Toolkit.Templates['component/config/property_' + schema.type] || Toolkit.Templates['component/config/property'];

			context.append(Mustache.to_html(propertyTemplate, {
				schema: schema,
				name: this.titleForProperty(name, schema),
				input: input,
				id: path
			}));
			return context;
		},

		events: _.extend({
			"click .add": "onClickAdd",
			"click .remove": "onClickRemove",
			"change": "onChange",
			"change .optionalsection": "onToggleOptionalSection",
			"change .saveintoolconfig": "onToggleSaveInToolConfig"
		}, CloseableView.prototype.events),

		onClickAdd: function (e) {
			fixBootstrapButtonEvent(e);
			var path = $(e.target).data('path').split('__');
			this.model = addItem(path, this.model, this.options.schema);
			this.options.parent.save(this.model);
			this.render();
		},

		onClickRemove: function (e) {
			fixBootstrapButtonEvent(e);
			var path = $(e.target).data('path').split('__');
			this.model = removeItem(path, this.model);
			this.options.parent.save(this.model);
			this.render();
		},

		onChange: function (e) {
			var value;
			console.log(e.target);
			if (e.target.type === "checkbox") {
				value = e.target.checked;
			} else if (e.target.tagName === "TEXTAREA") {
				try {
					value = JSON.parse(e.target.value);
				} catch (ex) {}
			} else {
				value = e.target.value;
			}
			var path = e.target.id ? e.target.id.split('__') : [];
			console.log("onChange -> ", path, value);
			this.model = setItem(path, this.model, this.options.schema, value);
			this.options.parent.save(this.model);
			console.log("onChange saved", path, value, this.model);
		},

		onToggleSaveInToolConfig: function (e) {
			// prevent this.onChange() from firing
			e.stopPropagation();

			var sensitive = !e.target.checked;
			var attributePath = $(e.target).attr('data-attribute-path').split('__')[1];
			this.options.sensitive[attributePath] = sensitive;
		},

		onToggleOptionalSection: function (e) {
			var path = $(e.target).data('path');
			console.log("onToggleOptionalSection -> ", path, e.target.checked);
			if (!e.target.checked) {  // remove section from config
				path = path.split('__');
				this.model = removeSection(path, this.model);
				this.options.parent.save(this.model);
				// TODO: Don't do a full render
				this.render();
			} else {
				path = path.split('__');
				this.model = addSection(path, this.model, this.options.schema);
				this.options.parent.save(this.model);
				// TODO: Don't do a full render
				this.render();
			}
			return false;
		},

		// Filthy fallbacks for v14
		v14 : {
			dirty_walk: function (context, schema, config, name, path, self) {
				if (schema._v14_premium_module) {
					context = self.v14.renderPremiumModule(context, schema, config, name, path);
					return null;
				}  else {
					return context;
				}
			},
			renderPremiumModule: function (context, schema, config, name, path) {
				path = path.join("__");
				var html = Mustache.to_html(Toolkit.Templates['component/config/v14_premium_module'], {
					schema: schema,
					name: name,
					value: config,
					id: path,
				});
				context.append(html);
				return context;
			}
		}
	});

	/** the logic */

	// Return a valid config value for a schema (based on an optional provided value)
	function sanitizeValueForSchema(schema, value) {
		if (schema._hide) {
			// Don't modify hidden fields
			return value;
		}
		if (schema.type === "object") {
			if ((value === null || value === undefined) && schema.required === false) {
				return undefined;
			}
			if (!_.isObject(value)) {
				// We know this is required, create at least an empty object
				value = {};
			}
			if (_.isObject(schema.properties)) {
				_.each(schema.properties, function (subSchema, name) {
					var subValue = sanitizeValueForSchema(subSchema, value[name]);
					if (subValue !== undefined) {
						value[name] = subValue;
					} else {
						delete value[name];
					}
				});
			}
			if (schema.additionalProperties !== undefined) {
				var propertyKeys;
				if (_.isObject(schema.properties)) {
					propertyKeys = _.keys(schema.properties);
				} else {
					propertyKeys = [];
				}
				var currentKeys = _.keys(value);
				_.each(_.difference(currentKeys, propertyKeys), function (key) {
					if (schema.additionalProperties === false) {
						delete value[key];
					} else {
						var subValue = sanitizeValueForSchema(schema.additionalProperties, value[key]);
						if (subValue !== undefined) {
							value[key] = subValue;
						} else {
							delete value[key];
						}
					}
				});
			}
			return value;
		} else if (schema.type === "array") {
			if ((value === null || value === undefined) && schema.required === false) {
				return undefined;
			}
			if (!_.isArray(value)) {
				value = [];
			}
			if (_.isObject(schema.items)) {
				var newValue = [];
				_.each(value, function (item) {
					var subValue = sanitizeValueForSchema(schema.items, item);
					if (subValue !== undefined) {
						newValue.push(subValue);
					}
				});
				value = newValue;
			}
			return value;
		} else if (schema.type === "boolean") {
			if (schema.required !== false) {
				value = value === true;
			} else {
				value = (value === true) ? true : undefined;
			}
		} else if (schema.type === "string") {
			if (!_.isString(value)) {
				// Make sure we at least have a string
				if (schema.required && _.isArray(schema['enum'])) {
					value = _.first(schema['enum']);
				} else {
					value = "";
				}
			}
			if (schema.required !== false) {
				if (schema.blank === true) {
					return value;
				} else {
					if (value.length === 0) {
						return value;
					} else {
						return value;
					}
				}
			} else {
				if (schema.blank === true) {
					return value;
				} else {
					if (value.length === 0) {
						return undefined;
					} else {
						return value;
					}
				}
			}
		}
		return value;
	}

	// sets the item at 'path' to 'value'
	function setItem (path, config, schema, value) {
		var nodeSchema = digSchema(_.tail(path), schema);
		var node = path.pop();

		value = sanitizeValueForSchema(nodeSchema, value);

		if (node === undefined) {
			config = value;
		} else {
			var target = digConfig(path, config);

			if (value === undefined) {
				delete target[node];
			} else {
				target[node] = value;
			}
		}
		return config;
	}

	// adds a new instance of the item at 'path' to the config
	function addItem (path, config, schema) {
		var nodeSchema = digSchema(_.tail(path), schema);
		var parent = path.pop();
		var target = digConfig(path, config);
		if (!_.isArray(target[parent])) {
			target[parent] = [];
		}
		target = target[parent];
		target.push(sanitizeValueForSchema(nodeSchema.items));
		return config;
	}

	// removes item at 'path' from config
	function removeItem (path, config) {
		var new_config = digConfig(_.initial(path), config);
		new_config.splice(_.last(path), 1);
		return config;
	}

	// remove section at 'path' from config
	function removeSection (path, config) {
		var target = _.last(path);
		var new_config = digConfig(_.initial(path), config);
		console.log("removeSection - trimming: ", path, new_config, target);
		if (new_config) {
			delete new_config[target];
		} else {
			console.log("removeSection - no config for: ", path);
		}
		return config;
	}

	// add an optional section
	function addSection (path, config, schema) {
		var nodeSchema = digSchema(_.tail(path), schema);
		var node = path.pop();
		var target = digConfig(path, config);

		target[node] = sanitizeValueForSchema(nodeSchema, {});

		return config;
	}

	// return true if 'n' is a number for any type of 'n'
	function is_definitely_a_number(n) {
		return !isNaN(n);
	}

	// convert an element path to a schema entry
	function digSchema (path, schema) {
		var value = _.reduce(path, function(schema, node) {
			if (!schema[node] && is_definitely_a_number(node) && schema.type === "array") {
				return schema.items;
			} else if (schema.type === "object" && schema.name === node) {
				return schema.properties;
			} else if (schema[node] && schema[node].type === "array") {
				return schema[node];
			} else if (schema.type === "object" && schema.properties[node]) {
				return schema.properties[node];
			} else if (schema[node]) {
				return schema[node];
			}
			console.log("case: oops schema", node, schema);
			return schema;
		}, schema);
		return value;
	}

	// convert an element path to a config entry
	function digConfig (path, config, schema) {
		path = _.rest(path); // configs are routed one level shallower than schemas
		var value = _.reduce(path, function(_config, node) {
			if (_config && _config[node]) {
				return _config[node];
			}
			console.log("oops config", node, config, schema);
			return undefined;
		}, config);
		return value;
	}

	// returns a list of property names in the given schema ordered by the '_order' tag
	function ordered (schema) {
		var names = _.has(schema, "properties") ? _.keys(schema.properties) : [];
		names = _.sortBy(names, function (name) {
			return _.has(schema.properties[name], "_order") ? schema.properties[name]._order : name;
		});
		return names;
	}

	/** premium module support to hold us over until 2.0 */
	Toolkit.Models.Premium = Toolkit.Models.RpcModel.extend({
		method: 'api.get',
		args: {
			cookies: Toolkit.cookies,
			method: 'premium'
		},
		inject: {
			modules: function (modules) {
				return new Backbone.Collection(modules);
			}
		}
	});
	Toolkit.premium = new Toolkit.Models.Premium();

	Toolkit.events.on('cookies', function () {
		Toolkit.registerPreReadyFunction(function (done) {
			Toolkit.premium.load();
			Toolkit.premium.once('loaded', done);
		});
	});


	/** exports */
	Toolkit.Views.ConfigurationView = ConfigurationView;
})();
