/*global Toolkit*/
/*global $*/
/*global _*/
/*global Backbone*/
/*global _superApply*/
/*global Mustache*/
(function () {
	/**
	 * Wraps an RPC in a Backbone model so views can easily observe it
	 */
	Toolkit.Models.ConsoleTask = Backbone.Model.extend({
		toJSON: function () {
			// RPC is a circular object so don't return it for JSON stringification.
			var attr = $.extend({}, this.attributes);
			delete attr['rpc'];
			return attr;
		},

		initialize: function (attributes, options) {
			var preventNavigation = true;
			if (options && 'preventNavigation' in options) {
				preventNavigation = options.preventNavigation;
			}

			this.set({
				'status': 'running',
				'output': this.get('output') || new Backbone.Collection(),
				'verbose': this.get('verbose') || false,
				'restarting': false,
				'unexpectedError': false
			});

			var levelOverride;
			var typeOverride;

			var me = this;
			var rpc = this.get('rpc');

			rpc.on('question', function (e, respond) {
				me.trigger('question', e, respond);
			});

			var forgeLogReceived = false;
			rpc.on('log', function (e) {
				// Default values
				var level = e.level;
				var content = e.message;
				var type = "tools";

				// Special cases for platform specific output
				// Android
				var match = content.match(/(.)\/Forge {3}\([0-9 ]+\): (.*)/);
				if (match) {
					level = 'INFO';
					switch (match[1]) {
					case "E":
						level = 'ERROR';
						break;
					case "W":
						level = 'WARNING';
						break;
					case "D":
						level = 'DEBUG';
						break;
					case "V":
						// Don't display Android verbose output
						return;
					case "I":
						level = 'INFO';
						break;
					}
					content = match[2];
					type = "forge";
				}

				// iOS
				match = content.match(/Forge(?:Inspector)?\[[^\]]+\] \[([^\]]+)\] (.*)/);
				if (match) {
					levelOverride = match[1];
					content = match[2];
					typeOverride = "forge";
				}

				// Hack for multiline iOS
				if (levelOverride) {
					level = levelOverride;
				}
				if (typeOverride) {
					type = typeOverride;
				}

				if (type == 'forge' && !forgeLogReceived) {
					forgeLogReceived = true;
					if (me.has('onForgeLogReceived')) {
						me.get('onForgeLogReceived')();
					}
				}

				me.get('output').add({
					level: level,
					content: content,
					type: type
				});
			});

			rpc.on('success', function () {
				if (me.get('restarting')) {
					me._runAgain();
				} else {
					me.set('status', 'success');
				}
			});

			rpc.on('error', function (e) {
				console.error('Task error: ', e.message);
				me.set('error', e);
				if (me.get('restarting')) {
					me._runAgain();
				} else {
					if (e.error_type == "CallInterrupted") {
						me.set('status', 'finished');
						me.get('output').add({
							level: 'INFO',
							content: 'Task finished: interrupted by user',
							type: 'tools'
						});
					} else {
						me.set('status', 'failed');
						me.get('output').add({
							level: 'ERROR',
							content: e.message,
							type: 'tools'
						});
					}

					me.get('output').add({
						level: 'DEBUG',
						content: e.traceback,
						type: 'tools'
					});

					if (!e.expected) {
						me.set('unexpectedError', true);
					}
				}
			});

			if (preventNavigation) {
				Toolkit.preventNavigationFromTask(me);
			}
		},

		messageWhenPreventingNavigation: function () {
			var message = this.get('messageWhenPreventingNavigation');
			if (message) {
				return message;
			}
			return 'Leaving the current page may interrupt the currently running task: ' + this.get('rpc').methodName;
		},

		killable: function () {
			var status = this.get('status');
			return status === 'running';
		},

		/**
		 * Send a kill signal to the task
		 */
		kill: function () {
			if (this.killable()) {
				this.set('status', 'killing');
				this.get('rpc').interrupt();
			}
		},

		/**
		 * Restart the task immediately, or send a kill signal and schedule
		 * a restart when it's finished.
		 */
		restart: function () {
			if (this.get('status') === 'running') {
				// kill and prepare to run again
				this.kill();
				this.set('restarting', true);
			} else {
				// run again immediately
				this._runAgain();
			}
		},

		// precondition:
		// current RPC has finished, so we don't have to worry about it pushing into console output
		_runAgain: function () {
			var oldRpc = this.get('rpc');

			// clear output
			this.get('output').reset([]);

			// create a new rpc call and output to observe
			this.set({
				rpc: Toolkit.rpcSession.remoteCall(oldRpc.methodName, oldRpc.params)
			});

			this.initialize();
		},

		outputAsText: function () {
			return this.get('output').map(function (line) {
				return '['+line.attributes.level+'] '+line.attributes.content;
			}).join("\n");
		}
	});


	var ViewWithTemplate = Toolkit.Views.ViewWithTemplate;

	Toolkit.Views.ConsoleControlsView = ViewWithTemplate.extend({
		className: 'console-controls',
		template: 'component/console/controls',

		templateAttributes: function () {
			var attributes = this.modelAttributes();
			attributes.killable = this.model.killable();
			return attributes;
		}
	});

	var ConsoleView = Toolkit.Views.ConsoleView = ViewWithTemplate.extend({
		className: 'console',
		template: 'component/console/layout',

		events: {
			"click .kill": "_killClicked",
			"click .again": "_againClicked",
			"click .report": "_errorReportClicked",
			"change .verbose": "_verbosityClicked"
		},

		_killClicked: function (e) {
			e.preventDefault();
			e.stopPropagation();
			if (this.model.killable()) {
				this.model.kill();
			}
		},

		_againClicked: function (e) {
			e.preventDefault();
			this.model.restart();
		},

		_errorReportClicked: function (e) {
			e.preventDefault();
			Toolkit.sendErrorReport(this.model);
		},

		_verbosityClicked: function (e) {
			var verbose = $(e.currentTarget).attr('checked');
			this.model.set('verbose', verbose);
		},

		initialize: function () {
			_superApply(ConsoleView, this, 'initialize', arguments);

			_.bindAll(this, '_addLine', '_addAll', '_verbosityClicked');

			// watch for changes in task state
			this.model.on('change', this._settingsChanged, this);
			this.model.on('question', this._handleQuestion, this);

			// flag used to indicate whether we should immediately run the console task
			this._runAgainOnError = false;

			// display output in console
			this.model.get('output')
				.on('add', this._addLine, this)
				.on('reset', this._addAll, this);

			// element to append lines to
			this.$lines = null;
		},

		_handleQuestion: function (e, respond) {
			var modal = this._createModal(e);

			modal.on('submit', function (response) {
				respond(response.model, response.sensitive);
			});

			modal.on('cancel', function () {
				respond(null);
			});
		},

		_createModal: function (e) {
			var modal = new Toolkit.Views.ModalSchemaForm({
				model: e.schema
			});
			modal.render();
			modal.modalView.show();
			return modal;
		},

		_settingsChanged: function () {
			// toggle DEBUG level lines
			if (this.model.get('verbose')) {
				$('.log-level-DEBUG', this.$lines).show();
			} else {
				$('.log-level-DEBUG', this.$lines).hide();
			}

			// render controls
			var controls = this.controls = new Toolkit.Views.ConsoleControlsView({
				model: this.model
			}).render();

			this.$('#console-controls').html(controls.el.innerHTML);
			var checkbox = $(this.el).find('input').get(0);
			checkbox.checked = this.model.get('verbose');
		},

		_addLine: function (line) {
			var template = Toolkit.Templates['component/console/line'];

			// First make sure the content is escaped
			var content = Mustache.to_html("{{content}}", {
				content: line.attributes.content
			});
			// Then make any links clickable
			content = content.replace(/\[(https?:\/\/[^\]]+)\]/g, '<a href="$1" target="_blank">$1</a>');

			// Then render the full line
			var templateAttributes = _.extend(line.attributes, {
				content: content
			});
			var html = Mustache.to_html(template, templateAttributes);
			var $html = $(html);

			if (line.attributes.level == 'DEBUG' && !this.model.get('verbose')) {
				$html.hide();
			}

			var lines = this.$lines.get(0);
			if ((lines.scrollHeight - (lines.scrollTop + this.$lines.height())) < 50) {
				setTimeout(function () {
					lines.scrollTop = lines.scrollHeight;
				});
			}
			this.$lines.append($html);
		},

		_addAll: function () {
			this.$lines = this.$('.output');
			this.$lines.html('');
			this.model.get('output').each(this._addLine);
		},

		render: function () {
			ViewWithTemplate.prototype.render.apply(this, arguments);
			this._addAll();
			this._settingsChanged();

			return this;
		}
	});
}());
