/* global _, Toolkit */
(function () {
	Toolkit.dependencies = {};

	function moduleVersionTemplate(moduleVersion) {
		return "<strong>"+moduleVersion.module+" v"+moduleVersion.version+"</strong>";
	}

	var version_in_range = Toolkit.dependencies.version_in_range = function (version, min, max) {
		if (!version) {
			return false;
		}
		var i;
		var ver_parts = version.split(".");
		if (min) {
			var min_parts = min.split(".");
			for (i = 0; i < Math.max(ver_parts.length, min_parts.length); i++) {
				if ((ver_parts[i] | 0) < (min_parts[i] | 0)) {
					return false;
				}
				if ((ver_parts[i] | 0) > (min_parts[i] | 0)) {
					break;
				}
			}
		}
		if (max) {
			var max_parts = max.split(".");
			for (i = 0; i < Math.max(ver_parts.length, max_parts.length); i++) {
				if ((ver_parts[i] | 0) > (max_parts[i] | 0)) {
					return false;
				}
				if ((ver_parts[i] | 0) < (max_parts[i] | 0)) {
					break;
				}
			}
		}
		return true;
	};

	Toolkit.dependencies.check_dependencies_and_clashes = function (dependencies, module_versions, platform_version) {
		var dependencyResults = Toolkit.dependencies.check_dependencies(dependencies, module_versions, platform_version);

		// don't bother checking for clashes if dependencies need fixing
		if (dependencyResults["status"] !== "ok") {
			return dependencyResults;
		}

		// check for clashes in resolved config
		var resolvedModuleVersions = dependencyResults["module_versions"];
		var foundClashes = {};
		var problemDicts = [];

		for (var i = 0; i < resolvedModuleVersions.length; i++) {
			var resolvedModule = resolvedModuleVersions[i];
			var dependencyInfoForResolvedModule = _.findWhere(dependencies, {module: resolvedModule["module"]});

			var clashesForResolvedModule = dependencyInfoForResolvedModule["clashes"];
			for (var j = 0; j < clashesForResolvedModule.length; j++) {
				var clashForResolvedModule = clashesForResolvedModule[j];

				// see if this clash is in the resolved modules
				var clashingModule = _.findWhere(resolvedModuleVersions, {
					module: clashForResolvedModule["module"],
					version: clashForResolvedModule["version"]
				});

				if (clashingModule) {
					// will need to record a clash each way
					if (!foundClashes[resolvedModule["module"]]) {
						foundClashes[resolvedModule["module"]] = {};
					}

					if (!foundClashes[clashingModule["module"]]) {
						foundClashes[clashingModule["module"]] = {};
					}

					var foundClash = foundClashes[resolvedModule["module"]];
					
					// skip if clash already recorded between these modules
					if (foundClash[clashingModule["module"]]) {
						continue;
					}

					// record both directions of the clash
					foundClashes[resolvedModule["module"]][clashingModule["module"]] = true;
					foundClashes[clashingModule["module"]][resolvedModule["module"]] = true;

					problemDicts.push({
						"message": "Incompatible modules: "+moduleVersionTemplate(resolvedModule)+" and "+moduleVersionTemplate(clashingModule),
						"tags": ["module-"+resolvedModule["module"], "module-"+clashingModule["module"]]
					});
				}
			}
		}

		if (_.isEmpty(problemDicts)) {
			return dependencyResults;
		}

		return {
			"status": "fail",
			// TODO: code review, is the "problems" field necessary/used?
			"problems": _.pluck(problemDicts, 'message'),
			"problem_dicts": problemDicts
		};
	};

	Toolkit.dependencies.check_dependencies = function (dependencies, module_versions, platform_version) {
		if (module_versions.length === 0) {
			return {"status": "ok", "module_versions": []};
		}
		// If we have a specific platform version then filter dependencies to only include compatible modules
		var possibleDependencies = _.filter(dependencies, function (dep) {
			if (platform_version !== undefined && platform_version.substr(0,1) === "v") {
				return version_in_range(
					platform_version.substr(1),
					dep.min_platform_version && dep.min_platform_version.substr(1),
					dep.max_platform_version && dep.max_platform_version.substr(1)
				);
			}
			return true;
		});
		var recurse_dependencies = function (add, to_add, added) {
			if (add.version) {
				// Specific version
				var add_dep = _.findWhere(possibleDependencies, {"module": add.module, "version": add.version});
				if (!add_dep || !add_dep.dependencies) {
					return false;
				}
				add_dep = add_dep.dependencies;
				_.each(add_dep, function (dep, key) {
					to_add.unshift(_.extend({}, {
						"module": key,
						"minimum_version": dep.minimum_version,
						"maximum_version": dep.maximum_version
					}));
				});
			} else {
				// Range of versions
				var added_version = _.findWhere(added, {"module": add.module});
				if (added_version) {
					// Already added, check version
					if (!version_in_range(added_version.version, add.minimum_version, add.maximum_version)) {
						return false;
					}
				} else {
					// Not added, find set of addable versions
					var result = false;
					_.each(_.where(possibleDependencies, {"module": add.module}), function (version) {
						if (result !== false) {
							return;
						}
						if (version_in_range(version.version, add.minimum_version, add.maximum_version)) {
							var cur_added = _.clone(added);
							cur_added.push({
								"module": version.module,
								"version": version.version
							});
							result = recurse_dependencies({
								"module": version.module,
								"version": version.version
							}, to_add, cur_added);
						}
					});
					return result;
				}
			}
			if (to_add.length !== 0) {
				return recurse_dependencies(to_add.pop(), to_add, added);
			} else {
				return added;
			}
		};

		var list_missing_dependencies = function (module_versions) {
			var problems = [];
			_.each(module_versions, function (version) {
				var deps = _.findWhere(possibleDependencies, {"module": version.module, "version": version.version});
				var allDeps = _.findWhere(dependencies, {"module": version.module, "version": version.version});
				if (!deps) {
					if (allDeps) {
						var dep = _.findWhere(dependencies, {'module': version.module, 'version': version.version});
						console.warn(dep);
						var message = "";
						if (dep.min_platform_version && dep.max_platform_version) {
							message = "between versions <strong>"+dep.min_platform_version+"</strong> and <strong>"+dep.max_platform_version+"</strong>";
						} else if (dep.min_platform_version) {
							message = "is at least version <strong>"+dep.min_platform_version+"</strong>";
						} else {
							message = "is at most version <strong>"+dep.max_platform_version+"</strong>";
						}
						problems.push({
							message: "Module "+moduleVersionTemplate(version)+" requires your platform is "+message+", your platform is version <strong>"+platform_version+"</strong>.",
							tags: ["module-"+version.module]
						});
					} else {
						problems.push({
							message: "Module "+moduleVersionTemplate(version)+" does not exist or you do not have access to it.",
							tags: ["module-"+version.module]
						});
					}
				} else {
					_.each(deps.dependencies, function (dep, key) {
						dep.module = key;
						var added_dep = _.findWhere(module_versions, {"module": dep.module});
						if (!added_dep) {
							var potential_deps = _.where(possibleDependencies, {"module": dep.module});
							var found = false;
							_.each(potential_deps, function (potential_dep) {
								if (version_in_range(potential_dep.version, dep.minimum_version, dep.maximum_version)) {
									found = true;
								}
							});
							if (found) {
								problems.push({
									message: "Module "+moduleVersionTemplate(version)+" requires module <strong>"+dep.module+"</strong> between versions <strong>"+(dep.minimum_version || 'any')+"</strong> and <strong>"+(dep.maximum_version || 'any')+"</strong>, but it cannot be added, due to conflicting or missing dependencies in all possible versions of that module.",
									tags: ["module-"+version.module]
								});
							} else {
								problems.push({
									message: "Module "+moduleVersionTemplate(version)+" requires module <strong>"+dep.module+"</strong> between versions <strong>"+(dep.minimum_version || 'any')+"</strong> and <strong>"+(dep.maximum_version || 'any')+"</strong>, but it cannot be added, either because it does not exist or you do not have access to it.",
									tags: ["module-"+version.module]
								});
							}
						} else if (!version_in_range(added_dep.version, dep.minimum_version, dep.maximum_version)) {
							problems.push({
								message: "Module "+moduleVersionTemplate(version)+" requires module <strong>"+dep.module+"</strong> is between versions <strong>"+(dep.minimum_version || 'any')+"</strong> and <strong>"+(dep.maximum_version || 'any')+"</strong> but you have version <strong>"+added_dep.version+"</strong>.",
								tags: ["module-"+version.module, "module-"+dep.module]
							});
						}
					});
				}
			});
			return problems;
		};
		var added = _.clone(module_versions);
		var result = recurse_dependencies(module_versions.pop(), module_versions, added);
		if (!result) {
			var problem_dicts = list_missing_dependencies(added);
			return {"status": "fail", "problems": _.pluck(problem_dicts, 'message'), "problem_dicts": problem_dicts};
		} else if (result.length === added.length) {
			return {"status": "ok", "module_versions": result};
		} else {
			var modules = _.map(added, function (module) { return module.module; });
			return {"status": "missing", "module_versions": result, "missing": _.filter(result, function (module) { return !_.contains(modules, module.module); })};
		}
	};

	// Methods to help deal with dependencies
	// Intended to be used both by apps and inspector config
	Toolkit.Mixins.Dependencies = {
		disableModule: function (module) {
			this.toggleModule(module, false);
		},
		enabledModule: function (module) {
			this.toggleModule(module, true);
		},
		toggleModule: function (module, enabled) {
			var me = this;
			var config = this.getConfig();
			if (enabled === undefined) {
				enabled = config.modules[module].disabled;
			}
			this.getProject().loadDependencies(function (project) {
				if (!config.modules || config.config_version < 4) {
					return;
				}
				var modules = [];
				var modulesBefore = [];
				_.each(config.modules, function (data, cur_module) {
					if (!data.disabled) {
						modulesBefore.push({
							"module": cur_module,
							"version": data.version
						});
						if (cur_module != module) {
							modules.push({
								"module": cur_module,
								"version": data.version
							});
						}
					}
					if (enabled && cur_module == module) {
						modules.push({
							"module": cur_module,
							"version": data.version
						});
					}
				});
				var before = Toolkit.dependencies.check_dependencies_and_clashes(project.get("dependencies"), modulesBefore, me.getPlatformVersion());
				var after = Toolkit.dependencies.check_dependencies_and_clashes(project.get("dependencies"), modules, me.getPlatformVersion());

				var toggleModule = function () {
					var config = me.getConfig();
					if (enabled) {
						delete config.modules[module]['disabled'];
					} else {
						config.modules[module]['disabled'] = true;
					}
					me.saveConfig();
				};

				if (before.status == "ok" && after.status != "ok") {
					if (after.status == "missing") {
						after.problems = [];
						_.each(after.missing, function (missing) {
							after.problems.push("Missing dependency: "+missing.module+" v"+missing.version);
						});
					}
					var modal = new Toolkit.Views.ModalDialog({
						title: 'Toggle "' + module + '"?',
						content: "<p>After toggling this module builds will not work due to the following problems:<br>"+after.problems.join("<br>")+"</p>",
						actions: [{
							type: "cancel",
							text: "Cancel"
						}, {
							type: "danger",
							text: enabled ? "Enable" : "Disable",
							onClicked: function () {
								toggleModule();
								modal.hide();
							}
						}]
					});
					modal.show();
				} else {
					toggleModule();
				}
			});
		},
		removeModule: function (module) {
			var me = this;
			var config = this.getConfig();
			this.getProject().loadDependencies(function (project) {
				if (!config.modules || config.config_version < 4) {
					return;
				}
				var modules = [];
				var modulesBefore = [];
				_.each(config.modules, function (data, cur_module) {
					if (!data.disabled) {
						modulesBefore.push({
							"module": cur_module,
							"version": data.version
						});
						if (cur_module != module) {
							modules.push({
								"module": cur_module,
								"version": data.version
							});
						}
					}
				});
				var before = Toolkit.dependencies.check_dependencies_and_clashes(project.get("dependencies"), modulesBefore, me.getPlatformVersion());
				var after = Toolkit.dependencies.check_dependencies_and_clashes(project.get("dependencies"), modules, me.getPlatformVersion());
				
				if (after.status == "missing") {
					after.problems = [];
					_.each(after.missing, function (missing) {
						after.problems.push("Missing dependency: "+missing.module+" v"+missing.version);
					});
				}
				var content = "<p>If you remove this module, the functionality it provides will no longer be available in your app.</p>";
				if (before.status == "ok" && after.status != "ok") {
					content += "<p>After removing this module builds will not work due to the following problems:<br>"+after.problems.join("<br>")+"</p>";
				}
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Remove "' + module + '"?',
					content: content,
					actions: [{
						type: "cancel",
						text: "Cancel"
					}, {
						type: "danger",
						text: "Remove",
						onClicked: function () {
							var config = me.getConfig();
							delete config.modules[module];
							me.saveConfig();
							modal.hide();
						}
					}]
				});
				modal.show();
			});
		},
		addModule: function (module, version) {
			this.updateModule(module, version);
		},
		updateModule: function (module, version) {
			var me = this;
			var config = this.getConfig();
			var adding = true;
			if (config.modules[module]) {
				if (config.modules[module].version == version) {
					return;
				}
				adding = false;
			}
			this.getProject().loadDependencies(function (project) {
				if (!config.modules || config.config_version < 4) {
					return;
				}
				var modules = [];
				var modulesBefore = [];
				_.each(config.modules, function (data, cur_module) {
					if (!data.disabled) {
						modulesBefore.push({
							"module": cur_module,
							"version": data.version
						});
						if (cur_module != module) {
							modules.push({
								"module": cur_module,
								"version": data.version
							});
						}
					}
				});
				modules.push({
					"module": module,
					"version": version
				});
				var before = Toolkit.dependencies.check_dependencies_and_clashes(project.get("dependencies"), modulesBefore, me.getPlatformVersion());
				var after = Toolkit.dependencies.check_dependencies_and_clashes(project.get("dependencies"), modules, me.getPlatformVersion());
				
				if (after.status == "missing") {
					after.problems = [];
					_.each(after.missing, function (missing) {
						after.problems.push("Missing dependency: "+missing.module+" v"+missing.version);
					});
				}
				var updateModule = function () {
					var config = me.getConfig();
					if (config.modules[module]) {
						config.modules[module].version = version;
					} else {
						config.modules[module] = {
							version: version
						};
					}
					me.saveConfig();
				};
				var modal, content;
				if (before.status == "ok" && after.status == "missing") {
					// This is something we can automatically fix
					content = "<p>This module version requires the following dependencies are added/enabled, this will be done automatically. Some of them may require configuration:<br>"+after.problems.join("<br>")+"</p>";
					modal = new Toolkit.Views.ModalDialog({
						title: (adding ? 'Add "' : 'Change version for "') + module + '"?',
						content: content,
						actions: [{
							type: "cancel",
							text: "Cancel"
						}, {
							type: "warning",
							text: adding ? "Add" : "Change version",
							onClicked: function () {
								var config = me.getConfig();
								_.each(after.missing, function (missing) {
									if (config.modules[missing.module]) {
										delete config.modules[missing.module].disabled;
									} else {
										config.modules[missing.module] = {
											version: missing.version
										};
									}
								});
								updateModule();
								modal.hide();
							}
						}]
					});
					modal.show();
				} else if (before.status == "ok" && after.status != "ok") {
					content = "<p>After changing this module's version builds will not work due to the following problems:<br>"+after.problems.join("<br>")+"</p>";
					modal = new Toolkit.Views.ModalDialog({
						title: (adding ? 'Add "' : 'Change version for "') + module + '"?',
						content: content,
						actions: [{
							type: "cancel",
							text: "Cancel"
						}, {
							type: "danger",
							text: adding ? "Add" : "Change version",
							onClicked: function () {
								updateModule();
								modal.hide();
							}
						}]
					});
					modal.show();
				} else {
					updateModule();
				}
			});
		},

		checkDependenciesAndSetStatus: function (remote, statusType, dependencies, modules, platformVersion) {
			var local = this;
			var config = local.getConfig();

			if (!config) {
				return;
			}
			
			_.each(config.modules, function (data, module) {
				if (!data.disabled) {
					modules.push({
						"module": module,
						"version": data.version
					});
				}
			});

			var dependencyCheck = Toolkit.dependencies.check_dependencies_and_clashes(
				dependencies,
				modules,
				platformVersion
			);
			
			if (dependencyCheck.status == "ok") {
				//pass
			} else if (dependencyCheck.status == "missing") {
				var missing = [];
				_.each(dependencyCheck.missing, function (module) {
					missing.push(module.module+" ("+module.version+")");
				});
				remote.setStatus({
					type: statusType,
					id: 'dependencies',
					severity: 'error',
					message: 'Missing or disabled module dependencies: '+missing.join(", ")+".",
					tags: [],
					action: {
						text: 'Add missing dependencies',
						callback: function () {
							_.each(dependencyCheck.missing, function (module) {
								if (config.modules[module.module]) {
									delete config.modules[module.module].disabled;
								} else {
									config.modules[module.module] = {
										version: module.version
									};
								}
							});
							Toolkit.rpcSession.remoteCall('file.write', {
								path: local.getConfigPath(),
								data: JSON.stringify(config, undefined, "\t")
							});
						}
					}
				});
			} else if (dependencyCheck.status == "fail") {
				var count = 0;
				_.each(dependencyCheck.problem_dicts, function (problem) {
					remote.setStatus({
						type: statusType,
						id: 'dependencies-'+count++,
						severity: 'error',
						message: problem.message,
						tags: problem.tags
					});
				});
			}
		}
	};
	/*

	var dependencies = [
		{"module": "a", "version": "0.1", "dependencies": {"b": {}}},
		{"module": "b", "version": "0.1", "dependencies": {"c": {}, "d": {}, "e": {maximum_version: "0.3"}}},
		{"module": "c", "version": "0.1", "dependencies": {"a": {}}},
		{"module": "d", "version": "0.1", "dependencies": {"e": {minimum_version: "0.2"}}},
		{"module": "e", "version": "0.3", "dependencies": {}},
		{"module": "e", "version": "0.2", "dependencies": {}},
		{"module": "e", "version": "0.1", "dependencies": {}},

		{"module": "f", "version": "0.1", "dependencies": {"g": {}}},
	];
	

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "e", 'version': "0.1"}
	]), null, 4));

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "a", 'version': "0.1"},
		{'module': "b", 'version': "0.1"},
		{'module': "c", 'version': "0.1"},
		{'module': "d", 'version': "0.1"},
		{'module': "e", 'version': "0.2"}
	]), null, 4));

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "a", 'version': "0.1"},
		{'module': "b", 'version': "0.1"},
		{'module': "c", 'version': "0.1"},
		{'module': "d", 'version': "0.1"},
		{'module': "e", 'version': "0.3"}
	]), null, 4));

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "a", 'version': "0.1"}
	]), null, 4));

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "a", 'version': "0.1"},
		{'module': "e", 'version': "0.1"}
	]), null, 4));

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "d", 'version': "0.1"},
		{'module': "e", 'version': "0.1"}
	]), null, 4));


	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "f", 'version': "0.1"}
	]), null, 4));

	console.log(JSON.stringify(check_dependencies(dependencies, [
		{'module': "g", 'version': "0.1"}
	]), null, 4));*/
}());
