/*global $*/
/*global _*/
/*global Toolkit*/
/*global printStackTrace*/

// helper to ratelimit API calls during development
_.rateLimit = function(f, rate, async) {
	var queue = [];
	var busy = false;
	var processQueue = function() {
		if (queue.length) {
			busy = true;
			_.delay(function() {
				if (async) {
					var fn = queue.shift();
					if (fn) {
						_.defer(function() { fn.call(); });
					}
				} else {
					queue.shift().call();
				}
				processQueue();
			}, rate);
		} else {
			busy = false;
		}
	};
	return function() {
		var args = _.map(arguments, function(e) { return e; });
		queue.push(_.bind.apply(this, [f, this].concat(args)));
		if (!busy) { processQueue(); }
	};
};
//var ajax_throttled = _.rateLimit($.ajax, 250, true);

function RpcCall(session, callId, methodName, params) {
	this.session = session;
	this.callId = callId;
	this.callbacks = {};
	this.methodName = methodName;
	this.params = params;
}

RpcCall.prototype.interrupt = function () {
	$.ajax({
		url: '/session/interruptcall',
		type: 'POST',
		data: {
			sessionId: this.session.sessionId,
			callId: this.callId
		},
		dataType: 'json',
		success: function (response) {
			if (!response.ok) {
				console.error('Failed to interrupt RPC', response);
			}
		},

		error: function (e) {
			console.error('Failed to interrupt RPC', e);
		}
	});
};

RpcCall.prototype.on = function (eventType, callback) {
	if (!this.callbacks[eventType]) {
		this.callbacks[eventType] = [];
	}
	this.callbacks[eventType].push(callback);
	return this;
};

RpcCall.prototype.trigger = function (event) {
	var me = this;
	var type = event.type;
	var callbacks = this.callbacks[type];

	if (callbacks) {
		var callback = function (data, sensitiveData) {
			me.input({
				eventId: event.eventId,
				data: data,
				sensitive_data: sensitiveData
			});
		};
		for (var i = 0; i < callbacks.length; ++i) {
			if (window.DEBUG) {
				callbacks[i].call(this, event, callback);
			} else {
				try {
					callbacks[i].call(this, event, callback);
				} catch (err) {
					var message = 'Error in callback for '+this.methodName+': '+err.message;
					console.error(message, err);
					Toolkit.showError({
						message: message,
						traceback: printStackTrace({e: err}).join("\n")
					});
				}
			}
		}
	} else {
		var fallback = this.defaults[type];
		if (fallback) {
			fallback.call(this, event, function (data) {
				me.input({
					eventId: event.eventId,
					data: data
				});
			});
		}
	}
};

RpcCall.prototype.defaults = {
	error: function (e) {
		if (e.extra && e.extra.content) {
			try {
				var content = JSON.parse(e.extra.content);
				if (content.type && content.type === 'login_required') {
					Toolkit.checkAuth();
					return;
				}
			} catch (ex) {}
		}
		Toolkit.captureException(new Error(e.message), { tags: { "subsystem": "rpc" } });
		Toolkit.showError({
			message: 'Error in remote call to '+this.methodName+': '+e.message,
			traceback: e.traceback
		});
	}
};

RpcCall.prototype.input = function (response) {
	var me = this;
	$.ajax({
		url: '/session/input',
		type: 'POST',
		dataType: 'json',
		data: {
			sessionId: me.session.sessionId,
			callId: me.callId,
			response: JSON.stringify(response)
		}
	});
};

function UUID() {
	return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g, function(c) {
		var r = Math.random() * 16 | 0;
		var v = c == "x" ? r : (r & 0x3 | 0x8);
		return v.toString(16);
	}).toUpperCase();
}

function RpcSession() {
	this.sessionId = undefined;
	this.emitters = {};
	this.errorCount = 0;
}

RpcSession.prototype.init = function (success) {
	var me = this;
	$.ajax({
		url: '/session/new?chrome_is_evil=' + Math.random(),
		type: 'POST',
		dataType: 'json',
		success: function (response) {
			me.sessionId = response.content;
			console.log("Created session with id: ", me.sessionId);
			me.get();
			success && success();
		}
	});
};

RpcSession.prototype.get = function () {
	var me = this;

	var error = function (e) {
		console.error('Failed to get events', e);
		me.errorCount++;

		if (me.errorCount > 5) {
			Toolkit.hasTimedout = true;
			if (!Toolkit.hasUpdated) {
				window.location.reload();
			}
		} else {
			me.get();
		}
	};

	$.ajax({
		url: '/session/pollevents',
		type: 'POST',
		data: {
			sessionId: me.sessionId
		},
		dataType: 'json',
		success: function (response) {
			if (response.ok) {
				me.errorCount = 0;
				var events = response.content;
				for (var i = 0; i < events.length; ++i) {
					var event = events[i];

					var emitter = me.emitters[event.callId];

					if (event.type === 'updatestart') {
						console.log("Received updatestart event from call");
						Toolkit.showUpdateViewAndShutdownOnSuccess(emitter);
					}

					if (!emitter) {
						console.error("Got an event for a call we didn't recognise: ", event);
						continue;
					}
					emitter.trigger(event);

					// we don't want to keep a reference around to the RpcCall once it's finished
					if (event.type === 'success' || event.type === 'error') {
						me.emitters[event.callId].trigger({type:'finish'});
						delete me.emitters[event.callId];
					}
				}

				me.get();
			} else {
				error(response);
			}
		},

		error: error
	});
};


RpcSession.prototype.remoteCall = function (methodName, params) {
	params = params || {};
	var me = this;
	var callId = UUID();
	var emitter = new RpcCall(this, callId, methodName, params);

	this.emitters[callId] = emitter;

	if (this.calls === undefined) {
		this.calls = [];
	}
	
	var makeCall = function () {
		if (me.calls.length === 0) {
			return;
		}
		$.ajax({
			url: '/session/call',
			type: 'POST',
			data: {
				sessionId: me.sessionId,
				calls: JSON.stringify(me.calls)
			},
			dataType: 'json',
			success: function (response) {
				if (!response.ok) {
					console.error('RPC failed', response);
				}
			},

			error: function (e) {
				console.error('RPC failed', e);
			}
		});
		me.calls = [];
	};

	if (window.DEBUG) {
		setTimeout(function () {
			me.calls.push({
				callId: callId,
				methodName: methodName,
				params: params
			});
			makeCall();
		}, 500 + 500 * Math.random());
	} else {
		this.calls.push({
			callId: callId,
			methodName: methodName,
			params: params
		});
		setTimeout(makeCall,10);
	}

	return emitter;
};
