/*global Toolkit*/
/*global _, $*/
(function () {
	Toolkit.Views.StatusBarView = Toolkit.Views.ViewWithTemplate.extend({
		template: 'component/statusbar',
		initialize: function () {
			this.listenTo(this.model, 'change:status', this.render);
		},
		templateAttributes: function () {
			return _.extend(this.model.getStatusSummary(), {
				"url": this.options.url
			});
		}
	});

	Toolkit.Views.StatusPageView = Toolkit.Views.ViewWithTemplate.extend({
		events: {
			"click button[data-type]": "clickedButton",
		},
		clickedButton: function (e) {
			this.model.getStatusForTemplating()[$(e.target).attr('data-type')][$(e.target).attr('data-index')].action.callback();
		},
		templateAttributes: function () {
			var attributes = this.modelAttributes();
			attributes.status = this.model.getStatusForTemplating();
			return attributes;
		},
		render: function () {
			Toolkit.Views.StatusPageView.__super__.render.apply(this, arguments);
			return this;
		},
		delegateSubViewEvents: function () {
			// no subviews to manage
		}
	});

	Toolkit.Mixins.Status = {
		/*
		 * @param type The type of status (i.e. 'development' or 'versions' for plugins)
		 * @param id A unique string to reference this status (so it can be updated later)
		 * @param severity One of 'error', 'warning', 'good' or 'info'
		 * @param message A html snippet to display to the user for this status
		 * @param action A dictionary containing text and callback for an action button related to the status
		 * tags An array of string tags to be able to find the status by later
		 */
		setStatus: function (type, id, severity, message, action) {
			var tags = [];
			if (arguments.length === 1) {
				// Dict form
				var params = type;
				type = params.type;
				id = params.id;
				severity = params.severity;
				message = params.message;
				action = params.action;
				tags = params.tags;
			}
			var statusDict = this.get('status');
			if (!(type in statusDict)) {
				statusDict[type] = {};
			}
			statusDict[type][id] = {
				id: id,
				severity: severity,
				message: message,
				action: action,
				tags: tags
			};
			this.set('status', statusDict);
			this.trigger('change change:status');
		},

		clearStatus: function (type, id) {
			var statusDict = this.get('status');
			if (type in statusDict) {
				if (id in statusDict[type]) {
					delete statusDict[type][id];
				}
			}
			this.set('status', statusDict);
			this.trigger('change change:status');
		},

		clearAllStatusesWithType: function (type) {
			var statusDict = this.get('status');
			if (type in statusDict) {
				delete statusDict[type];
			}
			this.set('status', statusDict);
			this.trigger('change change:status');
		},

		getStatusSummary: function () {
			var statusDict = JSON.parse(JSON.stringify(this.get('status')));
			var allStatuses = _.reduce(_.values(statusDict), function (memo, obj) { return _.union(memo, _.values(obj)); }, []);
			var problemCount = {
				total: 0,
				warning: 0,
				error: 0
			};
			var lastError, lastWarning;
			_.each(allStatuses, function (status) {
				if (status.severity == 'warning') {
					lastWarning = status.message;
					problemCount.warning++;
					problemCount.total++;
				} else if (status.severity == 'error') {
					lastError = status.message;
					problemCount.error++;
					problemCount.total++;
				}
			});
			if (problemCount.total === 0) {
				return {
					'severity': 'good',
					'message': 'No problems found',
					'class': 'success'
				};
			} else if (problemCount.total === 1) {
				if (problemCount.error === 1) {
					return {
						'severity': 'error',
						'message': 'Error: '+lastError,
						'class': 'error'
					};
				} else {
					return {
						'severity': 'warning',
						'message': 'Warning: '+lastWarning,
						'class': 'warning'
					};
				}
			} else {
				if (problemCount.error === 0) {
					return {
						'severity': 'warning',
						'message': problemCount.warning+' warnings found, see status page for details.',
						'class': 'warning'
					};
				} else {
					return {
						'severity': 'error',
						'message': problemCount.error+' errors'+(problemCount.warning > 0 ? ' and '+problemCount.warning+' warnings' : '')+' found, see status page for details.',
						'class': 'error'
					};
				}
			}
		},

		getStatusForTemplating: function () {
			var statusDict = {};
			_.each(this.get('status'), function (element, type) {
				var count = 0;
				statusDict[type] = _.sortBy(_.map(_.values(element), function (problem) {
					problem.type = type;
					problem.index = count++;
					if (problem.severity == 'info') {
						problem['class'] = 'info';
						problem.icon = 'info-sign';
					} else if (problem.severity == 'good') {
						problem['class'] = 'success';
						problem.icon = 'ok-sign';
					} else if (problem.severity == 'warning') {
						problem['class'] = 'warning';
						problem.icon = 'warning-sign';
					} else if (problem.severity == 'error') {
						problem['class'] = 'error';
						problem.icon = 'exclamation-sign';
					}
					return problem;
				}), function (problem) {
					if (problem.severity == 'error') {
						return 0;
					} else if (problem.severity == 'warning') {
						return 1;
					} else if (problem.severity == 'info') {
						return 2;
					} else if (problem.severity == 'good') {
						return 4;
					}
				});
			});
			return statusDict;
		},

		getStatusesWithTag: function (tag) {
			var statuses = [];
			_.each(this.get("status"), function (element) {
				_.each(element, function (problem) {
					if (problem.tags.indexOf(tag) !== -1) {
						if (problem.severity == 'info') {
							problem['class'] = 'info';
							problem.icon = 'info-sign';
						} else if (problem.severity == 'good') {
							problem['class'] = 'success';
							problem.icon = 'ok-sign';
						} else if (problem.severity == 'warning') {
							problem['class'] = 'warning';
							problem.icon = 'warning-sign';
						} else if (problem.severity == 'error') {
							problem['class'] = 'error';
							problem.icon = 'exclamation-sign';
						}
						statuses.push(problem);
					}
				});
			});

			return statuses;
		}
	};
}());
