/*global Toolkit*/
/*global Backbone*/
/*global _*/
(function () {
	var Models = Toolkit.Models;

	_(Backbone.Model.prototype).extend({
		// set an attribute that you don't want to be serialized by toJSON
		setTemp: function (key, value) {
			if (!this._temp) {
				this._temp = {};
			}
			this._temp[key] = 1;
			this.set(key, value);
			return this;
		},

		// remove an attribute that you added with setTemp
		deleteTemp: function (key) {
			delete this._temp[key];
			this.set(key, undefined);
			return this;
		},

		// omit attributes set through setTemp
		toJSON: function () {
			var args = [this.attributes];
			args = args.concat(this._temp ? _(this._temp).keys() : []);
			args.push("local");
			return _.omit.apply(undefined, args);
		}
	});

	var RpcModel = Models.RpcModel = Backbone.Model.extend({
		collectErrors: false,

		initialize: function () {
			RpcModel.__super__.initialize.apply(this, arguments);
			this._temp = {};
		},

		// omit parent model to prevent mutual recursion with parent's toJSON
		// toJSON any child collections
		toJSON: function () {
			var superAttrs = RpcModel.__super__.toJSON.call(this);

			if (!this.parent.attribute) {
				return superAttrs;
			}

			var attrs = _.omit(superAttrs, this.parent.attribute);

			for (var propName in this.children) {
				attrs[propName] = attrs[propName] && attrs[propName].toJSON();
			}

			return attrs;
		},

		loadSet: function (keyOrObject, valueOrOptions, maybeOptions) {
			var attrs, options;

			if (_.isObject(keyOrObject)) {
				attrs = keyOrObject;
				options = valueOrOptions;
			} else {
				(attrs = {})[keyOrObject] = valueOrOptions;
				options = maybeOptions;
			}

			// this will be used for the final this.set call
			var setter = {};
			for (var attribute in attrs) {
				var existingValue = this.get(attribute);
				var newValue = attrs[attribute];

				// if the attribute we're setting is meant to be a nested collection
				var childModel = this.children[attribute];
				if (childModel) {
					var me = this;
					// update each child to have a reference to their parent
					/*jshint -W083 */
					_(newValue).each(function (child) {
						child[childModel.prototype.parent.attribute] = me;
						child[childModel.prototype.parent.idAttribute] = me.id;
					});
				}

				if (existingValue instanceof Backbone.Collection) {
					existingValue.update(newValue);
					continue;
				}

				var injectFunction = this.inject[attribute];
				if (!existingValue && injectFunction) {
					newValue = injectFunction.call(this, newValue);
				}

				// if we're setting a new parent id:
				//   make sure the parent model exists
				//   store the parent in an attribute
				//   make sure we're in the collection for the parent
				if (!existingValue && this.parent.idAttribute === attribute) {
					// construct parent
					var Parent = Toolkit.Models[this.parent.model];
					var parentAttrs = {};
					parentAttrs[Parent.prototype.idAttribute] = newValue;
					var parent = new Parent(parentAttrs);

					// store reference to parent
					this.set(this.parent.attribute, parent);

					// ensure parent has collection
					if (!parent.get(this.parent.collectionAttribute)) {
						parent.loadSet(this.parent.collectionAttribute, []);
					}

					// ensure the app is in the right collection
					var parentCollection = parent.get(this.parent.collectionAttribute);
					if (!this.collection || this.collection !== parentCollection) {
						parentCollection.add(this);
					}

					// load data for the parent
					parent.load();
				}

				// Model.prototype.set doesn't trigger change events if
				// nothing has changed.
				setter[attribute] = newValue;
			}
			this.set(setter, options);

			return this;
		},

		invalidate: function () {
			this.load(true);
		},

		load: function (force) {
			// don't reload unless forced
			if (this.get('_loaded') && !force) {
				return this;
			}
			this.set('_loaded', undefined);

			// if currently updating, no need to start another one
			if (this.get('_loading')) {
				return this;
			}
			this.set('_loading', true);

			var rpc = Toolkit.rpcSession.remoteCall(this['method'], _.result(this, 'args'));
			var me = this;
			rpc.on('success', function (e) {
				if (e.data._invalidate_on) {
					_.each(e.data._invalidate_on,function (invalidate_on) {
						if (Toolkit.invalidationRules[invalidate_on] === undefined) {
							Toolkit.invalidationRules[invalidate_on] = [me];
						} else if (Toolkit.invalidationRules[invalidate_on].indexOf(me) === -1) {
							Toolkit.invalidationRules[invalidate_on].push(me);
						}
					});
				}

				var newAttributes = e.data['data'];

				_.each(newAttributes, function (newValue, attribute) {
					me.loadSet(attribute, newValue);
				});
			});

			if (this.collectErrors) {
				rpc.on('error', function (e) {
					me.set('error', e.message);
				});
			}

			rpc.on('finish', function () {
				me.set({
					'_loading': false,
					'_loaded': true
				});
				me.trigger('loaded');
			});

			return this;
		},

		// methods of this object will be used when creating new attributes
		// in this model
		inject: {},

		// details about ownership
		children: {},
		parent: {}
	});

	// Model that can watch files/directories.
	// files is a dict of property_name to path
	// When the file at path changes it will be read in, parsed as json and the results put in property_name
	// directories is a similar dict, where the property will be populated with the live directory tree under the given path
	Models.WatchingModel = Backbone.Model.extend({
		initialize: function () {
			this._fileWatches = [];
			this._loadedFiles = {};
			if (!this.get('files')) {
				this.set('files', {});
			}
			if (!this.get('rawfiles')) {
				this.set('rawfiles', {});
			}
			this._setupFileWatches();
			this.on("change:files", this._setupFileWatches, this);
			this.on("change:rawfiles", this._setupFileWatches, this);
			if (!this.get('directories')) {
				this.set('directories', {});
			}
			this._setupDirWatches();
			this.on("change:directories", this._setupDirWatches, this);
		},
		_setupFileWatches: function () {
			var me = this;
			var setupFileWatch = function (path, name) {
				// Load any new files
				if (!me._loadedFiles[name] || me._loadedFiles[name] != path) {
					me._updateFilesForPath(path);
					me._loadedFiles[name] = path;
				}
				// Add watches on any new paths
				if (me._fileWatches.indexOf(path) === -1 ) {
					var watch = Toolkit.rpcSession.remoteCall('watch.file', {
						path: path,
						dir_path: me.watchDirPath ? me.watchDirPath() : null
					});
					watch.on('change', function () {
						me._updateFilesForPath(path);
					});
					me._fileWatches.push(path);
				}
			};

			_.each(this.get('files'), setupFileWatch);
			_.each(this.get('rawfiles'), setupFileWatch);
		},
		// load the contents at the given path and update all files/rawfiles for
		// that path
		_updateFilesForPath: function (path) {
			var me = this;
			var rpc = Toolkit.rpcSession.remoteCall('file.contents', {path: path});
			rpc.on('success', function (e) {
				var parsedFile = null;

				_.each(me.get('files'), function (cur_path, name) {
					// no point updating files if we couldn't parse as JSON
					if (parsedFile && parsedFile.failed) {
						return;
					}

					if (path === cur_path) {
						// at least one file expects JSON, need to parse the file as JSON
						if (!parsedFile) {
							try {
								parsedFile = {contents: JSON.parse(e.data)};
							} catch (jsonErr) {
								// Prevent JSON error from affecting non-json raw files
								parsedFile = {failed: true};
								console.error("Failed to parse JSON file: "+path);
								return;
							}
						}

						me.set(name, parsedFile.contents);
					}
				});

				_.each(me.get('rawfiles'), function (cur_path, name) {
					if (path === cur_path) {
						me.set(name, e.data);
					}
				});
			});
		},
		_dirWatches: [],
		_loadedDirs: {},
		_setupDirWatches: function () {
			var me = this;
			_.each(this.get('directories'), function (path, name) {
				// Load any new files
				if (!me._loadedDirs[name] || me._loadedDirs[name] != path) {
					me._dirChanged(path);
					me._loadedDirs[name] = path;
				}
				// Add watches on any new paths
				if (me._dirWatches.indexOf(path) === -1 ) {
					var watch = Toolkit.rpcSession.remoteCall('watch.directory', {path: path});
					watch.on('change', function (e) {
						var watchEvent = e['watch_event'];
						me.trigger('watch:'+watchEvent['event_type']+":"+watchEvent['src_path'], watchEvent);
						me.trigger('watch:'+watchEvent['event_type'], watchEvent);
						me._dirChanged(path);
					});
					me._dirWatches.push(path);
				}
			});
		},
		_dirChanged: function (path) {
			var me = this;
			var rpc = Toolkit.rpcSession.remoteCall('directory.list_recursive', {path: path});
			rpc.on('success', function (e) {
				var result = e.data;
				_.each(me.get('directories'), function (cur_path, name) {
					if (path === cur_path) {
						me.set(name, result);
					}
				});
			});
		}
	});

	Models.AuthStatus = Backbone.Model.extend();
}());


