/*global Toolkit, _, CloseableView, Backbone, $, endsWith, makeSet*/
(function () {
    Toolkit.Models.Clash = Toolkit.Models.RpcModel.extend({
        idAttribute: 'id',
        method: 'api.get',
        args: function () {
            return {
                cookies: Toolkit.cookies,
                method: 'clash/'+this.id
            };
        }
    });

    Toolkit.Models.Module = Toolkit.Models.RpcModel.extend({
        idAttribute: 'name',
        cacheName: 'Module',
        method: 'api.get',
        args: function () {
            return {
                cookies: Toolkit.cookies,
                method: 'module/'+this.id
            };
        },
        parent: {
            attribute: 'project',
            idAttribute: 'project_id',
            model: 'Project',
            collectionAttribute: 'modules'
        },
        initialize: function () {
            Toolkit.Models.Module.__super__.initialize.call(this);
            this.set("config_modules", new Toolkit.Collections.ModuleVersionCollection());
            this.set('status', {});

            this.doWhenAvailable(this, 'local', function (local) {
                this.listenTo(this, 'change:project_id', function () {
                    this.updateConfigModules();
                    local.checkModuleStatus();
                });
                this.listenTo(this.get("local"), 'change:inspector_config change:manifest change:identity change:config_schema', function () {
                    this.updateConfigModules();
                    local.checkModuleStatus();
                });
            });
        },
        updateConfigModules: function () {
            var local = this.get("local");
            if (!local) {
                return;
            }

            var projectId = this.get("project_id");
            var thisModuleName = this.get('name');
            var config = local.getConfig();

            if (!config || !projectId) {
                return;
            }

            var config_modules = this.get("config_modules");
            var idsToRemove = config_modules.pluck('id');
            var modelsToAdd = [];

            _.each(config.modules, function (data, moduleName) {
                // inspector_config.json must contain config
                // for this module, but that has a separate UI
                if (moduleName === thisModuleName) {
                    return;
                }

                var id = projectId+"-"+moduleName+"-"+data.version;
                idsToRemove = _.without(idsToRemove, id);
                modelsToAdd.push(new Toolkit.Models.ModuleVersion({
                    id: id,
                    module__name: moduleName,
                    version: data.version,
                    project_id: projectId
                }).load());
            });

            var modelsToRemove = _.map(idsToRemove, function (id) {
                return config_modules.get(id);
            });

            config_modules.add(modelsToAdd);
            config_modules.remove(modelsToRemove);
        },
        updateVersion: function (versionToChange, action, success) {
            // precondition: versions have been loaded
            var me = this;
            var rpc = Toolkit.rpcSession.remoteCall('api.get', {
                method: "module/"+me.get('name')+"/version/"+versionToChange+"/"+action,
                cookies: Toolkit.cookies
            });

            rpc.on('success', function (e) {
                var versions = _.clone(me.get('versions'));
                _.each(versions, function (version) {
                    if (version.version == versionToChange) {
                        version.current = e.data.current;
                        version.approval = e.data.approval;
                        version.published = e.data.published;
                    } else if (e.data.current) {
                        version.current = false;
                    }
                });
                me.set('versions', versions);
                success && success();
            });
        }
    });

    _.extend(Toolkit.Models.Module.prototype, Toolkit.Mixins.Status, {
        // provide sensible default descriptions for when there are no
        // statuses for these
        getStatusForTemplating: function () {
            var status = Toolkit.Mixins.Status.getStatusForTemplating.call(this);
            if (!status.module || !status.module.length) {
                status.module = [{
                    icon: 'ok-sign',
                    class: 'success',
                    message: 'No problems detected with local module files.'
                }];
            }

            if (!status.config || !status.config.length) {
                status.config = [{
                    icon: 'ok-sign',
                    class: 'success',
                    message: 'No problems detected with inspector config.'
                }];
            }

            return status;
        }
    });

    // represents a module on the local file system
    Toolkit.Models.LocalModule = Toolkit.Models.WatchingModel.extend({
        idAttribute: 'path',

        cacheName: 'LocalModule',

        // return path to a watched file relative to the module root
        relativeFilePath: function (fileName) {
            var fullPath = this.get('files')[fileName];
            var prefix = this.get('path');

            if (!fullPath || !prefix) {
                return undefined;
            }

            return fullPath.slice(prefix.length + 1);
        },

        // load details about the local module
        load: function () {
            var me = this;

            var rpc = Toolkit.rpcSession.remoteCall('modules.state', {
                cookies: Toolkit.cookies,
                path: this.get('path')
            });
            rpc.on('success', function (e) {
                me.set(e.data);
            });
        },

        getName: function () {
            if (this.get("identity") && this.get("identity").name) {
                return this.get("identity").name;
            } else if (this.get("manifest") && this.get("manifest").name) {
                return this.get("manifest").name;
            } else {
                return false;
            }
        },

        initialize: function () {
            Toolkit.Models.LocalModule.__super__.initialize.apply(this, arguments);
            this.load();

            var me = this;

            // set attributes on LocalModule and Module that allow getting between them
            this.on('change:manifest change:identity', function () {
                if (this.getName()) {
                    var module = new Toolkit.Models.Module({name: this.getName()});
                    var previousModule = me.get('_module');
                    if (previousModule !== module) {
                        if (previousModule) {
                            me.stopListening(previousModule);
                            previousModule.set("local", undefined);
                        }
                        module.set("local", me);
                        me.setTemp('_module', module);

                        // some of our validation depends on the state of the Module
                        me.listenTo(module, 'change:versions', this.checkForUploadedVersion);
                        me.listenTo(module, 'change:project', this.checkModuleStatus);
                    }
                }
            });

            this.on('watch:modified watch:created watch:deleted', function (e) {
                me.updateFileHash(e['src_path']);
            });
            this.on('watch:moved', function (e) {
                me.updateFileHash(e['src_path']);
                me.updateFileHash(e['dest_path']);
            });
            this.set("module_directory_hashes", {});
            this.set("module_directory_validations", {});
            this.once('change:module_directory', function () {
                var recurse_directory = function (dir, path) {
                    _.each(dir, function (value, key) {
                        if (value === "file") {
                            if (path.length > 0) {
                                me.updateFileHash(path.join('/') + '/' + key);
                            } else {
                                me.updateFileHash(key);
                            }
                        } else {
                            var curPath = _.clone(path);
                            curPath.push(key);
                            recurse_directory(value, curPath);
                        }
                    });
                };
                recurse_directory(me.get('module_directory'), []);
            });
            this.on('change:module_directory', function () {
                me.updateValidationMessages();
            });
            this.on('change:manifest change:identity', function () {
                me.checkModuleStatus();
            });
            me.checkModuleStatus();
        },
        // return a dict of features for the platform version used with this
        // module
        platformFeatures: function () {
            var manifest = this.get("manifest");

            if (!manifest) {
                return {};
            }

            var platforms = Toolkit.platforms.get("platforms");

            if (!platforms) {
                return {};
            }

            var platformVersionModel = platforms.get(manifest.platform_version);

            if (!platformVersionModel) {
                return {};
            }

            var platformFeatures = platformVersionModel.get("features");
            return makeSet.apply(this, platformFeatures);
        },
        canUseFeature: function (featureName) {
            var platformFeatures = this.platformFeatures();
            return !!platformFeatures[featureName];
        },
        getPlatformVersion: function () {
            return this.get("manifest").platform_version;
        },
        // caller of getConfig expects to be able to modify the returned config
        // then call saveConfig() to persist the changes
        getConfig: function () {
            // inspector_config.json not supported
            // OR enough files haven't been loaded yet
            // in either case the caller should do nothing
            var config = this.get('inspector_config');
            var manifest = this.get('manifest');
            var manifestVersion = manifest && manifest['version'];

            if (config === undefined || manifestVersion === undefined) {
                return undefined;
            }

            var moduleName = this.getName();
            if (moduleName === false) {
                return undefined;
            }

            var needsSave = false;

            // platform version supports inspector config
            // inspector_config.json isn't there
            if (config === null) {
                config = {};
                needsSave = true;
            }

            if (!config.this_module) {
                config.this_module = {};
                needsSave = true;
            }

            if (!config.this_module['config']) {
                config.this_module['config'] = {};
                needsSave = true;
            }

            if (!config.modules) {
                config.modules = {};
                needsSave = true;
            }

            // so saveConfig() works later
            this.set('inspector_config', config, {silent: true});

            if (needsSave) {
                this.saveConfig();
            }

            return config;
        },
        getConfigPath: function () {
            return this.get("files").inspector_config;
        },
        saveConfig: function (config) {
            if (config === undefined) {
                config = this.getConfig();
            }
            var rpc = Toolkit.rpcSession.remoteCall('file.write', {
                path: this.getConfigPath(),
                data: JSON.stringify(config, undefined, "\t")
            });

            rpc.on('success', function () {
                $(".saved").fadeIn();
                setTimeout(function () {
                    $(".saved").fadeOut();
                }, 3000);
            });
        },
        getProject: function () {
            return new Toolkit.Models.Module({name: this.getName()}).get("project");
        },

        // Called when a file change event happens, hashes the file and triggers other updates if the hash has changed
        updateFileHash: function (path) {
            if (this.dontValidate) {
                return;
            }
            if (!path) {
                return;
            }
            var me = this;
            var path_segments = path.replace("\\", "/");
            path_segments = path_segments.split("/");
            var hashes = this.get("module_directory_hashes");
            _.each(_.initial(path_segments), function (segment) {
                if (!_.isObject(hashes[segment])) {
                    hashes[segment] = {};
                }
                hashes = hashes[segment];
            });
            var rpc = Toolkit.rpcSession.remoteCall('file.hash', {path: this.get("path") + "/module/" + path});
            rpc.on("success", function (e) {
                if (e.data !== null) {
                    if (hashes[_.last(path_segments)] != e.data) {
                        hashes[_.last(path_segments)] = e.data;
                        me.updateValidators(path_segments);
                        me.updateValidationMessages();
                    }
                } else if (_.last(path_segments) in hashes) {
                    delete hashes[_.last(path_segments)];
                    me.updateValidationMessages();
                }
            });
        },

        // Called when a file hash changes, updates any validator information for the file
        updateValidators: function (path) {
            if (this.dontValidate) {
                return;
            }
            var me = this;
            if (this.has("module_structure")) {
                var notfound = false;
                var structure = this.get("module_structure");
                _.each(path, function (segment) {
                    if (!notfound) {
                        if (structure.type === "directory" && structure['children']) {
                            structure = structure['children'][segment];
                            if (!structure) {
                                notfound = true;
                            }
                        }
                    }
                });
                if (!notfound && structure['validators']) {
                    var validations = this.get("module_directory_validations");
                    _.each(_.initial(path), function (segment) {
                        if (!_.isObject(validations[segment])) {
                            validations[segment] = {};
                        }
                        validations = validations[segment];
                    });
                    validations[_.last(path)] = [];
                    validations = validations[_.last(path)];
                    _.each(structure['validators'], function (extra_params, validator) {
                        var params = _.extend(extra_params, {
                            file: path.join("/"),
                            manifest: me.get('manifest')
                        });
                        var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
                            'python_path': me.get('module_dynamic_path'),
                            'method_name': 'validate.' + validator,
                            params: params
                        });
                        rpc.on("error", function (e) {
                            validations.push({message: "Validation for '"+path.join("/")+"' failed: "+e.message, type: "error"});
                            me.updateValidationMessages();
                        });
                    });
                }
            }
        },

        // helper to get Module from LocalModule
        getModule: function () {
            return this.getName() && new Toolkit.Models.Module({name: this.getName()});
        },

        // Called when the file structure changes or a validation changes
        // This should not make any RPC calls (it should be synchronous)
        // Any RPC work should be done then trigger this method
        updateValidationMessages: function () {
            if (this.dontValidate) {
                return;
            }

            this.checkInspectorHash();
            var local = this;

            var module = local.getModule();
            if (!module) {
                return;
            }

            module.clearAllStatusesWithType('module');

            // Check platform_version
            if (!this.get('manifest')) {
                module.setStatus('module', 'no-local', 'error', 'No local copy of module.');
            } else if (!this.get('manifest')['platform_version']) {
                module.setStatus('module', 'no-platform', 'error', 'No platform version in manifest.');
            } else if ("undefined" === typeof(this.get('dynamic_platform_version'))) {
                // .trigger/platform_version.txt not read in yet
            } else if (this.get('dynamic_platform_version') != this.get('manifest')['platform_version']) {
                var platforms = Toolkit.platforms;
                if (platforms.has('platforms')) {
                    var platformModel = platforms.get('platforms').get(local.get('manifest')['platform_version']);
                    if (platformModel && platformModel.get('module_support')) {
                        module.setStatus('module', 'platform-changed', 'error', 'Platform version in manifest has changed, downloading new dynamic module code...');
                        local.dontValidate = true;
                        var rpc = Toolkit.rpcSession.remoteCall('modules.update_module_dynamic', {
                            cookies: Toolkit.cookies,
                            'path':  local.get('local_path'),
                            'new_platform_version': local.get('manifest')['platform_version']
                        });
                        rpc.on('success', function () {
                            // XXX: File watchers get upset by this, so we reload, should do something better.
                            window.location.reload();
                        });
                    } else {
                        module.setStatus('module', 'platform-bad', 'error', 'Platform version in manifest is not a supported platform version.');
                    }
                }
            } else if (!this.get('module_structure') || !this.get('module_directory')) {
                module.setStatus('module', 'no-structure', 'error', 'No module structure found to validate module. Try refreshing the browser, if this problem persists please contact support@trigger.io.');
            } else {
                var validateStructure = function (structure, schema, path) {
                    if (schema.validators) {
                        var validations = local.get("module_directory_validations");
                        var notfound = false;
                        _.each(path.split('/'), function (segment) {
                            if (!notfound) {
                                if (!_.isObject(validations[segment])) {
                                    notfound = true;
                                } else {
                                    validations = validations[segment];
                                }
                            }
                        });
                        if (!notfound) {
                            _.each(validations, function (validation_error) {
                                module.setStatus('module', Math.random()+"", validation_error.type, validation_error.message);
                            });
                        }
                    }
                    if (schema.type == "file") {
                        if (structure == "file") {
                            return;
                        } else {
                            module.setStatus('module', Math.random()+"", "error", "The directory '"+path+"' was expected to be a file.");
                        }
                    } else if (schema.type == "directory") {
                        if (structure == "file") {
                            module.setStatus('module', Math.random()+"", "error", "The file '"+path+"' was expected to be a directory.");
                            return;
                        }
                        if (path !== "") {
                            path += "/";
                        }
                        if (schema.children) {
                            for (var schema_child in schema.children) {
                                if (schema.children[schema_child].required && !structure[schema_child]) {
                                    // XXX: hacky way of improving missing .a / .jar message
                                    // this *should* be handled better with improved logic for what is and is not a valid structure for a module
                                    var message = "The path '" + path + schema_child + "' is required and was not found.";
                                    if (endsWith(schema_child, '.aar')) {
                                        message = "Built module '" + path + schema_child +
                                            "' is required: if you don't want to build for Android, remove the " + path +
                                            " directory.";
                                    } else if (endsWith(schema_child, 'module.a')) {
                                        message = "Built module '" + path + schema_child +
                                            "' is required: if you don't want to build for iOS, remove the " + path +
                                            " directory.";
                                    }
                                    module.setStatus('module', Math.random()+"", "error", message);
                                }
                            }
                        }
                        for (var child in structure) {
                            if (schema.children && schema.children[child]) {
                                validateStructure(structure[child], schema.children[child], path+child);
                            } else if (schema.additionalChildren !== undefined) {
                                if (schema.additionalChildren === false && child.substring(0,1) != '.') {
                                    module.setStatus('module', Math.random()+"", "warning", "The path '"+path+child+"' was found but not expected.");
                                } else {
                                    validateStructure(structure[child], schema.additionalChildren, path+child);
                                }
                            }
                        }
                    }
                };
                validateStructure(this.get('module_directory'), this.get('module_structure'), '');

                if (local.toJSON().targets.length === 0) {
                    module.setStatus('module', "no-targets", "error", "Neither an Android nor iOS module has been created.");
                }
            }

            local.checkForUploadedVersion();
        },

        checkForUploadedVersion: function () {
            var local = this;
            var module = this.getModule();
            var manifestVersion = local.get('manifest') && local.get('manifest').version;

            if (manifestVersion && module) {
                // lookup the Module for this LocalModule
                if (module.has('versions')) {
                    var manifestVersionAlreadyUploaded = _(module.get('versions')).where(
                        {version: manifestVersion}
                    ).length !== 0;

                    if (manifestVersionAlreadyUploaded) {
                        module.setStatus('module', "already-uploaded", "warning", "A " + manifestVersion + " version of your module has already been uploaded.");
                    }
                }
            }
        },

        checkInspectorHash: function () {
            if (this.dontValidate) {
                return;
            }
            if (!this.debouncedCheckInspector) {
                this.debouncedCheckInspector = _.debounce(function (me) {
                    if (me.dontValidate) {
                        return;
                    }
                    var check_inspector_hash = function (platform) {
                        var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
                            'python_path': me.get('module_dynamic_path'),
                            'method_name': 'inspector.check_'+platform+'_hash',
                            params: {}
                        });
                        rpc.on("success", function (e) {
                            me.set('inspector_'+platform, e.data);
                            var module = me.getName() && new Toolkit.Models.Module({name: me.getName()});
                            module && module.setStatus('inspector', platform, e.data.type, e.data.message);
                        });
                    };
                    if (me.toJSON().targets.length > 0) {
                        for (var platform in me.toJSON().targets) {
                            check_inspector_hash(me.toJSON().targets[platform]);
                        }
                    }
                    // Clear old status
                    me.set('inspector_ios', undefined);
                    me.set('inspector_android', undefined);

                    var name = me.get('manifest') && me.get('manifest').name;
                    var module = name && new Toolkit.Models.Module({name: name});

                    if (module) {
                        module.clearAllStatusesWithType("inspector");
                    }
                }, 250);
            }
            this.debouncedCheckInspector(this);
        },

        checkModuleStatus: function () {
            var local = this;
            var moduleName = this.getName();
            var module = moduleName && new Toolkit.Models.Module({name: moduleName});
            var manifest = local && local.get("manifest");
            var inspectorConfig = local && local.getConfig();

            if (!module || !local || !manifest || !inspectorConfig) {
                return;
            }

            module.clearAllStatusesWithType('config');

            if (manifest["platform_version"].substr(0,2) == "v1") {
                // Ignore for pre v2
                return;
            }

            // validate config using schema in local module files
            var localSchema = local.get("config_schema");
            if (localSchema) {
                var localModuleConfig = inspectorConfig['this_module']['config'];
                var v = Toolkit.validateConfig(localModuleConfig, localSchema); // TODO: can this mutate its arguments?
                if (!v.valid) {
                    var message = "<strong>" + moduleName + "</strong> module config issues:<ul><li>";
                    message += _.map(v.problems, function (problem) {
                        return "<strong>"+problem.field.join(".")+"</strong> "+problem.message;
                    }).join("</li><li>")+"</li></ul>";

                    module.setStatus({
                        'type': 'config',
                        'id': moduleName+"-config",
                        'severity': "error",
                        'message': message,
                        'tags': ['module-'+moduleName]
                    });
                }
            }

            // validate config for other modules to be included in inspector
            var moduleVersions = module.get("config_modules");
            this.doWhenAll(moduleVersions.models, '_loaded', function () {
                moduleVersions.each(function (moduleVersion) {
                    if (inspectorConfig.modules[moduleVersion.get("module__name")].disabled) {
                        return;
                    }

                    if (moduleVersion.get("type") == "public" && moduleVersion.get("version") != moduleVersion.get("current_version")) {
                        module.setStatus({
                            'type': 'config',
                            'id': moduleVersion.get("module__name")+"-version",
                            'severity': "warning",
                            'message': "<strong>" + moduleVersion.get("module__name") + "</strong> module is not set to the current version, current version is <strong>"+moduleVersion.get("current_version")+"</strong>, your version is <strong>"+moduleVersion.get("version")+"</strong>.",
                            'tags': ['module-'+moduleVersion.get("module__name")]
                        });
                    }

                    var schema = moduleVersion.get("schema");

                    if (_.isEmpty(schema)) {
                        // No config for module.
                        return;
                    }

                    var config = inspectorConfig.modules[moduleVersion.get("module__name")].config;
                    var validate = Toolkit.validateConfig(config, schema);

                    if (validate.changed) {
                        inspectorConfig.modules[moduleVersion.get("module__name")].config = validate.config;
                        local.saveConfig();
                    }

                    if (!validate.valid) {
                        var message = "<strong>" + moduleVersion.get("module__name") + "</strong> module config issues:<ul><li>";
                        message += _.map(validate.problems, function (problem) {
                            return "<strong>"+problem.field.join(".")+"</strong> "+problem.message;
                        }).join("</li><li>")+"</li></ul>";

                        module.setStatus({
                            'type': 'config',
                            'id': moduleVersion.get("module__name")+"-config",
                            'severity': "error",
                            'message': message,
                            'tags': ['module-'+moduleVersion.get("module__name")]
                        });
                    }
                });
                local.checkDependencies();
            });
        },

        /**
         * Check module's inspector config using dependencies specified in
         * manifest.json.
         */
        checkDependencies: function () {
            // if the module can't use inspector config, don't check dependencies
            if (!this.canUseFeature('modules-inspector-config')) {
                return;
            }

            var local = this;
            var module = this.getName() && new Toolkit.Models.Module({name: this.getName()});
            var project = module && module.get('project');

            // make sure enough has been loaded to check dependencies
            if (!module || !project) {
                return;
            }

            project.loadDependencies(function () {
                // override any existing dependency information
                // for this module version
                var dependencies = _.clone(project.get("dependencies"));

                var indexOfExisting;
                for (var i = 0; i < dependencies.length; ++i) {
                    if (dependencies[i].module === module.id) {
                        indexOfExisting = i;
                        break;
                    }
                }

                var dependencyInfoFromManifest = {
                    module: module.id,
                    version: local.get("manifest").version,
                    dependencies: local.get("manifest").dependencies,
                    clashes: []
                };

                if (indexOfExisting !== undefined) {
                    dependencies[indexOfExisting] = dependencyInfoFromManifest;
                } else {
                    dependencies.push(dependencyInfoFromManifest);
                }

                var modules = [{
                    "module": module.id,
                    "version": local.get("manifest").version
                }];

                local.checkDependenciesAndSetStatus(
                    module,
                    'config',
                    dependencies,
                    modules,
                    local.get("manifest").platform_version
                );
            });
        },

        toJSON: function () {
            var dirs = this.attributes['module_directory'] ? _(this.attributes['module_directory']).keys() : [];
            return _.extend({
                // Filter directories by possible targets
                targets: _.intersection(dirs, ["android", "ios"])
            }, this.attributes);
        }
    });

    _.extend(Toolkit.Models.LocalModule.prototype, Toolkit.Mixins.Dependencies);

    // collection of local modules on the file system
    Toolkit.Collections.LocalModuleCollection = Backbone.Collection.extend({
        model: Toolkit.Models.LocalModule
    });

    // model which watches local_modules.json and maintains a collection of
    // all LocalModules on the file system
    Toolkit.Models.LocalModulesFile = Toolkit.Models.WatchingModel.extend({
        // return LocalModule for Module
        getLocal: function (module) {
            var name = module.get('name');
            var available = this.availableNames();
            return available[name];
        },

        availableNames: function () {
            var available = {};
            this.get('collection').map(function (localModule) {
                // might be in the middle of fetching the manifest file,
                // can't say what it's name is yet in that case
                if (localModule.getName()) {
                    available[localModule.getName()] = localModule;
                }
            });
            return available;
        },

        initialize: function () {
            Toolkit.Models.WatchingModel.prototype.initialize.apply(this, arguments);
            this.set({
                collection: new Toolkit.Collections.LocalModuleCollection()
            });
            this.listenTo(this, 'change:localModules', function () {
                this.get('collection').update(
                    _(this.get('localModules').module_locations).map(function (path) {
                        return {path: path};
                    })
                );
            });
            var me = this;
            Toolkit.rpcSession.remoteCall('modules.remove_local_not_on_filesystem').on('success', function () {
                me.set('files', {
                    localModules: 'build-tools/local_modules.json'
                });
            });
        }
    });

    Toolkit.events.once('rpc', function () {
        Toolkit.registerPreReadyFunction(function (done) {
            Toolkit.localModulesFile = new Toolkit.Models.LocalModulesFile();
            Toolkit.localModulesFile.once('change:localModules', done);
        });
    });

    Toolkit.Views.ModuleCreateView = Toolkit.Views.FormView.extend({
        template: 'projects/create_module',

        submit: function (form) {
            var me = this;

            if (form.name.match(/^[a-z][a-z0-9_]+$/) === null) {
                me.tempModel.set('errors', {'name': ['Name must start with a lowercase letter and only contain lowercase letters, numbers and _.']});
                me._submitFinish();
                return;
            }

            var rpc = Toolkit.rpcSession.remoteCall('api.post', {
                method: 'module/',
                data: {
                    name: form.name,
                    project: this.model.get('id')
                },
                cookies: Toolkit.cookies
            });

            rpc.on('error', this._handleFormErrors());
            rpc.on('success', function (e) {
                me.model.get('modules').add(e.data.data);
                me.trigger("cancel");
            });
        }
    });

    Toolkit.Views.CreateModuleTemplate = Toolkit.Views.FormView.extend({
        template: 'modules/create_template',

        changedVersion: function (e) {
            if (_.indexOf(Toolkit.platforms.get("platforms").get($(e.target).val()).get("features"), 'modules-namespace') == -1) {
                this.$('#namespace').prop('disabled', true);
                this.$('#namespace').val(this.model.get("module").get("name"));
            } else {
                this.$('#namespace').prop('disabled', false);
            }
        },

        templateAttributes: function () {
            var superAttrs = Toolkit.Views.CreateModuleTemplate.__super__.templateAttributes.call(this);
            var platforms = Toolkit.platforms;
            var possiblePlatforms = [];
            var latestStable = platforms.get('platforms').get(platforms.get('latest_stable'));
            platforms.get('platforms').each(function (platform) {
                if (platform.get('module_support')) {
                    possiblePlatforms.push({
                        name: platform.get('version'),
                        selected: (platform === latestStable)
                    });
                }
            });

            return _(superAttrs).extend({
                name: this.model.get('module').get('name'),
                platforms: possiblePlatforms,
                namespace_disabled: _.indexOf(latestStable.get("features"), 'modules-namespace') == -1
            });
        },

        submit: function (form) {
            var me = this;
            var rpc = Toolkit.rpcSession.remoteCall('modules.create_module_template', {
                cookies: Toolkit.cookies,
                path: form.path,
                namespace: form.namespace,
                core_version: form['core_version'],
                name: me.model.get('module').get('name')
            });

            rpc.on('error', this._handleFormErrors());
            rpc.on('success', function (e) {
                var localModule = _(e.data).extend({
                    path: e.data['local_path']
                });
                Toolkit.localModulesFile.get('collection').add(localModule);
                me.trigger("cancel");
            });
            rpc.on('finish', function () {
                me._submitFinish();
            });
        }
    });
    Toolkit.Views.CreateModuleTemplate.prototype.events = _.extend({
        "change #core_version": "changedVersion"
    }, Toolkit.Views.CreateModuleTemplate.prototype.events);

    Toolkit.Views.ImportModuleTemplate = Toolkit.Views.FormView.extend({
        template: 'modules/import',

        submit: function (form) {
            var me = this;

            var possiblePlatforms = [];
            Toolkit.platforms.get('platforms').each(function (platform) {
                if (_.indexOf(platform.get('features'), 'modules-namespace') != -1) {
                    possiblePlatforms.push(platform.get('version'));
                }
            });

            var rpc = Toolkit.rpcSession.remoteCall('modules.import_module', {
                cookies: Toolkit.cookies,
                path: form.path,
                name: me.model.get('module').get('name'),
                platform_versions_with_namespace: possiblePlatforms
            });

            rpc.on('error', this._handleFormErrors());
            rpc.on('success', function () {
                me.trigger('finish');
            });
            rpc.on('finish', function () {
                me._submitFinish();
            });
        }
    });

    Toolkit.Views.UpdateModuleInspector = Toolkit.Views.ViewWithTemplate.extend({
        template: 'modules/update_inspector_template',

        events: {
            "click .update": "_clickUpdate"
        },

        initialize: function (options) {
            Toolkit.Views.UpdateModuleInspector.__super__.initialize.call(this, options);
            this.listenTo(this.model, "change:inspectorUpdateProblem", this.render);
        },

        _clickUpdate: function () {
            var me = this;
            var local = this.model.get("local");
            this.$('.spinner').addClass('spinner-on');

            var update_inspector = function (dependencies) {
                var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
                    python_path: local.get('module_dynamic_path'),
                    method_name: 'inspector.update_'+me.options.platform,
                    params: {
                        cookies: Toolkit.cookies,
                        dependencies: dependencies
                    }
                });
                rpc.on('success', function () {
                    local.checkInspectorHash();
                    me.trigger('cancel');
                });

                rpc.on('error', function (err) {
                    me.$('.spinner').removeClass('spinner-on');
                    me.model.set('inspectorUpdateProblem', {message: err.message});
                    console.error(err);
                });
            };


            var inspector_config = local.getConfig();
            if (inspector_config && inspector_config.modules) {
                var modules = {};
                _.each(inspector_config.modules, function (data, module) {
                    if (module == me.model.id) {
                        return;
                    }
                    modules[module] = data.version;
                });

                var rpc = Toolkit.rpcSession.remoteCall('api.get', {
                    method: "modules/archive_links/?project="+this.model.get("project").id+"&modules="+encodeURIComponent(JSON.stringify(modules)),
                    cookies: Toolkit.cookies
                });

                rpc.on('success', function (e) {
                    update_inspector(e.data.data);
                });
            } else {
                update_inspector({});
            }
        },

        close: function () {
            this.model.set('inspectorUpdateProblem', false);
        }
    });

    var GenericModuleView = Toolkit.Views.GenericModuleView = CloseableView.extend({
        breadcrumbs: function () {
            return this.moduleBreadcrumbs(this.getModule());
        },
        getModule: function () {
            if (this.model.has("module")) {
                return this.model.get("module");
            } else {
                return this.model;
            }
        },
        initialize: function () {
            var module = this.getModule();
            // ContentView should be overidden by subclasses
            this.content = new this.ContentView(this.options);

            // create component views
            this.statusbar = new Toolkit.Views.StatusBarView({model: module, url: "modules/"+module.id});
            this.sidebar = new Toolkit.Views.SideBarView({
                model: new Backbone.Model(this.sidebarContents())
            });
            this.sidebarlayout = new Toolkit.Views.SideBarLayoutView({
                sidebar: this.sidebar,
                content: this.content
            });

            // keep breadcrumbs up to date
            this.listenTo(this.getModule(), 'change:name', this.updateBreadcrumbs);
            this.listenToWhenAvailable(this.getModule(), 'project', 'change:name', this.updateBreadcrumbs);

            // Update sidebar if required
            this.listenToWhenAvailable(this.getModule(), 'local', 'change', function () {
                this.sidebar.model.set(this.sidebarContents());
                this.sidebar.renderIfNecessary();
            });
        },

        close: function () {
            this.statusbar.close();
            this.sidebarlayout.close();
        },

        sidebarContents: function () {
            var contents = [
                {
                    "section": true,
                    "text": "Module"
                },
                {
                    "item": true,
                    "text": "Manifest",
                    "link": "/manifest"
                },
                {
                    "item": true,
                    "text": "Status",
                    "link": ""
                }
            ];

            var local = this.getModule().get("local");

            if (local && local.has('manifest') && local.get("manifest").platform_version) {
                var features = Toolkit.platforms.get("platforms").get(local.get("manifest").platform_version).get("features");
                if (_.contains(features, "modules-inspector-config")) {
                    contents.push({
                        "section": true,
                        "text": "Inspector config"
                    });
                    contents.push({
                        "item": true,
                        "text": "This module",
                        "link": "/thisModule"
                    });
                    contents.push({
                        "item": true,
                        "text": "Public",
                        "link": "/publicModules"
                    });
                    contents.push({
                        "item": true,
                        "text": "Private",
                        "link": "/projectModules"
                    });
                }
            }

            contents.push({
                "section": true,
                "text": "Versions"
            });
            contents.push({
                "item": true,
                "text": "Existing",
                "link": "/versions"
            });
            contents.push({
                "item": true,
                "text": "Upload",
                "link": "/upload"
            });

            var matches = _.where(contents, {text: this.active});
            if (matches.length > 0) {
                matches[0].active = true;
            } else {
                console.error("asked to activate non-existent section", this, this.active);
            }
            return {
                prefix: "#modules/"+this.getModule().id,
                contents: contents
            };
        },

        render: function () {
            // complete render, attach subviews to this.el
            this.$el.append(this.statusbar.render().el);
            this.$el.append(this.sidebarlayout.render().el);
            this.updateBreadcrumbs();
            return this;
        }
    });

    Toolkit.Views.ModuleStatus = GenericModuleView.extend({
        active: 'Status',
        ContentView: Toolkit.Views.StatusPageView.extend({
            template: 'modules/status',

            events: _.extend({
                "click .create": "_clickedCreate",
                "click .import": "_clickedImport",
                "click .android-inspector": "_clickedAndroidInspector",
                "click .ios-inspector": "_clickediOSInspector",
                "click .gen-docs": "generateDocs"
            }, Toolkit.Views.StatusPageView.prototype.events),

            generateDocs: function (e) {
                e.preventDefault();
                var local = this.model.get("local");
                var rpc = Toolkit.rpcSession.remoteCall('modules.generate_docs', {
                    path: local.get('path')
                });
                rpc.on('success', function () {

                });
            },

            _updateInspector: function (platform) {
                var view = new Toolkit.Views.UpdateModuleInspector({
                    model: this.model,
                    platform: platform
                });
                var modal = new Toolkit.Views.ModalDialog({
                    title: 'Update inspector: '+platform,
                    contentView: view,
                    clickOutsideToCancel: true
                });
                view.on('finish', function () {
                    modal.hide(function () { modal.remove(); });
                });
                modal.show();
                return false;
            },

            _clickedAndroidInspector: function () {
                this._updateInspector('android');
            },

            _clickediOSInspector: function () {
                this._updateInspector('ios');
            },

            _clickedImport: function () {
                var createView = new Toolkit.Views.ImportModuleTemplate({
                    model: new Backbone.Model({
                        module: this.model
                    })
                });
                var modal = new Toolkit.Views.ModalDialog({
                    title: 'Import module from existing location',
                    contentView: createView,
                    clickOutsideToCancel: true
                });
                createView.on('finish', function () {
                    modal.hide(function () { modal.remove(); });
                });
                modal.show();
                return false;
            },

            _clickedCreate: function () {
                var me = this;
                var rpc = Toolkit.rpcSession.remoteCall('directory.get_home_directory');

                rpc.on('success', function (e) {
                    var sep = e.data.separator;
                    var path = e.data.userdir + sep + 'forge-workspace' + sep + 'modules' + sep + me.model.get('name');

                    var createView = new Toolkit.Views.CreateModuleTemplate({
                        model: new Backbone.Model({
                            path: path,
                            module: me.model
                        })
                    });
                    var modal = new Toolkit.Views.ModalDialog({
                        title: 'Create a new template for this module',
                        contentView: createView,
                        clickOutsideToCancel: true
                    });
                    createView.on('finish', function () {
                        modal.hide(function () { modal.remove(); });
                    });
                    modal.show();
                });

                return false;
            },

            initialize: function () {
                Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
                this.listenTo(this.model, 'all', this.render);
                this.listenToWhenAvailable(this.model, 'project', 'change:name change:features', this.render);
                this.listenToWhenAvailable(this.model, 'local', 'change', this.render);
            },

            _prepareProblemForMustache: function (problem) {
                if (problem.type == 'good') {
                    problem['class'] = 'success';
                    problem.icon = 'ok-sign';
                } else if (problem.type == 'warning') {
                    problem['class'] = 'warning';
                    problem.icon = 'warning-sign';
                } else if (problem.type == 'error' || problem.type == 'fatal') {
                    problem['class'] = 'error';
                    problem.icon = 'exclamation-sign';
                }
                return problem;
            },

            templateAttributes: function () {
                var local = this.model.get("local");
                var attributes = _({
                    canUseInspector: local && local.canUseFeature('modules-inspector-config')
                }).extend(this.modelAttributes(), local && local.attributes);

                if (attributes.inspector_android) {
                    attributes.inspector_android = this._prepareProblemForMustache(attributes.inspector_android);
                }
                if (attributes.inspector_ios) {
                    attributes.inspector_ios = this._prepareProblemForMustache(attributes.inspector_ios);
                }
                attributes.status = this.model.getStatusForTemplating();

                return attributes;
            },

            render: function () {
                Toolkit.Views.ViewWithTemplate.prototype.render.apply(this, arguments);
                this.renderFeatureStyle(this.model.get('project'));
                return this;
            }
        })
    });

    function createClash(module, versionNumber, clashId) {
        var version = _.findWhere(module.get('versions'), {version: versionNumber});
        var clashAttrs = _.findWhere(version['clashes'], {id: clashId});
        var fullClashAttrs = _.extend({
            module_version_from__module__name: module.get('name'),
            module_version_from__version: versionNumber
        }, clashAttrs);
        return new Toolkit.Models.Clash(fullClashAttrs).load();
    }

    function clickClashDetails(view, module, e) {
        e.preventDefault();
        e.stopPropagation();
        var $a = $(e.target);
        var clashId = parseInt($a.attr('data-details-clash'), 10);
        var versionNumber = $a.attr('data-details-version');

        var clash = createClash(module, versionNumber, clashId);
        var modal = new Toolkit.Views.ModalDialog({
            title: 'Clash details',
            contentView: new ClashDetailsView({model: clash})
        });
        modal.show();
        view.autoClose(modal);
    }

    /**
     * @param model {Toolkit.Models.Clash}
     */
    var ClashDetailsView = Toolkit.Views.ViewWithTemplate.extend({
        template: 'modules/clash_detail',

        initialize: function () {
            this.listenTo(this.model, 'change', this.render);
        }
    });

    /**
     * Displays a list of uploaded module versions for the given Module,
     * allowing for each:
     *
     *  - Changing the published/current-ness of that version.
     *  - Examining incompatibilities with versions of other modules.
     *
     * @param model {Toolkit.Models.Module}
     */
    Toolkit.Views.ModuleVersions = GenericModuleView.extend({
        active: 'Existing',
        ContentView: Toolkit.Views.ViewWithTemplate.extend({
            template: 'modules/versions',

            events: {
                'click .set-current': '_clickedSetCurrent',
                'click [data-panel-expand]': '_expand',
                'click [data-details-clash]': '_clashDetails'
            },

            _expand: function (e) {
                e.preventDefault();
                e.stopPropagation();
                $(e.currentTarget)
                    .find('i')
                    .toggleClass('icon-expand-alt icon-collapse-alt');
                $(e.currentTarget)
                    .next()
                    .slideToggle(300);
            },

            _clashDetails: function (e) {
                clickClashDetails(this, this.model, e);
            },

            _performAction: function (e, action, confirmTitle, confirmText) {
                var versionToChange = $(e.target).parent().attr('data-version');
                var me = this;
                var modal = new Toolkit.Views.ModalDialog({
                    title: confirmTitle || 'Confirm Action',
                    content: confirmText || '<p><p>',
                    actions: [{
                        type: "cancel",
                        text: "Cancel"
                    }, {
                        type: "danger",
                        text: '<i class="icon-cog icon-spin spinner"></i> Confirm',
                        onClicked: function (e) {
                            $(e.target).find('.spinner').addClass('spinner-on');
                            me.model.updateVersion(versionToChange, action, function () {
                                me.render();
                                modal.hide();
                            });
                        }
                    }]
                });
                modal.show();
            },

            _clickedSetCurrent: function (e) {
                this._performAction(e, 'set_current', 'Set as current version?', '<p>Setting this version as the current version will many anyone using other versions will be prompted to update to this one, and any future users will be given this version by default.</p>');
            },

            initialize: function () {
                Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
                this.listenTo(this.model, 'change:_loaded change:versions', this.render);
                this.listenToWhenAvailable(this.model, 'project', 'change:features', this.render);
            },

            templateAttributes: function () {
                var modelAttributes = this.modelAttributes();
                Toolkit.preprocess.module(modelAttributes);

                var attributes = _({}).extend(modelAttributes, {
                    onlyOnePublished: _.reduce(_.pluck(this.model.get('versions'), 'published'), function (m, value) { return m + (value ? 1 : 0); }, 0) == 1,
                    publishing_agreement_url: Toolkit.websiteUrl() + "modules/_/publishing_agreement/"
                });

                return attributes;
            },

            render: function () {
                Toolkit.Views.ViewWithTemplate.prototype.render.apply(this, arguments);
                this.renderFeatureStyle(this.model.get('project'));
                return this;
            }
        })
    });

    Toolkit.Views.ModulePublish = GenericModuleView.extend({
        active: 'Existing',
        ContentView: Toolkit.Views.FormView.extend({
            template: 'modules/publish',
            events: _.extend({
                'click [data-details-clash]': 'clashDetails'
            }, Toolkit.Views.FormView.prototype.events),

            clashDetails: function (e) {
                clickClashDetails(this, this.model.get('module'), e);
            },

            initialize: function () {
                Toolkit.Views.FormView.prototype.initialize.call(this);
                this.listenTo(this.model, 'change', this.render);
            },

            templateAttributes: function () {
                var superMethod = Toolkit.Views.FormView.prototype.templateAttributes;
                var superAttrs = superMethod.call(this);
                if (superAttrs['moduleVersion']) {
                    Toolkit.preprocess.moduleVersion(superAttrs['moduleVersion']);
                }
                superAttrs['name'] = this.model.get('module').get('name');
                return superAttrs;
            },

            submit: function () {
                var module = this.model.get('module');
                module.updateVersion(
                    this.model.get('moduleVersion')['version'],
                    'publish',
                    function () {
                        Toolkit.router.navigate(
                            '#modules/'+module.get('name')+'/versions',
                            true
                        );
                    }
                );
            }
        })
    });

    Toolkit.Views.ModuleUnpublish = GenericModuleView.extend({
        active: 'Existing',
        ContentView: Toolkit.Views.FormView.extend({
            template: 'modules/unpublish',
            initialize: function () {
                Toolkit.Views.FormView.prototype.initialize.call(this);
                this.listenTo(this.model, 'change', this.render);
            },
            submit: function () {
                var module = this.model.get('module');
                module.updateVersion(
                    this.model.get('moduleVersion')['version'],
                    'unpublish',
                    function () {
                        Toolkit.router.navigate(
                            '#modules/'+module.get('name')+'/versions',
                            true
                        );
                    }
                );
            }
        })
    });

    Toolkit.Views.ModuleUpload = GenericModuleView.extend({
        active: 'Upload',
        ContentView: Toolkit.Views.ViewWithTemplate.extend({
            template: 'modules/upload',
            events: {
                "click .upload": "_clickedUpload",
                "click .update": "_clickedUpdate"
            },

            _clickedUpload: function (e) {
                var me = this;
                e.preventDefault();
                this.$('.spinner').addClass('spinner-on');
                this._postFiles(false, function () {
                    me.$('.spinner').removeClass('spinner-on');
                    // TODO: Update model => rerender
                    window.location.reload();
                });
            },
            _clickedUpdate: function (e) {
                e.preventDefault();
                var me = this;
                var modal = new Toolkit.Views.ModalDialog({
                    title: "Update existing version",
                    content: '<p>You have already uploaded a version of this module with this version number.</p><p>This means you can only update the <strong>docs</strong>, <strong>description</strong>, <strong>changes</strong> and <strong>minimum</strong> / <strong>maximum</strong> supported platform versions.</p><p>To upload new code you must change your version number.<p>',
                    actions: [{
                        type: "cancel",
                        text: "Cancel"
                    }, {
                        type: "warning",
                        text: '<i class="icon-cog icon-spin spinner"></i> Confirm',
                        onClicked: function () {
                            $('.spinner', this).addClass('spinner-on');
                            me._postFiles(true, function () {
                                $('.spinner', this).removeClass('spinner-on');
                                // TODO: Update model => rerender
                                window.location.reload();
                            });
                        }
                    }]
                });
                modal.show();
            },
            _postFiles: function (is_update, success) {
                var me = this;
                var local = this.model.get("local");
                var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
                    'python_path': local.get('module_dynamic_path'),
                    'method_name': 'module.create_upload_zip',
                    params: {
                        path: local.get('path')
                    }
                });
                rpc.on('success', function (e) {
                    var zipPath = e.data;
                    var rpc = Toolkit.rpcSession.remoteCall('api.post_files', {
                        cookies: Toolkit.cookies,
                        method: 'module/'+me.model.id+'/version',
                        data: {
                            is_update: is_update
                        },
                        files: {
                            module_zip: zipPath
                        }
                    });
                    rpc.on('success', function () {
                        success();
                    });
                });
            },
            initialize: function () {
                Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
                this.doWhenAvailable(this.model, 'local', function (local) {
                    this.listenTo(local, 'change', this.render);
                });
                this.listenToWhenAvailable(this.model, 'project', 'change:features', this.render);
            },
            templateAttributes: function () {
                var local = this.model.get("local");
                var attributes =  _({}).extend(this.modelAttributes(), local && local.attributes);

                attributes.problemsPreventUpload = (this.model.getStatusSummary().severity == "error");
                attributes.versionAlreadyUploaded = _(this.model.getStatusForTemplating()["module"]).where({
                    id: "already-uploaded"
                }).length !== 0;

                return attributes;
            },
            render: function () {
                Toolkit.Views.ViewWithTemplate.prototype.render.apply(this, arguments);
                this.renderFeatureStyle(this.model.get('project'));
                return this;
            }
        })
    });

    /**
     * @param model {Toolkit.Models.Module}
     */
    Toolkit.Views.ModuleManifest = GenericModuleView.extend({
        active: 'Manifest',
        ContentView: Toolkit.Views.ViewWithTemplate.extend({
            template: 'modules/manifest',
            events: {
                "change .manifest-input": "inputChanged",
                "click .add-dep": "addDependency",
                "click .remove": "removeDependency",
                'change .dependency-version': 'versionChanged',
                "click .change-version": "clickedChangeVersion"
            },
            clickedChangeVersion: function () {
                var me = this;
                var view = new (Toolkit.Views.ViewWithTemplate.extend({
                    template: 'modules/change_platform_version',
                    events: {
                        "click .update-stable": "updateStable",
                        "click .update-selection": "updateSelection"
                    },
                    updateToVersion: function (version) {
                        var local = me.model.get("local");
                        local.get('manifest').platform_version = version;
                        local.get('manifest').min_platform_version = version;
                        me.saveManifest();
                        me.render();
                        this.trigger("cancel");
                    },
                    updateStable: function () {
                        this.updateToVersion(Toolkit.platforms.get('latest_stable'));
                    },
                    updateSelection: function () {
                        this.updateToVersion(this.$('select').val());
                    },
                    templateAttributes: function () {
                        var local = me.model.get("local");
                        var current_version = local && local.get('manifest').platform_version;
                        var stable_version = Toolkit.platforms.get('latest_stable');
                        var possiblePlatforms = [];
                        Toolkit.platforms.get('platforms').each(function (platform) {
                            if (platform.get('module_support')) {
                                possiblePlatforms.push(platform.get('version'));
                            }
                        });
                        return {
                            'current_version': current_version,
                            'stable_version': stable_version,
                            'current_is_stable': current_version == stable_version,
                            'available_platforms': possiblePlatforms
                        };
                    }
                }))();
                var modalView = new Toolkit.Views.ModalDialog({
                    title: 'Change Trigger.io platform version',
                    contentView: view
                });
                modalView.show();
            },
            removeDependency: function (e) {
                e.preventDefault();
                var local = this.model.get("local");
                var manifest = local.get("manifest");
                delete manifest.dependencies[$(e.currentTarget).attr('data-module')];
                this.saveManifest();
                this.render();
            },
            addDependency: function (e) {
                e.preventDefault();
                var dependencies = this.model.get("project") && this.model.get("project").get("dependencies");
                var module = _.findWhere(dependencies, {module: $('#dependency').val()});
                if (module) {
                    var local = this.model.get("local");
                    var manifest = local.get("manifest");
                    if (!manifest.dependencies) {
                        manifest.dependencies = {};
                    }
                    if (!manifest.dependencies[module.module]) {
                        manifest.dependencies[module.module] = {};
                        this.saveManifest();
                        this.render();
                    }
                    $('#dependency').val('');
                } else {
                    var modal = new Toolkit.Views.ModalDialog({
                        title: 'Module not found',
                        content: "<p>Module named '"+ $('#dependency').val()+"' not found, it either doesn't exist or you do not have permission to use it<p>",
                        actions: [{
                            type: "cancel",
                            text: "Ok"
                        }]
                    });
                    modal.show();
                }
            },
            inputChanged: function (e) {
                var local = this.model.get("local");
                var manifest = local.get("manifest");
                var value = $(e.target).val();
                if (value == "any" && $(e.target).hasClass('any-is-undefined')) {
                    delete manifest[$(e.target).attr('id')];
                } else {
                    manifest[$(e.target).attr('id')] = value;
                }
                this.saveManifest();
            },
            versionChanged: function (e) {
                var local = this.model.get("local");
                var manifest = local.get("manifest");
                var value = $(e.target).val();
                if (value == 'any') {
                    delete manifest.dependencies[$(e.target).attr('data-module')][$(e.target).attr('data-field')];
                } else {
                    manifest.dependencies[$(e.target).attr('data-module')][$(e.target).attr('data-field')] = value;
                }
                this.saveManifest();
            },
            saveManifest: function () {
                var local = this.model.get("local");
                var rpc = Toolkit.rpcSession.remoteCall('file.write', {
                    path: local.get('files').manifest,
                    data: JSON.stringify(local.get('manifest'), undefined, "\t")
                });
                var self = this;
                rpc.on('success', function () {
                    self.$(".saved").fadeIn();
                    setTimeout(function () {
                        self.$(".saved").fadeOut();
                    }, 3000);
                });
            },
            initialize: function () {
                Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
                this.doWhenAvailable(this.model, 'project', function (project) {
                    project.loadDependencies();
                    this.listenTo(project, 'change:dependencies', this.render);
                });
            },
            templateAttributes: function () {
                var local = this.model.get('local');

                return _.extend(
                    {platform_features: local && local.platformFeatures()},
                    Toolkit.preprocess.moduleManifest(this.model),
                    this.modelAttributes(),
                    local && local.attributes
                );
            }
        })
    });

    Toolkit.Views.ModuleTrustTemplate = Toolkit.Views.FormView.extend({
        template: 'modules/trust',
    });

    Toolkit.Views.ModuleTrustWarning = Toolkit.Views.ModalDialog.extend({
        initialize: function () {
            var self = this;
            this.options.title = "3rd Party Code Warning";
            this.options.contentView = new Toolkit.Views.ModuleTrustTemplate({});
            this.options.actions = [ {
                type: "cancel",
                text: "Cancel"
            }, {
                type: "danger",
                text: "I Accept",
                onClicked: function () {
                    self.hide(function () { self.remove(); });
                    var rpc = Toolkit.rpcSession.remoteCall('api.post', {
                        method: 'auth/thirdparty_module_warning',
                        data: {
                            thirdparty_module_warning: true
                        },
                        cookies: Toolkit.cookies
                    });
                    rpc.on('success', function () {
                        Toolkit.authStatus.set("thirdparty_module_warning", true);
                        self.options.onAccept && self.options.onAccept();
                    });
                }
            } ];
            Toolkit.Views.ModalDialog.prototype.initialize.apply(this, arguments);
        }
    });
}());
