/*global Toolkit*/
/*global Backbone*/
/*global _*/
/*global _superApply*/
/*global _superCall*/
/*global CloseableView*/
/* global $ */
(function () {
	// represents a plugin owned by this team
	Toolkit.Models.Plugin = Toolkit.Models.RpcModel.extend({
		idAttribute: 'uuid',
		cacheName: 'Plugin',
		method: 'api.get',
		args: function () {
			return {
				cookies: Toolkit.cookies,
				method: 'plugin/'+this.id
			};
		},
		parent: {
			attribute: 'project',
			idAttribute: 'project_id',
			model: 'Project',
			collectionAttribute: 'plugins'
		},

		initialize: function () {
			Toolkit.Models.Plugin.__super__.initialize.call(this);
			// LocalPlugin validation relies on Plugin versions being available,
			// force re-validation when they are
			if (this.get('versions')) {
				this.setStatus('versions', 'numberofversions', 'info', this.get('versions').length+' uploaded plugin versions');
			}
			this.set('status', {});
			this.listenTo(this, 'change:versions', this.updateVersions);
			Toolkit.localPluginsFile.doWhenLocalAvailable(this, this, function (local) {
				this.listenTo(local, 'change', this.updateInspectorStatus);
			});
			this.setStatus('development', 'development-problems', 'warning', "No local development files found.");
			this.setStatus('general', 'plugins-deprecated', 'warning', "Plugins are deprecated" +
				'<a class="pull-right btn btn-small btn-primary toModule" href="javascript:void(0);">' +
				'Convert to a module</a>');
		},

		updateInspectorStatus: function (local) {
			if (local.get('inspector_android')) {
				this.setStatus('development', 'android-inspector', local.get('inspector_android').type, local.get('inspector_android').message);
			}
			if (local.get('inspector_ios')) {
				this.setStatus('development', 'ios-inspector', local.get('inspector_ios').type, local.get('inspector_ios').message);
			}
		},

		updateVersions: function () {
			this.setStatus('versions', 'numberofversions', 'info', this.get('versions').length+' uploaded plugin versions');
		}
	});

	_.extend(Toolkit.Models.Plugin.prototype, Toolkit.Mixins.Status);

	// represents a plugin on the local file system
	Toolkit.Models.LocalPlugin = Toolkit.Models.WatchingModel.extend({
		idAttribute: 'path',

		cacheName: 'LocalPlugin',

		// load details about the local plugin
		load: function () {
			var me = this;

			var rpc = Toolkit.rpcSession.remoteCall('plugins.state', {
				cookies: Toolkit.cookies,
				path: this.get('path')
			});
			rpc.on('success', function (e) {
				me.set(e.data);
			});
		},

		initialize: function () {
			_superApply(Toolkit.Models.LocalPlugin, this, 'initialize', arguments);
			this.load();

			var me = this;
			// Short timeout to allow any file watchers to update the model
			this.on('watch:modified watch:created watch:deleted', function (e) {
				me.updateFileHash(e['src_path']);
			});
			this.on('watch:moved', function (e) {
				me.updateFileHash(e['src_path']);
				me.updateFileHash(e['dest_path']);
			});
			this.set("plugin_directory_hashes", {});
			this.set("plugin_directory_validations", {});
			this.once('change:plugin_directory', function () {
				var recurse_directory = function (dir, path) {
					_.each(dir, function (value, key) {
						if (value === "file") {
							if (path.length > 0) {
								me.updateFileHash(path.join('/') + '/' + key);
							} else {
								me.updateFileHash(key);
							}
						} else {
							var curPath = _.clone(path);
							curPath.push(key);
							recurse_directory(value, curPath);
						}
					});
				};
				recurse_directory(me.get('plugin_directory'), []);
			});
			this.on('change:plugin_directory', function () {
				me.updateValidationMessages();
			});
		},

		// Called when a file change event happens, hashes the file and triggers other updates if the hash has changed
		updateFileHash: function (path) {
			if (this.dontValidate) {
				return;
			}
			if (!path) {
				return;
			}
			var me = this;
			var path_segments = path.replace("\\", "/");
			path_segments = path_segments.split("/");
			var hashes = this.get("plugin_directory_hashes");
			_.each(_.initial(path_segments), function (segment) {
				if (!_.isObject(hashes[segment])) {
					hashes[segment] = {};
				}
				hashes = hashes[segment];
			});
			var rpc = Toolkit.rpcSession.remoteCall('file.hash', {path: this.get("path") + "/plugin/" + path});
			rpc.on("success", function (e) {
				if (e.data !== null) {
					if (hashes[_.last(path_segments)] != e.data) {
						hashes[_.last(path_segments)] = e.data;
						me.updateValidators(path_segments);
						me.updateValidationMessages();
					}
				} else if (_.last(path_segments) in hashes) {
					delete hashes[_.last(path_segments)];
					me.updateValidationMessages();
				}
			});
		},

		// Called when a file hash changes, updates any validator information for the file
		updateValidators: function (path) {
			if (this.dontValidate) {
				return;
			}
			var me = this;
			if (this.has("plugin_structure")) {
				var notfound = false;
				var structure = this.get("plugin_structure");
				_.each(path, function (segment) {
					if (!notfound) {
						if (structure.type === "directory" && structure['children']) {
							structure = structure['children'][segment];
							if (!structure) {
								notfound = true;
							}
						}
					}
				});
				if (!notfound && structure['validators']) {
					var validations = this.get("plugin_directory_validations");
					_.each(_.initial(path), function (segment) {
						if (!_.isObject(validations[segment])) {
							validations[segment] = {};
						}
						validations = validations[segment];
					});
					validations[_.last(path)] = [];
					validations = validations[_.last(path)];
					_.each(structure['validators'], function (extra_params, validator) {
						var params = _.extend(extra_params, {
							file: path.join("/"),
							manifest: me.get('manifest')
						});
						var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
							'python_path': me.get('plugin_dynamic_path'),
							'method_name': 'validate.' + validator,
							params: params
						});
						rpc.on("error", function (e) {
							validations.push({message: "Validation for '"+path.join("/")+"' failed: "+e.message, type: "error"});
							me.updateValidationMessages();
						});
					});
				}
			}
		},

		// Called when the file structure changes or a validation changes
		// This should not make any RPC calls (it should be synchronous)
		// Any RPC work should be done then trigger this method
		updateValidationMessages: function () {
			if (this.dontValidate) {
				return;
			}
			this.checkInspectorHash();
			var me = this;
			// TODO: problems as a collection, rerender on update?
			var problems = [];

			// Check platform_version
			if (!this.get('manifest')) {
				problems.push({message: "No local copy of plugin.", type: "error"});
			} else if (!this.get('manifest')['platform_version']) {
				problems.push({message: "No platform version in manifest.", type: "error"});
			} else if ("undefined" === typeof(this.get('dynamic_platform_version'))) {
				// .trigger/platform_version.txt not read in yet
			} else if (this.get('dynamic_platform_version') != this.get('manifest')['platform_version']) {
				var platforms = Toolkit.platforms;
				if (platforms.has('platforms')) {
					var platformModel = platforms.get('platforms').get(me.get('manifest')['platform_version']);
					if (platformModel && platformModel.get('plugin_support')) {
						problems.push({message: "Platform version in manifest has changed, downloading new dynamic plugin code...", type: "error"});
						me.dontValidate = true;
						var rpc = Toolkit.rpcSession.remoteCall('plugins.update_plugin_dynamic', {
							cookies: Toolkit.cookies,
							'path':  me.get('local_path'),
							'new_platform_version': me.get('manifest')['platform_version']
						});
						rpc.on('success', function () {
							// XXX: File watchers get upset by this, so we reload, should do something better.
							window.location.reload();
						});
					} else {
						problems.push({message: "Platform version in manifest is not a supported platform version.", type: "error"});
					}
				}
			} else if (!this.get('plugin_structure') || !this.get('plugin_directory')) {
				problems.push({message: "No plugin structure found to validate plugin. Try refreshing the browser, if this problem persists please contact support@trigger.io.", type: "error"});
			} else {
				var validateStructure = function (structure, schema, path) {
					if (schema.validators) {
						var validations = me.get("plugin_directory_validations");
						var notfound = false;
						_.each(path.split('/'), function (segment) {
							if (!notfound) {
								if (!_.isObject(validations[segment])) {
									notfound = true;
								} else {
									validations = validations[segment];
								}
							}
						});
						if (!notfound) {
							_.each(validations, function (validation_error) {
								problems.push(validation_error);
							});
						}
					}
					if (schema.type == "file") {
						if (structure == "file") {
							return;
						} else {
							problems.push({message: "The directory '"+path+"' was expected to be a file.", type: "error"});
						}
					} else if (schema.type == "directory") {
						if (structure == "file") {
							problems.push({message: "The file '"+path+"' was expected to be a directory.", type: "error"});
							return;
						}
						if (path !== "") {
							path += "/";
						}
						if (schema.children) {
							for (var schema_child in schema.children) {
								if (schema.children[schema_child].required && !structure[schema_child]) {
									problems.push({message: "The path '"+path+schema_child+"' is required and was not found.", type: "error"});
								}
							}
						}
						for (var child in structure) {
							if (schema.children && schema.children[child]) {
								validateStructure(structure[child], schema.children[child], path+child);
							} else if (schema.additionalChildren !== undefined) {
								if (schema.additionalChildren === false && child.substring(0,1) != '.') {
									problems.push({message: "The path '"+path+child+"' was found but not expected.", type: "warning"});
								} else {
									validateStructure(structure[child], schema.additionalChildren, path+child);
								}
							}
						}
					}
				};
				validateStructure(this.get('plugin_directory'), this.get('plugin_structure'), '');

				if (me.toJSON().targets.length === 0) {
					problems.push({type: "fatal", message: "Neither an Android nor iOS plugin has been created."});
				}
			}

			// have we already uploaded the plugin at this version?
			var local = this;
			var manifestVersion = local.get('manifest') && local.get('manifest').version;

			var plugin = local.get('manifest') && local.get('manifest').uuid && new Toolkit.Models.Plugin({uuid: local.get('manifest').uuid});

			if (manifestVersion && plugin) {
				// lookup the Plugin for this LocalPlugin
				if (plugin.has('versions')) {
					var manifestVersionAlreadyUploaded = _(plugin.get('versions')).where(
						{version: manifestVersion}
					).length !== 0;

					if (manifestVersionAlreadyUploaded) {
						problems.push({
							type: "fatal",
							message: "A " + manifestVersion + " version of your plugin has already been uploaded."
						});
					}
				}
			}

			me.set('problems', problems);
			if (plugin) {
				var hasError = false;
				var hasWarning = false;
				_.each(problems, function (problem) {
					if (problem.type == "fatal" || problem.type == "error") {
						hasError = true;
					} else if (problem.type == "warning") {
						hasWarning = true;
					}
				});
				if (hasError) {
					plugin.setStatus('development', 'development-problems', 'error', "Problem in local plugin files, see development tab for details.");
				} else if (hasWarning) {
					plugin.setStatus('development', 'development-problems', 'warning', "Problem in local plugin files, see development tab for details.");
				} else {
					plugin.setStatus('development', 'development-problems', 'good', "No problems found in local plugin files.");
				}
			}
		},

		checkInspectorHash: function () {
			if (this.dontValidate) {
				return;
			}
			if (!this.debouncedCheckInspector) {
				this.debouncedCheckInspector = _.debounce(function (me) {
					if (me.dontValidate) {
						return;
					}
					var check_inspector_hash = function (platform) {
						var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
							'python_path': me.get('plugin_dynamic_path'),
							'method_name': 'inspector.check_'+platform+'_hash',
							params: {}
						});
						rpc.on("success", function (e) {
							me.set('inspector_'+platform, e.data);
						});
					};
					if (me.toJSON().targets.length > 0) {
						for (var platform in me.toJSON().targets) {
							check_inspector_hash(me.toJSON().targets[platform]);
						}
					}
				}, 250);
			}
			this.debouncedCheckInspector(this);
		},

		toJSON: function () {
			var dirs = this.attributes['plugin_directory'] ? _(this.attributes['plugin_directory']).keys() : [];
			return _.extend({
				// Filter directories by possible targets
				targets: _.intersection(dirs, ["android", "ios"])
			}, this.attributes);
		}
	});

	// collection of local plugins on the file system
	Toolkit.Collections.LocalPluginCollection = Backbone.Collection.extend({
		model: Toolkit.Models.LocalPlugin
	});

	// model which watches local_plugins.json and maintains a collection of
	// all LocalPlugins on the file system
	Toolkit.Models.LocalPluginsFile = Toolkit.Models.WatchingModel.extend({
		// return LocalPlugin for Plugin
		getLocal: function (plugin) {
			var name = plugin.get('name');
			var available = this.availableNames();
			return available[name];
		},

		// execute callback when local app available
		// only execute once
		doWhenLocalAvailable: function (plugin, listener, callback) {
			var me = this;

			var local = me.getLocal(plugin);
			if (local) {
				callback.call(listener, local);
				return;
			}

			var doIfAvailable = function () {
				var local = me.getLocal(plugin);
				if (local) {
					me.get('collection').off(null, doIfAvailable);
					callback.call(listener, local);
				}
			};

			Toolkit.localPluginsFile.get('collection');
			listener.listenTo(this.get('collection'), 'all', doIfAvailable);
		},

		availableNames: function () {
			var available = {};
			this.get('collection').map(function (localPlugin) {
				var manifest = localPlugin.get('manifest');
				// might be in the middle of fetching the manifest file,
				// can't say what it's name is yet in that case
				if (manifest) {
					available[manifest.name] = localPlugin;
				}
			});
			return available;
		},

		initialize: function () {
			Toolkit.Models.WatchingModel.prototype.initialize.apply(this, arguments);
			this.set({
				collection: new Toolkit.Collections.LocalPluginCollection()
			});
			this.listenTo(this, 'change:localPlugins', function () {
				this.get('collection').update(
					_(this.get('localPlugins').plugin_locations).map(function (path) {
						return {path: path};
					})
				);
			});
			var me = this;
			Toolkit.rpcSession.remoteCall('plugins.remove_local_not_on_filesystem').on('success', function () {
				me.set('files', {
					localPlugins: 'build-tools/local_plugins.json'
				});
			});
		}
	});

	Toolkit.events.once('rpc', function () {
		Toolkit.registerPreReadyFunction(function (done) {
			Toolkit.localPluginsFile = new Toolkit.Models.LocalPluginsFile();
			Toolkit.localPluginsFile.once('change:localPlugins', done);
		});
	});

	var GenericPluginView = Toolkit.Views.GenericPluginView = CloseableView.extend({
		breadcrumbs: function () {
			return this.pluginBreadcrumbs(this.model);
		},
		initialize: function () {
			// ContentView should be overidden by subclasses
			this.content = new this.ContentView(this.options);

			// create component views
			this.statusbar = new Toolkit.Views.StatusBarView({model: this.model, url: "plugins/"+this.model.id});
			this.sidebar = new Toolkit.Views.SideBarView({
				model: new Backbone.Model(this.sidebarContents())
			});
			this.sidebarlayout = new Toolkit.Views.SideBarLayoutView({
				sidebar: this.sidebar,
				content: this.content
			});

			// keep breadcrumbs up to date
			this.listenTo(this.model, 'change:name', this.updateBreadcrumbs);
			this.listenToWhenAvailable(this.model, 'project', 'change:name', this.updateBreadcrumbs);
		},

		close: function () {
			this.statusbar.close();
			this.sidebarlayout.close();
		},

		sidebarContents: function () {
			var contents = [
				{
					"item": true,
					"text": "Status",
					"link": ""
				},
				{
					"item": true,
					"text": "Development",
					"link": "/development"
				},
				{
					"item": true,
					"text": "Versions",
					"link": "/versions"
				}
			];

			var matches = _.where(contents, {text: this.active});
			if (matches.length > 0) {
				matches[0].active = true;
			} else {
				console.error("asked to activate non-existent section", this, this.active);
			}
			return {
				prefix: "#plugins/"+this.model.id,
				contents: contents
			};
		},

		render: function () {
			// complete render, attach subviews to this.el
			this.$el.append(this.statusbar.render().el);
			this.$el.append(this.sidebarlayout.render().el);
			this.updateBreadcrumbs();
			return this;
		}
	});


	Toolkit.Views.PluginCreateView = Toolkit.Views.FormView.extend({
		template: 'projects/create_plugin',

		submit: function (form) {
			var me = this;
		
			if (form.name.match(/^[a-z][a-z0-9_]+$/) === null) {
				me.tempModel.set('errors', {'name': ['Name must start with a lowercase letter and only contain lowercase letters, numbers and _.']});
				me._submitFinish();
				return;
			}
		
			var rpc = Toolkit.rpcSession.remoteCall('api.post', {
				method: 'plugin/',
				data: {
					name: form.name,
					project: this.model.get('id')
				},
				cookies: Toolkit.cookies
			});

			rpc.on('error', this._handleFormErrors());
			rpc.on('success', function (e) {
				me.model.get('plugins').add(e.data['data']);
				me.trigger("cancel");
			});
		}
	});

	Toolkit.Views.CreatePluginTemplate = Toolkit.Views.FormView.extend({
		template: 'plugins/create_template',

		templateAttributes: function () {
			var superAttrs = Toolkit.Views.CreatePluginTemplate.__super__.templateAttributes.call(this);
			var platforms = Toolkit.platforms;
			var possiblePlatforms = [];
			var latestStable = platforms.get('platforms').get(platforms.get('latest_stable'));
			platforms.get('platforms').each(function (platform) {
				if (platform.get('plugin_support')) {
					possiblePlatforms.push({
						name: platform.get('version'),
						selected: (platform === latestStable)
					});
				}
			});

			return _(superAttrs).extend({
				platforms: possiblePlatforms
			});
		},

		submit: function (form) {
			var me = this;
			var rpc = Toolkit.rpcSession.remoteCall('plugins.create_plugin_template', {
				uuid: me.model.get('plugin').get('uuid'),
				cookies: Toolkit.cookies,
				path: form.path,
				core_version: form['core_version'],
				name: me.model.get('plugin').get('name')
			});

			rpc.on('error', this._handleFormErrors());
			rpc.on('success', function (e) {
				var localPlugin = _(e.data).extend({
					path: e.data['local_path']
				});
				Toolkit.localPluginsFile.get('collection').add(localPlugin);
				me.trigger("cancel");
			});
			rpc.on('finish', function () {
				me._submitFinish();
			});
		}
	});

	Toolkit.Views.ImportPluginTemplate = Toolkit.Views.FormView.extend({
		template: 'plugins/import',

		submit: function (form) {
			var me = this;

			var rpc = Toolkit.rpcSession.remoteCall('plugins.import_plugin', {
				cookies: Toolkit.cookies,
				path: form.path,
				name: me.model.get('plugin').get('name')
			});

			rpc.on('error', this._handleFormErrors());
			rpc.on('success', function () {
				me.trigger('finish');
			});
			rpc.on('finish', function () {
				me._submitFinish();
			});
		}
	});

	Toolkit.Views.UpdatePluginInspector = Toolkit.Views.ViewWithTemplate.extend({
		template: 'plugins/update_inspector_template',

		events: {
			"click .update": "_clickUpdate"
		},

		initialize: function (options) {
			_superCall(Toolkit.Views.UpdatePluginInspector, this, 'initialize', options);
			this.listenTo(this.model, "change:inspectorUpdateProblem", this.render);
		},

		_clickUpdate: function () {
			var me = this;
			var local = Toolkit.localPluginsFile.getLocal(this.model);
			this.$('.spinner').addClass('spinner-on');
			var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
				python_path: local.get('plugin_dynamic_path'),
				method_name: 'inspector.update_'+this.options.platform,
				params: {
					cookies: Toolkit.cookies
				}
			});
			rpc.on('success', function () {
				local.checkInspectorHash();
				me.trigger('cancel');
			});

			rpc.on('error', function (err) {
				me.$('.spinner').removeClass('spinner-on');
				me.model.set('inspectorUpdateProblem', {message: err.message});
			});
		},

		close: function () {
			this.model.set('inspectorUpdateProblem', false);
		}
	});

	Toolkit.Views.PluginToModule = Toolkit.Views.ViewWithTemplate.extend({
		template: 'plugins/to_module',

		events: {
			"click .convert": "_clickedConvert"
		},

		initialize: function () {
			_superApply(Toolkit.Views.PluginToModule, this, 'initialize', arguments);

			// create our own model to store stuff not passed to us
			this.tempModel = new Backbone.Model();
			this.listenTo(this.tempModel, 'change:modules-dir', this.render);

			// get modules-dir
			var rpc = Toolkit.rpcSession.remoteCall('directory.get_home_directory');
			var me = this;
			rpc.on('success', function (e) {
				var sep = e.data.separator;
				me.tempModel.set({
					'modules-dir': e.data.userdir + sep + 'forge-workspace' + sep + 'modules' + sep,
					'workspace-dir': e.data.userdir + sep + 'forge-workspace' + sep
				});
			});
		},

		_clickedConvert: function () {
			var me = this,
				local = Toolkit.localPluginsFile.getLocal(this.model),
				modulePath = $('#module-path').val();

			this.$('.spinner').addClass('spinner-on');
			this.$('.error').hide();

			var rpc = Toolkit.rpcSession.remoteCall('plugins.convert_to_module', {
				cookies: Toolkit.cookies,
				uuid: this.model.get('uuid'),
				plugin_path: local.get('path'),
				module_path: modulePath
			});
			rpc.on('success', function (res) {
				var moduleName = res.data.name;
				// create module in DB
				var rpc = Toolkit.rpcSession.remoteCall('api.post', {
					method: 'module/',
					data: {
						name: moduleName,
						project: me.model.get('project_id')
					},
					cookies: Toolkit.cookies
				});

				rpc.on('success', function () {
					var rpc = Toolkit.rpcSession.remoteCall('modules.import_module', {
						cookies: Toolkit.cookies,
						path: modulePath,
						name: moduleName
					});

					rpc.on('error', Toolkit.showError);
					rpc.on('success', function () {
						me.trigger('finish');
						Toolkit.showNotification(moduleName + " module created");
						Toolkit.router.navigate('modules/' + moduleName, true);
					});
				});
				rpc.on('error', Toolkit.showError);
			});

			rpc.on('error', function (err) {
				me.$('.spinner').removeClass('spinner-on');
				Toolkit.showError(err);
			});
			return false;
		},

		templateAttributes: function () {
			var attributes = this.modelAttributes();
			if (this.tempModel.get('modules-dir')) {
				_.extend(attributes, {
					defaultPath: this.tempModel.get('modules-dir') + this.model.get('name'),
					workspacePath: this.tempModel.get('workspace-dir')
				});
			}
			return attributes;
		},

		render: function () {
			_superApply(Toolkit.Views.PluginToModule, this, 'render', arguments);
			var rpc = Toolkit.rpcSession.remoteCall("api.get", {
				method: "/module/permitted/" + this.model.get('name'),
				cookies: Toolkit.cookies
			});
			var me = this;

			function nameNotPermitted () {
				me.$("i.name-status").removeClass('icon-cog icon-spin');
				me.$("i.name-status").addClass('icon-ban-circle text-error');
				me.$('.when-name-not-permitted').show();
			}

			rpc.on("success", function (result) {
				me.$('i.name-status').removeClass('icon-cog icon-spin');
				if (result.data.data.permitted) {
					me.$('i.name-status').addClass('icon-ok text-success');
					me.$('.when-name-permitted').show();
				} else {
					nameNotPermitted();
				}
			});
			rpc.on("error", nameNotPermitted);
		}
	});

	Toolkit.Views.PluginStatus = GenericPluginView.extend({
		active: 'Status',
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: 'plugins/status',

			events: {
				"click .toModule": "_clickedToModule"
			},

			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
				this.listenTo(this.model, 'all', this.render);
				this.listenToWhenAvailable(this.model, 'project', 'change:name', this.render);
			},
			templateAttributes: function () {
				var attributes = this.modelAttributes();
				attributes.status = this.model.getStatusForTemplating();
				return attributes;
			},

			_clickedToModule: function () {
				var me = this;

				var view = new Toolkit.Views.PluginToModule({
					model: me.model
				});
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Convert plugin to a module: '+me.model.get('name'),
					contentView: view,
					clickOutsideToCancel: false
				});
				view.on('finish', function () {
					modal.hide(function () { modal.remove(); });
				});
				modal.show();
				return false;
			}
		})
	});
	
	Toolkit.Views.PluginDevelopment = GenericPluginView.extend({
		active: 'Development',
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: 'plugins/development',

			events: {
				"click .upload": "_clickedUpload",
				"click .create": "_clickedCreate",
				"click .import": "_clickedImport",
				"click .android-inspector": "_clickedAndroidInspector",
				"click .ios-inspector": "_clickediOSInspector"
			},

			_updateInspector: function (platform) {
				var view = new Toolkit.Views.UpdatePluginInspector({
					model: this.model,
					platform: platform
				});
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Update inspector: '+platform,
					contentView: view,
					clickOutsideToCancel: true
				});
				view.on('finish', function () {
					modal.hide(function () { modal.remove(); });
				});
				modal.show();
				return false;
			},

			_clickedAndroidInspector: function () {
				this._updateInspector('android');
			},

			_clickediOSInspector: function () {
				this._updateInspector('ios');
			},

			_clickedImport: function () {
				var createView = new Toolkit.Views.ImportPluginTemplate({
					model: new Backbone.Model({
						plugin: this.model
					})
				});
				var modal = new Toolkit.Views.ModalDialog({
					title: 'Import plugin from existing location',
					contentView: createView,
					clickOutsideToCancel: true
				});
				createView.on('finish', function () {
					modal.hide(function () { modal.remove(); });
				});
				modal.show();
				return false;
			},

			_clickedCreate: function () {
				var me = this;
				var rpc = Toolkit.rpcSession.remoteCall('directory.get_home_directory');

				rpc.on('success', function (e) {
					var sep = e.data.separator;
					var path = e.data.userdir + sep + 'forge-workspace' + sep + 'plugins' + sep + me.model.get('name');
				
					var createView = new Toolkit.Views.CreatePluginTemplate({
						model: new Backbone.Model({
							path: path,
							plugin: me.model
						})
					});
					var modal = new Toolkit.Views.ModalDialog({
						title: 'Create a new template for this module',
						contentView: createView,
						clickOutsideToCancel: true
					});
					createView.on('finish', function () {
						modal.hide(function () { modal.remove(); });
					});
					modal.show();
				});

				return false;
			},

			_clickedUpload: function (e) {
				e.preventDefault();
				this.$('.spinner').addClass('spinner-on');
				var local = Toolkit.localPluginsFile.getLocal(this.model);
				var rpc = Toolkit.rpcSession.remoteCall('dynamic.make_call', {
					'python_path': local.get('plugin_dynamic_path'),
					'method_name': 'plugin.upload',
					params: {
						cookies: Toolkit.cookies,
						path: local.get('path')
					}
				});
				rpc.on('success', function () {
					// TODO: Update model => rerender
					window.location.reload();
				});
			},

			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
				this.listenTo(this.model, 'all', this.render);
				this.listenToWhenAvailable(this.model, 'project', 'change:name', this.render);

				Toolkit.localPluginsFile.doWhenLocalAvailable(this.model, this, function (local) {
					this.listenTo(local, 'change', this.render);
				});
			},

			_prepareProblemForMustache: function (problem) {
				if (problem.type == 'good') {
					problem['class'] = 'success';
					problem.icon = 'ok-sign';
				} else if (problem.type == 'warning') {
					problem['class'] = 'warning';
					problem.icon = 'warning-sign';
				} else if (problem.type == 'error' || problem.type == 'fatal') {
					problem['class'] = 'error';
					problem.icon = 'exclamation-sign';
				}
				return problem;
			},

			templateAttributes: function () {
				var local = Toolkit.localPluginsFile.getLocal(this.model);
				var attributes = _({}).extend(this.modelAttributes(), local && local.attributes);

				if (attributes.inspector_android) {
					attributes.inspector_android = this._prepareProblemForMustache(attributes.inspector_android);
				}
				if (attributes.inspector_ios) {
					attributes.inspector_ios = this._prepareProblemForMustache(attributes.inspector_ios);
				}
				if (attributes.problems) {
					attributes.problems = _.map(attributes.problems, this._prepareProblemForMustache);
				}

				attributes.problemsPreventUpload = attributes.problems && _(attributes.problems).any(function (problem) {
					return problem.type === 'fatal';
				});

				return attributes;
			}
		})
	});

	Toolkit.Views.PluginVersions = GenericPluginView.extend({
		active: 'Versions',
		ContentView: Toolkit.Views.ViewWithTemplate.extend({
			template: 'plugins/versions',

			initialize: function () {
				Toolkit.Views.ViewWithTemplate.prototype.initialize.apply(this, arguments);
				this.listenTo(this.model, 'all', this.render);
				this.listenToWhenAvailable(this.model, 'project', 'change:name', this.render);
			},

			templateAttributes: function () {
				var attributes = _({}).extend(this.modelAttributes(), {
					onlyOnePublished: _.reduce(_.pluck(this.model.get('versions'), 'published'), function (m, value) { return m + (value ? 1 : 0); }, 0) == 1
				});

				return attributes;
			}
		})
	});
}());
