/*global Toolkit*/
/*global $*/
/*global _*/
/*global Backbone*/
/*global _superApply*/
(function () {
	Toolkit.Models.Project = Toolkit.Models.RpcModel.extend({
		method: "api.get",
		cacheName: 'Project',

		args: function () {
			return {
				cookies: Toolkit.cookies,
				method: "project/" + this.id
			};
		},

		load: function () {
			// the result from the server doesn't come with a "user id" of the parent,
			// but we only ever have one user Model, which is just used for the top level
			// of all the remotely fetched data
			var result = Toolkit.Models.Project.__super__.load.apply(this, arguments);
			this.loadSet('user_id', 1);
			return result;
		},

		loadDependencies: function (cb) {
			var project = this;
			if (this.get("_dependencies_loaded")) {
				cb && cb(this);
			} else {
				if (this.get("_dependencies_loading")) {
					this.get("_dependencies_loading_cbs").push(cb);
				} else {
					project.set("_dependencies_loading", true);
					project.set("_dependencies_loading_cbs", [cb]);
					var rpc = Toolkit.rpcSession.remoteCall('api.get', {
						method: "modules/dependencies/?project="+this.id,
						cookies: Toolkit.cookies
					});
					rpc.on('success', function (e) {
						if (e.data._invalidate_on) {
							var invalidator = {
								invalidate: function () {
									// Invalidate currently loaded dependencies but don't immediately reload
									project.set("_dependencies_loaded", false);
								}
							};
							_.each(e.data._invalidate_on,function (invalidate_on) {
								if (Toolkit.invalidationRules[invalidate_on] === undefined) {
									Toolkit.invalidationRules[invalidate_on] = [invalidator];
								} else if (Toolkit.invalidationRules[invalidate_on].indexOf(invalidator) === -1) {
									Toolkit.invalidationRules[invalidate_on].push(invalidator);
								}
							});
						}

						project.set("dependencies", e.data.data);
						project.set("_dependencies_loaded", true);
						_.each(project.get("_dependencies_loading_cbs"), function (cb) {
							cb && cb(project);
						});
						project.set("_dependencies_loading", false);
						project.set("_dependencies_loading_cbs", undefined);
					});
				}
			}
		},

		parent: {
			attribute: 'user',
			idAttribute: 'user_id',
			model: 'User',
			collectionAttribute: 'projects'
		},

		children: {
			apps: Toolkit.Models.App,
			plugins: Toolkit.Models.Plugin,
			modules: Toolkit.Models.Module
		},

		inject: {
			apps: function (apps) {
				return new Toolkit.Collections.AppCollection(apps);
			},

			plugins: function (plugins) {
				return new Backbone.Collection(plugins, {
					model: Toolkit.Models.Plugin
				});
			},

			modules: function (modules) {
				return new Backbone.Collection(modules, {
					model: Toolkit.Models.Module,
					comparator: function (model) {
						return model.id;
					}
				});
			},

			memberships: function (memberships) {
				return new Backbone.Collection(memberships);
			}
		}
	});

	Toolkit.Collections.ProjectCollection = Backbone.Collection.extend({
		model: Toolkit.Models.Project,
		comparator: function (project) {
			return project.get("name") && project.get("name").toLowerCase();
		}
	});

	// top level of the remotely fetched models
	Toolkit.Models.User = Toolkit.Models.RpcModel.extend({
		method: "api.get",
		cacheName: 'User',

		args: {
			cookies: Toolkit.cookies,
			method: "project"
		},

		children: {
			projects: Toolkit.Models.Project
		},

		inject: {
			projects: function (projects) {
				return new Toolkit.Collections.ProjectCollection(projects);
			}
		}
	});

	Toolkit.user = new Toolkit.Models.User({id: 1});

	var ProjectListView = Toolkit.Views.ProjectListView = Toolkit.Views.ViewWithTemplate.extend({
		className: "view-projects",

		template: "index",

		initialize: function () {
			_superApply(ProjectListView, this, 'initialize', arguments);
			this.listenTo(this.model, 'loaded', this.render);
			this.listenTo(this.model, 'loaded', this.updateBreadcrumbs);
		},

		breadcrumbs: function () {
			return this.rootBreadcrumbs();
		}
	});

	Toolkit.Views.ProjectDetailsView = Toolkit.Views.ViewWithTemplate.extend({
		className: "view-project",

		template: "projects/index",

		events: {
			"click .appImport": "clickedAppImport",
			"click .appCreate": "clickedAppCreate",
			"click .moduleCreate": "clickedModuleCreate",
			"click .pluginCreate": "clickedPluginCreate"
		},

		clickedAppImport: function () {
			var modal = new Toolkit.Views.ModalDialog({
				title: 'Import an app',
				contentView: new Toolkit.Views.AppImportView({
					model: this.model
				})
			});
			modal.show();
		},

		clickedAppCreate: function () {
			var modal = new Toolkit.Views.ModalDialog({
				title: 'Create a new app',
				contentView: new Toolkit.Views.AppCreateView({
					model: this.model
				})
			});
			modal.show();
		},

		clickedModuleCreate: function () {
			var modal = new Toolkit.Views.ModalDialog({
				title: 'Create a new module',
				contentView: new Toolkit.Views.ModuleCreateView({
					model: this.model
				})
			});
			modal.show();
		},

		clickedPluginCreate: function () {
			var modal = new Toolkit.Views.ModalDialog({
				title: 'Create a new plugin',
				contentView: new Toolkit.Views.PluginCreateView({
					model: this.model
				})
			});
			modal.show();
		},

		initialize: function () {
			this.listenTo(this.model, 'all', this.render);
			this.listenTo(Toolkit.localAppsFile.get('collection'), 'all', this.render);
			this.listenToWhenAvailable(this.model, 'apps', this.render);
			this.listenToWhenAvailable(this.model, 'plugins', this.render);
			this.listenToWhenAvailable(this.model, 'modules', this.render);
			this.listenTo(this.model, 'change:name all', this.updateBreadcrumbs);
		},

		templateAttributes: function () {
			var apps = this.model.get('apps');
			return _.extend({
				appsReady: !!apps,
				appList: !!apps && _.map(apps.withLocalApps(), function (app) {
					var local = app.get("local");
					return $.extend({}, app.attributes, {
						name: local.get('config') ? local.get('config').name : 'Unknown app'
					});
				})
			}, this.modelAttributes());
		},
		
		breadcrumbs: function () {
			return this.projectBreadcrumbs(this.model);
		},

		render: function() {
			this.renderTemplate();
			this.updateBreadcrumbs();
			this.renderFeatureStyle(this.model);
			return this;
		}
	});
})();
