/*global $*/
/*global _*/
/*global Backbone*/
/*global RpcSession*/
/*global Mustache*/
/*global Raven*/
var Toolkit = (function () {
    var Models = {};
    var Views = {};
    // window.ToolkitTemplates is generated by /static/templates
    var Templates = window.ToolkitTemplates;
    var Collections = {};

    var preReadyCount = 0;
    var isReady = false;
    function registerPreReadyFunction(func) {
        if (isReady) {
            throw Error("Called preready after ready fired");
        }
        preReadyCount++;
        func(function () {
            preReadyCount--;
            if (preReadyCount === 0) {
                isReady = true;
                Toolkit.events.trigger("ready");
            }
        });
    }

    // swap out page view and don't allow any more page transitions to occur
    function permanentPageTransition(view) {
        this.pageTransition(view);
        Toolkit.router.navigate = function () {};
        this.pageLocked = true;
    }

    function startNavigation() {
        if (!this.pageLocked) {
            $('#page-container').html('');
            Backbone.history.start();
        }
    }

    // swap out the current page view with the given view, cleaning up the old view
    function pageTransition(view) {
        if (this.pageLocked) {
            console.error('Tried to transition to view while page was locked');
            return;
        }

        if (this.currentPageView) {
            this.currentPageView.close();
        }

        this.currentPageView = view;

        $('#page-container').append(this.currentPageView.el);
    }

    function init() {
        Toolkit.events.trigger('init');

        $(function () {
            Toolkit.events.trigger('dom');
        });
    }

    function websiteFailedToload() {
        Toolkit.pageTransition(new Toolkit.Views.FailedToLoadWebsiteView().render());
    }

    function checkAuth(success) {
        Toolkit.authStatus.set({status: 'checking'});

        var rpc = Toolkit.rpcSession.remoteCall('auth.check', {
            cookies: Toolkit.cookies
        });

        rpc.on('success', function (e) {
            if (e.data.authenticated) {
                console.log('Already logged in', e.data);
                Toolkit.authStatus.set({user: e.data.user, validated_email: e.data.validated_email, status:'loggedin', thirdparty_module_warning: e.data.thirdparty_module_warning});
            } else {
                console.log('Not logged in');
                Toolkit.authStatus.set({status:'loggedout'});
                window.location.hash = 'auth/login';
            }
            success && success();
        });
    }

    function showNotification(text) {
        var notification = $(Mustache.to_html(Toolkit.Templates['component/notification'], {content: text}));
        $('body').append(notification);
        notification.fadeIn();
        setTimeout(function () {
            notification.fadeOut(function () {
                notification.detach();
            });
        }, 3000);
    }

    function showError(e) {
        var view = new Views.ErrorView({
            model: new Backbone.Model({
                message: e.message,
                traceback: e.traceback
            })
        });
        var modalView = new Toolkit.Views.ModalDialog({
            title: "Unexpected error",
            contentView: view
        });
        modalView.show();
        // Make sure the iframe/window is big enough to see the error
        $('body').css('min-height', '600px');
    }

    var conditionToPreventNavigation;
    var messageWhenPreventingNavigation;
    var postNavigationHook;
    var previousFragment;

    function shouldPreventNavigation() {
        // if the toolkit has just updated or if polling has failed, the toolkit has shutdown, no need to warn the user
        return !Toolkit.hasUpdated && !Toolkit.hasTimedout && conditionToPreventNavigation && conditionToPreventNavigation();
    }

    function shouldNavigate() {
        if (shouldPreventNavigation()) {
            var message = messageWhenPreventingNavigation();
            var leave = confirm(message + "\n\nAre you sure you want to leave this page?");

            if (!leave) {
                Toolkit.router.navigate(previousFragment, false);
                return false;
            }

            postNavigationHook && postNavigationHook();
        }

        conditionToPreventNavigation = undefined;
        messageWhenPreventingNavigation = undefined;
        postNavigationHook = undefined;
        previousFragment = undefined;

        return true;
    }


    window.onbeforeunload = function (e) {
        if (shouldPreventNavigation()) {
            var message = messageWhenPreventingNavigation();
            e = e || window.event;

            // For IE and Firefox prior to version 4
            if (e) {
                e.returnValue = message;
            }

            // For Chrome, Safari and Opera 12+
            return message;
        }
    };

    // the thing currently preventing navigation
    // usually a view
    var preventing;

    function preventNavigationFromTask(task, owner) {
        conditionToPreventNavigation = function () { return task.get('status') === 'running'; };
        messageWhenPreventingNavigation = function () { return task.messageWhenPreventingNavigation(); };
        postNavigationHook = function () { return task.kill(); };
        previousFragment = window.location.hash.substring(1);
        preventing = owner;
    }

    function preventNavigation(warning, owner) {
        conditionToPreventNavigation = function () { return true; };
        messageWhenPreventingNavigation = function () { return warning; };
        postNavigationHook = undefined;
        previousFragment = window.location.hash.substring(1);
        preventing = owner;
    }

    function allowNavigation(owner) {
        if (preventing !== owner) {
            console.error("Something tried allow navigation that wasn't the last thing to prevent it", owner);
            return;
        }
        preventing = undefined;
        conditionToPreventNavigation = undefined;
        messageWhenPreventingNavigation = undefined;
        postNavigationHook = undefined;
        previousFragment = undefined;
    }

    function shutdown() {
        console.error("SHUTDOWN");
        Toolkit.rpcSession.remoteCall('util.shutdown');
        setTimeout(function () {
            window.location.reload();
        }, 2000);
    }

    return {
        Models: Models,
        Views: Views,
        Templates: Templates,
        Collections: Collections,
        Mixins: {},

        // utils
        init: init,
        pageTransition: pageTransition,
        permanentPageTransition: permanentPageTransition,
        pageLocked: false,

        showNotification: showNotification,
        showError: showError,
        shouldNavigate: shouldNavigate,
        startNavigation: startNavigation,
        preventNavigationFromTask: preventNavigationFromTask,
        preventNavigation: preventNavigation,
        allowNavigation: allowNavigation,
        shutdown: shutdown,
        websiteFailedToload: websiteFailedToload,
        registerPreReadyFunction: registerPreReadyFunction,

        cookies: new Backbone.Model(),

        // cached views
        globalNavigationView: null,
        browseAppsView: null,
        currentPageView: null,

        checkAuth: checkAuth,
        events: _.extend({}, Backbone.Events),
        websiteUrl: function () {
            // chop off "api/"
            return Toolkit.server.substring(0, Toolkit.server.length - 4);
        },

        // wrapper for raven if we should decide to use something else
        captureException: function () {
            if (window.DEBUG) {
                return;
            }
            // return Raven.captureException.apply(this, arguments);
        }
    };
}());

// init, (dom), session, cookies, auth, ready

Toolkit.events.on('init', function () {
    // Init some useful globals
    Toolkit.authStatus = new Toolkit.Models.AuthStatus({status: 'loggedout'});
    Toolkit.router = new Toolkit.Router();
    Toolkit.breadcrumbsCollection = new Backbone.Collection([]);

    Toolkit.rpcSession = new RpcSession();
    Toolkit.rpcSession.init(function () {
        Toolkit.events.trigger('rpc');
        Toolkit.registerPreReadyFunction(function (done) {
            done();
        });
    });
});

Toolkit.events.on('dom', function () {
    Toolkit.globalNavigationView = new Toolkit.Views.GlobalNavigationView().render();
    $('#masthead-container').prepend(Toolkit.globalNavigationView.el);

    var breadcrumbsView = new Toolkit.Views.BreadcrumbsView({collection: Toolkit.breadcrumbsCollection});
    $('#breadcrumbs-container').html(breadcrumbsView.el);

    $('#page-container').html('<p class="loading"><i class="icon-cog icon-spin"></i> Loading...</p>');
});

Toolkit.events.on('rpc', function () {
    // Always check for an update first thing
    Toolkit.rpcSession.remoteCall('util.check_for_update');
});

Toolkit.events.on('rpc', function () {
    // Setup automatic reloading of invalidated models
    Toolkit.invalidationRules = {};
    Toolkit.rpcSession.remoteCall('invalidator.monitor').on('invalidate', function (data) {
        var tmp;
        console.log("Invalidated: "+data.name);
        if (data.name == '*') {
            tmp = Toolkit.invalidationRules;
            Toolkit.invalidationRules = {};
            _.each(_.unique(_.flatten(_.values(tmp))), function (invalidated) {
                invalidated.invalidate();
            });
        } else {
            if (Toolkit.invalidationRules[data.name] !== undefined) {
                tmp = Toolkit.invalidationRules[data.name];
                Toolkit.invalidationRules[data.name] = undefined;
                _.each(tmp, function (invalidated) {
                    invalidated.invalidate();
                });
            }
        }
    });
});

Toolkit.events.on('rpc', function () {
    Toolkit.registerPreReadyFunction(function (done) {
        var rpc = Toolkit.rpcSession.remoteCall('api.server');
        rpc.on("success", function (e) {
            Toolkit.server = e.data;

            var frameReady = function () {
                Toolkit.websiteFrame.postMessage({method: 'getSession'}, Toolkit.server.substr(0,Toolkit.server.length-4));

                // Setup method to forward messages to website analytics
                Toolkit.trackPageview = function (page) {
                    Toolkit.websiteFrame.postMessage({method: 'trackPageview', page: page}, Toolkit.server.substr(0,Toolkit.server.length-4));
                };
            };

            if (window.top == window) {
                // We're the top level, make an iframe to the website
                var view = new Toolkit.Views.IframeSessionView({
                    model: new Backbone.Model({
                        server: e.data,
                        port: window.location.port
                    })
                }).render();
                $('body').append(view.el);
                var loaded = false;
                var session = false;
                setTimeout(function () {
                    if (!loaded) {
                        Toolkit.websiteFailedToload();
                    }
                }, 5000);
                view.$el.find('iframe').on('load', function () {
                    loaded = true;
                    setTimeout(function () {
                        if (!session) {
                            Toolkit.websiteFailedToload();
                        }
                    }, 1000);
                    var iframe = view.$el.find('iframe').get(0);

                    Toolkit.websiteFrame = iframe.contentWindow;

                    frameReady();
                });
            } else {
                // We're embedded, assume our parent is the website
                Toolkit.websiteFrame = window.parent;

                // Let the website know we're here
                Toolkit.websiteFrame.postMessage({
                    method: "init"
                }, Toolkit.server.substr(0,Toolkit.server.length-4));

                frameReady();
            }
            window.addEventListener("message", function (evt) {
                if (e.data.indexOf(evt.origin) === 0) {
                    if (evt.data.method == 'getSession') {
                        // Session data recieved, save and begin routing.
                        session = true;
                        Toolkit.cookies.set(evt.data.data);
                        Toolkit.events.trigger("cookies");
                        Toolkit.checkAuth(function () {
                            Toolkit.events.trigger("auth");
                            done();
                        });
                    } else if (evt.data.method == 'loggedIn') {
                        // Login event recieved, update auth
                        Toolkit.cookies.set(evt.data.data);
                        Toolkit.checkAuth(function () {
                            Toolkit.router.navigate('/', true);
                        });
                    } else if (evt.data.method == 'hello') {
                        // Used when embedding the website so the website knows we're here
                        evt.source.postMessage({method: "init"}, Toolkit.server.substr(0,Toolkit.server.length-4));
                    } else if (evt.data.method == "resizeEmbed") {
                        $("iframe").each(function (i, iframe) {
                            if (iframe.contentWindow === evt.source) {
                                $(iframe).css("height", evt.data.height + "px");
                                if (window != window.top) { // bubble resize message to parent
                                    window.parent.postMessage({
                                        method: "resizeEmbed",
                                        height: $('body').height()
                                    }, Toolkit.server);
                                }
                            }
                        });
                    } else if (evt.data.method == "scrollEmbed") {
                        $("iframe").each(function (i, iframe) {
                            if (iframe.contentWindow === evt.source) {
                                var scrollTop = $(iframe).offset().top + evt.data.top;

                                // if we're not the top window, we shouldn't scroll
                                if (window != window.top) {
                                    window.parent.postMessage({
                                        method: "scrollEmbed",
                                        top: scrollTop
                                    }, Toolkit.server);
                                    return;
                                }

                                window.scrollTo(0, scrollTop);
                            }
                        });
                    }
                }
            }, false);
            // notify iframe's parents of size change
            if (window != window.top) {
                document.body.addEventListener("DOMSubtreeModified", function() {
                    window.parent.postMessage({
                        method: "resizeEmbed",
                        height: $('body').height()
                    }, Toolkit.server);
                }, false);
            }
        });
    });
});

Toolkit.events.on('ready', function () {
    Toolkit.startNavigation();
});
