/*global Backbone*/
/*global _*/
/*global Toolkit*/

window.NO_TIMEZONE_FORMAT = 'YYYY-MM-DDTHH:mm:ss';

// helper that hopefully makes calling a method in the parent prototype less error prone
function _super(callMethod, args) {
	var subtype = args[0];
	var context = args[1];
	var methodName = args[2];

	var superObject = subtype.__super__;
	if (!superObject) {
		throw new Error("Tried to call super method for a type that doesn't have a super type.");
	}

	var superMethod = superObject[methodName];

	if (!superMethod) {
		throw new Error("Tried to call super method '"+methodName+"' for a type, but the parent didn\'t have such a method.");
	}

	if (!context.constructor || !context.constructor.__super__) {
		throw new Error("Tried to call super method '"+methodName+"' with an object that doesn\'t look like a Backbone type.");
	}

	var superArgs;
	if (callMethod === 'call') {
		superArgs = Array.prototype.slice.call(args, 3);
	} else {
		superArgs = Array.prototype.slice.call(args, 3, 4)[0];
	}

	return superMethod.apply(context, superArgs);
}

window._superCall = function _superCall() {
	return _super('call', arguments);
};

function _superApply() {
	return _super('apply', arguments);
}

// view that comes with a helper function for cleaning up registered event handlers
var BreadcrumbsAwareView = Backbone.View.extend({
	// The default breadcrumbs
	breadcrumbs: function () {
		return false;
	},
	// Called to update the global breadcrumbsCollection
	// Expects to find this.breadcrumbs which returns a list of breadcrumbs sections
	// See rootBreadcrumbs, projectBreadcrumbs and friends...
	updateBreadcrumbs: function () {
		if (this.breadcrumbs() !== false) {
			Toolkit.breadcrumbsCollection.reset(this.breadcrumbs());
		}
	},

	rootBreadcrumbs: function (extra) {
		return [{
			link: "#",
			text: "Projects",
			current: !extra
		}].concat(extra || []);
	},

	projectBreadcrumbs: function (project, extra) {
		return this.rootBreadcrumbs([
			{
				link: (project && "#projects/"+project.id),
				text: (project && project.get('name')),
				current: !extra
			}
		].concat(extra || []));
	},

	// helper to construct a set of breadcrumbs leading up to an app
	appBreadcrumbs: function (app, extra) {
		var local = app.get("local");
		var config = local && local.get('config');
		var project = app.get('project');
		return this.projectBreadcrumbs(project, [
			{
				link: (app && "#apps/"+app.get('uuid')),
				text: (config && config.name),
				current: !extra
			}
		].concat(extra || []));
	},

	pluginBreadcrumbs: function (plugin, extra) {
		var project = plugin.get('project');
		return this.projectBreadcrumbs(project, [
			{
				link: (plugin && "#plugins/"+plugin.id),
				text: (plugin && "Plugin: "+plugin.get('name')),
				current: !extra
			}
		].concat(extra || []));
	},

	moduleBreadcrumbs: function (plugin, extra) {
		var project = plugin.get('project');
		return this.projectBreadcrumbs(project, [
			{
				link: (plugin && "#modules/"+plugin.id),
				text: (plugin && "Module: "+plugin.get('name')),
				current: !extra
			}
		].concat(extra || []));
	},

	render: function () {
		_superApply(BreadcrumbsAwareView, this, 'render', arguments);
		this.updateBreadcrumbs();
	}
});

window.CloseableView = BreadcrumbsAwareView.extend({
	// store a reference to a view to be closed when this view is closed
	autoClose: function (view) {
		if (!this._autoClose) {
			this._autoClose = {};
		}
		this._autoClose[view.cid] = view;
	},
	close: function () {
		if (this._autoClose) {
			for (var cid in this._autoClose) {
				this._autoClose[cid].close();
			}
		}

		// removes this.el from the page
		// also removes any callbacks set on other views/models/collections
		this.remove();

		// removes any listeners for events emitted by the view itself
		this.unbind();
	}
});

function allModels(models, attribute, predicate) {
	return _.all(models, function (m) { return predicate(m.get(attribute)); });
}

function doWhenAll(models, attribute, predicate, callback) {
	var listener = this;

	if (!callback) {
		callback = predicate;
		predicate = _.identity;
	}

	if (allModels(models, attribute, predicate)) {
		callback.call(listener);
		return;
	}

	var me = this;
	var called = false;

	_.each(models, function (model) {
		me.doWhen(model, attribute, predicate, function () {
			if (!called && allModels(models, attribute, predicate)) {
				// prevent other calls
				called = true;
				callback.call(listener);
			}
		});
	});
}

function doWhen(model, attribute, predicate, callback) {
	var listener = this;

	if (!callback) {
		callback = predicate;
		predicate = _.identity;
	}

	var wrappedCallback = function () {
		var value = model.get(attribute);
		if (!predicate(value)) {
			return;
		}
		callback.call(listener, value);
		model.off('change:'+attribute, wrappedCallback);
	};


	if (model.has(attribute) && predicate(model.get('attribute'))) {
		var value = model.get(attribute);
		callback.call(listener, value);
		return;
	}

	listener.listenTo(model, 'change:'+attribute, wrappedCallback);
}

// execute the given callback when an attribute is available on a model
function doWhenAvailable(model, attribute, callback) {
	var listener = this;

	var wrappedCallback = function () {
		var value = model.get(attribute);
		callback.call(listener, value);
		model.off('change:'+attribute, wrappedCallback);
	};


	if (model.has(attribute)) {
		var value = model.get(attribute);
		callback.call(listener, value);
		return;
	}

	listener.listenTo(model, 'change:'+attribute, wrappedCallback);
	return this;
}

// listen for events on a model attribute that can emit events,
// as soon as it is available
function listenToWhenAvailable(model, attribute, event, callback) {
	if (!callback) {
		callback = event;
		event = 'all';
	}

	var listener = this;
	listener.doWhenAvailable(model, attribute, function (emitter) {
		listener.listenTo(emitter, event, callback);
	});
}

// helper for passing attributes of a nested model into a template
// e.g. app.getAttributes('project')
// e.g. buildEvent.getAttributes('app.project')
function getAttributes(dotted) {
	var keys = dotted.split('.');
	var model = this;
	for (var i = 0; i < keys.length; ++i) {
		var key = keys[i];
		model = model.get(key);
		if (!model) {
			break;
		}
	}
	return model ? model.attributes : undefined;
}

var constructors = [Backbone.Model, Backbone.Collection, Backbone.View, Backbone.Router];
for (var i = 0; i < constructors.length; ++i) {
	constructors[i].prototype.doWhen = doWhen;
	constructors[i].prototype.doWhenAll = doWhenAll;
	constructors[i].prototype.doWhenAvailable = doWhenAvailable;
	constructors[i].prototype.listenToWhenAvailable = listenToWhenAvailable;
	constructors[i].prototype.getAttributes = getAttributes;
}


/**
 * If you click on the icon inside a bootstrap button the event.target is set to
 * the icon, and not the button.
 *
 * Usage:
 *
 *    someButtonEventHandler: function (e) {
 *        fixBootstrapButtonEvent(e);
 *        ... handle event normally ...
 *    }
 */
window.fixBootstrapButtonEvent = function (e) {
	if (e.target.tagName === "I") {
		e.target = e.target.parentElement;
	}
};

/**
 * Listen for changes on one attribute and keep another one up to date using a function.
 * @this {Backbone.Model}
 * @param dependency {string}
 * @param computer {function}
 */
function computedAttribute(dependency, computer) {
	this.listenTo(this, 'change:'+dependency, computer);
	computer.call(this);
}

Backbone.Model.prototype.computedAttribute = computedAttribute;

/**
 * Make an object with the given arguments as keys, all mapped to true, to act
 * like a set.
 */
window.makeSet = function () {
	var set = {};
	for (var i = 0; i < arguments.length; ++i) {
		set[arguments[i]] = true;
	}
	return set;
};

/**
 * Return true iff the last characters in haystack are identical to the needle passed in
 * @param haystack {string}
 * @param needle {string}
 */
function endsWith (haystack, needle) {
	// see http://aganov.github.io/underscore-strings/docs/underscore.strings.html
	var d = haystack.length - needle.length;
	return d >= 0 && haystack.indexOf(needle, d) === d;
}

/**
 * Returns true if object has nested property
 */
function hasNestedProperty(obj, path) {
	if (obj === null || typeof obj === "undefined") {
		return false;
	}
	var apath = path.split(".");
	for (var i = 0; i < apath.length; ++i) {
		if (obj !== null && typeof obj !== "undefined" && obj.hasOwnProperty(apath[i])) {
			obj = obj[apath[i]];
		} else {
			return false;
		}
	}
	return true;
}
