import os
import shutil
import logging
import markdown
import pystache
import re
import codecs
import urlparse
from BeautifulSoup import BeautifulSoup

import markdown_api
import markdown_note

logging.basicConfig(level=logging.DEBUG, format='[%(levelname)7s] %(asctime)s -- %(message)s')
log = logging.getLogger(__name__)

def generate(type, output, input, module, base_url='https://trigger.io/'):
	api_ext = markdown_api.ApiExtension(configs={})
	note_ext = markdown_note.NoteExtension(configs={})

	# Build config
	if module:
		docs_base_url = base_url + "modules/x/current/docs/"
	else:
		docs_base_url = base_url + "docs/current/"
	sections = not module
	safe = module
	output_type = type
	output_folder = os.path.abspath(output)
	input_folder = os.path.abspath(input)
	if safe:
		allowed_raw_files = ('.png', '.gif', '.jpg')
		allow_html = False
	else:
		allowed_raw_files = ('.png', '.gif', '.jpg', '.html', '.css', '.js', '.xml', '.zip')
		allow_html = True

	with codecs.open(os.path.abspath(os.path.join(__file__, '..', 'type', output_type, 'template.html')), encoding="utf8") as template_file:
		template = template_file.read()

	with codecs.open(os.path.abspath(os.path.join(__file__, '..', 'type', output_type, 'template.html.nav')), encoding="utf8") as template_nav_file:
		template_nav = template_nav_file.read()

	def file_to_structure(path):
		structure = {}
		with codecs.open(path, encoding="utf8") as file:
			str = file.read()

		structure['markdown'] = str
		structure['filename'] = os.path.splitext(os.path.split(path)[-1])[0]
		if structure['filename'] == 'index':
			structure['slug'] = os.path.split(os.path.split(path)[0])[1]
		else:
			structure['slug'] = structure['filename']

		# Remove any number prefix from filenames.
		structure['slug'] = re.sub(r'^[0-9]+_(.*)$', r'\1', structure['slug'])
		structure['filename'] = re.sub(r'^[0-9]+_(.*)$', r'\1', structure['filename'])

		structure['title'] = str.split("\n")[0]
			
		return structure

	def directory_to_structure(source):
		files = os.listdir(source)
		files = sorted(files)
		if "index.md" in files:
			files.remove("index.md")
			structure = file_to_structure(os.path.join(source, "index.md"))
			structure['children'] = []
			for file in files:
				file_path = os.path.join(source, file)
				if os.path.isdir(file_path):
					child_structure = directory_to_structure(file_path)
					if child_structure is not None:
						structure['children'].append(child_structure)
				elif file_path.endswith('.md'):
					structure['children'].append(file_to_structure(file_path))
			return structure
		else:
			log.info("No index.md in "+source+" not looking any deeper for markdown")
			return None

	def copy_raw_files(source, dest):
		for dirpath, dirnames, filenames in os.walk(source):
			for filename in filenames:
				if os.path.splitext(filename)[1] in allowed_raw_files:
					if not os.path.isdir(re.sub(r'[0-9]+_([^\/]*)', r'\1', os.path.join(dest, dirpath[len(source)+1:]))):
						os.makedirs(re.sub(r'[0-9]+_([^\/]*)', r'\1', os.path.join(dest, dirpath[len(source)+1:])))
					
					shutil.copy2(os.path.join(dirpath, filename), re.sub(r'[0-9]+_([^\/]*\/)', r'\1', os.path.join(dest, dirpath[len(source)+1:], filename)))
				elif not filename.endswith('.md'):
					log.warn("Ignoring not-whitelisted filetype: "+os.path.join(dirpath, filename))

	def generate_html(md, nav, path, depth, dest):
		if not os.path.exists(os.path.split(dest)[0]):
			os.makedirs(os.path.split(dest)[0])
			
		def nav_for_dict(nav, path, ul_class="nav nav-list", top=False):
			if "top" in nav:
				nexttop = True
				out = ''
			else:
				nexttop = False
				li_class = ''
				if "active" in nav:
					li_class = 'active'
				if "section" in nav:
					li_class = 'nav-header'
					
				rel_path = [".."] * len(depth) + path[1:] + [nav['slug']]
				if nav['filename'] == 'index':
					rel_path += ['index']
				
				rel_path = [x for x in rel_path if x != ""]
				
				title = markdown.markdown(nav['title'], [], safe_mode=False)
				# Stripe <p> and </p>
				title = title[3:-4]

				if "section" in nav:
					link = title
				else:
					link = '<a href="%s.html">%s</a>' % ("/".join(rel_path), title)

				out = '<ul class="%s"><li class="%s">%s</li><li>' % (ul_class, li_class, link)
			if "children" in nav:
				for child in nav["children"]:
					out += nav_for_dict(child, path + [nav['slug']], top=nexttop)
			if "top" not in nav:
				out += '</li></ul>'
			return out

		def child_nav(nav, path):
			out = ''
			if "children" in nav:
				for child in nav["children"]:
					if "active" in nav:
						out += nav_for_dict(child, path + [nav['slug']], ul_class='unstyled')
					else:
						out += child_nav(child, path + [nav['slug']])
			return out
			
		if allow_html:
			markdown_safe_mode = False
		else:
			markdown_safe_mode = 'escape'

		# can't use all of 'extra' - attr_list allows injection of JS, but can include the rest
		extensions = ['sane_lists', 'headerid', 'abbr', 'def_list', 'fenced_code', 'footnotes', 'tables', 'smart_strong', 'codehilite', api_ext, note_ext]

		processed_md = markdown.markdown(md, extensions, safe_mode=markdown_safe_mode, output_format='html5').replace('<table>', '<table class="table">').replace('<dl>', '<dl class="dl-wide">').replace('[TOC]', child_nav(nav, []))

		if module:
			soup = BeautifulSoup(processed_md)
			index = ''
			if len(depth) == len(path):
				if len(depth) > 0:
					index = '/index'
				else:
					index = 'index'
			page_url = "%s%s%s.html" % (docs_base_url, "/".join(path), index)
			for a in soup.findAll('a'):
				if a.has_key('href'):
					if a['href'].startswith('#'):
						continue
					abs_url = urlparse.urljoin(page_url, a['href'])
					if abs_url == page_url:
						continue
					if abs_url.startswith("%smodules/" % base_url):
						# Embeddable link
						page_rel = page_url[len(base_url):].split("/")
						link_rel = abs_url[len(base_url):].split("/")
						while page_rel[0] == link_rel[0]:
							page_rel.pop(0)
							link_rel.pop(0)
						rel_url = "../"*(len(page_rel)-1) + "/".join(link_rel)
						a['href'] = rel_url
					elif abs_url.startswith("%s" % base_url):
						# Not embeddable, but to a Trigger page
						a['href'] = abs_url
						a['target'] = "_top"
					else:
						# External link
						a['href'] = abs_url
						a['target'] = "_blank"

			processed_md = unicode(soup)
		
		with codecs.open(dest+".html", 'w', encoding="utf8") as output_file:
			log.debug("Writing: "+dest+".html")
			output_file.write(pystache.render(template, {
				"nav": nav_for_dict(nav, []),
				"docs": processed_md,
				"depth": depth
			}))
		
		with codecs.open(dest+".html.nav", 'w', encoding="utf8") as output_file:
			log.debug("Writing: "+dest+".html.nav")
			output_file.write(pystache.render(template_nav, {
				"nav": nav_for_dict(nav, []),
				"docs": processed_md,
				"depth": depth
			}))
		with codecs.open(dest+".txt", 'w', encoding="utf8") as output_file:
			log.debug("Writing: "+dest+".txt")
			for x in BeautifulSoup(processed_md, convertEntities=BeautifulSoup.HTML_ENTITIES).findAll(text=True):
				output_file.write(unicode(x))

	def generate_html_from_structure(structure, path, dest):
		nav = {
			'filename': 'index',
			'slug': '',
			'title': 'Docs',
			'children': []
		}
		if sections:
			nav['top'] = True  # Don't include in generated output.
			nav['children'].append({
				'filename': 'index',
				'slug': '',
				'title': 'Docs home',
				'children': []
			})
			if len(path) == 0:
				nav['children'][0]['active'] = True
		elif len(path) == 0:
			nav['active'] = True

		nextnav = nav
		nextsub = structure
		for (i, part) in enumerate(path):
			curnav = nextnav
			sub = nextsub
			for child in sub['children']:
				childnav = {
					'filename': child['filename'],
					'slug': child['slug'],
					'title': child['title'],
					'children': []
				}
				
				if i == 0 and sections:
					childnav['section'] = True

				if child['slug'] == part:
					if i == len(path)-1:
						childnav['active'] = True
					nextsub = child
					nextnav = childnav
				elif i == 0 and sections:
					for child in child['children']:
						child_childnav = {
							'filename': child['filename'],
							'slug': child['slug'],
							'title': child['title'],
							'children': []
						}
						childnav['children'].append(child_childnav)
			
				curnav['children'].append(childnav)
		
		curnav = nextnav
		sub = nextsub
		if "children" in sub:
			for child in sub['children']:
				childnav = {
					'filename': child['filename'],
					'slug': child['slug'],
					'title': child['title'],
					'children': []
				}

				if len(path) == 0 and sections:
					childnav['section'] = True
					for child in child['children']:
						child_childnav = {
							'filename': child['filename'],
							'slug': child['slug'],
							'title': child['title'],
							'children': []
						}
						childnav['children'].append(child_childnav)

				curnav['children'].append(childnav)

		depth = range(0, len(path)-1)
		if sub['filename'] == 'index' and len(path) > 0:
			depth += [0]
		if len(path) != 1 or not sections:
			generate_html(sub['markdown'], nav, path, depth, os.path.join(dest, sub['filename']))
		
		if "children" in sub:
			for child in sub['children']:
				if child['filename'] == 'index':
					generate_html_from_structure(structure, path + [child['slug']], os.path.join(dest, child['slug']))
				else:
					generate_html_from_structure(structure, path + [child['slug']], dest)

	if os.path.exists(output_folder):
		log.info("Removing existing output folder")
		shutil.rmtree(output_folder)

	if os.path.exists(os.path.abspath(os.path.join(__file__, '..', 'type', output_type, 'static'))):
		log.info("Copying static files")
		shutil.copytree(os.path.abspath(os.path.join(__file__, '..', 'type', output_type, 'static')), output_folder)

	log.info("Generating tree structure")
	structure = directory_to_structure(input_folder)

	log.info("Generating html")
	generate_html_from_structure(structure, [], output_folder)

	log.info("Copying non-markdown files from source.")
	copy_raw_files(input_folder, output_folder)
