import os
import logging
import json
import urlparse
import shutil
import codecs
import errno
from contextlib import contextmanager

from forge import async
from forge import build_config, defaults, ForgeError
from forge.lib import cd
import forge.lib as forge_lib
from forge import main as forge_main
from forge.remote import RequestError

from trigger.api import expose, run_in_process
from trigger import forge_tool
from trigger import lib
from trigger.lib import ensure_folder_exists


LOG = logging.getLogger(__name__)

@contextmanager
def _open_local_apps_file(*args, **kwargs):
    encoding = kwargs.get('encoding')

    active_apps_filename = os.path.join(defaults.FORGE_ROOT, 'local_apps.json')

    with async.current_call().io_lock:
        if not os.path.exists(active_apps_filename):
            LOG.info('No active apps file, creating')
            with codecs.open(active_apps_filename, mode='w', encoding=encoding) as apps_file:
                apps_file.write(u'{}')

        with codecs.open(active_apps_filename, *args, **kwargs) as f:
            yield f

def _get_local_apps():
    """Loads the locally registered apps from storage and returns them as a dict"""
    try:
        with _open_local_apps_file() as apps_file:
            return json.load(apps_file)
    except:
        return dict()

@contextmanager
def _app_config_file_for_path(path):
    """Returns a handle to the config file for the app at the given location"""
    path_to_config_for_app = os.path.join(path, 'src', 'config.json')
    with async.current_call().io_lock:
        yield forge_lib.open_file(path_to_config_for_app)

def _app_config_for_path(path):
    """Returns the configuration (as a dict) for an app at the given location"""
    with _app_config_file_for_path(path) as app_config_file:
        app_config = json.load(app_config_file)
    return app_config

def _set_local_apps(apps):
    """Dumps the given dict to storage"""
    with _open_local_apps_file('w') as apps_file:
        return json.dump(apps, apps_file, indent=4)

def set_settings(uuid, settings):
    """Updates the stored settings for an app, identified by its uuid,
    e.g. updating packaging settings

    :param uuid: {str} uuid of the app to update
    :param settings: {dict} dictionary of the local settings for this app
    """
    apps = _get_local_apps()
    if uuid not in apps:
        apps[uuid] = {}

    old_settings = apps.get(uuid)

    old_settings.update(settings)
    _set_local_apps(apps)

@expose
def import_app(cookies, path, project):
    """Attempts to use the path given as an app directory.

    1) Checks for src/config.json in the given folder
    2) Extracts uuid field
    3) See if it correlates with an app that this user has access to
    """
    path = os.path.expanduser(path)
    project = int(project)
    app_paths = [details['path'] for details in _get_local_apps().values()]
    if path in app_paths:
        raise Exception('Already have an app imported from %s' % path)

    if not os.path.isdir(path):
        raise Exception('No such folder "%s"' % path)

    try:
        config_dict = build_config.load_app(path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise Exception("%s is missing" % defaults.APP_CONFIG_FILE)
        raise

    uuid = config_dict['uuid']
    platform_version = config_dict['platform_version']
    name = config_dict.get('name')

    if project: # check with server that this app does indeed belong to the project
        app = forge_tool.singleton.remote._api_get('app/' + uuid, cookies=cookies)
        if app['data']['project_id'] != project:
            raise Exception('You cannot import an app which does not belong to this project')

    if name is None:
        name = '(Unititled App)'

    set_settings(uuid, {'path': path})

    return {
            "uuid": uuid,
            "path": path,
            "name": name,
            "platform_version": platform_version
    }


@expose
def forget(uuid):
    apps = _get_local_apps()
    del apps[uuid]
    _set_local_apps(apps)

@expose
def clean(path):
    # TODO: possibly need to check there are no running tasks
    # for the given app? might be unable to clean otherwise...
    template_dir = os.path.join(path, defaults.TEMPLATE_DIR)
    LOG.info('Removing templates dir: {0}'.format(template_dir))
    try:
        shutil.rmtree(template_dir)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise ForgeError("There was nothing to clean.")


@run_in_process
def migrate(cookies, path):
    """Rewrites src/config.json appropriately for a new schema

    :param path: Path to migrate the config file for
    """
    # Override cookies for any new remote objects
    from forge import remote
    remote.override_cookies = cookies

    with cd(path):
        # TODO: HAAACK, should extract an interface to the build tools
        # that both the CLI and web interfaces back onto instead of
        # simulating calls to the CLI
        forge_main.handle_primary_options(['migrate', '-v'])
        forge_main.handle_secondary_options('migrate', [])
        forge_main.migrate(['--general.interactive', 'no'])


def _classify_request_error(e):
    try:
        error_text = json.loads(e.response.content).get('text')

        if error_text is None:
            raise ValueError

        if error_text == 'App matching query does not exist.':
            return {
                    'type': 'uuid_unrecognised',
                    'data': {
                            'server': urlparse.urlparse(e.response.url).hostname
                    }
            }

        else:
            return {
                    'type': 'remote_error',
                    'data': {
                            'server': urlparse.urlparse(e.response.url).hostname,
                            'message': error_text
                    }
            }

    except ValueError:
        return {
                'type': 'remote_error',
                'data': {
                        'server': urlparse.urlparse(e.response.url).hostname,
                        'message': e.message
                }
        }

def _missing_file_problems(path):
    if not os.path.isdir(path):
        return {
                'type': 'folder_missing'
        }

    elif not os.path.isfile(os.path.join(path, 'src', 'config.json')):
        return {
                'type': 'config_missing'
        }

    elif not os.path.isfile(os.path.join(path, 'src', 'identity.json')):
        return {
                'type': 'identity_missing'
        }

    return False

@expose
def state(cookies, path):
    """Returns the state of an app, which includes:

    {
            'platform_version': platform version in config.json.
            'problems': a list of problems which should be displayed in the app status box.
            'should_rebuild': whether the app needs to be rebuilt due to config or platform changes.
            'did_migrate': whether the app was updated in some way to make it compatible with the platform version in config.json.
    }

    :param path: Path to the app we want the state of.
    """
    file_problem = _missing_file_problems(path)
    if file_problem:
        return dict(
                problem=file_problem,
                should_rebuild=False
        )

    app_config = build_config.load_app(path)
    app_platform = app_config.get('platform_version')
    if app_platform is None:
        return dict(
                problem={
                        'type': 'platform_version_missing'
                },
                should_rebuild=False
        )

    try:
        response_dict = forge_tool.singleton.remote.server_says_should_rebuild(path, cookies=cookies)
    except RequestError as e:
        return dict(
                platform_version=app_platform,
                problem=_classify_request_error(e),
                should_rebuild=False
        )

    if not response_dict['should_rebuild']:
        did_migrate = lib.make_dynamic_call(
                [os.path.join(path, '.template'), os.path.join(path, '.template', '.lib')],
                'generate_dynamic.customer_goals.migrate_app_to_current',
                {'path': path},
                fail_if_missing=False
        )
        if did_migrate:
            async.current_call().emit('migration')

    return dict(
            platform_version=app_platform,
            should_rebuild=response_dict['should_rebuild'],
    )

def _mark_module_unavailable(module_name, config_schema):
    modules = config_schema.get('properties', {}).get('modules', {}).get('properties', {})

    if module_name not in modules:
        return

    modules[module_name]['_unavailable'] = True

def _remove_stream_handlers():
    """Just remove all StreamHandler instances from root.
    This is used to stop there being multiple lines of output in the console when running build processes
    """
    for h in logging.root.handlers:
        if isinstance(h, logging.StreamHandler):
            logging.root.removeHandler(h)

@run_in_process
def serve(cookies, path, target):
    # Override cookies for any new remote objects
    from forge import remote
    remote.override_cookies = cookies

    build(cookies, path, target)
    with cd(path):
        # TODO: HAAACK, should extract an interface to the build tools
        # that both the CLI and web interfaces back onto instead of
        # simulating calls to the CLI
        #forge_main.handle_primary_options(['serve', '-v'])
        forge_main.handle_secondary_options('serve', [target])
        forge_main.serve(['--general.interactive', 'no'])

@run_in_process
def package(cookies, path, target):
    # Override cookies for any new remote objects
    from forge import remote
    remote.override_cookies = cookies

    build(cookies, path, target)
    with cd(path):
        # TODO: HAAACK, should extract an interface to the build tools
        # that both the CLI and web interfaces back onto instead of
        # simulating calls to the CLI
        forge_main.handle_primary_options(['package', '-v'])
        forge_main.handle_secondary_options('package', [target])
        forge_main.package(['--general.interactive', 'no'])

@run_in_process
def run(cookies, path, target):
    # Override cookies for any new remote objects
    from forge import remote
    remote.override_cookies = cookies

    build(cookies, path, target)
    with cd(path):
        # TODO: HAAACK, should extract an interface to the build tools
        # that both the CLI and web interfaces back onto instead of
        # simulating calls to the CLI
        forge_main.handle_primary_options(['run', '-v'])
        forge_main.handle_secondary_options('run', [target])
        forge_main.run(['--general.interactive', 'no'])

@run_in_process
def build(cookies, path, target=None):
    # Override cookies for any new remote objects
    from forge import remote
    remote.override_cookies = cookies

    with cd(path):
        # TODO: HAAACK, should extract an interface to the build tools
        # that both the CLI and web interfaces back onto instead of
        # simulating calls to the CLI
        forge_main.handle_primary_options(['build', '-v'])
        forge_main.handle_secondary_options('build', [])
        if target is None:
            forge_main.development_build(['--general.interactive', 'no'], has_target=False)
        else:
            forge_main.development_build([target, '--general.interactive', 'no'])
        _remove_stream_handlers()

@expose
def create(cookies, name, path, project):
    """Creates an app on the Forge website and
    downloads the initial files for the app
    """
    path = os.path.expanduser(path)
    ensure_folder_exists(path)
    if os.path.exists(os.path.join(path, defaults.SRC_DIR)):
        raise Exception('Source folder "%s" already exists, if you really want to create a new app you will need to remove it!' % defaults.SRC_DIR)

    remote = forge_tool.singleton.remote
    content_dict = remote._api_post('app', data={
            'name': name,
            'project': project,
    }, cookies=cookies)
    uuid = content_dict['data']['uuid']
    remote.fetch_initial(uuid, path, cookies=cookies)
    set_settings(uuid, {'path': path})
    return content_dict

@expose
def remove_local_apps_not_on_filesystem():
    local_apps = _get_local_apps()
    to_remove = set()
    for uuid, details in local_apps.iteritems():
        if not os.path.isdir(details['path']):
            to_remove.add(uuid)

    for uuid in to_remove:
        del local_apps[uuid]

    _set_local_apps(local_apps)

@expose
def platform_targets(path):
    """Returns the targets supported by the current platform"""
    try:
        targets = lib.make_dynamic_call(
                [os.path.join(path, '.template'), os.path.join(path, '.template', '.lib')],
                'generate_dynamic.build.enabled_platforms',
                { 'build_type_dir': path },
                fail_if_missing = True
        )
    except:
        targets = [     'safari', 'osx', 'firefox', 'reload', 'web', 'chrome', 'ios',
                                'android', 'wp', 'ie', 'ios-native', 'android-native' ]

    return targets
