import os
import subprocess
import sys
import logging
import traceback
import ctypes

from trigger.api import expose
from trigger import lib
from forge import lib as forge_lib


LOG = logging.getLogger(__name__)

def _has_hidden_attribute(filepath):
    try:
        attrs = ctypes.windll.kernel32.GetFileAttributesW(unicode(filepath))
        assert attrs != -1
        result = bool(attrs & 2)
    except (AttributeError, AssertionError):
        result = False
    return result

def _listdir(directory, show_hidden):
    if not show_hidden:
        if sys.platform.startswith("win"):
            files = []
            for p in os.listdir(directory):
                if p.startswith("."):
                    continue
                if _has_hidden_attribute(os.path.join(directory, p)):
                    continue
                files.append(p)
            return files
        else:
            return [p for p in os.listdir(directory) if not p.startswith(".")]
    return os.listdir(directory)

@expose
def list_contents(path, root, show_hidden):
    """List the contents of a directory.

    :param path: Path to the directory to get contents of. If this is None or the empty string,
            we get the contents of the user's home directory.

    Return value takes is a dict with the form:

    {
            'path': list representation of path to requested directory, including drive
            'contents': list of the files contained in the requested directory
            'separator': path separator for host OS
    }
    """
    if not path:
        path = os.path.expanduser(u'~')

    if not root:
        root = os.path.abspath(u'/')

    def _is_folder(f):
        return os.path.isdir(os.path.join(path, f))

    directory_contents = _listdir(path, show_hidden=show_hidden)
    sorted_contents = sorted(directory_contents, key=lambda x: x.lower())
    categorised_contents = [(f, _is_folder(f)) for f in sorted_contents]

    path_parts = lib.path_components(path)
    root_parts = lib.path_components(root)

    return {
            "root_parts": root_parts,
            "path": path_parts,
            "contents": categorised_contents,
            "separator": os.sep
    }

def _splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

@expose
def list_recursive(path):
    path = os.path.abspath(path)
    contents = {}
    for dir_path, dirnames, filenames in os.walk(path):
        cur_contents = contents
        if dir_path != path:
            relative_path = dir_path[len(path)+1:]
            parts = _splitall(relative_path)
            while len(parts) > 0:
                part = parts.pop(0)
                if part in cur_contents:
                    cur_contents = cur_contents[part]
                else:
                    cur_contents[part] = {}
                    cur_contents = cur_contents[part]
        for dirname in dirnames:
            cur_contents[dirname] = {}
        for filename in filenames:
            cur_contents[filename] = "file"
    return contents

@expose
def create(path):
    os.makedirs(path)


@expose
def open_in_filebrowser(path):
    lib.open_in_filebrowser(path)


@expose
def get_home_directory():
    return {
            'userdir': os.path.expanduser('~'),
            'separator': os.path.sep,
    }


@expose
def reveal_file(path):
    normedpath = os.path.normpath(path)
    if sys.platform.startswith("win"):
        forge_lib.PopenWithoutNewConsole(r'explorer.exe /select,"%s"' % normedpath)
    elif sys.platform.startswith("darwin"):
        forge_lib.PopenWithoutNewConsole(['open', '-R', normedpath])
