import logging
import json
import os
import traceback
import zipfile
import webbrowser

from send2trash import send2trash

import docgen
from forge import defaults, remote
from forge import async
from trigger.lib import make_dynamic_call
from trigger.api import expose
from trigger import forge_tool

LOG = logging.getLogger(__name__)

def _update_module_dynamic_if_needed(cookies, path, manifest_platform_version):
    dynamic_platform_version = ""
    dynamic_platform_version_path = os.path.join(path, ".trigger", "platform_version.txt")
    if os.path.exists(dynamic_platform_version_path):
        with async.current_call().io_lock:
            with open(dynamic_platform_version_path, 'r') as dynamic_platform_version_file:
                dynamic_platform_version = dynamic_platform_version_file.read()

    if dynamic_platform_version != manifest_platform_version:
        _update_module_dynamic(cookies, path, manifest_platform_version)

def _update_module_dynamic(cookies, path, core_version):
    if os.path.exists(os.path.join(path, '.trigger')):
        for dirpath, dirnames, filenames in os.walk(os.path.join(path, '.trigger'), topdown=False):
            for filename in filenames:
                send2trash(os.path.join(dirpath, filename))
            for dirname in dirnames:
                send2trash(os.path.join(dirpath, dirname))
        send2trash(os.path.join(path, '.trigger'))

    os.makedirs(os.path.join(path, ".trigger"))

    module_dynamic_file = os.path.join(path, ".trigger", "module_instructions.zip")
    forge_tool.singleton.remote._get_file(forge_tool.singleton.remote.make_api_url("platform/%s/module_instructions" % core_version), module_dynamic_file, cookies=cookies)

    try:
        with zipfile.ZipFile(module_dynamic_file) as module_dynamic_zip:
            module_dynamic_zip.extractall(os.path.join(path, ".trigger"))
    finally:
        os.unlink(module_dynamic_file)

def _add_module_location(path):
    local_modules_path = os.path.join(defaults.FORGE_ROOT, 'local_modules.json')

    local_modules = {}
    if os.path.exists(local_modules_path):
        with async.current_call().io_lock:
            with open(local_modules_path, 'r') as local_modules_file:
                local_modules = json.load(local_modules_file)

    if "module_locations" in local_modules:
        if not path in local_modules["module_locations"]:
            local_modules["module_locations"].append(path)
    else:
        local_modules["module_locations"] = [path]

    with async.current_call().io_lock:
        with open(local_modules_path, 'w') as local_modules_file:
            json.dump(local_modules, local_modules_file)

@expose
def state(cookies, path):
    # assumption that manifest is always at module/manifest.json
    module_model = {}
    module_manifest_path = os.path.join(path, "module", "manifest.json")

    with async.current_call().io_lock:
        try:
            with open(module_manifest_path) as module_manifest_file:
                try:
                    module_manifest_dict = json.load(module_manifest_file)
                except ValueError:
                    raise ValueError("Couldn't parse JSON file: %s" % module_manifest_path)
        except IOError:
            return module_model

    if "platform_version" in module_manifest_dict:
        try:
            _update_module_dynamic_if_needed(cookies, path, module_manifest_dict['platform_version'])

            module_model.update(make_dynamic_call(
                    os.path.join(path, ".trigger", "module_dynamic"),
                    'module.load',
                    {"path": path, "manifest": module_manifest_dict}
            ))
        except Exception:
            module_model['problems'] = [{
                    "raw_message": """<p class='bigger'>Failed to download platform code</p>
                    <p>You probably have a bad platform_version in your module manifest, make sure the file '%s' contains a valid platform_version property.</p>""" % module_manifest_path
            }]
    else:
        # No platform version means we can't use dynamic code to load the module
        # XXX: I don't think this error will ever be seen, local modules with no name are effectively ignored by the toolkit
        module_model['problems'] = [{
                "raw_message": """<p class='bigger'>No platform version in manifest</p>
                <p>A platform version is required in the module manifest, make sure the file '%s' contains a platform_version property.</p>""" % module_manifest_path
        }]

    return module_model

@expose
def create_module_template(cookies, name, path, namespace, core_version):
    '''Create a template for a module and return a local module object (similar to list_local) to be immediately used'''
    path = os.path.abspath(path)
    if not os.path.exists(path):
        os.makedirs(path)

    if os.listdir(path):
        raise Exception("module can only be created in an empty directory.")

    _update_module_dynamic(cookies, path, core_version)

    # Run create template from module_dynamic
    result = make_dynamic_call(os.path.join(path, ".trigger", "module_dynamic"), 'module.create_template', {"name": name, "path": path, "namespace": namespace})

    _add_module_location(path)

    return result

@expose
def import_module(cookies, name, path, platform_versions_with_namespace=[]):
    path = os.path.abspath(path)
    if not os.path.isdir(path):
        raise remote.FormError({
                "__all__": ["Path is not a directory."]
        })

    if not os.path.exists(os.path.join(path, 'module', 'manifest.json')):
        raise remote.FormError({
                "__all__": ["manifest.json file does not exist in that directory."]
        })

    with async.current_call().io_lock:
        with open(os.path.join(path, 'module', 'manifest.json'), 'r') as manifest_file:
            try:
                manifest = json.load(manifest_file)
            except ValueError:
                raise remote.FormError({
                        "__all__": ["manifest.json file is not value JSON."]
                })

        if os.path.exists(os.path.join(path, 'module', 'identity.json')):
            with open(os.path.join(path, 'module', 'identity.json'), 'r') as identity_file:
                try:
                    identity = json.load(identity_file)
                except ValueError:
                    raise remote.FormError({
                            "__all__": ["identity.json file is not value JSON."]
                    })
        else:
            if not "name" in manifest:
                raise remote.FormError({
                        "__all__": ["manifest.json does not contain a name."]
                })
            else:
                identity = {
                        'name': manifest['name']
                }

        if not "name" in identity:
            raise remote.FormError({
                    "__all__": ["identity.json does not contain a name."]
            })

    if not "platform_version" in manifest:
        raise remote.FormError({
                "__all__": ["manifest.json does not contain a platform_version."]
        })

    if name != identity["name"]:
        if manifest['platform_version'] not in platform_versions_with_namespace:
            raise remote.FormError({
                    "__all__": ["manifest.json contains a mismatching name: manifest.json must contain the name for this module."]
            })

        # Update name/namespace
        if 'namespace' not in manifest:
            manifest['namespace'] = identity['name']
        manifest['name'] = name
        with async.current_call().io_lock:
            # Let identity be regenerated by module dynamic
            if os.path.exists(os.path.join(path, 'module', 'identity.json')):
                os.remove(os.path.join(path, 'module', 'identity.json'))
            with open(os.path.join(path, 'module', 'manifest.json'), 'w') as manifest_file:
                json.dump(manifest, manifest_file, indent=4, sort_keys=True)

    _update_module_dynamic_if_needed(cookies, path, manifest['platform_version'])

    _add_module_location(path)

    return make_dynamic_call(
            os.path.join(path, ".trigger", "module_dynamic"),
            'module.load',
            {"path": path, "manifest": manifest}
    )

@expose
def update_module_dynamic(cookies, path, new_platform_version):
    _update_module_dynamic_if_needed(cookies, path, new_platform_version)

@expose
def remove_local_not_on_filesystem():
    local_modules_path = os.path.join(defaults.FORGE_ROOT, 'local_modules.json')

    local_modules = {}
    if os.path.exists(local_modules_path):
        with async.current_call().io_lock:
            with open(local_modules_path, 'r') as local_modules_file:
                local_modules = json.load(local_modules_file)

    if "module_locations" in local_modules:
        local_modules["module_locations"] = [x for x in local_modules["module_locations"] if os.path.isdir(x)]

    with async.current_call().io_lock:
        with open(local_modules_path, 'w') as local_modules_file:
            json.dump(local_modules, local_modules_file, indent=4)

@expose
def generate_docs(path):
    try:
        docgen.generate("module", os.path.join(path, '.trigger', 'docs-build'), os.path.join(path, 'module', 'docs'), module=True)
    except Exception as e:
        LOG.error(traceback.format_exc())
        raise Exception('Failed to process your documentation: %s' % e)
    webbrowser.open("file://"+os.path.abspath(os.path.join(path, '.trigger', 'docs-build', 'index.html')))
