import codecs
import logging
import json
import os
import shutil
import zipfile

from send2trash import send2trash

from forge import defaults, remote
from forge import async
from trigger.lib import make_dynamic_call
from trigger.api import expose
from trigger import forge_tool

LOG = logging.getLogger(__name__)

def _update_plugin_dynamic_if_needed(cookies, path, manifest_platform_version):
    dynamic_platform_version = ""
    dynamic_platform_version_path = os.path.join(path, ".trigger", "platform_version.txt")
    if os.path.exists(dynamic_platform_version_path):
        with async.current_call().io_lock:
            with open(dynamic_platform_version_path, 'r') as dynamic_platform_version_file:
                dynamic_platform_version = dynamic_platform_version_file.read()

    if dynamic_platform_version != manifest_platform_version:
        _update_plugin_dynamic(cookies, path, manifest_platform_version)

def _update_plugin_dynamic(cookies, path, core_version):
    if os.path.exists(os.path.join(path, '.trigger')):
        for dirpath, dirnames, filenames in os.walk(os.path.join(path, '.trigger'), topdown=False):
            for filename in filenames:
                send2trash(os.path.join(dirpath, filename))
            for dirname in dirnames:
                send2trash(os.path.join(dirpath, dirname))
        send2trash(os.path.join(path, '.trigger'))

    os.makedirs(os.path.join(path, ".trigger"))

    plugin_dynamic_file = os.path.join(path, ".trigger", "plugin_instructions.zip")
    forge_tool.singleton.remote._get_file("%splatform/%s/plugin_instructions" % (forge_tool.singleton.remote.server, core_version), plugin_dynamic_file, cookies=cookies)

    try:
        with zipfile.ZipFile(plugin_dynamic_file) as plugin_dynamic_zip:
            plugin_dynamic_zip.extractall(os.path.join(path, ".trigger"))
    finally:
        os.unlink(plugin_dynamic_file)

def _add_plugin_location(path):
    local_plugins_path = os.path.join(defaults.FORGE_ROOT, 'local_plugins.json')

    local_plugins = {}
    if os.path.exists(local_plugins_path):
        with async.current_call().io_lock:
            with open(local_plugins_path, 'r') as local_plugins_file:
                local_plugins = json.load(local_plugins_file)

    if "plugin_locations" in local_plugins:
        if not path in local_plugins["plugin_locations"]:
            local_plugins["plugin_locations"].append(path)
    else:
        local_plugins["plugin_locations"] = [path]

    with async.current_call().io_lock:
        with open(local_plugins_path, 'w') as local_plugins_file:
            json.dump(local_plugins, local_plugins_file)

@expose
def state(cookies, path):
    # assumption that manifest is always at plugin/manifest.json
    plugin_model = {}
    plugin_manifest_path = os.path.join(path, "plugin", "manifest.json")

    with async.current_call().io_lock:
        with open(plugin_manifest_path) as plugin_manifest_file:
            try:
                plugin_manifest_dict = json.load(plugin_manifest_file)
            except ValueError:
                raise ValueError("Couldn't parse JSON file: %s" % plugin_manifest_path)

    if "platform_version" in plugin_manifest_dict:
        _update_plugin_dynamic_if_needed(cookies, path, plugin_manifest_dict['platform_version'])

        plugin_model.update(make_dynamic_call(
                os.path.join(path, ".trigger", "plugin_dynamic"),
                'plugin.load',
                {"path": path, "manifest": plugin_manifest_dict}
        ))
    elif "uuid" in plugin_manifest_dict:
        # No platform version means we can't use dynamic code to load the plugin
        plugin_model['problems'] = [{
                "raw_message": """<p class='bigger'>No platform version in manifest</p>
                <p>A platform version is required in the plugin manifest, make sure the file '%s' contains a platform_version property.</p>""" % plugin_manifest_path
        }]
    else:
        # No way to identify plugin
        raise Exception("No plugin uuid found in manifest: %s" % plugin_manifest_path)

    return plugin_model

@expose
def create_plugin_template(cookies, uuid, name, path, core_version):
    '''Create a template for a plugin and return a local plugin object (similar to list_local) to be immediately used'''
    path = os.path.abspath(path)
    if not os.path.exists(path):
        os.makedirs(path)

    if os.listdir(path):
        raise Exception("Plugin can only be created in an empty directory.")

    _update_plugin_dynamic(cookies, path, core_version)

    # Run create template from plugin_dynamic
    result = make_dynamic_call(os.path.join(path, ".trigger", "plugin_dynamic"), 'plugin.create_template', {"uuid": uuid, "name": name, "path": path})

    _add_plugin_location(path)

    return result

@expose
def convert_to_module(cookies, uuid, plugin_path, module_path):
    '''Copy a plugin to a new module location, moving files on the way'''
    # check validity of destination path
    if not os.path.exists(module_path):
        os.makedirs(module_path)
    elif not os.path.isdir(module_path):
        raise Exception ("{path} must be an empty directory".format(path=module_path))
    else:
        if os.listdir(module_path):
            raise Exception ("{path} must be an empty directory".format(path=module_path))

    # copy android and ios plugin directories
    for target in ("android", "ios"):
        if os.path.isdir(os.path.join(plugin_path, "plugin", target)):
            shutil.copytree(
                    os.path.join(plugin_path, "plugin", target),
                    os.path.join(module_path, "module", target)
            )
        else:
            os.makedirs(os.path.join(module_path, "module", target))

    # copy android and ios inspector directories
    for target in ("an-inspector", "ios-inspector"):
        if os.path.isdir(os.path.join(plugin_path, "inspector", target)):
            shutil.copytree(
                    os.path.join(plugin_path, "inspector", target),
                    os.path.join(module_path, "inspector", target)
            )

    # create docs/index.md
    os.makedirs(os.path.join(module_path, "module", "docs"))
    with async.current_call().io_lock:
        with codecs.open(os.path.join(module_path, "module", "docs", "index.md"), "w", "utf8") as index:
            index.write("""Docs
====

This is where you should document your module.

You can create other .md files in this folder, or even more folders containing index.md and other .md files.

You can also include `.jpg`, `.gif`, and `.png` files if you would like to use images.
""")

    # copy javascript/plugin.js to javascript/module.js
    if os.path.isdir(os.path.join(plugin_path, "plugin", "javascript")):
        os.makedirs(os.path.join(module_path, "module", "javascript"))
        if os.path.isfile(os.path.join(plugin_path, "plugin", "javascript", "plugin.js")):
            shutil.copyfile(
                    os.path.join(plugin_path, "plugin", "javascript", "plugin.js"),
                    os.path.join(module_path, "module", "javascript", "module.js" ))

    # remove plugin binaries
    if os.path.isfile(os.path.join(module_path, "module", "android", "plugin.jar")):
        os.remove(os.path.join(module_path, "module", "android", "plugin.jar"))
    if os.path.isfile(os.path.join(module_path, "module", "ios", "plugin.a")):
        os.remove(os.path.join(module_path, "module", "ios", "plugin.a"))

    # create manifest.json
    with async.current_call().io_lock:
        with codecs.open(os.path.join(plugin_path, "plugin", "manifest.json"), "r", "utf8") as manifest_file:
            manifest = json.load(manifest_file)
            new_manifest = {
                    "description": manifest.get("description", ""),
                    "name": manifest["name"],
                    "version": manifest.get("version", "0.1"),
                    "platform_version": "v2.0.0"
            }
        with codecs.open(os.path.join(module_path, "module", "manifest.json"), "w", "utf8") as new_manifest_file:
            json.dump(new_manifest, new_manifest_file)

    # copy tests
    if os.path.isdir(os.path.join(plugin_path, "plugin", "tests")):
        shutil.copytree(os.path.join(plugin_path, "plugin", "tests"),
                os.path.join(module_path, "module", "tests"))

    return {
            "name": manifest["name"],
            "path": module_path
    }

@expose
def import_plugin(cookies, name, path):
    path = os.path.abspath(path)
    if not os.path.isdir(path):
        raise remote.FormError({
                "__all__": ["Path is not a directory."]
        })

    if not os.path.exists(os.path.join(path, 'plugin', 'manifest.json')):
        raise remote.FormError({
                "__all__": ["manifest.json file does not exist in that directory."]
        })

    with async.current_call().io_lock:
        with open(os.path.join(path, 'plugin', 'manifest.json'), 'r') as manifest_file:
            try:
                manifest = json.load(manifest_file)
            except ValueError:
                raise remote.FormError({
                        "__all__": ["manifest.json file is not value JSON."]
                })

    if not "name" in manifest:
        raise remote.FormError({
                "__all__": ["manifest.json does not contain a name."]
        })

    if name != manifest["name"]:
        raise remote.FormError({
                "__all__": ["manifest.json contains a mismatching name, manifest.json must contain the name for this plugin."]
        })

    if not "platform_version" in manifest:
        raise remote.FormError({
                "__all__": ["manifest.json does not contain a platform_version."]
        })

    _update_plugin_dynamic_if_needed(cookies, path, manifest['platform_version'])

    _add_plugin_location(path)

    return make_dynamic_call(
            os.path.join(path, ".trigger", "plugin_dynamic"),
            'plugin.load',
            {"path": path, "manifest": manifest}
    )

@expose
def update_plugin_dynamic(cookies, path, new_platform_version):
    _update_plugin_dynamic_if_needed(cookies, path, new_platform_version)

@expose
def remove_local_not_on_filesystem():
    local_plugins_path = os.path.join(defaults.FORGE_ROOT, 'local_plugins.json')

    local_plugins = {}
    if os.path.exists(local_plugins_path):
        with async.current_call().io_lock:
            with open(local_plugins_path, 'r') as local_plugins_file:
                local_plugins = json.load(local_plugins_file)

    if "plugin_locations" in local_plugins:
        local_plugins["plugin_locations"] = [x for x in local_plugins["plugin_locations"] if os.path.isdir(x)]

    with async.current_call().io_lock:
        with open(local_plugins_path, 'w') as local_plugins_file:
            json.dump(local_plugins, local_plugins_file)
