import os
import multiprocessing
import uuid
import sys
from threading import Thread

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from trigger.api import expose
from trigger import lib
from forge import async
import workspace

watcher_process_queue = None
watcher_process_out_queue = multiprocessing.Queue()

def watcher(in_queue, out_queue):
    with open(workspace.path('watcher.log'), mode='w') as watcher_output:
        watcher_output.write('Starting up file watcher daemon using %s\n' % Observer)
        try:
            watches = {
                    'recursive': {},
                    'single': {}
            }

            observer = Observer()
            observer.start()

            class RPCEventHandler(FileSystemEventHandler):
                def __init__(self, call_id, dir_path, file_path=None):
                    self.call_id = call_id
                    self.file_path = file_path
                    self.dir_path = dir_path

                def on_any_event(self, event):
                    watcher_output.write('Watchdog event %s %s %s\n' % (self.dir_path, self.file_path, event))
                    # Ignore directory modified events (we see a file modified event too)
                    if event.event_type == "modified" and event.is_directory is True:
                        return
                    if self.file_path is None or event.src_path == self.file_path:
                        result = {
                                "event_type": event.event_type,
                                "is_directory": event.is_directory,
                                "src_path": event.src_path[len(self.dir_path)+1:]
                        }
                        if hasattr(event, "dest_path"):
                            result['dest_path'] = event.dest_path
                        try:
                            out_queue.put((self.call_id, result))
                        except Exception:
                            pass

            while True:
                try:
                    watch = in_queue.get()
                    call_id = watch[0]
                    dir_path = watch[1]
                    file_path = watch[2]
                    recursive = watch[3]
                    event_handler = RPCEventHandler(call_id, dir_path, file_path)
                    watcher_output.write("Got file watch request for %s %s\n" % (dir_path, file_path))

                    if recursive:
                        type = 'recursive'
                    else:
                        type = 'single'

                    if dir_path not in watches[type]:
                        watches[type][dir_path] = observer.schedule(event_handler, path=dir_path, recursive=recursive)
                    else:
                        observer.add_handler_for_watch(event_handler, watches[type][dir_path])

                except Exception as e:
                    watcher_output.write("Exception handling watch request: %s\n" % e)
        except Exception as e:
            watcher_output.write("Exception in outer watcher: %s\n" % e)

def start_watcher_process():
    global watcher_process_queue

    watcher_process_queue = multiprocessing.Queue()
    watcher_process = multiprocessing.Process(target=watcher, args=(watcher_process_queue, watcher_process_out_queue))
    watcher_process.daemon = True
    watcher_process.start()

calls = {}

def handle_out_queue(out_queue):
    while True:
        result = out_queue.get()
        if result[0] in calls:
            calls[result[0]].emit('change', watch_event=result[1])

out_queue_handler = Thread(target=handle_out_queue, args=(watcher_process_out_queue,))
out_queue_handler.daemon = True
out_queue_handler.start()

def _watch(dir_path, file_path=None, recursive=False):
    if watcher_process_queue is None:
        start_watcher_process()

    call = async.current_call()
    call_id = uuid.uuid4()
    calls[call_id] = call

    watcher_process_queue.put((call_id, dir_path, file_path, recursive))

    # TODO: Actually remove event handlers for this call?
    call.at_exit(lambda: calls.pop(call_id))
    call.never_return()

@expose
def file(path, dir_path):
    path = lib.normalized_path_relative_to_toolkit(path)

    if sys.platform.startswith("darwin"):
        # ignore passed dir_path on osx
        dir_path = os.path.dirname(path)
        recursive = False
    else:
        recursive = True

    # Windows/Linux can only watch directories, if no dir_path provided
    # then watch the directory the file is in
    if dir_path is None:
        dir_path = os.path.dirname(path)

    _watch(dir_path, path, recursive=recursive)

@expose
def directory(path):
    path = lib.normalized_path_relative_to_toolkit(path)

    _watch(path, recursive=True)
