import inspect
import logging
import errno
import traceback
import os
import bottle
import json
import glob
import shutil
import subprocess
import sys
import workspace
import hashlib
import importlib
import threading

from forge import lib as forge_lib, async

LOG = logging.getLogger(__name__)


def post_params(*param_names):
    """Construct a decorator to specify necessary post parameters"""
    def decorator(view):
        args = inspect.getargspec(view).args
        if len(args) != len(param_names):
            raise Exception("View decorated by post_params should take the same number of parameters as provided")

        def with_required_parameters():
            params = []
            for name in param_names:
                p = bottle.request.POST.get(name)
                if p is None:
                    raise Exception("You need to provide '{param}' as a POST parameter".format(param=name))
                params.append(p)

            # finally, call the view with the extracted POST parameters
            return view(*params)

        return with_required_parameters

    return decorator

def json_view(view):
    """Used as a decorator to encode the value returned by a view as JSON"""
    def with_json_response(*args, **kwargs):
        bottle.response.content_type = 'application/json'
        try:
            response_content = view(*args, **kwargs)
            return json.dumps({
                    "ok": 1,
                    "content": response_content
            })

        except Exception as e:
            LOG.error("Error rendering JSON view: {0}".format(e))
            LOG.debug("Exception info:", exc_info=True)
            return json.dumps({
                    "error": 1,
                    "message": unicode(e),
                    "traceback": traceback.format_exc(),
                    "error_type": type(e).__name__,
            })

    return with_json_response

def ensure_folder_exists(path):
    """Returns True if the operation had to make a folder"""
    try:
        os.makedirs(path)
        return True
    except OSError as e:
        if e.errno == errno.EEXIST:
            return False
        else: raise

def nested_dict_to_dot_notation(dictionary, accumulator=None, prefix=''):
    if accumulator is None:
        accumulator = {}

    for k,v in dictionary.items():
        if isinstance(v, dict):
            nested_dict_to_dot_notation(v, accumulator, prefix='%s%s.' % (prefix, k))
        else:
            accumulator['%s%s' % (prefix, k)] = v

    return accumulator

def dot_notation_to_nested_dict(dictionary, accumulator=None):
    top_level = {}

    for k,v in dictionary.items():
        crumbs = k.split('.')

        current_level = top_level
        for c in crumbs[:-1]:
            if c not in current_level:
                current_level[c] = {}

            current_level = current_level[c]
        current_level[crumbs[-1]] = v

    return top_level

def incremental_backup_file(path_to_backup):
    """Renames a file foo to foo.1, as well as moving
    foo.1 -> foo.2, foo.2 -> foo.3 ... foo.4 -> foo.5 beforehand.

    :param path_to_backup: Path to file to create a backup of
    """
    file_name = os.path.basename(path_to_backup)

    if os.path.exists(path_to_backup):
        backups_one_to_five = list(sorted(glob.glob('%s.*' % path_to_backup)))[:4]

        # move file.3 -> file.4, file.2 -> file.3, file.1 -> file.2
        for backup in reversed(backups_one_to_five):
            backup_number = int(backup.split('.')[-1])
            shutil.move(backup, path_to_backup + '.' + str(backup_number + 1))

        shutil.move(path_to_backup, path_to_backup + '.1')


def disable_file_inheritance(file_object):
    """Sets flags on the low level parts of a file to prevent it from being
    inherited by spawned processes (including multiprocessing.Processes).

    Currently only does anything on windows.

    :param file_object: File object to set as non-inheritable.
    """
    if sys.platform.startswith("win"):
        import ctypes
        import msvcrt
        HANDLE_FLAG_INHERIT = 1
        ctypes.windll.kernel32.SetHandleInformation(
                msvcrt.get_osfhandle(file_object.fileno()),
                HANDLE_FLAG_INHERIT,
                0
        )

def open_in_filebrowser(path):
    """Attempt to open the provided path in the file browser of the host OS"""
    if sys.platform.startswith('darwin'):
        forge_lib.PopenWithoutNewConsole(['open', path], stdout=subprocess.PIPE, stderr=subprocess.STDOUT).wait()
    elif sys.platform.startswith('win'):
        # 'start' seems to need shell=True to be found (probably a builtin)
        forge_lib.PopenWithoutNewConsole('start /D "{0}" .'.format(path), shell=True)
    elif sys.platform.startswith('linux'):
        forge_lib.PopenWithoutNewConsole(['xdg-open', path], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

def fork_and_run(target):
    # do the UNIX double-fork magic, see Stevens' "Advanced
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError, e:
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror)
        sys.exit(1)

    # decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent, print eventual PID before
            print "Daemon PID %d" % pid
            sys.exit(0)
    except OSError, e:
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror)
        sys.exit(1)

    target()

def path_components(path_string):
    path_string = os.path.normpath(path_string)
    path_parts = []
    while True:
        path_string, component = os.path.split(path_string)
        if component:
            path_parts.append(component)
        else:
            if path_string:
                drive, _ = os.path.splitdrive(path_string)
                if drive:
                    path_parts.append(drive)
                else:
                    path_parts.append(path_string)
            break

    path_parts.reverse()
    return path_parts

def make_hash(o):
    """
    Makes a hash from any json stringifyable object.
    """
    return hashlib.sha1(json.dumps(o, separators=(',',':'), sort_keys=True)).hexdigest()

dynamic_lock = threading.Lock()
def make_dynamic_call(python_path, method_name, params, fail_if_missing=True):
    if isinstance(python_path, basestring):
        python_path = [python_path]

    dynamic_lock.acquire()
    try:
        current_modules = set(sys.modules.keys())
        try:
            method_parts = method_name.split(".")

            for p in python_path:
                sys.path.insert(0, p)

            try:
                target = importlib.import_module(method_parts.pop(0))
            finally:
                for p in python_path:
                    sys.path.remove(p)

            for name in method_parts:
                target = getattr(target, name, None)

            if target is None:
                if not fail_if_missing:
                    return
                raise Exception("Tried to call method %s, but it didn't exist" % method_name)

            params = {k: params[k] for k in params if k != ""}   # strip anonymous form fields
            given_arg_names = set(params.keys())
            expected_arg_names = set(inspect.getargspec(target)[0])

            incorrect_invocation_message = 'invoking {method} with {given}, expected {expected}'.format(
                            method=method_name, given=given_arg_names, expected=expected_arg_names)
            if expected_arg_names != given_arg_names:
                LOG.warning(incorrect_invocation_message)

            return target(**params)
        finally:
            new_modules = set(sys.modules.keys()) - current_modules
            for module in new_modules:
                del sys.modules[module]
    finally:
        dynamic_lock.release()


file_handlers = {}

def close_log_to_file(filename):
    """Stop logging to a log file setup using setup_log_to_file."""
    if filename in file_handlers:
        file_handler = file_handlers[filename]
        logging.root.removeHandler(file_handler)
        del file_handlers[filename]

class RpcAwareFormatter(logging.Formatter):
    def __init__(self, formatter, *args, **kwargs):
        super(RpcAwareFormatter, self).__init__(*args, **kwargs)
        self._inner_formatter = formatter

    def format(self, record):
        inner_format = self._inner_formatter.format(record)
        call = async.current_call()
        if call is not None:
            target = call._target
            return "RPC:%s.%s %s" % (target.__module__, target.__name__, inner_format)
        return inner_format

def setup_log_to_file(filename, backup):
    """Setup logging to a file in the directory the Toolkit is installed into."""
    path_to_log_to = workspace.path(filename)
    try:
        if backup:
            incremental_backup_file(path_to_log_to)

        file_handler = logging.FileHandler(path_to_log_to)

        # adb seems to get a lock on the log file if we don't have this
        disable_file_inheritance(file_handler.stream)

        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(RpcAwareFormatter(logging.Formatter('%(asctime)s PID:%(process)d [%(name)s:%(levelname)7s] %(message)s')))
        logging.root.addHandler(file_handler)
        file_handlers[filename] = file_handler
    except OSError:
        LOG.error("Couldn't create %s for log output, carrying on anyway." % filename)

def setup_toolkit_log(backup):
    return setup_log_to_file('toolkit.log', backup)

# TODO: taken straight from generate_dynamic/web_tasks, might want to make this shared between them somehow
# TODO: can we replace this with import webbrowser, webbrowser.open() (which is in the standard library)
def open_url(url):
    '''Attempt to open the provided URL in the default browser'''
    LOG.info('Attempting to open browser at %s' % url)
    if sys.platform.startswith('darwin'):
        subprocess.call(['open', url], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    elif sys.platform.startswith('win'):
        # 'start' seems to need shell=True to be found (probably a builtin)
        cmd = subprocess.list2cmdline(['start', url])
        subprocess.call(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    elif sys.platform.startswith('linux'):
        subprocess.call(['xdg-open', url], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

def normalized_path_relative_to_toolkit(path):
    return os.path.join(workspace.path(), os.path.normpath(path))
