import uuid
import Queue
import time
import multiprocessing
import logging
import threading
import inspect
import traceback
import sys

from trigger import forge_tool
from trigger import lib
from forge import async


LOG = logging.getLogger(__name__)


def _transfer_between_queues_until_success_or_error(producer, consumer):
    while True:
        next = producer.get()
        type = next['type']
        consumer.put(next)

        if type in ('success', 'error'):
            return

        if type == 'log':
            logging_level = getattr(logging, next.get('level', 'DEBUG'))
            LOG.log(logging_level, next.get('message', ''))

# SessionManagerProxy relies on this to run a SessionManager and proxy input/output for it
def _worker(worker_input, worker_output, shutdown_when_no_sessions):
    try:
        # need to setup logging on windows again, on UNIX it is inherited from the parent process.
        if sys.platform.startswith("win"):
            logging.basicConfig(level='DEBUG', format='[%(levelname)7s] %(message)s')
            lib.setup_toolkit_log(backup=False)

        m = SessionManager(worker_input, worker_output, multiprocessing.RLock(), shutdown_when_no_sessions)
        m.start()
    except:
        LOG.error("_worker died", exc_info=True)
        raise
    finally:
        # this will only be reached if something has gone wrong
        worker_output.put({'type': 'shutdown'})


class SessionManagerProxy(object):
    def __init__(self, shutdown, shutdown_when_no_sessions):
        self._shutdown = shutdown

        self._worker_input = multiprocessing.Queue()
        self._worker_output = multiprocessing.Queue()
        self._session_queues = {}

        worker_proc = multiprocessing.Process(
                target=_worker,
                args=(self._worker_input, self._worker_output, shutdown_when_no_sessions)
        )
        worker_proc.start()

        distribution_thread = threading.Thread(
                target=self._distribute_worker_output,
        )
        distribution_thread.daemon = True
        distribution_thread.start()

    def _distribute_worker_output(self):
        while True:
            next_event = self._worker_output.get()
            if next_event['type'] == 'shutdown':
                self._worker_input.put(next_event)
                self._shutdown()
                return

            session_id = next_event['sessionId']
            self._session_queues[session_id].put(next_event)

    def get_session_events(self, session_id, limit, timeout=5):
        self._cancel_session_timeout(session_id)
        output_queue = self._session_queues[session_id]

        got_item = False
        def timeout_queue():
            time.sleep(timeout)
            if not got_item:
                output_queue.put(None)

        threading.Thread(target=timeout_queue).start()

        events = []
        next = output_queue.get(block=True)
        got_item = True
        if next is not None:
            events.append(next)

        # see if there are any other events we can quickly get to reduce round trips
        for i in xrange(limit - 1):
            try:
                next = output_queue.get_nowait()
                if next is not None:
                    events.append(next)
            except Queue.Empty:
                pass

        return events

    def _cancel_session_timeout(self, session_id):
        self._worker_input.put({
                'type': 'sessionCancelTimeout',
                'sessionId': session_id,
        })

    def create_session(self):
        session_id = str(uuid.uuid4())
        self._session_queues[session_id] = Queue.Queue()

        self._worker_input.put({
                'type': 'sessionCreate',
                'sessionId': session_id,
        })

        return session_id

    def call(self, session_id, call_id, method_name, params):
        self._worker_input.put({
                'type': 'callCreate',
                'sessionId': session_id,
                'callId': call_id,
                'methodName': method_name,
                'params': params,
        })

    def interrupt_call(self, session_id, call_id):
        self._worker_input.put({
                'type': 'callInterrupt',
                'sessionId': session_id,
                'callId': call_id,
        })

    def input(self, session_id, call_id, response):
        self._worker_input.put({
                'type': 'callInput',
                'sessionId': session_id,
                'callId': call_id,
                'response': response,
        })


class RequirementMissing(Exception):
    def __init__(self, requirement, *args, **kwargs):
        self.requirement = requirement
        Exception.__init__(self, *args, **kwargs)


def require(d, *keys):
    vs = []
    for k in keys:
        if k not in d:
            raise RequirementMissing(k, "Missing %s" % k)
        vs.append(d[k])

    if len(keys) == 1:
        return vs[0]
    else:
        return vs


class SessionManager(object):
    def __init__(self, input, output, io_lock, shutdown_when_no_sessions):
        self._shutdown_when_no_sessions = shutdown_when_no_sessions
        self._input = input
        self._output = output
        self._sessions = {}
        self._watcher_rate = 10
        self._watcher_limit = 3
        self._watcher_thread = threading.Thread(
                target=self._watch_sessions,
        )
        self._watcher_thread.daemon = True
        self._watcher_thread.start()
        self._io_lock = io_lock

    def _watch_sessions(self):
        had_no_active_sessions = 0
        ever_had_session = False

        while True:
            if self._sessions:
                had_no_active_sessions = 0
                ever_had_session = True
            else:
                had_no_active_sessions += 1
            session_ids = self._sessions.keys()
            LOG.info('%d sessions: %s' % (len(session_ids), session_ids))

            # if we never got any active sessions, shutdown
            if (self._shutdown_when_no_sessions and had_no_active_sessions > self._watcher_limit):
                LOG.info('No sessions opened in %d seconds, shutting down...' % (self._watcher_limit * self._watcher_rate))
                self._shutdown()
                return

            time.sleep(self._watcher_rate)
            to_remove = set()

            for id, sesh in self._sessions.items():
                if sesh.has_timedout():
                    sesh.interrupt_calls()
                    to_remove.add(id)

            for id in to_remove:
                del self._sessions[id]

            # if we just removed all sessions, don't bother polling again
            if self._shutdown_when_no_sessions and not self._sessions and ever_had_session:
                LOG.info('Last active session timedout, shutting down...')
                self._shutdown()
                return

    def _shutdown(self):
        self._output.put({'type': 'shutdown'})

    def start(self):
        """Process all input to the session until receiving a shutdown event."""
        while True:
            try:
                next_input = self._input.get()
                event_type = require(next_input, 'type')

                if event_type == 'shutdown':
                    return

                elif event_type == 'sessionCreate':
                    session_id = require(next_input, 'sessionId')
                    self._create_session(session_id)

                elif event_type == 'callCreate':
                    session_id, call_id, method_name, params = require(next_input, 'sessionId', 'callId', 'methodName', 'params')
                    self._call(session_id, call_id, method_name, params)

                elif event_type == 'callInput':
                    session_id, call_id, response = require(next_input, 'sessionId', 'callId', 'response')
                    self._call_input(session_id, call_id, response)

                elif event_type == 'sessionCancelTimeout':
                    session_id = require(next_input, 'sessionId')
                    call = require(self._sessions, session_id)
                    call.cancel_timeout()

                elif event_type == 'callInterrupt':
                    session_id, call_id = require(next_input, 'sessionId', 'callId')
                    self._call_interrupt(session_id, call_id)

            except RequirementMissing as e:
                LOG.debug('Expected %s' % e.requirement)
                LOG.debug(traceback.format_exc(e))
                continue

    def _call_interrupt(self, session_id, call_id):
        session = self._sessions[session_id]
        session.interrupt_call(call_id)

    def _call_input(self, session_id, call_id, response):
        if session_id not in self._sessions:
            LOG.warning("Tried to input to session that doesn't exist: %s" % session_id)
            return

        session = self._sessions[session_id]
        session.input_to_call(call_id, response)

    def _create_session(self, session_id):
        self._sessions[session_id] = Session(session_id=session_id, output=self._output, io_lock=self._io_lock)

    def _call(self, session_id, call_id, method_name, params):
        if session_id not in self._sessions:
            LOG.warning("Tried to call from a session that doesn't exist: %s" % session_id)
            return

        self._sessions[session_id].call(call_id, method_name, params)


class Session(object):
    """Responsible for creating Calls and allowing them to emit events
    back to the invoking side.

    Keeps track of calls using their call IDs, so the invoking side
    can send responses to events emitted by the call by providing the call ID.
    """
    def __init__(self, session_id, output, io_lock):
        self._output = output
        self._session_id = session_id

        self.cancel_timeout()
        self._timeout = 30
        self._poll_timeout = 5
        self._calls = {}
        self._procs = {}
        self._io_lock = io_lock

    def cancel_timeout(self):
        self._last_request_for_events = int(time.time())

    def has_timedout(self):
        cur_time = int(time.time())
        return self._last_request_for_events < cur_time - self._timeout

    def call(self, call_id, method_name, params):
        from trigger import api

        target = api
        for name in method_name.split('.'):
            target = getattr(target, name, None)

        if target is None:
            self._run_error(call_id, "Tried to call method %s, but it didn't exist" % method_name)
            return

        # strip anonymous form fields
        params =  {k: params[k] for k in params if k != ""}
        given_arg_names = set(params.keys())
        expected_arg_names = set(inspect.getargspec(target)[0])

        if expected_arg_names != given_arg_names:
            self._run_error(call_id, "Tried to call %s with params: %s, but expected %s" % (method_name, given_arg_names, expected_arg_names))
            return

        required_isolation_level = api.isolation_level_for_target(target)

        if required_isolation_level == api.PROCESS:
            LOG.info('spawning proc for %s' % method_name)
            self._run_in_process(call_id, target, params)

        elif required_isolation_level == api.THREAD:
            LOG.info('spawning thread for %s' % method_name)
            self._run_in_thread(call_id, target, params)

    def get_call(self, call_id):
        return self._calls[call_id]

    def input_to_call(self, call_id, response):
        if call_id not in self._calls:
            LOG.warning("Tried to input to call %s which didn't exist!" % call_id)
            LOG.warning(str(response))
            return

        call = self._calls[call_id]
        call.input(response)

    def _raise(self, message):
        """Just used when we want to immediately indicate that an RPC call has failed."""
        raise Exception(message)

    def _run_error(self, call_id, message):
        """Create a Call which just errors immediately, and run it synchronously."""
        async.Call(
                call_id=call_id,
                stamp=dict(sessionId=self._session_id),
                target=self._raise,
                args=(message,),
                output=self._output,
                input=Queue.Queue(),
                io_lock=self._io_lock
        ).run()

    def _run_in_thread(self, call_id, target, params):
        input_queue = Queue.Queue()

        call = async.Call(
                call_id=call_id,
                stamp=dict(sessionId=self._session_id),
                target=target,
                kwargs=params,
                output=self._output,
                input=input_queue,
                io_lock=self._io_lock
        )
        self._calls[call_id] = call

        threading.Thread(
                target=_run_call,
                args=(call, forge_tool.singleton_params),
                kwargs=dict(remove_loggers=False)
        ).start()

    def _run_in_process(self, call_id, target, params):
        process_output = multiprocessing.Queue()
        process_input = multiprocessing.Queue()

        call = async.Call(
                call_id=call_id,
                stamp=dict(sessionId=self._session_id),
                target=target,
                kwargs=params,
                output=process_output,
                input=process_input,
                io_lock=self._io_lock
        )
        self._calls[call_id] = call

        proc = multiprocessing.Process(
                target=_run_call,
                args=(call, forge_tool.singleton_params),
                kwargs=dict(remove_loggers=True, running_in_new_process=True),
        )
        proc.start()

        self._procs[call_id] = proc

        transfer_thread = threading.Thread(
                target=_transfer_between_queues_until_success_or_error,
                kwargs=dict(producer=process_output, consumer=self._output)
        )
        transfer_thread.daemon = True
        transfer_thread.start()

    def interrupt_call(self, call_id):
        POLLS_BEFORE_HARD_INTERRUPT = 4

        # send a soft interrupt to the call
        # relies on the api logic checking for an interrupt
        if call_id not in self._calls:
            LOG.warning("Tried to interrupt call that doesn't exist: %s" % call_id)
            return

        call = self._calls[call_id]
        call.interrupt()

        # if the Call was run in another process, we can terminate it
        # more forcefully if needs be
        proc = self._procs.get(call_id)

        if proc:
            polls = 0
            while proc.is_alive():
                time.sleep(1)
                polls += 1
                if polls == POLLS_BEFORE_HARD_INTERRUPT:
                    proc.terminate()
                    call.emit('error', message='Forcefully terminated', check_for_interrupt=False)

        # remove references to Call/Process as we're done with this RPC
        try:
            del self._calls[call_id]
            del self._procs[call_id]
        except KeyError:
            pass

    def interrupt_calls(self):
        for call_id in self._calls.keys():
            self.interrupt_call(call_id)


def _run_call(call, forge_tool_singleton_params, remove_loggers=False, running_in_new_process=False):
    """We use this as our starting point for RPC calls. We need a top level function
    as a starting point in the case where we want to start a call in a new process.
    """
    handler = None
    try:
        async.set_current_call(call, thread_local=(not running_in_new_process))

        # when spawning a new process, we want to clear all the existing loggers to
        # prevent duplicate output
        if remove_loggers:
            for h in logging.root.handlers[:]:
                logging.root.removeHandler(h)

        if running_in_new_process:
            # setup a logging.Handler which captures logging activity from this process
            handler = async.CallHandler(call)
            handler.setLevel(logging.DEBUG)
            logging.root.addHandler(handler)
            logging.root.setLevel(logging.DEBUG)

        # when spawning a new process on windows, we want to try to ensure that
        # we re-use the same forge_build.json
        forge_tool.init(forge_tool_singleton_params)

        call.run()
    finally:
        if handler:
            logging.root.removeHandler(handler)
