"""Runs a webserver mapping various urls to tool actions"""
import json
import logging
import time
import thread
from os import path
from uuid import uuid1
import sys
import threading
import os
import codecs

import bottle
from bottle import static_file
import pystache

from trigger import lib
from trigger.lib import json_view, post_params
from trigger.session import SessionManagerProxy
from forge import build_config

import workspace

LOG = logging.getLogger(__name__)

app = bottle.Bottle()
server = None
session_manager = None

@app.hook('after_request')
def set_cache_control():
    """Tell the browser that it should not cache anything served up by bottle."""
    bottle.response.set_header('Cache-Control', 'no-cache')

@app.route('/session/new', method='POST')
@json_view
def session_new():
    session_id = session_manager.create_session()
    return session_id


@app.route('/session/pollevents', method='POST')
@json_view
@post_params('sessionId')
def session_events(session_id):
    return session_manager.get_session_events(session_id, limit=10)


@app.route('/session/call', method='POST')
@json_view
@post_params('sessionId', 'calls')
def session_call(session_id, calls):
    calls = json.loads(calls)
    for call in calls:
        session_manager.call(session_id, call['callId'], call['methodName'], call['params'])


@app.route('/session/interruptcall', method='POST')
@json_view
@post_params('sessionId', 'callId')
def session_interrupt_call(session_id, call_id):
    session_manager.interrupt_call(session_id, call_id)


@app.route('/session/input', method='POST')
@json_view
@post_params('sessionId', 'callId', 'response')
def session_input(session_id, call_id, response):
    session_manager.input(session_id, call_id, json.loads(response))

@app.route('/static/templates')
def serve_templates():
    bottle.response.content_type = 'text/javascript'

    templates = {}

    for dirpath, dirnames, filenames in os.walk(workspace.path('static/templates')):
        for filename in filenames:
            rel_path = path.join(dirpath, filename)[len(workspace.path('static/templates'))+1:]
            if rel_path.endswith('.html'):
                template_name = rel_path[:-5].replace('\\', '/')

                with open(path.join(dirpath, filename)) as template_file:
                    templates[template_name] = template_file.read()

    return "var ToolkitTemplates = " + json.dumps(templates)

# changes every time the toolkit is run, so that restarting
# invalidates the appcache.
_run_id = None

@app.route('/static/manifest-v2.appcache')
def serve_manifest():
    """Serve up an appcache manifest with a timestamp inserted.

    This appcache means that attempts to revisit a Toolkit page while it's offline
    can be handled by the browser.
    """
    global _run_id
    if not _run_id:
        _run_id = uuid1()
    bottle.response.content_type = 'text/cache-manifest'
    return """CACHE MANIFEST

# Update the cache each time the Toolkit restarts
# {run_id}

# If the toolkit isn't running show dead.html rather than index.html
FALLBACK:
/ /static/dead.html
/embed /static/dead.html

# Try not to cache things.
NETWORK:
*""".format(run_id=_run_id)

def static_file_uncached(*args, **kw):
    """Pass through to bottle's static_file, not allowing caching"""
    bottle.request.environ['HTTP_IF_MODIFIED_SINCE'] = None
    return static_file(*args, **kw)

@app.route('/static/<filename:path>')
def serve_static(filename):
    """Serve up any other static file."""
    if filename.endswith(".png"):
        return static_file_uncached(filename, root=workspace.path('static'), mimetype='image/png')
    return static_file_uncached(filename, root=workspace.path('static'))

@app.route('/')
def index():
    bottle.request.environ['HTTP_IF_MODIFIED_SINCE'] = None
    bottle.response.content_type = 'text/html'

    with codecs.open(os.path.join(workspace.path('static'), 'index.html'), 'r') as index_file:
        return pystache.render(index_file.read(), {'embed': False, 'debug': build_config.load().get('debug', False)})

@app.route('/embed')
def embed():
    bottle.request.environ['HTTP_IF_MODIFIED_SINCE'] = None
    bottle.response.content_type = 'text/html'

    with codecs.open(os.path.join(workspace.path('static'), 'index.html'), 'r') as index_file:
        return pystache.render(index_file.read(), {'embed': True, 'debug': build_config.load().get('debug', False)})

port_file_path = workspace.path('.trigger_port')

def _run_server(launch_browser, port):
    global server, session_manager

    port_file = open(port_file_path, 'w')
    port_file.write(str(port))
    port_file.close()
    url = 'https://%s:%d/' % ('toolkit-local.com', port)

    LOG.debug('Toolkit startup complete, closing startup.log')
    lib.close_log_to_file('startup.log')
    lib.setup_toolkit_log(backup=True)

    from cherrypy import wsgiserver
    server_address = ('127.0.0.1', port)
    LOG.info('Starting up webserver at {0:s}'.format(server_address))
    server = wsgiserver.CherryPyWSGIServer(server_address, app, numthreads=30, shutdown_timeout=0.01)
    server.ssl_certificate = workspace.path('https/toolkit-local_com.crt')
    server.ssl_certificate_chain = workspace.path('https/toolkit-local_com-chain.crt')
    server.ssl_private_key = workspace.path('https/toolkit-local_com.key')

    def sd():
        LOG.debug('Starting shutdown of Toolkit web server: %s' % time.time())
        thread.interrupt_main()
    session_manager = SessionManagerProxy(shutdown=sd, shutdown_when_no_sessions=('--testing' not in sys.argv))

    timer = None
    if launch_browser:
        timer = threading.Timer(2, lib.open_url, args=(url,))
        timer.start()

    try:
        server.start()
    finally:
        LOG.debug('Web server process quitting %s' % time.time())
        try:
            if timer:
                timer.cancel()
        except Exception as e:
            LOG.error("Error in finally", exc_info=True)
            pass

def run_server(launch_browser):
    if '--testing' in sys.argv:
        port = 5000
    else:
        port = 38394

        with open(workspace.path('build-tools', 'forge_build.json'), 'r') as forge_build_file:
            try:
                forge_build_dict = json.load(forge_build_file)
                port = forge_build_dict['port']
            except Exception:
                pass

    if sys.platform.startswith("darwin") and ('--testing' not in sys.argv):
        # important to fork here, motivation is that we want
        # to be able to run TriggerToolkit.app repeatedly on OSX
        lib.fork_and_run(lambda: _run_server(launch_browser, port))

    else:
        _run_server(launch_browser, port)
